Package: cNORM
Title: Continuous Norming
Version: 3.6.2
Authors@R: c(
    person("Alexandra", "Lenhard", 
           email = "lenhard@psychometrica.de",
           role = "aut",
           comment = c(ORCID = "0000-0001-8680-4381")), 
    person("Wolfgang", "Lenhard", 
           email = "wolfgang.lenhard@uni-wuerzburg.de",
           role = c("cre", "aut"), 
           comment = c(ORCID = "0000-0002-8184-6889")),
    person("Sebastian", "Gary", 
           role = "aut"),
    person("WPS", "Publisher", 
           role = "fnd",
           comment = "https://www.wpspublish.com/"))
Description: Generates continuous test norms in 
     psychometrics and biometrics, and analyzing model fit. The package offers 
     both distribution-free modeling using Taylor polynomials and parametric 
     modeling using the beta-binomial and the 'Sinh-Arcsinh' distribution. 
     Originally developed for achievement tests, it is applicable to a wide 
     range of mental, physical, or other test scores dependent on continuous or 
     discrete explanatory variables. The package provides several advantages: 
     It minimizes deviations from representativeness in subsamples, interpolates 
     between discrete levels of explanatory variables, and significantly reduces 
     the required sample size compared to conventional norming per age group. 
     cNORM enables graphical and analytical evaluation of model fit, 
     accommodates a wide range of scales including those with negative and 
     descending values, and as well supports conventional norming. It generates 
     norm tables including confidence intervals. Methods for addressing 
     representativeness issues are available through Iterative Proportional 
     Fitting.  Based on Lenhard et al. (2016) 
    <doi:10.1177/1073191116656437>, Lenhard et al. (2019) 
    <doi:10.1371/journal.pone.0222279>, Lenhard and Lenhard (2021) 
    <doi:10.1177/0013164420928457> and Gary et al. (2023) 
    <doi:10.1007/s00181-023-02456-0>.
License: AGPL-3
URL: https://www.psychometrica.de/cNorm_en.html,
        https://github.com/WLenhard/cNORM
BugReports: https://github.com/WLenhard/cNORM/issues
Depends: R (>= 4.0.0)
Imports: ggplot2 (>= 3.5.0), leaps (>= 3.1), grDevices, parallel,
        stats, utils
Suggests: DT, haven, foreign, knitr, markdown, numDeriv, readxl,
        rmarkdown, shiny, shinycssloaders, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
LazyData: true
LazyDataCompression: xz
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-20 11:10:14 UTC; gbpa005
Author: Alexandra Lenhard [aut] (ORCID:
    <https://orcid.org/0000-0001-8680-4381>),
  Wolfgang Lenhard [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-8184-6889>),
  Sebastian Gary [aut],
  WPS Publisher [fnd] (https://www.wpspublish.com/)
Maintainer: Wolfgang Lenhard <wolfgang.lenhard@uni-wuerzburg.de>
Repository: CRAN
Date/Publication: 2026-07-20 11:40:02 UTC

# Imports from base packages
importFrom(grDevices, rainbow, dev.off)
importFrom(stats,
           mad, model.matrix, terms, setNames, qbeta, optim, pf, xtabs,
           as.formula, summary.lm, pbeta, approx, poly, na.omit,
           complete.cases, predict, predict.lm, t.test, formula, ave,
           aggregate, median, sd, qnorm, rnorm, dnorm, pnorm, lm, .lm.fit,
           runif, optimize, cor, quantile, coef)
importFrom(utils, menu, head, tail)

importFrom(parallel, clusterCall)
importFrom(parallel, clusterExport)
importFrom(parallel, detectCores)
importFrom(parallel, makeCluster)
importFrom(parallel, parLapply)
importFrom(parallel, stopCluster)
importFrom(utils, flush.console)

# Imports from suggested packages (with checks in code)
importFrom(leaps, regsubsets)

# Imports from ggplot2
importFrom(ggplot2,
           .data, aes, annotate, element_blank, element_line, element_rect,
           element_text, facet_wrap, geom_abline, geom_contour,
           geom_contour_filled, geom_hline, geom_line, geom_point, geom_label,
           geom_smooth, geom_text, geom_tile, geom_vline, ggplot, ggsave,
           guide_legend, guides, labs, margin, scale_color_brewer,
           scale_color_manual, scale_color_viridis_d, scale_fill_gradientn,
           scale_shape_manual, scale_x_continuous, scale_y_continuous,
           scale_y_log10, theme, theme_minimal, unit, ylim, coord_cartesian,
           geom_step)

# Exported functions
export(
    autoselect.betabinomial,
    autoselect.shash,
    bestModel,
    betaCoefficients,
    buildCnormObject,
    checkConsistency,
    cNORM.GUI,
    cNORM.GUI2,
    cnorm,
    cnorm.betabinomial,
    cnorm.cv,
    cnorm.shash,
    compare,
    computePowers,
    computeWeights,
    derivationTable,
    derive,
    diagnostics.betabinomial,
    diagnostics.shash,
    dshash,
    getGroups,
    getNormCurve,
    getNormScoreSE,
    modelSummary,
    normTable,
    normTable.betabinomial,
    normTable.shash,
    plotCnorm,
    plotDensity,
    plotDerivative,
    plotNorm,
    plotNormCurves,
    plotPercentiles,
    plotPercentileSeries,
    plotRaw,
    plotSubset,
    predictMoments,
    predictNorm,
    predictRaw,
    prepareData,
    printSubset,
    pshash,
    qshash,
    rangeCheck,
    rankByGroup,
    rankBySlidingWindow,
    rawTable,
    regressionFunction,
    rshash,
    simulateRasch,
    standardize,
    taylorSwift,
    weightedAverageModel,
    weighted.quantile,
    weighted.quantile.harrell.davis,
    weighted.quantile.inflation,
    weighted.quantile.type7,
    weighted.rank
)

# S3 methods for cnorm class
S3method(plot, cnorm)
S3method(print, cnorm)
S3method(summary, cnorm)

# S3 methods for cnormBetaBinomial class
S3method(plot, cnormBetaBinomial)
S3method(predict, cnormBetaBinomial)
S3method(summary, cnormBetaBinomial)

# S3 methods for cnormBetaBinomial2 class
S3method(plot, cnormBetaBinomial2)
S3method(predict, cnormBetaBinomial2)
S3method(summary, cnormBetaBinomial2)

# S3 methods for cnormShash class
S3method(plot, cnormShash)
S3method(predict, cnormShash)
S3method(summary, cnormShash)

# S3 methods to predict moments
S3method(predictMoments, cnorm)
S3method(predictMoments, cnormBetaBinomial)
S3method(predictMoments, cnormBetaBinomial2)
S3method(predictMoments, cnormShash)
S3method(predictMoments, default)

<!-- badges: start -->
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/cNORM)](https://cran.r-project.org/package=cNORM)
[![R-CMD-check](https://github.com/WLenhard/cNORM/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/WLenhard/cNORM/actions/workflows/R-CMD-check.yaml)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/cNORM)](https://cran.r-project.org/package=cNORM)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/grand-total/cNORM?color=blue)](https://r-pkg.org/pkg/cNORM)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
<!-- badges: end -->
<img src="vignettes/logo.png" align=right style="border:0;">

# cNORM

 cNORM (A. Lenhard, W. Lenhard & S. Gary) is a package for the R environment for statistical computing that aims at generating continuous test norms in psychometrics and biometrics and to analyze the model fit. Originally, cNorm exclusively used an approach that makes no assumptions about the specific distribution of the raw data (A. Lenhard, Lenhard, Suggate & Segerer, 2016). Since version 3.2 (2024), however, the package also offers the option of parametric modeling using the beta-binomial distribution and since version 3.5 modeling with the Sinh-Arcsinh (ShaSh) distribution.

cNORM was developed specifically for achievement tests (e.g. vocabulary development: A. Lenhard, Lenhard, Segerer & Suggate, 2015; written language acquisition: W. Lenhard, Lenhard & Schneider, 2017). However, the package can be used wherever mental (e.g. reaction time), physical (e.g. body weight) or other test scores depend on continuous (e.g. age, duration of schooling) or discrete explanatory variables (e.g. sex, test form). In addition, the package can also be used for "conventional" norming based on individual groups, i.e. without including explanatory variables.

The package estimates percentiles as a function of the explanatory variable. This is done either parametrically on the basis of the beta-binomial or the Sinh-Arcsinh (ShaSh) distribution or distribution-free using Taylor polynomials. For an in-depth tutorial, visit the [project homepage](https://www.psychometrica.de/cNorm_en.html), try the [online demonstration](https://cnorm.shinyapps.io/cNORM/) and have a look at the vignettes.


## In a nutshell

A quick guide to distribution-free modeling with the essential cNORM functions:
```{r example}
## Basic example code for modeling the sample dataset
library(cNORM)

# Start the graphical user interface (needs shiny installed)
# The GUI includes the most important functions. For specific cases,
# please use cNORM on the console.
cNORM.GUI()  // distribution-free modelling with Taylor polynomials
cNORM.GUI2() // parametric modelling with beta-binomial or the Sinh-Arcsinh (ShaSh) distribution

# Using the syntax on the console: The function 'cnorm' performs
# all steps automatically. Please specify the raw score and the
# grouping variable. The resulting object contains the ranked data
# via object$data and the model via object$model.
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)

# ... and since we love pop music as much as Taylor polynomials, you
# can also use the taylorSwift function to swiftly compute a distribution-
# free Taylor polynomial model (which is however identical to cnorm). Here
# with the sample dataset 'ppvt':
model <- taylorSwift(ppvt$raw, ppvt$group)

# Plot different indicators of model fit depending on the number of
# predictors
plot(cnorm.elfe, "subset", type=0) # plot R2
plot(cnorm.elfe, "subset", type=3) # plot MSE

# NOTE! At this point, you usually select a good fitting model and rerun
# the process with a fixed number of terms, e. g. 4:
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group, terms = 4)

# Powers of age can be specified via the parameter 't'.
# Cubic modeling is usually sufficient, i.e., t = 3.
# In contrast, 'k' specifies the power of the person location.
# This parameter should be somewhat higher, e.g., k = 5.
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group, k = 5, t = 3)

# Visual inspection of the percentile curves of the fitted model
plot(cnorm.elfe, "percentiles")

# Visual inspection of the observed and fitted raw and norm scores
plot(cnorm.elfe, "norm")
plot(cnorm.elfe, "raw")

# In order to compare different models, generate a series of percentile
# plots with an ascending number of predictors, in this example between
# 5 and 14 predictors.
plot(cnorm.elfe, "series", start=5, end=14)

# Cross validation of number of terms with 80% of the data for training
# and 20% for validation. Due to the time intensity, the maximum
# number of terms is restricted to 10 in this example
# with 3 repetitions.
cnorm.cv(cnorm.elfe$data, max=10, repetitions=3)

# Cross validation with prespecified terms of an already
# existing model
cnorm.cv(cnorm.elfe, repetitions=3)

# Print norm table (in this case: 0, 3 or 6 months at grade level 3)
normTable(c(3, 3.25, 3.5), cnorm.elfe)

# The other way round: Print raw table (grade level 3; 0 months) together
# with 90% confidence intervalls for a test with a reliability of .94
rawTable(3, cnorm.elfe, CI = .9, reliability = .94)
```


Modelling norm data using beta-binomial distributions:
```{r example}
library(cNORM)
# cNORM can as well model norm data using the beta-binomial
# distribution, which usually performs well on tests with
# a fixed number of dichotomous items without time cutoff.
# Ideal use case: 1PL IRT scale / Rasch modelling
# The example uses the inbuilt ELFE demo dataset (reading comprehension test
# for elementary school with grade variable; 28 items).
model.betabinomial <- cnorm.betabinomial(elfe$group, elfe$raw)

# Adapt the power parameters for α and β to increase or decrease
# the fit:
model.betabinomial <- cnorm.betabinomial(elfe$group, elfe$raw, alpha = 4)

# Automatic grid search to determine the model with the lowest BIC
model.betabinomial <- autoselect.betabinomial(elfe$group, elfe$raw)


# Plot percentile curves and display manifest and modelled norm scores.
plot(model.betabinomial, elfe$group, elfe$raw)
plotNorm(model.betabinomial, elfe$group, elfe$raw, width = 1)

# Display fit statistics:
summary(model.betabinomial)

# Prediction of norm scores for new data
predict(model.betabinomial, c(2.0, 2.2, 2.4, 2.6), c(10, 15, 13, 22))

# generate norm tables
tables <- normTable.betabinomial(model.betabinomial, c(2, 3, 4),
                                 reliability=0.9)
```

Modelling norm data using Sinh-Arcsinh (ShaSh) distributions:
```{r example}
library(cNORM)
# The Sinh-Arcsinh (ShaSh) distribution is a flexible approach.
# It can handle raw score value ranges including zeros and negative
# values, which pose a problem to Box Cox distributions.
# Shape parameters mu, sigma, epsilon and delta can be adjusted as well.
# The following example uses the inbuilt PPVT4 demo dataset for receptive
# vocabulary development from 3 to 18.
model.shash <- cnorm.shash(ppvt$age, ppvt$raw)

# Automatic grid search to determine the model with the lowest BIC
model.shash <- autoselect.shash(ppvt$age, ppvt$raw)

# Plot percentile curves and display manifest and modelled norm scores.
plot(model.shash, ppvt$age, ppvt$raw)

# Display fit statistics:
summary(model.shash, ppvt$age, ppvt$raw)

# Prediction of norm scores for new data and generating norm tables
predict(model.shash, c(8.9, 10.1), c(153, 121))
tables <- normTable.shash(model.shash, c(10, 15),
                                 reliability=0.9)
```

Conventional norming:
```{r example}
library(cNORM)

# cNORM can as well be used for conventional norming:
cnorm(raw=elfe$raw)
```


Start vignettes in cNORM:
```{r example}
library(cNORM)

vignette("cNORM-Demo", package = "cNORM")
vignette("WeightedRegression", package = "cNORM")
vignette("BetaBinomial", package = "cNORM")
vignette("sinh", package = "cNORM")
```



## Sample Data
The package includes data from two large test norming projects, namely ELFE 1-6 (Lenhard & Schneider, 2006) and German adaption of the PPVT4 (A. Lenhard, Lenhard, Suggate & Seegerer, 2015), which can be used to run the analysis. Furthermore, large samples from the Center of Disease Control (CDC) on growth curves in childhood and adolescence (for computing Body Mass Index 'BMI' curves), Type `?elfe`, `?ppvt` or `?CDC` to display information on the data sets.

## Terms of use, license and declaration of interest
cNORM is licensed under GNU Affero General Public License v3 (AGPL-3.0). This means that copyrighted parts of cNORM can be used free of charge for commercial and non-commercial purposes that run under this same license, retain the copyright notice, provide their source code and correctly cite cNORM. Copyright protection includes, for example, the reproduction and distribution of source code or parts of the source code of cNORM. The integration of the package into a server environment in order to access the functionality of the software (e.g. for online delivery of norm scores) is also subject to this license. However, a regression function determined with cNORM, the norm tables and plots are not subject to copyright protection and may be used freely without preconditions. If you want to apply cNORM in a way that is not compatible with the terms of the AGPL 3.0 license, please do not hesitate to contact us to negotiate individual conditions. If you want to use cNORM for scientific publications, we would also ask you to quote the source.

The authors would like to thank WPS (<https://www.wpspublish.com/>) for providing funding for developing, integrating and evaluating weighting and post stratification in the cNORM package. The research project was conducted in 2022. 

## References

*   Gary, S., Lenhard, W., Lenhard, A. et al. A tutorial on automatic post-stratification and weighting in conventional and regression-based norming of psychometric tests. Behav Res (2023a). https://doi.org/10.3758/s13428-023-02207-0
*   Gary, S., Lenhard, A., Lenhard, W., & Herzberg, D. S. (2023b). Reducing the bias of norm scores in non-representative samples: Weighting as an adjunct to continuous norming methods. Assessment, 10731911231153832. https://doi.org/10.1177/10731911231153832
*   Lenhard, A., Lenhard, W., Segerer, R. & Suggate, S. (2015). Peabody Picture Vocabulary Test - Revision IV (Deutsche Adaption). Frankfurt a. M./Germany: Pearson Assessment.
*   Lenhard, A., Lenhard, W., Suggate, S. & Segerer, R. (2016). A continuous solution to the norming problem. Assessment, Online first, 1-14. https://doi.org/10.1177/1073191116656437
*   Lenhard, A., Lenhard, W., Gary, S. (2018). Continuous Norming (cNORM). The Comprehensive R Network, Package cNORM, available: https://CRAN.R-project.org/package=cNORM
*   Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. PLoS ONE, 14(9),  e0222279. https://doi.org/10.1371/journal.pone.0222279
*   Lenhard, W., & Lenhard, A. (2020). Improvement of Norm Score Quality via Regression-Based Continuous Norming. Educational and Psychological Measurement(Online First), 1-33. https://doi.org/10.1177/0013164420928457

## ----setup, include = FALSE-------------------------------
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
library(cNORM)
library(ggplot2)


## ----echo=FALSE, fig.align='center', fig.width=10---------
knitr::include_graphics("plos.png")

## ----eval=FALSE-------------------------------------------
#  ## Basic example code for modeling the sample dataset
# 
# library(cNORM)
# 
# # Start the graphical user interface (needs shiny installed)
# # The GUI includes the most important functions. For specific cases,
# # please use cNORM on the console.
# 
# cNORM.GUI()
# 
# # Using the syntax on the console: The function 'cnorm' performs
# # all steps automatically. Please specify the raw score and the
# # grouping variable. The resulting object contains the ranked data
# # via object$data and the model via object$model.
# 
# cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
# 
# # Plot different indicators of model fit depending on the number of
# # predictors
# 
# plot(cnorm.elfe, "subset", type=0) # plot R2
# plot(cnorm.elfe, "subset", type=3) # plot MSE
# 
# # NOTE! At this point, you usually select a good fitting model and rerun
# # the process with a fixed number of terms, e. g. 4. Avoid models
# # with a high number of terms:
# 
# cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group, terms = 4)
# 
# # Powers of age can be specified via the parameter 't'.
# # Cubic modeling is usually sufficient, i.e., t = 3.
# # In contrast, 'k' specifies the power of the person location.
# # This parameter should be somewhat higher, e.g., k = 5.
# 
# cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group, k = 5, t = 3)
# 
# # Visual inspection of the percentile curves of the fitted model
# 
# plot(cnorm.elfe, "percentiles")
# 
# # Visual inspection of the observed and fitted raw and norm scores
# 
# plot(cnorm.elfe, "norm")
# plot(cnorm.elfe, "raw")
# 
# # In order to compare different models, generate a series of percentile
# # plots with an ascending number of predictors, in this example between
# # 5 and 14 predictors.
# 
# plot(cnorm.elfe, "series", start=5, end=14)
# 
# # Cross validation in order to choose appropriate number of terms
# # with 80% of the data for training and 20% for validation. Due to
# # the time consumption, the maximum number of terms is limited to 10
# # in this example with 3 repetitions of the cross validation.
# 
# cnorm.cv(cnorm.elfe$data, max=10, repetitions=3)
# 
# # Cross validation with prespecified terms of an already
# # existing model
# 
# cnorm.cv(cnorm.elfe, repetitions=3)
# 
# # Print norm table (in this case: 0, 3 or 6 months at grade level 3)
# # (Note: The data is coded such that 3.0 represents the beginning and
# # 3.5 the middle of the third school year)
# 
# normTable(c(3, 3.25, 3.5), cnorm.elfe)
# 

## ----eval=FALSE-------------------------------------------
# # Application of cNORM for the generation of conventional norms
# # for a specific age group (in this case age group 3):
# 
# data <- elfe[elfe$group == 3,]
# cnorm(raw=data$raw)

## ----eval=FALSE-------------------------------------------
#     # Creates a grouping variable for the variable 'age'
#     # of the ppvt data set. In this example, 12 equidistant
#     # subgroups are generated.
# 
#     group <- getGroups(ppvt$age, 12)

## ----fig0, fig.height = 4, fig.width = 7------------------
library(cNORM)
# Models the 'raw' variable as a function of the discrete 'group' variable
model <- cnorm(raw = elfe$raw, group = elfe$group)

## ---------------------------------------------------------
print(model)

## ----eval=FALSE-------------------------------------------
# # Generates a series of percentile plots with increasing number of predictors
# 
# plotPercentileSeries(model, start = 1, end = 15)

## ----fig1, fig.height = 4, fig.width = 7------------------
    plot(model, "subset", type = 0)

## ---------------------------------------------------------
predictNorm(15, 4.7, model, minNorm = 25, maxNorm = 75)


## ---------------------------------------------------------
predictRaw(55, 4.5, model, minRaw = 0, maxRaw = 28)

# ... or for several norm scores and age levels ...
predictRaw(c(45, 50, 55), c(2.5, 3, 3.5), model) 

## ---------------------------------------------------------
normTable(3, model, minRaw = 0, maxRaw = 28, minNorm=30.5, maxNorm=69.5, step = 1)

## ---------------------------------------------------------
rawTable(3.5, model, minRaw = 0, maxRaw = 28, minNorm = 25, maxNorm = 75, step = 1, CI = .95, reliability = .89)

# generate several raw tables
table <- rawTable(c(2.5, 3.5, 4.5), model, minRaw = 0, maxRaw = 28)

## ----fig2, fig.height = 7, fig.width = 7------------------
plot(model, "raw", group = TRUE)

## ----fig3, fig.height = 7, fig.width = 7------------------
plot(model, "norm", group = TRUE, minNorm = 25, maxNorm = 75)

## ----fig4, fig.height = 4, fig.width = 7------------------
plot(model, "density", group = c (2, 3, 4))

## ----fig5, fig.height = 4, fig.width = 7------------------
plot(model, "derivative")

library(shiny)
library(cNORM)
library(DT)
library(ggplot2)

shinyServer(function(input, output, session) {

  # Reactive values to store data and model
  rv <- reactiveValues(
    data = NULL,
    model = NULL,
    normTables = NULL
  )

  # ============================================================================
  # DATA INPUT TAB
  # ============================================================================

  # Load data based on selection
  observeEvent(input$Example, {
    req(input$Example)

    if (input$Example == "elfe") {
      rv$data <- elfe
    } else if (input$Example == "ppvt") {
      rv$data <- ppvt
    } else if (input$Example == "CDC") {
      rv$data <- CDC
    }

    # Invalidate model when new data is loaded
    rv$model <- NULL
    rv$normTables <- NULL
  })

  # Load data from file
  observeEvent(input$file, {
    req(input$file)

    tryCatch({
      ext <- tools::file_ext(input$file$name)

      if (ext == "csv") {
        rv$data <- read.csv(input$file$datapath)
      } else if (ext %in% c("xlsx", "xls")) {
        rv$data <- readxl::read_excel(input$file$datapath)
      } else if (ext == "rda" || ext == "RData") {
        load(input$file$datapath)
        rv$data <- get(ls()[1])
      } else if (ext == "sav") {
        rv$data <- haven::read_sav(input$file$datapath)
      }

      # Invalidate model when new data is loaded
      rv$model <- NULL
      rv$normTables <- NULL

      showNotification("Data loaded successfully!", type = "message")
    }, error = function(e) {
      showNotification(paste("Error loading file:", e$message), type = "error")
    })
  })

  # Display data preview
  output$dataPreview <- DT::renderDT({
    req(rv$data)

    DT::datatable(
      rv$data,
      options = list(
        pageLength = 15,
        scrollX = TRUE,
        dom = 'Bfrtip'
      ),
      rownames = FALSE
    )
  })

  # ============================================================================
  # MODELING TAB - Dynamic UI
  # ============================================================================

  # Age variable selector
  output$ageVariable <- renderUI({
    req(rv$data)

    selectInput(
      "age_var",
      "Age Variable:",
      choices = names(rv$data),
      selected = names(rv$data)[1]
    )
  })

  # Score variable selector
  output$scoreVariable <- renderUI({
    req(rv$data)

    selectInput(
      "score_var",
      "Raw Score Variable:",
      choices = names(rv$data),
      selected = if(ncol(rv$data) > 1) names(rv$data)[2] else names(rv$data)[1]
    )
  })

  # Weight variable selector
  output$weightVariable <- renderUI({
    req(rv$data)

    selectInput(
      "weight_var",
      "Weight Variable (optional):",
      choices = c("None" = "", names(rv$data)),
      selected = ""
    )
  })

  # ============================================================================
  # MODELING TAB - Compute Model
  # ============================================================================

  observeEvent(input$computeModel, {
    req(rv$data, input$age_var, input$score_var)

    # Show progress
    withProgress(message = 'Fitting model...', value = 0, {

      tryCatch({
        # Extract variables
        age <- rv$data[[input$age_var]]
        score <- rv$data[[input$score_var]]

        # Handle weights
        weights <- NULL
        if (!is.null(input$weight_var) && input$weight_var != "") {
          weights <- rv$data[[input$weight_var]]
        }

        incProgress(0.2, detail = "Preparing data...")

        # Fit model based on distribution type
        if (input$distributionType == "shash") {

          # Determine delta parameter
          delta_deg <- if (input$fix_delta) NULL else input$delta_degree
          delta_val <- if (input$fix_delta) input$delta_value else 1

          incProgress(0.3, detail = "Fitting SHASH model...")

          rv$model <- cnorm.shash(
            age = age,
            score = score,
            weights = weights,
            mu_degree = input$mu_degree,
            sigma_degree = input$sigma_degree,
            epsilon_degree = input$epsilon_degree,
            delta_degree = delta_deg,
            delta = delta_val,
            scale = input$scale,
            plot = FALSE
          )

        } else if (input$distributionType == "betabinomial") {

          incProgress(0.3, detail = "Fitting Beta-Binomial model...")

          rv$model <- cnorm.betabinomial(
            age = age,
            score = score,
            weights = weights,
            alpha = input$alpha_degree,
            beta = input$beta_degree,
            scale = input$scale,
            plot = FALSE
          )
        }

        incProgress(1, detail = "Model fitted!")

        showNotification("Model computed successfully!", type = "message")

      }, error = function(e) {
        showNotification(
          paste("Error fitting model:", e$message),
          type = "error",
          duration = 10
        )
      })
    })
  })

  # ============================================================================
  # MODELING TAB - Outputs
  # ============================================================================

  # Model summary
  output$modelSummary <- renderPrint({
    req(rv$model)
    summary(rv$model, age = rv$data[[input$age_var]], score = rv$data[[input$score_var]])
  })

  # Percentile plot
  output$percentilePlot <- renderPlot({
    req(rv$model, rv$data, input$age_var, input$score_var)

    weights <- NULL
    if (!is.null(input$weight_var) && input$weight_var != "") {
      weights <- rv$data[[input$weight_var]]
    }

    plot(rv$model,
         age = rv$data[[input$age_var]],
         score = rv$data[[input$score_var]],
         weights = weights,
         points = TRUE)
  })

  # Model diagnostics
  output$modelDiagnostics <- renderPrint({
    req(rv$model)

    if (input$distributionType == "shash") {
      diag <- diagnostics.shash(rv$model,
                                age = rv$data[[input$age_var]],
                                score = rv$data[[input$score_var]])
    } else {
      diag <- diagnostics.betabinomial(rv$model,
                                       age = rv$data[[input$age_var]],
                                       score = rv$data[[input$score_var]])
    }

    print(diag)
  })

  # Derivative plot (SHASH only)
  output$derivativePlot <- renderPlot({
    req(rv$model, input$distributionType == "shash")

    plotDerivative(rv$model, age = rv$data[[input$age_var]])
  })

  # Download model
  output$downloadModel <- downloadHandler(
    filename = function() {
      paste0("cnorm_model_", input$distributionType, "_", Sys.Date(), ".rds")
    },
    content = function(file) {
      saveRDS(rv$model, file)
    }
  )

  # ============================================================================
  # NORM TABLES TAB
  # ============================================================================

  # Auto-populate score range when data/score variable changes
  observe({
    req(rv$data, input$score_var)

    score <- rv$data[[input$score_var]]

    updateNumericInput(
      session,
      "norm_start",
      value = floor(min(score, na.rm = TRUE))
    )

    updateNumericInput(
      session,
      "norm_end",
      value = ceiling(max(score, na.rm = TRUE))
    )
  })

  observeEvent(input$generateTables, {
    req(rv$model, input$norm_ages)

    withProgress(message = 'Generating norm tables...', value = 0, {

      tryCatch({
        # Parse ages
        ages_text <- gsub(",", " ", input$norm_ages)
        ages <- as.numeric(unlist(strsplit(ages_text, "\\s+")))
        ages <- ages[!is.na(ages)]

        if (length(ages) == 0) {
          showNotification("Please enter valid ages", type = "error")
          return(NULL)
        }

        incProgress(0.3, detail = "Computing percentiles...")

        # Set score range
        start_score <- input$norm_start
        end_score <- input$norm_end

        if (is.null(start_score)) start_score <- attr(rv$model$result, "min")
        if (is.null(end_score)) end_score <- attr(rv$model$result, "max")

        # Generate tables
        if (input$distributionType == "shash") {

          ci_val <- if (input$include_ci) input$ci_level else NULL
          rel_val <- if (input$include_ci) input$reliability else NULL

          rv$normTables <- normTable.shash(
            rv$model,
            ages = ages,
            start = start_score,
            end = end_score,
            step = input$norm_step,
            CI = ci_val,
            reliability = rel_val
          )

        } else if (input$distributionType == "betabinomial") {

          ci_val <- if (input$include_ci) input$ci_level else NULL
          rel_val <- if (input$include_ci) input$reliability else NULL

          rv$normTables <- normTable.betabinomial(
            rv$model,
            ages = ages,
            start = start_score,
            end = end_score,
            step = input$norm_step,
            CI = ci_val,
            reliability = rel_val
          )
        }

        incProgress(1, detail = "Done!")

        showNotification("Norm tables generated!", type = "message")

      }, error = function(e) {
        showNotification(
          paste("Error generating tables:", e$message),
          type = "error",
          duration = 10
        )
      })
    })
  })

  # Display norm tables
  output$normTablesUI <- renderUI({
    req(rv$normTables)

    table_tabs <- lapply(names(rv$normTables), function(age_name) {
      tabPanel(
        paste("Age", age_name),
        tags$br(),
        DT::DTOutput(paste0("normTable_", age_name))
      )
    })

    do.call(tabsetPanel, table_tabs)
  })

  # Render individual norm tables
  # Render individual norm tables
  observe({
    req(rv$normTables)

    lapply(names(rv$normTables), function(age_name) {
      output[[paste0("normTable_", age_name)]] <- DT::renderDT({

        # Identify columns to round
        cols_to_round <- c("Px", "Pcum", "Percentile", "z", "norm")

        # Add CI columns if they exist
        ci_cols <- c("lowerCI", "upperCI", "lowerCI_PR", "upperCI_PR")
        existing_ci_cols <- ci_cols[ci_cols %in% names(rv$normTables[[age_name]])]
        cols_to_round <- c(cols_to_round, existing_ci_cols)

        DT::datatable(
          rv$normTables[[age_name]],
          options = list(
            pageLength = 25,
            scrollX = TRUE,
            dom = 'Bfrtip'
          ),
          rownames = FALSE
        ) %>%
          DT::formatRound(columns = cols_to_round, digits = 2)
      })
    })
  })

  # Download norm tables
  output$downloadTables <- downloadHandler(
    filename = function() {
      paste0("norm_tables_", input$distributionType, "_", Sys.Date(), ".csv")
    },
    content = function(file) {
      req(rv$normTables)

      # Combine all tables with age column
      combined <- do.call(rbind, lapply(names(rv$normTables), function(age_name) {
        df <- rv$normTables[[age_name]]
        df$age <- as.numeric(age_name)
        df
      }))

      write.csv(combined, file, row.names = FALSE)
    }
  )
})

if(!require(shiny)){
  install.packages("shiny")
  require(shiny)
}

if(!require(markdown)){
  install.packages("markdown")
  require(markdown)
}

if(!require(foreign)){
  install.packages("foreign")
  require(foreign)
}

if(!require(readxl)){
  install.packages("readxl")
  require(readxl)
}

# Define server logic required for cNORM-Application
shinyServer(function(input, output, session) {

  # Stops program by closing browser tab
  session$onSessionEnded(function() {
    stopApp()
  })


  # Returns currently chosen file; returns null,if no file was chosen
  currentFile <- reactive({

    inFile <- input$file
    exData <- input$Example

    # Check for input file; if no file is chosen, check for chosen example data
    if(is.null(input$file)){

      if(is.null(exData)){
        return(NULL)
      }
      if(!is.null(exData)){

        if(exData == ""){
          return(NULL)
        }
        if(exData == "elfe"){
          currentData <- as.data.frame(elfe)
          return(currentData)
        }
        if(exData == "ppvt"){
          currentData <- as.data.frame(ppvt)
          return(currentData)
        }
        if(exData == "CDC"){
          currentData <- as.data.frame(CDC)
          return(currentData)
        }
      }
    }

    # If a file other than the examples is chosen, the file will be loaded depending on the file type
    currentData <- NULL
    currentFilePath <- input$file$datapath
    fileEnding <- strsplit(currentFilePath, ".", fixed = TRUE)[[1]][2]

    if(fileEnding == "csv"){
      currentData <- as.data.frame(read.csv2(currentFilePath))
    }
    if(fileEnding == "sav"){
      currentData <- as.data.frame(foreign::read.spss(currentFilePath))
    }
    if(fileEnding == "xlsx"){
      currentData <- as.data.frame(readxl::read_xlsx(currentFilePath))
    }
    if (fileEnding == "xls"){
      currentData <- as.data.frame(readxl::read_xls(currentFilePath))
    }
    if(fileEnding == "rda"){

      load_first_object <- function(fname){
        e <- new.env(parent = parent.frame())
        load(fname, e)
        return(e[[ls(e)[1]]])
      }
      currentData <- load_first_object(currentFilePath)

    }
    return(currentData)
  })

  # Returns column names of chosen data file
  variableNames <- reactive({

    if (is.null(currentFile())) {
      return()
    }
    return(names(currentFile()))

  })

  # Generates selectInput for choosing grouping variable for data analysis
  output$GroupingVariable <- renderUI({

    currentNames <- names(currentFile())
    if(is.element("group", currentNames)|| is.element("Group", currentNames))
    {
      if(is.element("group", currentNames)){
        selectInput(inputId = "InputGroupingVariable", "Grouping Variable", choices = variableNames(), selected = "group")
      }
      else{
        selectInput(inputId = "InputGroupingVariable", "Grouping Variable", choices = variableNames(), selected = "group")

      }
    }
    else{    selectInput(inputId = "InputGroupingVariable", "Grouping Variable", choices = variableNames(), selected = NULL)
    }
  })

  # Generates selectINput for choosing explanatory variable for data analysis
  output$ExplanatoryVariable <- renderUI({

    selectInput(inputId = "InputExplanatoryVariable", "Explanatory variable", choices = variableNames(), selected = input$InputGroupingVariable)
  })

  # Generates selectINput for choosing explanatory variable for data analysis
  output$WeightVariable <- renderUI({

    selectInput(inputId = "InputWeightVariable", "Weighting variable", choices = c("---", variableNames()))
  })

  # Generates selectInput for choosing raw values for data analysis
  output$RawValues <- renderUI({

    currentNames <- names(currentFile())
    if(is.element("raw", currentNames)||is.element("Raw", currentNames)){
      if(is.element("raw", currentNames)){
        selectInput(inputId = "InputRawValues", "Raw Score Variable", choices = variableNames(), selected = "raw")
      }
      else{
        selectInput(inputId = "InputRawValues", "Raw Score Variable", choices = variableNames(), selected = "Raw")

      }
    }
    else{
      selectInput(inputId = "InputRawValues", "Raw Score Variable", choices = variableNames(), selected = NULL)
    }

  })

  # Shows introduction text until a file was chosen
  output$introduction <- renderText({

    if(is.null(currentFile())){

      output <- readChar("www/introduction.html", file.info("www/introduction.html")$size)
      #output <- textreadr::read_html(file = "www/introduction.html")
      suppressWarnings(includeHTML("www/introduction.html"))
    }
    else{
      return(NULL)
    }



  })

  # Shows the current data file
  output$table <- renderDataTable({
    if(is.null(currentFile())){ return(NULL)}
    else{return(currentFile())}

  })

  # Updating chosen variables for group, raw, explanatory and powers

  # Retruns chosen variable for grouping
  chosenGrouping <- reactive({
    return(input$InputGroupingVariable)
  })

  # Returns chosen raw value
  chosenRaw <- reactive({
    return(input$InputRawValues)
  })

  # Returns chosen weighting
  chosenWeighting <- reactive({
    return(input$InputWeightVariable)
  })

  chosenDescend <- reactive({
    if(input$RankingOrder == "Descending")
      return(TRUE)
    else
      return(FALSE)
  })

  # Returns chosen explanatory variable
  chosenExplanatory <- reactive({

    return(input$InputExplanatoryVariable)
  })

  chosenMethod <- reactive({
    methods <- c("Blom (1985)", "Tukey (1949)", "Van der Warden (1952)", "Rankit (Bliss, 1967)", "Levenbach (1953)", "Filliben (1975)", "Yu & Huang (2001)")
    return(match(input$Method, methods))
  })

  chosenScale <- reactive({
    return(toString(input$Scale))
  })

  chosenNumberOfPowers <- reactive({
    return(as.numeric(input$NumberOfPowers))
  })

  chosenNumberOfPowersAge <- reactive({
    return(as.numeric(input$NumberOfPowersAge))
  })

  # Calculates prepared data by pressing action button
  preparedData <- eventReactive(input$DoDataPreparation, {
    # Data preperation for data analysis with cNORM
    # Returns prepared data, which can be used for calculating best model
    preparedData <- reactive({

      if(is.null(currentFile())){
        print("CurrentFile is null")
        return()}

      if(is.null(chosenGrouping())){
        print("GroupingVariable is null")
        return()}

      # Ranky by chosen group
      weights <- NULL
      if(chosenWeighting()!="---")
        weights <- chosenWeighting()

      print(weights)
      data_to_output <- cNORM::rankByGroup(currentFile(),
                                           group = chosenGrouping(),
                                           raw = chosenRaw(),
                                           method = chosenMethod(),
                                           scale = chosenScale(),
                                           descend = chosenDescend(),
                                           weights = weights)
      # Computation of powers and linear combinations
      data_to_output <- cNORM::computePowers(data_to_output,
                                             k = chosenNumberOfPowers(),
                                             t = chosenNumberOfPowersAge(),
                                             age = chosenGrouping(),
                                             norm = "normValue")

      return(data_to_output)
    })


    return(preparedData())
  })

  # Shows prepared data for data analysis with cNORM
  output$preparedData <- renderDataTable({
    if(is.null(preparedData())){return()}

    return(preparedData())
  })

  output$downloadData <- downloadHandler(
    filename = function() {
      paste(input$dataset, "data.RData", sep = "")
    },
    content = function(file) {
      data.cnorm <- preparedData()
      attr(data.cnorm, "descend") <- chosenDescend()

      save(data.cnorm, file = file)
    }
  )

  output$downloadModel <- downloadHandler(
    filename = function() {
      paste(input$dataset, "model.RData", sep = "")
    },
    content = function(file) {
      model.cnorm <- bestModel()
      save(model.cnorm, file = file)
    }
  )



  output$NumberOfTerms <- renderUI({

    k <- as.numeric(chosenNumberOfPowers())
    terms_max <- k*k + 2*k
    numericInput("InputNumberOfTerms", "Number of terms", value = NULL, min = 1, max = terms_max, step = 1)
  })

  chosenCoeffOfDet <- reactive({

    return(as.numeric(input$ChosenDetCoeff))
  })

  chosenNumberOfTerms <- reactive({

    if(!is.null(input$InputNumberOfTerms)){
      return(as.numeric(input$InputNumberOfTerms))

    }
    else{
      return(0)
    }
  })

  # Calculates best model using cNORM
  bestModel <- eventReactive(input$CalcBestModel, {

    # CAUTION: The value of a non-specified numericInput-value is NA and not NULL!
    # Therefore use is.na() to check, whether a specific number of terms were chosen or not
    if(!is.na(chosenNumberOfTerms()))
    {
      currentBestModel <- cNORM::bestModel(data = preparedData(),
                                           raw = chosenRaw(),
                                           k = chosenNumberOfPowers(),
                                           predictors = NULL,
                                           terms = chosenNumberOfTerms(),
                                           plot=FALSE)
    }
    else{
      currentBestModel <- cNORM::bestModel(data = preparedData(),
                                           raw = chosenRaw(),
                                           k = chosenNumberOfPowers(),
                                           predictors = NULL,
                                           R2 = chosenCoeffOfDet(),
                                           terms = 0,
                                           plot=FALSE)
    }



    return(currentBestModel)
  })

  modelDerivatives <- eventReactive(input$CalcBestModel, {
    cNORM::plotDerivative(bestModel())

  })


  modelDensity <- reactive({
    if(input$densities == ""){
      cNORM::plotDensity(bestModel())
    }
    else{
      densityList <- as.numeric(unlist(strsplit(input$densities, "\\, |\\,| ")))
      cNORM::plotDensity(bestModel(), group = densityList)
    }
  })



  chosenTypeOfPlotSubset <- reactive({
    method <- c("Adjusted R2 by Number of Predictors", "Log Transformed Mallow's Cp by Adjusted R2", "Bayesian Information Criterion (BIC) by Adjusted R2", "RMSE by Number of Predictors")
    return(match(input$chosenTypePlotSubset, method) - 1)
  })

  changeObject <- reactive({
    paste(input$CalcBestModel, input$chosenTypePlotSubset)
  })


  wlPlot <- eventReactive(changeObject(), {
    cNORM::plotSubset(bestModel(), type = chosenTypeOfPlotSubset())
  })

  # Generates and prints norm curves
  output$NormCurves <- renderPlot({
    return(normCurves())
  })

  # Generates and plots percentile curves
  output$modelPlot <- renderPlot({

    MIN_AGE <- min((currentFile())[chosenGrouping()])
    MAX_AGE <- max((currentFile())[chosenGrouping()])

    # data <- currentFile()
    # attr(data, "descend") <- chosenDescend()

    cNORM::plotPercentiles(buildCnormObject(preparedData(), bestModel()), raw = chosenRaw(),
                             group = chosenGrouping(),
                             minAge = MIN_AGE,
                             maxAge = MAX_AGE)

  })

  valuesPlot <- eventReactive(input$CalcBestModel, {

    cNORM::plotRaw(buildCnormObject(preparedData(), bestModel()), group = chosenGrouping(), raw = chosenRaw())
  })

  normScorePlot <- eventReactive(c(input$grouping, input$differences, input$normScoreButton), {
    type <- 0
    if(input$differences)
      type <- 1

    if(input$grouping){
      cNORM::plotNorm(buildCnormObject(preparedData(), bestModel()), group = T, type = type)
    }else{
      cNORM::plotNorm(buildCnormObject(preparedData(), bestModel()), type = type)
    }
  })


  crossValidation <- eventReactive(input$CrossValidation, {
    rep <- input$RepetitionsCV
    norm <- input$NormsCV
    maxT <- input$MaxTermsCV
    g <- chosenGrouping()
    r <- chosenRaw()
    e <- chosenExplanatory()

        table <- cNORM::cnorm.cv(preparedData(), repetitions = rep, norms = norm, min = 1, max = maxT, group = g, raw = r, age = e)
        output$TableCV <- renderDataTable({table})
  })


  rawScorePlot <- eventReactive(c(input$grouping1, input$differences1, input$rawScoreButton), {
    type <- 0
    if(input$differences1)
      type <- 1

    if(input$grouping1){
      cNORM::plotRaw(buildCnormObject(preparedData(), bestModel()), group = bestModel()$group, type = type)
    }else{
      cNORM::plotRaw(buildCnormObject(preparedData(), bestModel()), type = type)
    }
  })

  # Prints best model
  output$BestModel1 <- renderText({
    return(bestModel()$report[1])
  })
  output$BestModel2 <- renderText({
    return(bestModel()$report[2])
  })
  output$BestModel3 <- renderText({
    return(bestModel()$report[3])
  })
  output$BestModel4 <- renderText({
    return(bestModel()$report[4])
  })
  output$BestModel5 <- renderText({
    return(bestModel()$report[5])
  })
  output$BestModel6 <- renderText({
    if(checkConsistency(bestModel())){
      return("WARNING! The model seems to be inconsistent. Please check the percentile plot for intersecting percentile curves and change the number of terms for a different solution.")
    }else{
      return("No violations of model consistency found within the boundaries of the original data.")
    }
  })
  output$BestModel7 <- renderText({
    return(rangeCheck(bestModel()))
  })

  # Plots model derivation
  output$PlotDerivatives <- renderPlot({
    return(modelDerivatives())
  })

  output$PlotDensity <- renderPlot({
    return(modelDensity())
  })

  output$PlotCV <- renderPlot({
    return(crossValidation())
  })

  output$PlotWL <- renderPlot({
    return(wlPlot())
  })

  output$PlotValues <- renderPlot({

    return(valuesPlot())
  })

  output$PlotNormScores <- renderPlot({

    return(normScorePlot())
  })

  output$PlotRawScores <- renderPlot({

    return(rawScorePlot())
  })

  output$PrintSubset <- renderDataTable({
    return(cNORM::printSubset(bestModel()))
  })

  output$Series <- renderPlot({

    return(cNORM::plotPercentileSeries(buildCnormObject(preparedData(), bestModel()), start = input$terms, end = input$terms))

  })


  chosenPercentilesForNormCurves <- reactive({


    if(input$PercentilesForNormCurves == ""){
      return(NULL)
  }
    else{
    (input$PercentilesForNormCurves)
  }

  })

  chosenPercentilesForPercentiles <- reactive({
    if(input$PercentilesForPercentiles == ""){
      return(NULL)
    }
    else{
      return(input$PercentilesForPercentiles)
    }
  })

  normCurves <- reactive({

    MIN_AGE <- min((currentFile())[chosenGrouping()])
    MAX_AGE <- max((currentFile())[chosenGrouping()])


    percentileList <- chosenPercentilesForNormCurves()

    if(is.null(chosenPercentilesForNormCurves())){
      percentileList <- c(.02276, 0.1587, 0.5000, 0.8413, 0.97724)
    }
    else{
      percentileList <- as.numeric(unlist(strsplit(chosenPercentilesForNormCurves(), "\\, |\\,| ")))/100
    }



    # Set different norm lists for different scales
      currentScale <- chosenScale()
      if(currentScale == "T"){
        normList <- qnorm(percentileList, 50, 10)
      }
      if(currentScale == "z")
      {
        normList <- normList <- qnorm(percentileList, 0, 1)
      }
      if(currentScale == "IQ"){
        normList <- normList <- qnorm(percentileList, 100, 15)
      }

      normList <- round(normList, digits = 2)

    currentNormCurves <- cNORM::plotNormCurves(bestModel(),
                                              minAge = MIN_AGE,
                                              maxAge = MAX_AGE,
                                              normList = normList)
    return(currentNormCurves)
  })

  # Generates and prints norm curves
  output$NormCurves <- renderPlot({

    return(normCurves())
  })

  # Generates and plots percentile curves
  output$PlotPercentiles <- renderPlot({

    MIN_AGE <- min((currentFile())[chosenGrouping()])
    MAX_AGE <- max((currentFile())[chosenGrouping()])
    # data <- currentFile()
    # attr(data, "descend") <- chosenDescend()

    percentileList <- chosenPercentilesForPercentiles()

    if(is.null(percentileList)){
      cNORM::plotPercentiles(buildCnormObject(preparedData(), bestModel()), raw = chosenRaw(),
                             group = chosenGrouping(),
                             minAge = MIN_AGE,
                             maxAge = MAX_AGE)
    }
    else{
      percentileList <- as.numeric(unlist(strsplit(chosenPercentilesForPercentiles(), "\\, |\\,| ")))/100

      cNORM::plotPercentiles(buildCnormObject(preparedData(), bestModel()), raw = chosenRaw(),
                             group = chosenGrouping(),
                             minAge = MIN_AGE,
                             maxAge = MAX_AGE,
                             percentiles = percentileList)
    }

  })


  output$InputNormValue <- renderUI({
    tagList(numericInput(inputId = "NormValueInputAge", label = "Choose age", value = NULL, min = bestModel()$minA1, max = bestModel()$maxA1),
            numericInput(inputId = "NormValueInputRaw", label = "Choose raw value", value = NULL, min = bestModel()$minRaw, max = bestModel()$maxRaw),
            actionButton(inputId = "CalcNormValue",label = "Norm Score"))
  })

  normValue <- eventReactive(input$CalcNormValue,{

    if(is.null(input$NormValueInputRaw) || is.null(input$NormValueInputRaw)){return()}
    else{
      currentAgeForNormValue <- input$NormValueInputAge
      currentRawForNormValue <- input$NormValueInputRaw
      currentBestModel <- bestModel()
      MIN_NORM <- currentBestModel$minL1
      MAX_NORM <- currentBestModel$maxL1
      MIN_RAW <- currentBestModel$minRaw
      MAX_RAW <- currentBestModel$maxRaw

      currentNormValue <- cNORM::predictNorm(currentRawForNormValue, currentAgeForNormValue, model =currentBestModel, minNorm = MIN_NORM, maxNorm = MAX_NORM)
      return(currentNormValue)
    }
  })


  output$NormValue <- renderText({
    if(is.null(normValue())){return()}

    return(paste("Predicted norm value:",toString(normValue())))
  })

  output$InputRawValue <- renderUI({
    tagList(numericInput(inputId = "RawValueInputAge", label = "Choose age", value = NULL, min = bestModel()$minA1, max = bestModel()$maxA1),
            numericInput(inputId = "RawValueInputNorm", label = "Choose norm value", value = NULL, min = bestModel()$minL1, max = bestModel()$maxL1),
            actionButton(inputId = "CalcRawValue",label = "Raw Score"))
  })

  rawValue <- eventReactive(input$CalcRawValue,{

    if(is.null(input$RawValueInputAge) || is.null(input$RawValueInputNorm)){return()}
    else{
      currentAgeForRawValue <- input$RawValueInputAge
      currentNormForRawValue <- input$RawValueInputNorm
      currentBestModel <- bestModel()
      MIN_NORM <- currentBestModel$minL1
      MAX_NORM <- currentBestModel$maxL1
      MIN_RAW <- currentBestModel$minRaw
      MAX_RAW <- currentBestModel$maxRaw

      currentRawValue <- cNORM::predictRaw(currentNormForRawValue, currentAgeForRawValue, coefficients = currentBestModel$coefficients,
                                           minRaw = MIN_RAW, maxRaw = MAX_RAW)
      return(currentRawValue)
    }
  })

  output$RawValue <- renderText({
    if(is.null(rawValue())){return()}

    return(paste("Predicted raw value:",toString(rawValue())))
  })


  output$InputNormTable <- renderUI({
    tagList(numericInput(inputId = "NormTableInput", label = "Choose age for prediction of norm values", value = NULL,min = 0, max = bestModel()$maxA1),
            numericInput(inputId = "NormTableInputStart", label = "Choose norm start value", value = bestModel()$scaleM - bestModel()$scaleSD*2.5),
            numericInput(inputId = "NormTableInputEnd", label = "Choose norm end value", value = bestModel()$scaleM + bestModel()$scaleSD*2.5),
            numericInput(inputId = "NormTableInputStepping", label = "Choose stepping value", value = NULL),
            numericInput(inputId = "NormTableCI", label = "Confidence Coefficient", value = .9),
            numericInput(inputId = "NormTableRel", label = "Reliability Coefficient", value = NULL),
            actionButton(inputId = "CalcNormTables",label = "Generate norm table", value = NULL),
            downloadButton("DownloadNormTable", "Download norm table"))

  })


  normTable <- eventReactive(input$CalcNormTables, {

    check_inputs <- is.null(input$NormTableInput) || is.null(input$NormTableInputStart) || is.null(input$NormTableInputEnd) || is.null(input$NormTableInputStepping)
    if(check_inputs){return()}
    else{
      currentAgeForNormTable <- input$NormTableInput
      currentBestModel <- bestModel()
      MIN_NORM <- as.numeric(input$NormTableInputStart)
      MAX_NORM <- as.numeric(input$NormTableInputEnd)

      if(is.null(input$NormTableRel)){
        REL <- NULL
      }else{
        REL <- as.numeric(input$NormTableRel)
      }

      if(is.na(REL)){
        REL <- NULL
      }

      CI <- as.numeric(input$NormTableCI)
      if(is.na(CI)){
        CI <- NULL
      }

      STEPPING <- as.numeric(input$NormTableInputStepping)
      MIN_RAW <- currentBestModel$minRaw
      MAX_RAW <- currentBestModel$maxRaw

      currentNormTable <- cNORM::normTable(currentAgeForNormTable, bestModel(), minNorm = MIN_NORM, maxNorm = MAX_NORM,
                                           minRaw = MIN_RAW, maxRaw = MAX_RAW, step = STEPPING, CI = CI, reliability = REL, pretty = TRUE)

      currentNormTable$raw <- round(currentNormTable$raw, digits = 2)
      currentNormTable$percentile <- round(currentNormTable$percentile, digits = 2)

      return(currentNormTable)}
  })



  output$NormTable <- renderDataTable({
    if(is.null(normTable())){return()}

    return(normTable())
  })

  output$DownloadNormTable <- downloadHandler(
    filename = function(){
      "NormTable.csv"
    },
    content = function(file){
      output <- as.data.frame(normTable())
      write.csv(output, file)
    }
  )


  output$InputRawTable <- renderUI({
    tagList(numericInput(inputId = "RawTableInput", label = "Choose age for prediction of raw values", value = NULL,min = 0, max = bestModel()$maxA1),
            numericInput(inputId = "RawTableInputStart", label = "Choose raw start value", value = bestModel()$minRaw),
            numericInput(inputId = "RawTableInputEnd", label = "Choose raw end value", value = bestModel()$maxRaw),
            numericInput(inputId = "RawTableInputStepping", label = "Choose stepping value", value = NULL),
            numericInput(inputId = "RawTableCI", label = "Confidence Coefficient", value = .9),
            numericInput(inputId = "RawTableRel", label = "Reliability Coefficient", value = NULL),
            actionButton(inputId = "CalcRawTables",label = "Generate raw table"),
            downloadButton("DownloadRawTable", label = "Download raw table"))
  })


  rawTable <- eventReactive(input$CalcRawTables, {

    checking_input <- is.null(input$RawTableInputStart) || is.null(input$RawTableInputStart) || is.null(input$RawTableInputEnd) || is.null(input$RawTableInputStepping)
    if(checking_input){return()}
    else{
      currentAgeForRawTable <- input$RawTableInput
      currentBestModel <- bestModel()
      MIN_NORM <- currentBestModel$minL1
      MAX_NORM <- currentBestModel$maxL1
      MIN_RAW <- input$RawTableInputStart
      MAX_RAW <- input$RawTableInputEnd
      STEPPING <- as.numeric(input$RawTableInputStepping)

      if(is.null(input$RawTableRel)){
        REL <- NULL
      }else{
        REL <- as.numeric(input$RawTableRel)
      }

      if(is.na(REL)){
        REL <- NULL
      }

      CI <- as.numeric(input$RawTableCI)
      if(is.na(CI)){
        CI <- NULL
      }

      currentRawTable <- cNORM::rawTable(currentAgeForRawTable, currentBestModel, minNorm = MIN_NORM, maxNorm = MAX_NORM,
                                         minRaw = MIN_RAW, maxRaw = MAX_RAW, step = STEPPING, CI = CI, reliability = REL, pretty = TRUE)
      currentRawTable$norm <- round(currentRawTable$norm, digits = 3)
      currentRawTable$percentile <- round(currentRawTable$percentile, digits = 2)


      return(currentRawTable)}
  })

  output$RawTable <- renderDataTable({
    if(is.null(rawTable())){return()}

    return(rawTable())
  })

  output$DownloadRawTable <- downloadHandler(
    filename = function(){
      "RawTable.csv"
    },
    content = function(file){
      output <- as.data.frame(rawTable())
      write.csv(output, file)
    }
  )

  output$exportData <- downloadHandler(
    filename = function() {
      paste("data.RData", sep = "")
    },
    content = function(file) {
      dat <- preparedData()
      save(dat, filename = file)
    }
  )

  # Generates and plots modelled against actual data values
  output$PlotValues <- renderPlot({

    cNORM::plotRaw(buildCnormObject(preparedData(), bestModel()), group = chosenGrouping(), raw = chosenRaw())
  })

  # Generates and plots contour plot of first derivation
  output$PlotDerivatives <- renderPlot({


    MIN_AGE <- min((currentFile())[chosenGrouping()])
    MAX_AGE <- max((currentFile())[chosenGrouping()])

    cNORM::plotDerivative(bestModel(),
                          minAge = MIN_AGE,
                          maxAge = MAX_AGE)
  })
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)

## ----fig0, fig.height = 4, fig.width = 7--------------------------------------
 ## Loads the package and displays the data

library(cNORM)
str(ppvt)
plot(ppvt$age, ppvt$raw, main="PPVT Raw Scores by Age",
              xlab="Age", ylab="Raw score")

## ----fig1, fig.height = 4, fig.width = 7--------------------------------------
# Models the data across a continuos explanatroy variable such as age,
# thereby assuming that the raw scores follow a beta-binomial distribution
# at a given age level:

model.betabinomial <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)

## -----------------------------------------------------------------------------
# Provides fit indices

diagnostics.betabinomial(model.betabinomial, age = ppvt$age, score = ppvt$raw)

## -----------------------------------------------------------------------------
# Provides norm scores for specified age levels and raw scores.
# If not specified otherwise in the model, the norm scores will
# be returned as T-scores.

predict(model.betabinomial, c(10.125, 10.375, 10.625, 10.875), c(200, 200, 200, 200))

## ----fig2, fig.height = 4, fig.width = 7--------------------------------------
# Calculates weights and models the data:

margins <- data.frame(variables = c("sex", "sex",
                    "migration", "migration"),
                    levels = c(1, 2, 0, 1),
                    share = c(.52, .48, .7, .3))
weights <- computeWeights(ppvt, margins)
model_weighted <- cnorm.betabinomial(ppvt$age, ppvt$raw, weights = weights)


## -----------------------------------------------------------------------------
# Generates norm tables for age 14.25 and 14.75 and computes 95%-confidence
# intervals with a reliability of .97.

tables <- normTable.betabinomial(model.betabinomial, c(14.25, 14.75), CI = .95, reliability = .97)
head(tables[[1]]) # head is used to show only the first few rows of the table

if(!require(shiny)){
  install.packages("shiny")
  require(shiny)
}

if(!require(shinycssloaders)){
  install.packages("shinycssloaders")
  require(shinycssloaders)
}

if(!require(cNORM)){
  install.packages("cNORM")
  require(cNORM)
}

if(!require(DT)){
  install.packages("DT")
  require(cNORM)
}
title <- "cNORM-Shiny"

# Define UI for cNORM-Application
shinyUI(fluidPage(

  # Set tab title
  title = "cNORM - Shiny",

  # Tabsetpanel for single tabs
  tabsetPanel(
    # Tab for data input
    tabPanel("Data Input", sidebarLayout(
      sidebarPanel(
        img(src = "logo.png", align = "right"), tags$h3("Data Input"), tags$p("Please choose a data set for your calculations. You can use an inbuilt example or load your own file:"), selectizeInput("Example",
          label = "Example:",
          choices = c("", "elfe", "ppvt", "CDC"), selected = character(0),
          multiple = FALSE
        ), hr(),
        fileInput("file", "Choose a file", multiple = FALSE, accept = c(".csv", ".xlsx", ".xls", ".rda", ".sav")),
        tags$p(tags$b("HINT: If you choose a file from your own directory, the chosen example will not be used during the session anymore!")),
        hr(),tags$p(tags$b("The built-in example data:")),
        tags$ul(tags$li("'elfe' includes norming sample data of a reading comprehension in elementary school, based on a multiple choice sentence completion task with raw scores ranging from 0 to 28 "),
        tags$li("'ppvt' uses data from the Peabody Picture Vocabulary Test (Revision IV) ranging from age 2.5 to 17.5 with raw scores from 0 to 228"),
        tags$li("'CDC' is a data collection of BMI, weight and growth development from 2 to 18; large sample based on epidemiological studies of the Center of Disease Control (CDC)"))
      ),
      mainPanel(htmlOutput("introduction"), DT::DTOutput("table"))
    )),

    # Panel for Data Preparation (choosing and visualizing)
    tabPanel(
      "Preparation",
      # Define layout for sidebar
      sidebarLayout(

        # Define sidebar panel for choosing grouping, raw values and scale
        sidebarPanel(
          tags$h3("Choose"),
          uiOutput("GroupingVariable"),
          tags$p("Variable which is used to divide the observations into groups, for example variable age by rounding to a half or a full year"),
          tags$br(),
          uiOutput("RawValues"),
          tags$br(),
          uiOutput("WeightVariable"),
          tags$div(class = "body", checked = NA,
          tags$span("Optional weighting variable, please consult vignette on "),
          tags$a(href="https://cran.r-project.org/web/packages/cNORM/vignettes/WeightedRegression.html", "WeightedRegression"),
          ),
          tags$br(),
          selectInput("Scale", label = "Norm Scale", choices = c("T", "IQ", "z"), selected = "T"),
          actionButton(inputId = "DoDataPreparation", label = "Prepare Data"),
          downloadButton("downloadData", "Download Data"),
          tags$br(),
          tags$br(),
          # Additional options (explanatory variable with default grouping, ranking method, number of powers)
          tags$h3("Additional options"),
          uiOutput("ExplanatoryVariable"),
          tags$p("As default, the explanatory variable is set to grouping variable. If available, use a continuous explanatory variable like age."),
          tags$b("HINT! The range of values of the explanatory variable have to correpsond with the values of the grouping variable, i. e. specify the group means."),
          tags$br(),
          tags$br(),
          tags$br(),
          selectInput(
            inputId = "NumberOfPowers",
            label = "Power degree for location",
            choices = c(1:5),
            selected = 5
          ),
          tags$p("This variable specifies the power parameter for the norm score in the Taylor polynomial. As default number of power is set to 4. Higher values might lead to a closer fit, but yield the danger of overfitting."),
          tags$br(),
          selectInput(
            inputId = "NumberOfPowersAge",
            label = "Power degree for age",
            choices = c(1:5),
            selected = 3
          ),
          tags$p("This variable specifies the power parameter for the explanatory variable in the Taylor polynomial, usually age. Since age trajectories are usually less complex, it can be reduced to 3 or 2 in most use cases, leading to simpler and more robust models."),
          tags$br(),
          selectInput("Method", label = "Ranking method", choices = c("Blom (1985)", "Tukey (1949)", "Van der Warden (1952)", "Rankit (Bliss, 1967)", "Levenbach (1953)", "Filliben (1975)", "Yu & Huang (2001)"), selected = "Rankit (Bliss, 1967)"),
          selectInput(
            inputId = "RankingOrder",
            label = "Ranking order of the raw score",
            choices = c("Ascending", "Descending"),
            selected = 1
          )
        ),
        # Main panel for showing prepared data
        mainPanel(withSpinner(DT::DTOutput("preparedData"), type = 5))
      )
    ),

    # Defines panel for best model output
    # Tab returns bestModel with information function and plot of real and predicted raw values
    navbarMenu(
      "Modeling & Validation",
      tabPanel(
      "Model",
      sidebarLayout(
        sidebarPanel(
          tags$h3("Model Data"),
          tags$p("Here, you can calculate a regression model that models the original data as close as possible, while smoothing the curves and eliminating noise. The plots display the percentile curves and the information criteria for the different models, beginning with the model with one terms up to the maximum. A high R2 with as few terms as possible is preferable."),
          tags$br(), tags$b("HINT: Please ensure that the data is loaded and prepared, before starting the modeling. In case of k > 4, the calculation will take a few seconds. Look out for an 'elbow' in the information function chart and try that number of terms for your model. Avoid intersecting percentile curves."),
          tags$br(), tags$br(),

          actionButton(
            inputId = "CalcBestModel",
            label = "Model Data"
          ),

          downloadButton("downloadModel", "Download Model"),

          tags$br(),
          # Additional options (R^2, terms, type for printSubset)
          tags$h3("Additional options"),
          numericInput(
            inputId = "ChosenDetCoeff",
            label = "Coefficient of Determination", value = 0.99, min = 0, max = 1, step = 0.01
          ),
          uiOutput("NumberOfTerms"),
          selectInput(inputId = "chosenTypePlotSubset", "Type of plot", choices = c("Adjusted R2 by Number of Predictors", "Log Transformed Mallow's Cp by Adjusted R2", "Bayesian Information Criterion (BIC) by Adjusted R2", "RMSE by Number of Predictors"), selected = "RMSE by Number of Predictors")
        ),
        mainPanel(
          withSpinner(verbatimTextOutput("BestModel1"), type = 5),
          verbatimTextOutput("BestModel2"),
          verbatimTextOutput("BestModel3"),
          verbatimTextOutput("BestModel4"),
          verbatimTextOutput("BestModel5"),
          #verbatimTextOutput("BestModel6"),
          #verbatimTextOutput("BestModel7"),
          tags$br(),
          # tags$h4("Information Function, Subset Specifics and Fitted Values"),
          # tags$p("The plot shows the informationcriteria for the different models, beginning with the model with one terms up to the maximum. The model should have a high R2 with as few terms as possible. The information of the plot is again displayed as a table below the chart. On the bottom of the page, you can see, how well the observed data are fitted by the model."),
          # plotOutput("PlotPercentiles", width = "100%", height = "600px"),
          withSpinner(plotOutput("modelPlot", width = "100%", height = "600px"), type = 5),
          tags$br(),

          withSpinner(plotOutput("PlotWL", width = "100%", height = "600px"), type = 5),
          tags$br(),
          withSpinner(DT::DTOutput("PrintSubset"), type = 5),
          tags$br()
          )
      )
    ),

    tabPanel(
      "Cross-Validation",
      sidebarLayout(
        sidebarPanel(
          tags$h3("Cross Validation"),
          tags$p("This function helps in selecting the number of terms for the model by doing repeated cross validation with 80 percent of the data as training data and 20 percent as the validation data. The cases are drawn randomly but stratified by norm group. Successive models are retrieved with increasing number of terms and the RMSE of raw scores (fitted by the regression model) is plotted for the training, validation and the complete dataset. Additionally to this analysis on the raw score level, it is possible (default) to estimate the mean norm score reliability and crossfit measures. "),
          tags$p("HINT: The function has a high computational load when computing norm scores and takes some time to finish. Time increases with number of maximum terms, sample size and number of repetitions."),
          actionButton(
            inputId = "CrossValidation",

            label = "Cross Validation"
          ),
          tags$br(),
          tags$h3("Additional options"),
          sliderInput("MaxTermsCV", "Maximum number of terms:",
                      min = 1, max = 24, value = 10
          ),
          tags$br(),
          checkboxInput("NormsCV", "Check norm scores:", TRUE),
          tags$br(),
          sliderInput("RepetitionsCV", "Repetitions:",
                      min = 1, max = 20, value = 1
          )
        ),
        mainPanel(
          withSpinner(plotOutput("PlotCV", width = "100%", height = "800px"), type = 5),
          tags$br(),
          withSpinner(DT::DTOutput("TableCV"), type = 5)
        )
      )
    )
    ),

    navbarMenu(
      "Visualization",
      tabPanel("Percentiles", sidebarLayout(
        sidebarPanel(
          tags$h3("Percentiles"), tags$p("The chart shows how well the model generally fits the observed data. The observed percentiles are represented as dots, the continuous norm curves as lines. In case of intersecting norm curves the model is inconsistent. Please change the number of terms in the 'Best Model' tab in order to find a consistent model. You can use the 'Series' option to look out for suitable parameters."),
          tags$br(),
          textInput(inputId = "PercentilesForPercentiles", "Choose percentiles"),
          tags$p("Please seperate the values by a comma or space.")
        ),
        mainPanel(plotOutput("PlotPercentiles", width = "100%", height = "800px"))
      )),

      tabPanel("Series", sidebarLayout(
        sidebarPanel(
          tags$h3("Percentile Series"),
          tags$p("In oder to facilitate model selection, the chart displays percentile curves of the different models."),
          tags$br(), sliderInput("terms", "Number of terms:",
            min = 1, max = 24, value = 5
          ), tags$br(), tags$br(),
          tags$p("Please use the slider to change the number of terms in the model. Please select a model with non-intersecting percentile curves. Avoid undulating curves, as these indicate model overfit."),
          tags$br(), tags$b("After choosing the best fitting number of terms, please rerun the model calculation with a fixed number of terms before generating norm tables.")
        ),
        mainPanel(plotOutput("Series", width = "100%", height = "800px"))
      )),


      tabPanel("Norm Curves", sidebarLayout(
        sidebarPanel(
          tags$h3("Norm Curves"), tags$p("The chart is comparable to the percentile plot. It only shows the norm curves for some selected norm scores."),
          textInput(inputId = "PercentilesForNormCurves", label = "Choose percentiles for norm curves", value = ""),
          tags$p("Please seperate the values by a comma or space. The percentile values are automatically transformed to the norm scale used in the data preparation. In order to get curves specific z values, you can use the following percentiles:"),
          tags$div(
            HTML("<div align=center><table width=100%><tr><td align = right><b>z</b></td><td align = right>-2</td><td align = right>-1</td><td align = right>0</td><td align = right>1</td><td align = right>2</td></tr><tr><td align = right><b>percentile</b></td><td align = right> 2.276</td><td align = right> 15.87</td><td align = right> 50.00</td><td align = right> 84.13</td><td align = right> 97.724</td></tr></table></div>")
          )
        ),
        mainPanel(plotOutput("NormCurves", width = "100%", height = "800px"))
      )),
      tabPanel("Density", sidebarLayout(sidebarPanel(
        tags$h3("Density"), tags$p("The plot shows the probability density function of the raw scores based on the regression model. Like the 'Derivative Plot', it can be used to identify violations of model validity or to better visualize deviations of the test results from the normal distribution. As a default, the lowest, highest and a medium group is shown."),
        tags$br(),
        textInput(inputId = "densities", "Choose groups"),
        tags$p("Please seperate the values by a comma or space.")
      ), mainPanel(
        withSpinner(
          plotOutput("PlotDensity", width = "100%", height = "800px"),
          type = 5
        )
      ))),
      tabPanel("Derivative Plot", sidebarLayout(sidebarPanel(tags$h3("Derivative Plot"), tags$p("To check whether the mapping between latent person variables and test scores is biunique, the regression function can be searched numerically within each group for bijectivity violations using the 'checkConsistency' function. In addition, it is also possible to plot the first partial derivative of the regression function to l and search for negative values. Look out for values lower than 0. These indicate violations of the model.")), mainPanel(
        withSpinner(plotOutput("PlotDerivatives", width = "100%", height = "600px"), type = 5)
      ))),

      tabPanel("Norm Scores", sidebarLayout(sidebarPanel(
        tags$h3("Norm Scores Plot"), tags$p("The plot shows the observed and predicted norm scores. You can identify, how well the model is able to predict the norm scores of the dataset. The duration of the computation increases with the size of the dataset."),
        tags$br(),
        checkboxInput("grouping", "Plot by group", FALSE),
        checkboxInput("differences", "Plot differences", FALSE), tags$br(),
        actionButton(
          inputId = "normScoreButton",

          label = "Update"
        )
      ), mainPanel(
        withSpinner(
          plotOutput("PlotNormScores", width = "100%", height = "800px"),
          type = 5
        )
      ))),

      tabPanel("Raw Scores", sidebarLayout(sidebarPanel(
        tags$h3("Raw Scores Plot"), tags$p("The plot shows the observed and predicted raw scores. You can identify, how well the model is able to predict the raw scores of the original dataset."),
        tags$br(),
        checkboxInput("grouping1", "Plot by group", FALSE),
        checkboxInput("differences1", "Plot differences", FALSE), tags$br(),
        actionButton(
          inputId = "rawScoreButton",

          label = "Update"
      )), mainPanel(
        withSpinner(
          plotOutput("PlotRawScores", width = "100%", height = "800px"),
          type = 5
        )
      )))
    ),


    # NavbarMenu for predicting norm and raw values and raw/norm tables
    navbarMenu(
      "Prediction",
      tabPanel("Norm value prediction", sidebarLayout(sidebarPanel(tags$h3("Prediction of singel norm values"), uiOutput("InputNormValue")), mainPanel(verbatimTextOutput("NormValue")))),
      tabPanel("Raw value prediction", sidebarLayout(sidebarPanel(tags$h3("Prediction of single raw values"), uiOutput("InputRawValue")), mainPanel(verbatimTextOutput("RawValue")))),
      tabPanel("Norm table", sidebarLayout(sidebarPanel(tags$h3("Norm table compilation"), uiOutput("InputNormTable")), mainPanel(DT::DTOutput("NormTable")))),
      tabPanel("Raw table", sidebarLayout(sidebarPanel(tags$h3("Raw table compilation"), uiOutput("InputRawTable")), mainPanel(DT::DTOutput("RawTable"))))
    )
  )
))

## ----setup, include = FALSE-------------------------------
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
options(width = 60)

#devtools::load_all(".")
library(cNORM)

## ----fig1, fig.height = 4, fig.width = 7------------------
# Fit shash model with custom settings
# The function automatically displays percentile plots
model.shash <- cnorm.shash(age = ppvt$age, score = ppvt$raw)


# Use print(model.shash), diagnostics(model.shash) or summary(model.shash)
# to retrieve information on the data fit.

## ----eval=FALSE-------------------------------------------
# # Conservative parameterization with fixed delta across age
# model.simple <- cnorm.shash(age = ppvt$age, score = ppvt$raw,
#                            mu_degree = 2,        # Quadratic location pattern
#                            sigma_degree = 1,     # Linear variability change
#                            epsilon_degree = 1,   # Linear skewness change
#                            delta_degree = NULL,  # deactivates polynimial fitting for delta
#                            delta = 1.1)          # Slightly heavy tails,
#                                                  # kept constant across age
# 
# # Example with more complex parameterization
# model.complex <- cnorm.shash(age = ppvt$age, score = ppvt$raw,
#                            mu_degree = 4,        # Quadric pattern
#                            sigma_degree = 3,     # Complex variability changes
#                            epsilon_degree = 2,   # Quadratic age-varying skewness
#                            delta_degree = 2)     # Changing tail weights across age (quadratic)
# 
# # Compare models
# compare(model.simple, model.complex, age = ppvt$age, score = ppvt$raw,
#         title = "ShaSh Model Comparison")
# 
# 

## ----eval=FALSE-------------------------------------------
# # Calculate post-stratification weights
# margins <- data.frame(variables = c("sex", "sex", "migration", "migration"),
#                      levels = c(1, 2, 0, 1),
#                      share = c(.52, .48, .7, .3))
# 
# weights <- computeWeights(ppvt, margins)
# 
# # Fit weighted ShaSh model
# model.weighted <- cnorm.shash(ppvt$age, ppvt$raw, weights = weights)
# 
# # Compare weighted vs. unweighted
# compare(model.shash, model.weighted, age = ppvt$age, score = ppvt$raw,
#         title = "Unweighted vs. Weighted ShaSh Models")

## ----eval=FALSE-------------------------------------------
# # Individual Norm Score Prediction:
# # Generate norm scores for specific age-score combinations
# ages <- c(10.25, 10.75, 11.25, 11.75)
# raw_scores <- c(180, 185, 190, 195)
# 
# norm_scores <- predict(model.shash, ages, raw_scores)
# prediction_table <- data.frame(
#   Age = ages,
#   Raw_Score = raw_scores,
#   Norm_Score = round(norm_scores, 1)
# )
# print(prediction_table)
# 
# 
# # Norm Score tables:
# # Generate detailed norm tables for multiple ages
# tables <- normTable.shash(model.shash,
#                          ages = c(10.25, 10.75),
#                          start = 150,
#                          end = 220,
#                          step = 1,
#                          CI = 0.95,
#                          reliability = 0.94)
# 
# # Display head from first table
# head(tables[[1]], 10)

## ----fig5, fig.height = 4, fig.width = 7------------------
# Compare shash with betabinomial models (BB). BB models should work worse,
# since the test has stop rules, leading to non-binomial distributions.
model.bb <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228, plot = FALSE)

# Model comparisons
compare(model.shash, model.bb, age = ppvt$group, score = ppvt$raw,
        title = "SinH-ArcSinH vs. Beta-Binomial")


# Compare distribution free Taylor model
model.taylor <- cnorm(group = ppvt$group, raw = ppvt$raw, plot=FALSE)

# Model comparisons shash versus taylor
compare(model.shash, model.taylor, age = ppvt$group, score = ppvt$raw,
        title = "SinH-ArcSinH vs. Taylor")

#' Computes the curve for a specific T value
#'
#' As with this continuous norming regression approach, raw scores are modeled as a function of age and norm score
#' (location), getNormCurve is a straightforward approach to show the raw score development over
#' age, while keeping the norm value constant. This way, e. g. academic performance or intelligence development
#' of a specific ability is shown.
#' @param norm The specific norm score, e. g. T value
#' @param model The model from the regression modeling obtained with the cnorm function
#' @param minAge Age to start from
#' @param maxAge Age to stop at
#' @param step Stepping parameter for the precision when retrieving of the values, lower
#' values indicate higher precision (default 0.1).
#' @param minRaw lower bound of the range of raw scores (default = 0)
#' @param maxRaw upper bound of raw scores
#' @return data.frame of the variables raw, age and norm
#' @examples
#' \dontrun{
#' # Generate cnorm object from example data
#' cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
#' getNormCurve(35, cnorm.elfe)
#' }
#'
#' @family predict
#' @export
getNormCurve <-
  function(norm,
           model,
           minAge = NULL,
           maxAge = NULL,
           step = 0.1,
           minRaw = NULL,
           maxRaw = NULL) {
    if (isTaylor(model)) {
      model <- model$model
    }

    if (!inherits(model, "cnorm") &&
        !inherits(model, "cnormModel")) {
      stop("Please provide a cnorm object.")
    }

    if (is.null(minAge)) {
      minAge <- model$minA1
    }

    if (is.null(maxAge)) {
      maxAge <- model$maxA1
    }

    if (is.null(minRaw)) {
      minRaw <- model$minRaw
    }

    if (is.null(maxRaw)) {
      maxRaw <- model$maxRaw
    }

    ages <- seq(minAge, maxAge, by = step)
    results <- lapply(ages, function(age) {
      r <- predictRaw(norm, age, model$coefficients, minRaw, maxRaw)
      data.frame(norm = paste(norm, "T"),
                 age = age,
                 raw = r)
    })
    curve <- do.call(rbind, results)

    return(curve)
  }


#' Predict raw values
#'
#' Most elementary function to predict raw score based on Location (L, T score),
#' Age (grouping variable) and the coefficients from a regression model.
#'
#' @param norm The norm score, e. g. a specific T score or a vector of scores
#' @param age The age value or a vector of scores
#' @param coefficients The a cnorm object or the coefficients from the regression model
#' @param minRaw Minimum score for the results; can be used for clipping unrealistic outcomes,
#' usually set to the lower bound of the range of values of the test (default: 0)
#' @param maxRaw Maximum score for the results; can be used for clipping unrealistic outcomes
#' usually set to the upper bound of the range of values of the test
#' @return the predicted raw score or a data.frame of scores in case, lists of norm scores or age is used
#' @examples
#' \dontrun{
#' # Prediction of single scores
#' model <- cnorm(raw = elfe$raw, group = elfe$group)
#' predictRaw(35, 3.5, model)
#' }
#'
#' @family predict
#' @export
predictRaw <-
  function(norm,
           age,
           coefficients,
           minRaw = -Inf,
           maxRaw = Inf) {
    if (inherits(coefficients, "cnorm")) {
      coef <- coefficients$model$coefficients
    } else if (inherits(coefficients, "cnormModel")) {
      coef <- coefficients$coefficients
    } else {
      coef <- coefficients
    }

    if (length(norm) != length(age)) {
      if (length(norm) == 1L)
        norm <- rep_len(norm, length(age))
      else if (length(age) == 1L)
        age  <- rep_len(age, length(norm))
      else
        stop("`norm` and `age` must have the same length, ",
             "or one of them must be of length 1.")
    }

    # Robust extraction of L (k) and A (t) max powers
    L_names <- grep("^L[0-9]+", names(coef), value = TRUE)
    A_names <- grep("A[0-9]+$", names(coef), value = TRUE)

    k <- if (length(L_names) > 0L)
      max(as.integer(sub("^L([0-9]+).*", "\\1", L_names)))
    else
      0L
    t <- if (length(A_names) > 0L)
      max(as.integer(sub(".*A([0-9]+)$", "\\1", A_names)))
    else
      0L

    # Belt-and-braces guard against -Inf
    k <- if (is.finite(k))
      k
    else
      0
    t <- if (is.finite(t))
      t
    else
      0

    # Prepare the matrix for new data
    X_new <- prepare_matrix(norm, age, k, t)

    # Get all variable names from the model
    all_vars <- c("Intercept", colnames(X_new))

    # Identify which variables are in the model
    model_vars <- names(coef)
    model_vars <- gsub("^.*Intercept.*$", "Intercept", model_vars)

    # Create a matrix with all possible variables, filled with zeros
    X_new_full <- matrix(0, nrow = nrow(X_new), ncol = length(all_vars))
    colnames(X_new_full) <- all_vars

    # Fill in the values for the variables that are present in X_new
    X_new_full[, colnames(X_new)] <- X_new

    # Add the intercept column
    X_new_full[, "Intercept"] <- 1

    # Subset X_new_full to include only the variables that are in the model.
    # drop = FALSE keeps it a matrix even when only one column survives.
    X_new_selected <- X_new_full[, model_vars, drop = FALSE]

    # Make predictions
    predictions <- as.vector(X_new_selected %*% coef)
    predictions[predictions < minRaw] <- minRaw
    predictions[predictions > maxRaw] <- maxRaw

    return(predictions)
  }


#' Create a norm table based on model for specific age
#'
#' This function generates a norm table for a specific age based on the regression
#' model by assigning raw scores to norm scores. Please specify the
#' range of norm scores, you want to cover. A T value of 25 corresponds to a percentile
#' of .6. As a consequence, specifying a range of T = 25 to T = 75 would cover 98.4 % of
#' the population. Please be careful when extrapolating vertically (at the lower and
#' upper end of the age specific distribution). Depending on the size of your standardization
#' sample, extreme values with T < 20 or T > 80 might lead to inconsistent results.
#' In case a confidence coefficient (CI, default .9) and the reliability is specified,
#' confidence intervals are computed for the true score estimates, including a correction for
#' regression to the mean (Eid & Schmidt, 2012, p. 272).
#'
#' @param A the age as single value or a vector of age values
#' @param model The regression model from the cnorm function
#' @param minNorm The lower bound of the norm score range
#' @param maxNorm The upper bound of the norm score range
#' @param minRaw clipping parameter for the lower bound of raw scores
#' @param maxRaw clipping parameter for the upper bound of raw scores
#' @param step Stepping parameter with lower values indicating higher precision
#' @param monotonuous corrects for decreasing norm scores in case of model inconsistencies (default)
#' @param CI confidence coefficient, ranging from 0 to 1, default .9
#' @param reliability coefficient, ranging between  0 to 1
#' @param pretty Format table by collapsing intervals and rounding to meaningful precision
#' @return either data.frame with norm scores, predicted raw scores and percentiles in case of simple A
#' value or a list of norm tables if vector of A values was provided
#' @seealso rawTable
#' @references Eid, M. & Schmidt, K. (2012). Testtheorie und Testkonstruktion. Hogrefe.
#' @examples
#' \dontrun{
#' # Generate cnorm object from example data
#' cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
#'
#' # create single norm table
#' norms <- normTable(3.5, cnorm.elfe, minNorm = 25, maxNorm = 75, step = 0.5)
#'
#' # create list of norm tables
#' norms <- normTable(c(2.5, 3.5, 4.5), cnorm.elfe,
#'   minNorm = 25, maxNorm = 75,
#'   step = 1, minRaw = 0, maxRaw = 26
#' )
#'
#' # conventional norming, set age to arbitrary value
#' model <- cnorm(raw = elfe$raw)
#' normTable(0, model)
#' }
#'
#' @family predict
#' @export
normTable <- function(A,
                      model,
                      minNorm = NULL,
                      maxNorm = NULL,
                      minRaw = NULL,
                      maxRaw = NULL,
                      step = NULL,
                      monotonuous = TRUE,
                      CI = .9,
                      reliability = NULL,
                      pretty = TRUE) {
  if (isBeta(model)) {
    return(normTable.betabinomial(model, A, CI = CI, reliability = reliability))
  } else if (isSHASH(model)) {
    return(normTable.shash(
      model,
      A,
      minRaw,
      maxRaw,
      step,
      CI = CI,
      reliability = reliability
    ))
  } else if (isTaylor(model)) {
    model <- model$model
  } else if (!inherits(model, "cnormModel")) {
    stop("Please provide a cnorm object.")
  }

  if (model$useAge && !is.numeric(A)) {
    stop("Please specify age.")
  }

  if (is.null(model)) {
    stop("No model specified")
  }

  if (is.null(CI) || is.na(CI)) {
    reliability <- NULL
  } else if (CI > .99999 || CI < .00001) {
    stop("Confidence coefficient (CI) out of range. Please specify value between 0 and 1.")
  }

  rel <- FALSE
  if (!is.null(reliability)) {
    if (reliability > .9999 || reliability < .0001) {
      stop("Reliability coefficient out of range. Please specify value between 0 and 1.")
    } else {
      se  <- qnorm(1 - ((1 - CI) / 2)) * sqrt(reliability * (1 - reliability))
      rel <- TRUE
    }
  }

  if (is.null(minNorm) || is.na(minNorm)) {
    minNorm <- model$scaleM - 2.5 * model$scaleSD
  }

  if (is.null(maxNorm) || is.na(maxNorm)) {
    maxNorm <- model$scaleM + 2.5 * model$scaleSD
  }

  if (is.null(minNorm) || is.null(maxNorm)) {
    stop("Please specify minNorm and maxNorm.")
  }

  if (is.null(step) || is.na(step)) {
    step <- model$scaleSD / 10
  }

  descend <- model$descend

  if (is.null(minRaw) || is.na(minRaw)) {
    minRaw <- model$minRaw
  }

  if (is.null(maxRaw) || is.na(maxRaw)) {
    maxRaw <- model$maxRaw
  }

  tables <- vector("list", length(A))

  for (x in seq_along(A)) {
    maxn <- maxNorm
    minn <- minNorm
    n_steps <- floor((maxn - minn) / step) + 1L
    norm <- vector("list", n_steps)
    raw  <- vector("list", n_steps)
    i <- 1L

    while (i <= n_steps) {
      r <- predictRaw(minn,
                      A[[x]],
                      model$coefficients,
                      minRaw = minRaw,
                      maxRaw = maxRaw)

      norm[[i]] <- minn
      raw[[i]]  <- r

      minn <- minn + step
      i    <- i + 1L
    }

    normTable <- do.call(rbind, Map(data.frame, norm = norm, raw = raw))

    if (!is.na(model$scaleM) && !is.na(model$scaleSD)) {
      normTable$percentile <- pnorm((normTable$norm - model$scaleM) / model$scaleSD) * 100
    }

    if (monotonuous) {
      normTable$raw <- enforce_monotone_middle(normTable$raw, descend = isTRUE(model$descend))
    }

    if (rel) {
      zPredicted <- reliability * (normTable$norm - model$scaleM) / model$scaleSD
      normTable$lowerCI    <- (zPredicted - se) * model$scaleSD + model$scaleM
      normTable$upperCI    <- (zPredicted + se) * model$scaleSD + model$scaleM
      normTable$lowerCI_PR <- pnorm(zPredicted - se) * 100
      normTable$upperCI_PR <- pnorm(zPredicted + se) * 100
    }

    if (pretty)
      tables[[x]] <- prettyPrint(normTable)
    else
      tables[[x]] <- normTable
  }

  names(tables) <- A

  if (length(A) == 1) {
    return(tables[[1]])
  } else {
    return(tables)
  }
}


#' Create a table with norm scores assigned to raw scores for a specific age based on the regression model
#'
#' This function is comparable to 'normTable', despite it reverses the assignment:
#' A table with raw scores and the according norm scores for a specific age based on the regression
#' model is generated. This way, the inverse function of the regression model is solved numerically with
#' brute force. Please specify the range of raw values, you want to cover. With higher precision
#' and smaller stepping, this function becomes computational intensive.
#' In case a confidence coefficient (CI, default .9) and the reliability is specified,
#' confidence intervals are computed for the true score estimates, including a correction for
#' regression to the mean (Eid & Schmidt, 2012, p. 272).
#' @param A the age, either single value or vector with age values
#' @param model The regression model or a cnorm object
#' @param minRaw The lower bound of the raw score range
#' @param maxRaw The upper bound of the raw score range
#' @param minNorm Clipping parameter for the lower bound of norm scores (default 25)
#' @param maxNorm Clipping parameter for the upper bound of norm scores (default 25)
#' @param step Stepping parameter for the raw scores (default 1)
#' @param monotonuous corrects for decreasing norm scores in case of model inconsistencies (default)
#' @param CI confidence coefficient, ranging from 0 to 1, default .9
#' @param reliability coefficient, ranging between  0 to 1
#' @param pretty Format table by collapsing intervals and rounding to meaningful precision
#' @return either data.frame with raw scores and the predicted norm scores in case of simple A value or a list
#' of norm tables if vector of A values was provided
#' @seealso normTable
#' @references Eid, M. & Schmidt, K. (2012). Testtheorie und Testkonstruktion. Hogrefe.
#' @examples
#' \dontrun{
#' # Generate cnorm object from example data
#' cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
#' # generate a norm table for the raw value range from 0 to 28 for the time point month 7 of grade 3
#' table <- rawTable(3 + 7 / 12, cnorm.elfe, minRaw = 0, maxRaw = 28)
#'
#' # generate several raw tables
#' table <- rawTable(c(2.5, 3.5, 4.5), cnorm.elfe, minRaw = 0, maxRaw = 28)
#'
#' # additionally compute confidence intervals
#' table <- rawTable(c(2.5, 3.5, 4.5), cnorm.elfe, minRaw = 0, maxRaw = 28, CI = .9, reliability = .94)
#'
#' # conventional norming, set age to arbitrary value
#' model <- cnorm(raw = elfe$raw)
#' rawTable(0, model)
#' }
#'
#' @family predict
#' @export
rawTable <- function(A,
                     model,
                     minRaw = NULL,
                     maxRaw = NULL,
                     minNorm = NULL,
                     maxNorm = NULL,
                     step = 1,
                     monotonuous = TRUE,
                     CI = .9,
                     reliability = NULL,
                     pretty = TRUE) {
  if (isTaylor(model)) {
    model <- model$model
  } else if (!inherits(model, "cnormModel")) {
    stop("Please provide a cnorm object.")
  }

  if (model$useAge && !is.numeric(A)) {
    stop("Please specify age.")
  }

  if (is.null(step) || is.na(step)) {
    step <- 1
  }

  if (is.null(model)) {
    stop("No model specified")
  }

  if (is.null(minNorm) || is.na(minNorm)) {
    minNorm <- model$minL1
  }

  if (is.null(maxNorm) || is.na(maxNorm)) {
    maxNorm <- model$maxL1
  }

  if (is.null(minRaw) || is.na(minRaw)) {
    minRaw <- model$minRaw
  }

  if (is.null(maxRaw) || is.na(maxRaw)) {
    maxRaw <- model$maxRaw
  }

  if (is.null(CI) || is.na(CI)) {
    reliability <- NULL
  } else if (CI > .99999 || CI < .00001) {
    stop("Confidence coefficient (CI) out of range. Please specify value between 0 and 1.")
  }

  rel <- FALSE
  if (!is.null(reliability)) {
    if (reliability > .9999 || reliability < .0001) {
      stop("Reliability coefficient out of range. Please specify value between 0 and 1.")
    } else {
      se  <- qnorm(1 - ((1 - CI) / 2)) * sqrt(reliability * (1 - reliability))
      rel <- TRUE
    }
  }

  tables <- vector("list", length(A))

  for (x in seq_along(A)) {
    maxr <- maxRaw
    minr <- minRaw
    n_steps <- floor((maxr - minr) / step) + 1L
    norm <- vector("list", n_steps)
    raw  <- vector("list", n_steps)
    i <- 0L
    while (minr <= maxr) {
      i <- i + 1L
      n <- predictNorm(minr, A[[x]], model, minNorm, maxNorm)
      norm[[i]] <- n
      raw[[i]]  <- minr

      minr <- minr + step
    }

    table <- do.call(rbind, Map(data.frame, raw = raw, norm = norm))

    if (!is.na(model$scaleM) && !is.na(model$scaleSD)) {
      table$percentile <- pnorm((table$norm - model$scaleM) / model$scaleSD) * 100
    }

    if (model$descend) {
      table <- table[order(table$raw, decreasing = TRUE), ]
    }

    if (monotonuous) {
      table$norm <- enforce_monotone_middle(table$norm, descend = FALSE)
      if (!is.null(table$percentile)) {
        table$percentile <- pnorm((table$norm - model$scaleM) / model$scaleSD) * 100
      }
    }

    if (rel) {
      zPredicted <- reliability * (table$norm - model$scaleM) / model$scaleSD
      table$lowerCI    <- (zPredicted - se) * model$scaleSD + model$scaleM
      table$upperCI    <- (zPredicted + se) * model$scaleSD + model$scaleM
      table$lowerCI_PR <- pnorm(zPredicted - se) * 100
      table$upperCI_PR <- pnorm(zPredicted + se) * 100
    }

    if (pretty)
      tables[[x]] <- prettyPrint(table)
    else
      tables[[x]] <- table
  }

  names(tables) <- A

  if (length(A) == 1) {
    return(tables[[1]])
  } else {
    return(tables)
  }
}


#' Create a table based on first order derivative of the regression model for specific age
#'
#' In order to check model assumptions, a table of the first order derivative of the model
#' coefficients is created.
#' @param A the age
#' @param model The regression model or a cnorm object
#' @param minNorm The lower bound of the norm value range
#' @param maxNorm The upper bound of the norm value range
#' @param step Stepping parameter with lower values indicating higher precision
#' @return data.frame with norm scores and the predicted scores based on the
#' derived regression function
#' @seealso plotDerivative, derive
#' @examples
#' \dontrun{
#' # Generate cnorm object from example data
#' cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
#'
#' # retrieve function for time point 6
#' d <- derivationTable(6, cnorm.elfe, step = 0.5)
#' }
#'
#' @family predict
#' @export
derivationTable <-
  function(A,
           model,
           minNorm = NULL,
           maxNorm = NULL,
           step = 0.1) {
    if (isTaylor(model)) {
      model <- model$model
    } else if (!inherits(model, "cnormModel")) {
      stop("Please provide a cnorm object.")
    }

    if (is.null(minNorm)) {
      minNorm <- model$minL1
    }

    if (is.null(maxNorm)) {
      maxNorm <- model$maxL1
    }

    n_steps <- floor((maxNorm - minNorm) / step) + 1L
    norm <- vector("list", n_steps)
    raw  <- vector("list", n_steps)
    i <- 0L
    coeff <- derive(model)
    while (minNorm <= maxNorm) {
      i <- i + 1L
      r <- predictRaw(minNorm, A, coeff)

      norm[[i]] <- minNorm
      raw[[i]]  <- r

      minNorm <- minNorm + step
    }
    normTable <- do.call(rbind, Map(data.frame, norm = norm, raw = raw))
    return(normTable)
  }


#' Retrieve norm value for raw score at a specific age
#'
#' This functions numerically determines the norm score for raw scores depending on the
#' level of the explanatory variable A, e. g. norm scores for raw scores at given ages.
#' @param raw The raw value, either single numeric or numeric vector
#' @param A the explanatory variable (e. g. age), either single numeric or numeric vector
#' @param model The regression model or a cnorm object
#' @param minNorm The lower bound of the norm score range
#' @param maxNorm The upper bound of the norm score range
#' @param force Try to resolve missing norm scores in case of inconsistent models
#' @param silent set to TRUE to suppress messages
#' @return The predicted norm score for a raw score, either single value or vector
#' @examples
#' \dontrun{
#' # Generate cnorm object from example data
#' cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
#'
#' # return norm value for raw value 21 for grade 2, month 9
#' specificNormValue <- predictNorm(raw = 21, A = 2.75, cnorm.elfe)
#' }
#'
#' @family predict
#' @export
predictNorm <-
  function(raw,
           A,
           model,
           minNorm = NULL,
           maxNorm = NULL,
           force = TRUE,
           silent = FALSE) {
    if (!inherits(model, "cnormModel") && !inherits(model, "cnorm")) {
      stop("Please provide a cnorm object.")
    }

    if (length(raw) == 1 && is.na(raw)) {
      return(rep(NA, times = length(A)))
    } else if (!is.numeric(raw)) {
      stop("Please provide a single numeric value or a numeric vector for the raw score.")
    }

    if (length(A) == 1 && is.na(A)) {
      return(rep(NA, times = length(raw)))
    } else if (!is.numeric(A)) {
      stop("Please provide a single numeric value or a numeric vector for A.")
    }

    if (length(A) > 1 &&
        length(raw) > 1 && length(raw) != length(A)) {
      stop("A and raw need to have the same length.")
    }

    if (length(A) == 0 || length(raw) == 0) {
      return(NULL)
    }

    if (isTaylor(model)) {
      if (is.null(minNorm)) {
        minNorm <- attributes(model$data)$scaleMean - (attributes(model$data)$scaleSD * 2.5)
      }

      if (is.null(maxNorm)) {
        maxNorm <- attributes(model$data)$scaleMean + (attributes(model$data)$scaleSD * 2.5)
      }

      model <- model$model
    } else if (inherits(model, "cnormModel")) {
      if (is.null(minNorm)) {
        minNorm <- attributes(model)$scaleMean - (attributes(model)$scaleSD * 2.5)
      }

      if (is.null(maxNorm)) {
        maxNorm <- attributes(model)$scaleMean + (attributes(model)$scaleSD * 2.5)
      }
    } else {
      if (is.null(minNorm) || is.null(maxNorm)) {
        stop("ERROR: Please specify minimum and maximum norm score")
      }
    }

    if (length(raw) == 1 && length(A) == 1) {
      return(predictNormByRoots(raw, A, model, minNorm, maxNorm, force = force))
    } else if (length(raw) > 1 || length(A) > 1) {
      if (length(raw) == 1) {
        raw <- rep(raw, length(A))
      } else if (length(A) == 1) {
        A <- rep(A, length(raw))
      }

      # Hash for de-duplication of redundant lookups
      hash <- paste0(raw, "_", A)

      normTable <- na.omit(data.frame(A = A, raw = raw, hash = hash))
      normTable <- normTable[!duplicated(normTable[, "hash"]), ]

      if (nrow(normTable) > 500 && !silent) {
        cat("Retrieving norm scores, please stand by ...\n")
      }
      raw2 <- normTable$raw
      A2   <- normTable$A
      n    <- length(raw2)
      values <- rep(NA_real_, n)

      for (i in seq_len(n)) {
        v <- predictNormByRoots(raw2[[i]], A2[[i]], model, minNorm, maxNorm, force = force)
        if (length(v) == 0) {
          v <- NA
        }
        values[[i]] <- v
      }

      values <- values[match(hash, normTable$hash)]
      return(values)
    } else {
      stop(
        "Please check raw and A value. Both have to be either single values or vectors of the same length."
      )
    }
  }


#' Format raw and norm tables
#'
#' Internal helper. Takes a raw or norm table, condenses repeated intervals at
#' the bottom and top, and rounds numeric columns to a meaningful precision.
#'
#' @param table the table to format
#' @return formatted table
#' @noRd
prettyPrint <- function(table) {
  if (colnames(table)[1] == "raw")
    tab <- table[match(unique(table$norm), table$norm), ]
  else
    tab <- table[match(unique(table$raw), table$raw), ]

  if (nrow(tab) == 1)
    return(tab)

  for (i in seq_len(nrow(tab))) {
    x <- which(tab[i, 2] == table[, 2])
    if (length(x) > 1) {
      label <- paste0(table[x[1], 1], " - ", table[x[length(x)], 1])
      tab[i, 1] <- label
    }
  }

  if (colnames(table)[1] == "raw")
    tab$norm <- round(tab$norm, digits = 2)
  else
    tab$raw <- round(tab$raw, digits = 2)

  tab$percentile <- round(tab$percentile, digits = 1)

  if (!is.null(tab$lowerCI_PR))
    tab$lowerCI_PR <- round(tab$lowerCI_PR, 1)

  if (!is.null(tab$upperCI_PR))
    tab$upperCI_PR <- round(tab$upperCI_PR, 1)

  if (!is.null(tab$lowerCI))
    tab$lowerCI <- round(tab$lowerCI, 2)

  if (!is.null(tab$upperCI))
    tab$upperCI <- round(tab$upperCI, 2)

  return(tab)
}


#' Internal helper to enforce monotonicity outward from the middle,
#' robust to NA values.
#'
#' @param v numerical vector
#' @param descend logical. If `FALSE` (default) the result is monotone
#'   non-decreasing; if `TRUE`, monotone non-increasing.
#' @return the input vector with monotonicity enforced outward from the
#'   median; `NA`s are preserved in place.
#' @noRd
enforce_monotone_middle <- function(v, descend = FALSE) {
  v <- as.numeric(v)
  n <- length(v)
  if (n < 2L)
    return(v)

  # Anchor at the middle of the *non-NA* observations
  non_na_idx <- which(!is.na(v))
  if (length(non_na_idx) < 2L)
    return(v)
  mid <- non_na_idx[(length(non_na_idx) + 1L) %/% 2L]

  # NA-safe running extrema
  running <- function(x, fun) {
    out <- x
    cur <- NA_real_
    for (i in seq_along(x)) {
      if (!is.na(x[i])) {
        cur <- if (is.na(cur))
          x[i]
        else
          fun(cur, x[i])
        out[i] <- cur
      }
    }
    out
  }

  if (!descend) {
    v[mid:n] <- running(v[mid:n], max)
    if (mid > 1L)
      v[1:mid] <- rev(running(rev(v[1:mid]), min))
  } else {
    v[mid:n] <- running(v[mid:n], min)
    if (mid > 1L)
      v[1:mid] <- rev(running(rev(v[1:mid]), max))
  }
  v
}

library(shiny)
library(shinycssloaders)
library(cNORM)
library(DT)
library(ggplot2)

# Define UI for Parametric Modeling Application
shinyUI(fluidPage(

  # Set tab title
  title = "cNORM - Parametric Modeling",

  # Tabsetpanel for main workflow
  tabsetPanel(

    # ------------------------------------------------------------------
    # Tab 1: Data Input
    # ------------------------------------------------------------------
    tabPanel(
      "Data Input",
      sidebarLayout(
        sidebarPanel(
          width = 3,
          tags$h3("Load Data"),

          tags$p("Please choose a data set for parametric modeling. You can use a built-in example or load your own file:"),

          selectizeInput(
            "Example",
            label = "Example:",
            choices = c("", "elfe", "ppvt", "CDC"),
            selected = character(0),
            multiple = FALSE
          ),

          hr(),

          fileInput(
            "file",
            "Choose a file",
            multiple = FALSE,
            accept = c(".csv", ".xlsx", ".xls", ".rda", ".sav")
          ),

          tags$p(tags$b("Note: If you upload a file, the selected example will be overridden.")),

          hr(),

          tags$h4("Built-in Examples:"),
          tags$ul(
            tags$li(tags$b("elfe:"), "Reading comprehension test (ages 6-10, scores 0-28)"),
            tags$li(tags$b("ppvt:"), "Peabody Picture Vocabulary Test (ages 2.5-17.5, scores 0-228)"),
            tags$li(tags$b("CDC:"), "Growth data from CDC (ages 2-18, BMI/weight/height)")
          )
        ),

        mainPanel(
          width = 9,
          tags$h3("Data Preview"),
          tags$p("With this app, you can model norm scores like percentiles, T scores and IQ scores for psychometric, biometric or
                 other data. It uses parametric modelling techniques like the beta-binomial distribution or the SinH-ArcsinH distribution
                 to derive continuous norms in dependence of an explanatory variable like age. After loading data, review it below before
                 proceeding to modeling. To fit the models, please select the 'Modelling' tab. To generate norm tables, please select 'Norm tables'.",
                 tags$br(), tags$br(),
                 "Further information on the methodology can be found at",
                 tags$a(href = "https://www.psychometrica.de/cNorm_betabinomial_en.html", "Psychometrica", target = "_blank"),
                 "and in the package vignettes on modelling with the ",
                 tags$a(href = "https://cran.r-project.org/web/packages/cNORM/vignettes/BetaBinomial.html", "Beta-Binomial Distribution", target = "_blank"),
                 "or the",
                 tags$a(href = "https://cran.r-project.org/web/packages/cNORM/vignettes/sinh.html", "SinH-ArcsinH Distribution (SHASH)", target = "_blank"),
                 "on CRAN."),
          withSpinner(DT::DTOutput("dataPreview"), type = 5)
        )
      )
    ),

    # ------------------------------------------------------------------
    # Tab 2: Modeling
    # ------------------------------------------------------------------
    tabPanel(
      "Modeling",
      sidebarLayout(
        sidebarPanel(
          width = 3,
          tags$h3("Model Configuration"),

          # Variable selection
          uiOutput("ageVariable"),
          uiOutput("scoreVariable"),
          uiOutput("weightVariable"),

          hr(),

          # Distribution type
          selectInput(
            "distributionType",
            "Distribution Type:",
            choices = c("SinH-ArcSinH (SHASH)" = "shash",
                        "Beta-Binomial"        = "betabinomial"),
            selected = "shash"
          ),

          hr(),

          # Conditional UI for SHASH parameters
          conditionalPanel(
            condition = "input.distributionType == 'shash'",
            tags$h4("SHASH Parameters"),
            tags$p("Specify polynomial degrees for each distribution parameter:"),

            sliderInput("mu_degree",      "Degree of \u03BC (location):",
                        min = 0, max = 5, value = 3, step = 1),
            sliderInput("sigma_degree",   "Degree of \u03C3 (scale):",
                        min = 0, max = 5, value = 2, step = 1),
            sliderInput("epsilon_degree", "Degree of \u03B5 (skewness):",
                        min = 0, max = 5, value = 2, step = 1),
            sliderInput("delta_degree",   "Degree of \u03B4 (tail weight):",
                        min = 0, max = 5, value = 1, step = 1),

            checkboxInput("fix_delta", "Fix delta (instead of polynomial)?", value = FALSE),

            conditionalPanel(
              condition = "input.fix_delta == true",
              numericInput("delta_value", "Fixed delta value:",
                           value = 1, min = 0.1, max = 5, step = 0.1)
            )
          ),

          # Conditional UI for Beta-Binomial parameters
          conditionalPanel(
            condition = "input.distributionType == 'betabinomial'",
            tags$h4("Beta-Binomial Parameters"),
            tags$p("Specify polynomial degrees for alpha and beta:"),

            sliderInput("alpha_degree", "Degree of \u03B1:",
                        min = 0, max = 5, value = 3, step = 1),
            sliderInput("beta_degree",  "Degree of \u03B2:",
                        min = 0, max = 5, value = 3, step = 1)
          ),

          hr(),

          # Scale selection
          selectInput(
            "scale",
            "Norm Scale:",
            choices = c("T-scores (M=50, SD=10)"  = "T",
                        "IQ-scores (M=100, SD=15)" = "IQ",
                        "z-scores (M=0, SD=1)"    = "z"),
            selected = "T"
          ),

          hr(),

          # Action buttons
          actionButton(
            "computeModel",
            "Compute Model",
            class = "btn-primary",
            style = "width: 100%;"
          ),

          br(), br(),

          downloadButton(
            "downloadModel",
            "Download Model",
            style = "width: 100%;"
          )
        ),

        mainPanel(
          width = 9,

          tags$h3("Model Results"),

          conditionalPanel(
            condition = "input.computeModel == 0",
            tags$div(
              style = "padding: 20px; background-color: #f8f9fa; border-radius: 5px;",
              tags$p(
                style = "font-size: 16px;",
                "Configure the model parameters in the sidebar and click 'Compute Model' to fit the parametric distribution."
              )
            )
          ),

          conditionalPanel(
            condition = "input.computeModel > 0",
            tabsetPanel(
              id = "modelTabs",
              tabPanel(
                "Percentile Plot",
                tags$br(),
                tags$p("Observed percentiles (points) vs. model predictions (lines). Wavy lines indicate overfit. Reduce polynomial degrees in this case."),
                tags$p(""),
                withSpinner(
                  plotOutput("percentilePlot", width = "100%", height = "700px"),
                  type = 5
                )
              ),
              tabPanel(
                "Summary",
                tags$br(),
                withSpinner(verbatimTextOutput("modelSummary"), type = 5)
              )
            )
          )
        )
      )
    ),

    # ------------------------------------------------------------------
    # Tab 3: Norm Tables
    # ------------------------------------------------------------------
    tabPanel(
      "Norm Tables",
      sidebarLayout(
        sidebarPanel(
          width = 3,
          tags$h3("Norm Table Configuration"),

          tags$p("Generate norm tables with confidence intervals based on the fitted model."),

          # Age specification
          textInput(
            "norm_ages",
            "Ages (comma or space separated):",
            value = ""
          ),

          tags$p(tags$small("Example: 7, 8, 9, 10 or 7 8 9 10")),

          hr(),

          # Score range - only relevant for SHASH (continuous distribution)
          conditionalPanel(
            condition = "input.distributionType == 'shash'",
            numericInput("norm_start", "Start raw score:", value = NULL, step = 1),
            numericInput("norm_end",   "End raw score:",   value = NULL, step = 1),
            numericInput("norm_step",  "Step size:",       value = 1, min = 0.1, step = 0.1)
          ),

          # Note for Beta-Binomial
          conditionalPanel(
            condition = "input.distributionType == 'betabinomial'",
            tags$div(
              style = "padding: 10px; background-color: #f8f9fa; border-left: 4px solid #0275d8; border-radius: 3px;",
              tags$p(tags$small(
                tags$b("Note:"), " For the Beta-Binomial distribution, the norm table is generated automatically
                across the entire item range (0 to n). Start/End/Step are not required."
              ))
            ),
            # Optional: filtering inputs (will simply trim the resulting table)
            tags$br(),
            numericInput("norm_start", "Optional: restrict table to start raw score:",
                         value = NULL, step = 1),
            numericInput("norm_end",   "Optional: restrict table to end raw score:",
                         value = NULL, step = 1),
            # Hidden default - keeps server reactive happy
            tags$div(style = "display: none;",
                     numericInput("norm_step", "step", value = 1))
          ),

          hr(),

          # Confidence intervals
          checkboxInput(
            "include_ci",
            "Include confidence intervals",
            value = FALSE
          ),

          conditionalPanel(
            condition = "input.include_ci == true",
            sliderInput(
              "ci_level",
              "Confidence level:",
              min = 0.80,
              max = 0.99,
              value = 0.90,
              step = 0.01
            ),
            numericInput(
              "reliability",
              "Reliability coefficient:",
              value = 0.90,
              min = 0.01,
              max = 0.99,
              step = 0.01
            )
          ),

          hr(),

          actionButton(
            "generateTables",
            "Generate Tables",
            class = "btn-primary",
            style = "width: 100%;"
          ),

          br(), br(),

          downloadButton(
            "downloadTables",
            "Download Tables (CSV)",
            style = "width: 100%;"
          )
        ),

        mainPanel(
          width = 9,
          tags$h3("Norm Tables"),

          conditionalPanel(
            condition = "input.generateTables == 0",
            tags$div(
              style = "padding: 20px; background-color: #f8f9fa; border-radius: 5px;",
              tags$p(
                style = "font-size: 16px;",
                "Ensure a model has been computed, then configure the table parameters and click 'Generate Tables'."
              )
            )
          ),

          conditionalPanel(
            condition = "input.generateTables > 0",
            uiOutput("normTablesUI")
          )
        )
      )
    )
  )
))

#' Internal function for retrieving regression function coefficients at specific age
#'
#' The function is an inline for searching zeros in the inverse regression
#' function. It collapses the regression function at a specific age and simplifies
#' the coefficients.
#' @param raw The raw value (subtracted from the intercept)
#' @param age The age
#' @param model The cNORM regression model
#'
#' @return The coefficients
calcPolyInL <- function(raw, age, model) {
  k <- model$k
  coeff <- model$coefficients
  return(calcPolyInLBase2(raw, age, coeff, k))
}


#' Internal function for retrieving regression function coefficients at specific
#' age
#'
#' The function is an inline for searching zeros in the inverse regression
#' function. It collapses the regression function at a specific age and
#' simplifies the coefficients. Optimized version of the prior 'calcPolyInLBase'
#' @param raw The raw value (subtracted from the intercept)
#' @param age The age
#' @param coeff The cNORM regression model coefficients
#' @param k The cNORM regression model power parameter
#'
#' @return The coefficients
calcPolyInLBase2 <- function(raw, age, coeff, k) {
  nam <- names(coeff)
  coeff <- as.numeric(coeff)

  # use regex to identify powers of A
  positionsA <- regexpr("A\\d", nam)
  positionsA[positionsA == -1] <- 0
  powerA <- as.numeric(gsub("A", "", regmatches(nam, positionsA)))
  powerA[is.na(powerA)] <- 0

  # modify coefficients by powers of A
  coeff <- coeff * (age^powerA)

  # use regex to identify powers of L
  positionsL <- rep("", length(nam))
  indices <- grep("^L", nam)
  positionsL[indices] <- substr(nam[indices], start=2, stop=2)
  positionsL[positionsL==""] <- "0"
  positionsL <- as.numeric(positionsL)
  coefficients <- rep(0, k + 1)

  # iterate through coefficients
  for(j in 0:k)
    coefficients[j + 1] <- sum(coeff[positionsL==j])

  coefficients[1] <-  coefficients[1] - raw
  return(coefficients)
}


predictNormByRoots <- function(raw, age, model,
                                minNorm, maxNorm,
                                polynom = NULL,
                                force = TRUE,
                                covariate = NULL) {

  if (!is.null(covariate)) {
    if (is.null(model$coefficients)) {
      stop("Covariate specified, but model does not include covariate")
    }
    # model$coefficients <- simplifyCoefficients(model$coefficients, covariate)
  }

  # Build the polynomial in L (norm) at the requested age.
  if (is.null(polynom)) {
    polynomForPrediction <- calcPolyInLBase2(
      raw   = raw,
      age   = age,
      coeff = model$coefficients,
      k     = model$k
    )
  } else {
    polynomForPrediction    <- polynom
    polynomForPrediction[1] <- polynomForPrediction[1] - raw
  }

  # Guard against degenerate polynomials (e.g. intercept-only models).
  if (length(polynomForPrediction) < 2L ||
      all(polynomForPrediction[-1L] == 0)) {
    return(NA_real_)
  }

  # Real roots only.
  roots  <- polyroot(polynomForPrediction)
  output <- Re(roots[abs(Im(roots)) < 1e-7])

  # Single in-range real root: done.
  if (length(output) == 1L && output >= minNorm && output <= maxNorm) {
    return(output)
  }

  # Direction of search: which side of the model's prediction at scaleM
  # is `raw` on?  IMPORTANT: do NOT clip, otherwise the comparison flips
  # near the clipping boundary.
  raw_at_scaleM <- predictRaw(model$scaleM, age, model$coefficients)

  if (raw > raw_at_scaleM) {
    output <- output[output > model$scaleM & output <= maxNorm]
  } else if (raw < raw_at_scaleM) {
    output <- output[output < model$scaleM & output >= minNorm]
  } else {
    return(model$scaleM)
  }

  if (length(output) == 1L) {
    return(output)
  } else if (length(output) > 1L) {
    # Multiple plausible roots: take the one closest to scaleM.
    return(output[which.min((output - model$scaleM)^2)])
  }

  # ------------------------------------------------------------------
  # Fallback: numerical search.  optimize() always returns a value in
  # [minNorm, maxNorm], so we must judge success by the *objective*,
  # not by the position.
  # ------------------------------------------------------------------
  functionToMinimize <- function(norm) {
    currentRawValue <- predictRaw(norm = norm, age = age,
                                  coefficients = model$coefficients)
    (currentRawValue - raw)^2
  }

  optimum <- optimize(functionToMinimize,
                      lower = minNorm, upper = maxNorm,
                      tol   = .Machine$double.eps^0.5)

  # Tolerance on the residual raw score.  Anything below ~1e-3 is a
  # genuine root for any realistic raw-score scale.
  raw_residual_tol <- 1e-3

  if (sqrt(optimum$objective) <= raw_residual_tol) {
    return(optimum$minimum)
  }

  # The raw score is not reachable on [minNorm, maxNorm] at this age.
  msg <- paste0("No plausible norm score available for raw = ", raw,
                " at age ", age, "; ")

  if (!force) {
    warning(msg, "returning NA.")
    return(NA_real_)
  }

  # force = TRUE: clip to the bound on the side of the distribution
  # that the raw score lies on.
  if (raw > raw_at_scaleM) {
    return(maxNorm)
  } else {
    return(minNorm)
  }
}

# ==========================================================================
# Beta-binomial continuous norming (parametric part of cNORM)
# ==========================================================================


# --------------------------------------------------------------------------
# Internal helpers
# --------------------------------------------------------------------------

#' Build a raw polynomial design matrix with intercept
#'
#' @param x A numeric vector.
#' @param degree Degree of the raw polynomial.
#' @return A numeric matrix with columns 1, x, x^2, ..., x^degree.
#' @keywords internal
bb_design_matrix <- function(x, degree) {
  cbind(1, outer(x, seq_len(degree), `^`))
}

#' Resolve a norm scale specification
#'
#' @param scale Either a character ("T", "IQ", "z", anything else = percentile)
#'   or a numeric vector of length 2 with mean and sd.
#' @return A numeric vector c(mean, sd); c(NA, NA) indicates percentile output.
#' @keywords internal
bb_resolve_scale <- function(scale) {
  if (is.numeric(scale) && length(scale) == 2) {
    return(c(scale[1], scale[2]))
  }
  if (is.character(scale) && length(scale) == 1) {
    return(switch(scale,
                  "IQ" = c(100, 15),
                  "z"  = c(0, 1),
                  "T"  = c(50, 10),
                  c(NA_real_, NA_real_)))
  }
  c(NA_real_, NA_real_)
}

#' Validate and clean age / score / weights input
#'
#' Performs the shared input validation for the beta-binomial fitting
#' functions: length checks, removal of non-finite cases, weight validation,
#' and checks for negative or non-integer scores as well as degenerate
#' predictor variance.
#'
#' @param age Numeric vector of predictor values.
#' @param score Numeric vector of response values.
#' @param weights Optional numeric vector of weights.
#' @return A list with the (possibly filtered) elements age, score, weights.
#' @keywords internal
bb_prepare_data <- function(age, score, weights = NULL) {
  if (length(age) != length(score)) {
    stop("Length of 'age' and 'score' must be the same.")
  }
  if (!is.null(weights) && length(weights) != length(age)) {
    stop("Length of 'weights' must match length of 'age' and 'score'.")
  }

  # reject invalid weights early
  if (!is.null(weights)) {
    if (any(weights < 0, na.rm = TRUE)) {
      stop("'weights' must be non-negative.")
    }
    if (all(weights == 0, na.rm = TRUE)) {
      stop("All 'weights' are zero. The model cannot be fitted.")
    }
  }

  vectors_to_check <- list(age = age, score = score)
  if (!is.null(weights)) {
    vectors_to_check$weights <- weights
  }

  needs_filtering <- any(vapply(vectors_to_check,
                                function(x) any(!is.finite(x)),
                                logical(1)))

  if (needs_filtering) {
    message("Vector(s) contained non-finite values (NA, NaN, Inf). These cases will be removed.")
    tmp <- do.call(filter_complete, c(vectors_to_check, verbose = FALSE))
    age <- tmp[[1]]
    score <- tmp[[2]]
    if (!is.null(weights)) {
      weights <- tmp[[3]]
    }
  }

  # Guard against degenerate input after filtering.
  if (length(age) < 2) {
    stop("Insufficient valid observations after removing non-finite cases.")
  }
  if (sd(age) == 0) {
    stop("'age' has zero variance. Age-dependent modelling requires ",
         "variability in the predictor.")
  }

  if (any(score < 0)) {
    stop("'score' contains negative values. ",
         "Beta-binomial modelling requires positive integers (including zero). ",
         "Please consider using Taylor polynomials (function 'cnorm') or ",
         "SinusH-ArcsinH distributions (function 'cnorm.shash') instead, ",
         "or transform your data to positive integers.")
  }

  if (any(score != floor(score))) {
    stop("'score' contains non-integer values. ",
         "Beta-binomial modelling requires positive integers (including zero). ",
         "Please consider using Taylor polynomials (function 'cnorm') or ",
         "SinusH-ArcsinH distributions (function 'cnorm.shash') instead, ",
         "or transform your data to positive integers.")
  }

  list(age = age, score = score, weights = weights)
}

#' Beta-binomial distribution on the full support 0:n
#'
#' Computes the (normalized) probability mass function, the cumulative
#' distribution function and mid-p percentiles of a beta-binomial
#' distribution in a numerically stable way (log space).
#'
#' @param a Alpha parameter (> 0).
#' @param b Beta parameter (> 0).
#' @param n Number of trials / maximum score.
#' @return A list with elements x (support 0:n), Px (pmf), cum (cdf) and
#'   perc (mid-p percentiles, i.e. P(X < x) + 0.5 * P(X = x)).
#' @keywords internal
bb_distribution <- function(a, b, n) {
  x <- 0:n
  if (!is.finite(a) || !is.finite(b) || a <= 0 || b <= 0) {
    na <- rep(NA_real_, n + 1)
    return(list(x = x, Px = na, cum = na, perc = na))
  }

  log_pmf <- lchoose(n, x) + lbeta(x + a, n - x + b) - lbeta(a, b)
  Px <- exp(log_pmf)
  s <- sum(Px)
  if (!is.finite(s) || s <= 0) {
    na <- rep(NA_real_, n + 1)
    return(list(x = x, Px = na, cum = na, perc = na))
  }
  Px <- Px / s               # normalize to guard against numerical drift
  cum <- cumsum(Px)
  cum[n + 1] <- 1            # enforce exact upper bound
  perc <- cum - 0.5 * Px     # mid-p percentiles

  list(x = x, Px = Px, cum = cum, perc = perc)
}


# --------------------------------------------------------------------------
# Likelihood functions and analytic gradients
# --------------------------------------------------------------------------

#' Calculate the negative log-likelihood for a beta binomial regression model
#'
#' This function computes the negative log-likelihood for a beta binomial
#' regression model where both the mean and standard deviation are modeled
#' as functions of predictors (mode 1, normal approximation on standardized
#' scores).
#'
#' @param params A numeric vector containing all model parameters. The first
#'   n_beta elements are coefficients for the mean model, and the remaining
#'   elements are coefficients for the log-standard deviation model.
#' @param X A matrix of predictors for the mean model.
#' @param Z A matrix of predictors for the log-standard deviation model.
#' @param y A numeric vector of response values.
#' @param weights A numeric vector of weights for each observation
#'   (NULL = equal weights).
#'
#' @return The negative log-likelihood of the model (large finite penalty
#'   if non-finite).
#' @keywords internal
log_likelihood <- function(params, X, Z, y, weights = NULL) {
  n_beta <- ncol(X)
  beta <- params[seq_len(n_beta)]
  gamma <- params[(n_beta + 1):length(params)]

  mu <- drop(X %*% beta)
  # Clamp the linear predictor of log(sigma) to keep exp() finite and to
  # stabilize BFGS line searches far away from the optimum
  log_sigma <- pmin(pmax(drop(Z %*% gamma), -20), 20)
  sigma <- exp(log_sigma)

  ll_i <- dnorm(y, mean = mu, sd = sigma, log = TRUE)
  ll_i[!is.finite(ll_i)] <- -709  # approx. log(.Machine$double.xmin)

  ll <- if (is.null(weights)) sum(ll_i) else sum(weights * ll_i)

  if (!is.finite(ll)) {
    return(1e10)  # large but finite penalty
  }
  -ll
}

#' Analytic gradient of the negative log-likelihood (mode 1)
#'
#' Computes the exact gradient of \code{\link{log_likelihood}} with respect
#' to all regression coefficients. With \eqn{r = (y - \mu)/\sigma}:
#' \deqn{\partial\ell/\partial\mu = r/\sigma, \qquad
#'       \partial\ell/\partial\log\sigma = r^2 - 1.}
#' Coefficient gradients follow via the chain rule through the design
#' matrices. Contributions of observations whose log(sigma) linear predictor
#' is clamped (see \code{log_likelihood}) are zeroed, which is the exact
#' subgradient of the clamped objective. Non-finite per-observation terms
#' (matching the penalty branch of the objective) are zeroed as well.
#'
#' @inheritParams log_likelihood
#' @return Numeric vector: gradient of the negative log-likelihood with
#'   respect to \code{params}.
#' @keywords internal
gradient_log_likelihood <- function(params, X, Z, y, weights = NULL) {
  n_beta <- ncol(X)
  beta <- params[seq_len(n_beta)]
  gamma <- params[(n_beta + 1):length(params)]

  mu <- drop(X %*% beta)
  eta_sigma <- drop(Z %*% gamma)
  sigma_free <- (eta_sigma > -20) & (eta_sigma < 20)
  sigma <- exp(pmin(pmax(eta_sigma, -20), 20))

  r <- (y - mu) / sigma
  d_mu <- r / sigma            # dl/dmu
  d_lsig <- r * r - 1          # dl/d(log sigma)

  # Zero out numerically failed observations (consistent with the
  # floored/penalized branch of the objective) and clamped predictors
  ok <- is.finite(d_mu) & is.finite(d_lsig)
  d_mu[!ok] <- 0
  d_lsig[!ok] <- 0
  d_lsig <- d_lsig * sigma_free

  w <- if (is.null(weights)) 1 else weights
  grad <- c(-drop(crossprod(X, w * d_mu)),
            -drop(crossprod(Z, w * d_lsig)))
  grad[!is.finite(grad)] <- 0
  grad
}

#' Calculate the negative log-likelihood for a beta-binomial regression model
#'
#' This function computes the negative log-likelihood for a beta-binomial
#' regression model where both the alpha and beta parameters are modeled as
#' functions of predictors (mode 2).
#'
#' @param params A numeric vector containing all model parameters. The first
#'   n_alpha elements are coefficients for the alpha model, and the remaining
#'   elements are coefficients for the beta model.
#' @param X A matrix of predictors for the alpha model.
#' @param Z A matrix of predictors for the beta model.
#' @param y A numeric vector of response values.
#' @param n The maximum score (number of trials).
#' @param weights A numeric vector of weights for each observation
#'   (NULL = equal weights).
#' @param lch Optional precomputed \code{lchoose(n, y)}. Since this term does
#'   not depend on the parameters, passing it once avoids recomputation in
#'   every optimizer iteration.
#'
#' @return The negative log-likelihood of the model (large finite penalty
#'   if non-finite).
#'
#' @details
#' Uses a numerically stable implementation of the beta-binomial
#' log-probability via \code{lbeta}. The linear predictors of
#' log(alpha) and log(beta) are clamped to [-20, 20].
#'
#' @keywords internal
log_likelihood2 <- function(params, X, Z, y, n, weights = NULL, lch = NULL) {
  n_alpha <- ncol(X)
  alpha_coef <- params[seq_len(n_alpha)]
  beta_coef <- params[(n_alpha + 1):length(params)]

  log_alpha <- pmin(pmax(drop(X %*% alpha_coef), -20), 20)
  log_beta <- pmin(pmax(drop(Z %*% beta_coef), -20), 20)

  alpha <- exp(log_alpha)
  beta <- exp(log_beta)

  if (is.null(lch)) {
    lch <- lchoose(n, y)
  }

  logp <- lch + lbeta(y + alpha, n - y + beta) - lbeta(alpha, beta)
  logp[!is.finite(logp)] <- -709  # approx. log(.Machine$double.xmin)

  ll <- if (is.null(weights)) sum(logp) else sum(weights * logp)

  if (!is.finite(ll)) {
    return(1e10)  # large but finite penalty
  }
  -ll
}

#' Analytic gradient of the negative log-likelihood (mode 2)
#'
#' Computes the exact gradient of \code{\link{log_likelihood2}} with respect
#' to all regression coefficients. With the per-observation log-probability
#' \deqn{\log p = \mathrm{lchoose}(n, y) + \mathrm{lbeta}(y+\alpha, n-y+\beta)
#'   - \mathrm{lbeta}(\alpha, \beta),}
#' the partial derivatives on the natural scale are (with digamma function
#' \eqn{\psi}):
#' \deqn{\partial \log p/\partial\alpha = \psi(y+\alpha) - \psi(\alpha)
#'   + \psi(\alpha+\beta) - \psi(n+\alpha+\beta)}
#' \deqn{\partial \log p/\partial\beta = \psi(n-y+\beta) - \psi(\beta)
#'   + \psi(\alpha+\beta) - \psi(n+\alpha+\beta)}
#' The chain rule through the log link multiplies by \eqn{\alpha} resp.
#' \eqn{\beta}; coefficient gradients follow as
#' \eqn{-X^\top(w \cdot \partial\ell/\partial\eta)} for the negative
#' log-likelihood. Contributions of observations whose linear predictor is
#' clamped (see \code{log_likelihood2}) are zeroed, which is the exact
#' subgradient of the clamped objective.
#'
#' Supplying this gradient to \code{\link[stats]{optim}} (L-BFGS-B) replaces
#' the finite-difference approximation, which required
#' \code{length(params) + 1} likelihood evaluations per gradient.
#'
#' @inheritParams log_likelihood2
#' @return Numeric vector: gradient of the negative log-likelihood with
#'   respect to \code{params}.
#' @keywords internal
gradient_log_likelihood2 <- function(params, X, Z, y, n,
                                     weights = NULL, lch = NULL) {
  # Note: 'lch' is accepted (optim passes identical ... to fn and gr)
  # but not needed, since the term is parameter-independent.
  n_alpha <- ncol(X)
  alpha_coef <- params[seq_len(n_alpha)]
  beta_coef <- params[(n_alpha + 1):length(params)]

  eta_alpha <- drop(X %*% alpha_coef)
  eta_beta <- drop(Z %*% beta_coef)
  alpha_free <- (eta_alpha > -20) & (eta_alpha < 20)
  beta_free <- (eta_beta > -20) & (eta_beta < 20)

  alpha <- exp(pmin(pmax(eta_alpha, -20), 20))
  beta <- exp(pmin(pmax(eta_beta, -20), 20))

  ab <- alpha + beta
  dig_ab <- digamma(ab)
  dig_nab <- digamma(n + ab)

  # dl/d(log alpha), dl/d(log beta) via chain rule through the log link
  d_lalpha <- alpha * (digamma(y + alpha) - digamma(alpha) + dig_ab - dig_nab)
  d_lbeta  <- beta  * (digamma(n - y + beta) - digamma(beta) + dig_ab - dig_nab)

  # Zero out numerically failed observations (consistent with the floored
  # branch of the objective) and clamped predictors (exact subgradient)
  ok <- is.finite(d_lalpha) & is.finite(d_lbeta)
  d_lalpha[!ok] <- 0
  d_lbeta[!ok] <- 0
  d_lalpha <- d_lalpha * alpha_free
  d_lbeta <- d_lbeta * beta_free

  w <- if (is.null(weights)) 1 else weights
  grad <- c(-drop(crossprod(X, w * d_lalpha)),
            -drop(crossprod(Z, w * d_lbeta)))
  grad[!is.finite(grad)] <- 0
  grad
}


# --------------------------------------------------------------------------
# Model fitting
# --------------------------------------------------------------------------

#' Fit a beta binomial regression model (mode 1, mu/sigma parameterization)
#'
#' This function fits a beta binomial regression model where both the mean and
#' standard deviation of the response variable are modeled as polynomial
#' functions of the predictor variable. While 'cnorm.betabinomial2' fits a
#' beta-binomial model directly on the basis of \eqn{\alpha} and \eqn{\beta},
#' this function fits \eqn{\mu} and \eqn{\sigma}, which are then used to
#' estimate the beta binomial distribution parameters. Kept for backwards
#' compatibility; 'cnorm.betabinomial2' is the recommended default.
#'
#' @param age A numeric vector of predictor values (e.g., age).
#' @param score A numeric vector of response values.
#' @param n Number of items in the test, resp. maximum score to be achieved.
#' @param weights A numeric vector of weights for each observation.
#'   Default is NULL (equal weights).
#' @param mu Integer specifying the degree of the polynomial for the mean
#'   model. Default is 3.
#' @param sigma Integer specifying the degree of the polynomial for the
#'   standard deviation model. Default is 3.
#' @param control A list of control parameters to be passed to the
#'   \code{optim} function. Any parameters not specified are filled with
#'   the defaults \code{list(reltol = 1e-8, maxit = 1000)}.
#' @param scale Type of norm scale, either T (default), IQ, z or percentile
#'   (= no transformation); a double vector with the mean and standard
#'   deviation can be provided as well, e.g. c(10, 3) for Wechsler scale
#'   index points.
#' @param plot Logical indicating whether to plot the model. Default is TRUE.
#'
#' @return A list of class "cnormBetaBinomial" containing:
#'   \item{beta_est}{Estimated coefficients for the mean model}
#'   \item{gamma_est}{Estimated coefficients for the log-standard deviation model}
#'   \item{se}{Standard errors of the estimated coefficients}
#'   \item{mu}{Degree of the polynomial for the mean model}
#'   \item{sigma}{Degree of the polynomial for the standard deviation model}
#'   \item{result}{Full result from the optimization procedure}
#'
#' @details
#' The function standardizes the input variables, fits polynomial models for
#' both the mean and standard deviation, and uses maximum likelihood
#' estimation with analytic gradients to find the optimal parameters. The
#' optimization is performed using the BFGS method.
#'
#' @keywords internal
cnorm.betabinomial1 <- function(age,
                                score,
                                n = NULL,
                                weights = NULL,
                                mu = 3,
                                sigma = 3,
                                control = NULL,
                                scale = "T",
                                plot = TRUE) {
  # Shared input validation and cleaning
  cleaned <- bb_prepare_data(age, score, weights)
  age <- cleaned$age
  score <- cleaned$score
  weights <- cleaned$weights

  # Resolve and validate n before fitting.
  if (is.null(n)) {
    n <- max(score)
  }
  if (max(score) > n) {
    stop("'score' contains values larger than n = ", n, ".")
  }

  # Guard against zero score variance (log(sd) = -Inf start value)
  if (sd(score) == 0) {
    stop("'score' has zero variance. The model cannot be fitted.")
  }

  # Standardize inputs
  age_std <- standardize(age)
  score_std <- standardize(score)

  # Design matrices (raw polynomials incl. intercept)
  X <- bb_design_matrix(age_std, mu)
  Z <- bb_design_matrix(age_std, sigma)
  y <- score_std

  # Sensible starting values on the standardized scale
  initial_params <- c(mean(y), rep(0, mu), log(sd(y)), rep(0, sigma))

  # Merge user control with defaults instead of all-or-nothing
  control_defaults <- list(reltol = 1e-8, maxit = 1000)
  control <- if (is.null(control))
    control_defaults
  else
    utils::modifyList(control_defaults, control)

  result <- optim(
    initial_params,
    log_likelihood,
    gr = gradient_log_likelihood,
    X = X,
    Z = Z,
    y = y,
    weights = weights,
    method = "BFGS",
    hessian = TRUE,
    control = control
  )

  if (result$convergence != 0) {
    warning("Optimization did not converge (code: ", result$convergence,
            "). Consider adjusting control parameters.")
  }

  # Store the analytic gradient at the solution; consumed by
  # diagnostics.betabinomial as a first-order optimality check.
  result$gradient <- gradient_log_likelihood(result$par, X = X, Z = Z,
                                             y = y, weights = weights)

  # Extract results and calculate standard errors
  beta_est <- result$par[1:(mu + 1)]
  gamma_est <- result$par[(mu + 2):length(result$par)]

  se <- tryCatch({
    d <- diag(solve(result$hessian))
    if (any(d < 0, na.rm = TRUE)) {
      warning("Hessian is not positive definite; standard errors are set to NA where invalid.")
      d[d < 0] <- NA_real_
    }
    sqrt(d)
  }, error = function(e) {
    warning("Could not compute standard errors: Hessian matrix issue")
    rep(NA_real_, length(result$par))
  })

  # Resolve norm scale
  scaleMSD <- bb_resolve_scale(scale)

  # Store attributes for usage in other functions
  attr(result, "age_mean") <- mean(age)
  attr(result, "age_sd") <- sd(age)
  attr(result, "ageMin") <- min(age)
  attr(result, "ageMax") <- max(age)
  attr(result, "score_mean") <- mean(score)
  attr(result, "score_sd") <- sd(score)
  attr(result, "max") <- n
  attr(result, "N") <- length(score)
  attr(result, "scaleMean") <- scaleMSD[1]
  attr(result, "scaleSD") <- scaleMSD[2]

  model <- list(
    beta_est = beta_est,
    gamma_est = gamma_est,
    se = se,
    mu = mu,
    sigma = sigma,
    result = result
  )
  class(model) <- "cnormBetaBinomial"

  if (plot) {
    p <- plot.cnormBetaBinomial(model, age, score, weights = weights)
    print(p)
  }

  model
}

#' Fit a beta-binomial regression model for continuous norming (mode 2)
#'
#' This function fits a beta-binomial regression model where both the alpha
#' and beta parameters of the beta-binomial distribution are modeled as
#' polynomial functions of the predictor variable (typically age). While
#' 'cnorm.betabinomial1' fits a beta-binomial model on the basis of \eqn{\mu}
#' and \eqn{\sigma}, this function fits the model directly on the basis of
#' \eqn{\alpha} and \eqn{\beta}. This is the recommended default approach
#' (see \code{\link{cnorm.betabinomial}}).
#'
#' @param age A numeric vector of predictor values (e.g., age).
#' @param score A numeric vector of response values.
#' @param n The maximum score (number of trials in the beta-binomial
#'   distribution). If NULL, max(score) is used.
#' @param weights A numeric vector of weights for each observation.
#'   Default is NULL (equal weights).
#' @param alpha_degree Integer specifying the degree of the polynomial for
#'   the alpha model. Default is 3.
#' @param beta_degree Integer specifying the degree of the polynomial for
#'   the beta model. Default is 3.
#' @param control A list of control parameters to be passed to the
#'   \code{optim} function (method "L-BFGS-B"). Any parameters not specified
#'   are filled with adaptive defaults based on problem size.
#' @param scale Type of norm scale, either "T" (default), "IQ", "z" or a
#'   double vector with the mean and standard deviation.
#' @param plot Logical indicating whether to plot the model. Default is TRUE.
#'
#' @return A list of class "cnormBetaBinomial2" containing:
#'   \item{alpha_est}{Estimated coefficients for the alpha model}
#'   \item{beta_est}{Estimated coefficients for the beta model}
#'   \item{se}{Standard errors of the estimated coefficients}
#'   \item{alpha_degree}{Degree of the polynomial for the alpha model}
#'   \item{beta_degree}{Degree of the polynomial for the beta model}
#'   \item{result}{Full result from the optimization procedure}
#'
#' @details
#' The function standardizes the input variables, fits polynomial models for
#' both the alpha and beta parameters, and uses maximum likelihood estimation
#' with analytic (digamma-based) gradients to find the optimal parameters.
#' The optimization is performed using the L-BFGS-B method with parameter
#' bounds to prevent numerical issues.
#'
#' @keywords internal
cnorm.betabinomial2 <- function(age,
                                score,
                                n = NULL,
                                weights = NULL,
                                alpha_degree = 3,
                                beta_degree = 3,
                                control = NULL,
                                scale = "T",
                                plot = TRUE) {
  # Shared input validation and cleaning
  cleaned <- bb_prepare_data(age, score, weights)
  age <- cleaned$age
  score <- cleaned$score
  weights <- cleaned$weights

  if (is.null(n)) {
    n <- max(score)
    message("Using max(score) = ", n, " as the maximum score.")
  }

  if (max(score) > n) {
    stop("'score' contains values larger than n = ", n, ".")
  }

  # Standardize predictor
  age_std <- standardize(age)

  # Design matrices (raw polynomials incl. intercept)
  X <- bb_design_matrix(age_std, alpha_degree)
  Z <- bb_design_matrix(age_std, beta_degree)
  y <- score

  # Precompute the parameter-independent part of the log-likelihood once;
  # this saves one lchoose() call per observation per optimizer evaluation
  lch <- lchoose(n, y)

  # Robust initial parameter calculation via method of moments
  initial_values <- tryCatch({
    vals <- suppressWarnings(betaCoefficients(y, n))
    vals[vals <= 0 | !is.finite(vals)] <- 1e-4
    vals
  }, error = function(e) {
    m <- max(mean(y), 1e-4)
    a <- 1.0
    b <- max((n - m) / m * a, 1e-4)
    c(a, b, m, sd(y), n)
  })

  initial_params <- c(log(initial_values[1]),
                      rep(1e-6, alpha_degree),
                      log(initial_values[2]),
                      rep(1e-6, beta_degree))

  # Adaptive control parameters based on problem size.
  # Note on 'factr': optim() multiplies factr with .Machine$double.eps
  # (~2.2e-16) to obtain the relative reduction tolerance. Values below 1
  # would demand sub-machine precision and cause spurious convergence
  # failures. The values below correspond to tolerances of ~2e-12 to ~2e-10,
  # i.e. very strict but numerically achievable.
  n_param <- alpha_degree + beta_degree + 2
  if (n <= 50) {
    factr <- 1e4
    maxit <- n_param * 100
  } else if (n <= 150) {
    factr <- 1e5
    maxit <- n_param * 150
  } else {
    factr <- 1e6
    maxit <- n_param * 200
  }
  control_defaults <- list(factr = factr,
                           maxit = maxit,
                           lmm = min(n_param, 20))

  # Merge user control with adaptive defaults
  control <- if (is.null(control))
    control_defaults
  else
    utils::modifyList(control_defaults, control)

  # Parameter bounds consistent with the clamping in log_likelihood2
  lower_bounds <- rep(-20, length(initial_params))
  upper_bounds <- rep(20, length(initial_params))

  run_optim <- function(par, ctrl) {
    optim(
      par,
      log_likelihood2,
      gr = gradient_log_likelihood2,
      X = X,
      Z = Z,
      y = y,
      n = n,
      weights = weights,
      lch = lch,
      method = "L-BFGS-B",
      lower = lower_bounds,
      upper = upper_bounds,
      hessian = TRUE,
      control = ctrl
    )
  }

  # First optimization attempt; retry with neutral start values on failure
  result <- tryCatch({
    run_optim(initial_params, control)
  }, error = function(e) {
    message("First optimization attempt failed. Trying with different parameters...")
    retry_params <- c(log(1.0), rep(0, alpha_degree),
                      log(3.0), rep(0, beta_degree))
    retry_control <- control
    if (!is.null(retry_control$factr)) {
      retry_control$factr <- retry_control$factr * 100
    }
    if (!is.null(retry_control$maxit)) {
      retry_control$maxit <- retry_control$maxit * 2
    }
    run_optim(retry_params, retry_control)
  })

  if (result$convergence != 0) {
    warning("Optimization did not converge (code: ", result$convergence,
            "). Consider adjusting control parameters.")
  }

  # Warn when parameters sit at their box constraints; standard errors
  # and Wald tests are not valid there.
  at_bound <- result$par <= lower_bounds + 1e-6 |
    result$par >= upper_bounds - 1e-6
  if (any(at_bound)) {
    warning(
      "Parameter(s) at index ",
      paste(which(at_bound), collapse = ", "),
      " reached their bounds. Standard errors and Wald tests for these ",
      "parameters are not valid. Consider reducing model complexity."
    )
  }

  result$gradient <- gradient_log_likelihood2(result$par, X = X, Z = Z,
                                              y = y, n = n,
                                              weights = weights)

  # Extract results
  alpha_est <- result$par[1:(alpha_degree + 1)]
  beta_est <- result$par[(alpha_degree + 2):length(result$par)]
  se <- tryCatch({
    d <- diag(solve(result$hessian))
    if (any(d < 0, na.rm = TRUE)) {
      warning("Hessian is not positive definite; standard errors are set to NA where invalid.")
      d[d < 0] <- NA_real_
    }
    sqrt(d)
  }, error = function(e) {
    warning("Could not compute standard errors: Hessian matrix issue")
    rep(NA_real_, length(result$par))
  })

  # Resolve norm scale
  scaleMSD <- bb_resolve_scale(scale)

  # Store attributes for usage in other functions
  attr(result, "age_mean") <- mean(age)
  attr(result, "age_sd") <- sd(age)
  attr(result, "ageMin") <- min(age)
  attr(result, "ageMax") <- max(age)
  attr(result, "score_mean") <- mean(score)
  attr(result, "score_sd") <- sd(score)
  attr(result, "max") <- n
  attr(result, "N") <- length(score)
  attr(result, "scaleMean") <- scaleMSD[1]
  attr(result, "scaleSD") <- scaleMSD[2]

  model <- list(
    alpha_est = alpha_est,
    beta_est = beta_est,
    se = se,
    alpha_degree = alpha_degree,
    beta_degree = beta_degree,
    result = result
  )
  class(model) <- "cnormBetaBinomial2"

  if (plot) {
    p <- plot.cnormBetaBinomial(model, age, score, weights = weights)
    print(p)
  }

  model
}

#' Fit a beta-binomial regression model for continuous norming
#'
#' This function fits a beta-binomial regression model where both the
#' \eqn{\alpha} and \eqn{\beta} parameters of the beta-binomial distribution
#' are modeled as polynomial functions of the predictor variable (typically
#' age). Setting mode to 1 fits a beta-binomial model on the basis of
#' \eqn{\mu} and \eqn{\sigma}, setting it to 2 (default and recommended)
#' fits a beta-binomial model directly on the basis of \eqn{\alpha} and
#' \eqn{\beta}.
#'
#' @param age A numeric vector of predictor values (e.g., age).
#' @param score A numeric vector of response values.
#' @param n The maximum score (number of trials in the beta-binomial
#'   distribution). If NULL, max(score) is used.
#' @param weights A numeric vector of weights for each observation.
#'   Default is NULL (equal weights).
#' @param mode Integer specifying the mode of the model, either 1 or 2.
#'   Default is 2 (direct modelling of \eqn{\alpha} and \eqn{\beta};
#'   recommended). If set to 1, the model is fitted on the basis of
#'   \eqn{\mu} and \eqn{\sigma}, the predicted mean and standard deviation
#'   over age (kept for backwards compatibility).
#' @param alpha Integer specifying the degree of the polynomial for the alpha
#'   model. Default is 3. If mode is set to 1, this parameter is used to
#'   specify the degree of the polynomial for the \eqn{\mu} model.
#' @param beta Integer specifying the degree of the polynomial for the beta
#'   model. Default is 3. If mode is set to 1, this parameter is used to
#'   specify the degree of the polynomial for the \eqn{\sigma} model.
#' @param control A list of control parameters to be passed to the
#'   \code{optim} function. Parameters not specified are filled with
#'   defaults, namely \code{list(reltol = 1e-8, maxit = 1000)} for mode 1
#'   and adaptive L-BFGS-B settings (factr between 1e4 and 1e6, maxit
#'   proportional to the number of parameters) for mode 2.
#' @param scale Type of norm scale, either "T" (default), "IQ", "z" or a
#'   double vector with the mean and standard deviation.
#' @param plot Logical indicating whether to plot the model. Default is TRUE.
#'
#' @return A list of class "cnormBetaBinomial" or "cnormBetaBinomial2".
#'   In case of mode 2 containing:
#'   \item{alpha_est}{Estimated coefficients for the alpha model}
#'   \item{beta_est}{Estimated coefficients for the beta model}
#'   \item{se}{Standard errors of the estimated coefficients}
#'   \item{alpha_degree}{Degree of the polynomial for the alpha model}
#'   \item{beta_degree}{Degree of the polynomial for the beta model}
#'   \item{result}{Full result from the optimization procedure}
#'
#' @details
#' The function standardizes the input variables, fits polynomial models for
#' both distribution parameters, and uses maximum likelihood estimation with
#' analytic gradients to find the optimal parameters. In mode 2, the
#' optimization is performed using the L-BFGS-B method.
#'
#' Please note that the model fit indices (log-likelihood, AIC, BIC) of
#' mode 1 and mode 2 models are not directly comparable, since mode 1
#' maximizes a normal likelihood on standardized scores, whereas mode 2
#' maximizes the discrete beta-binomial likelihood on the raw scores.
#'
#' @examples
#' \dontrun{
#' # Fit a beta-binomial regression model to the PPVT data
#' model <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228)
#' summary(model)
#'
#' # Use weights for post-stratification
#' marginals <- data.frame(var = c("sex", "sex", "migration", "migration"),
#'                         level = c(1,2,0,1),
#'                         prop = c(0.51, 0.49, 0.65, 0.35))
#' weights <- computeWeights(ppvt, marginals)
#' model <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228, weights = weights)
#' }
#' @export
cnorm.betabinomial <- function(age,
                               score,
                               n = NULL,
                               weights = NULL,
                               mode = 2,
                               alpha = 3,
                               beta = 3,
                               control = NULL,
                               scale = "T",
                               plot = TRUE) {
  if (length(age) != length(score)) {
    stop("Length of 'age' and 'score' must be the same.")
  }

  if (!(is.numeric(mode) && length(mode) == 1 && mode %in% c(1, 2))) {
    stop("'mode' must be 1 or 2.")
  }

  if (is.null(n)) {
    n <- max(score, na.rm = TRUE)
    message("n parameter not specified, using the maximum score in the data instead. Consider to provide n manually.")
  }

  if (!(all(score >= 0, na.rm = TRUE) &&
        all(score == floor(score), na.rm = TRUE))) {
    warning("The score variable needs to include only positive integers for modelling with beta-binomial distributions. Trying to use Taylor polynomials instead (function 'cnorm').")
    return(cnorm(
      raw = score,
      age = age,
      weights = weights,
      scale = scale,
      plot = plot
    ))
  }

  if (mode == 2) {
    model <- cnorm.betabinomial2(age, score, n, weights,
                                 alpha_degree = alpha,
                                 beta_degree = beta,
                                 control = control,
                                 scale = scale,
                                 plot = plot)
  } else {
    model <- cnorm.betabinomial1(age, score, n, weights,
                                 mu = alpha,
                                 sigma = beta,
                                 control = control,
                                 scale = scale,
                                 plot = plot)
  }

  model
}


# --------------------------------------------------------------------------
# Prediction of distribution parameters
# --------------------------------------------------------------------------

#' Predict mean and standard deviation for a beta binomial regression model
#'
#' This function generates predictions from a fitted beta binomial regression
#' model (mode 1) for new age points and converts them into alpha and beta
#' parameters of the beta-binomial distribution via method of moments.
#'
#' @param model An object of class "cnormBetaBinomial".
#' @param ages A numeric vector of age points at which to make predictions.
#' @param n The maximum score to be achieved.
#'
#' @return A data frame with columns:
#'   \item{age}{The input age points}
#'   \item{mu}{Predicted mean values}
#'   \item{sigma}{Predicted standard deviation values}
#'   \item{a}{Alpha parameters}
#'   \item{b}{Beta parameters}
#'
#' @details
#' The function applies the same standardization used in model fitting,
#' generates predictions on the standardized scale, and transforms these back
#' to the original scale. If the method of moments yields invalid (i.e.
#' non-positive or non-finite) parameters, a mean-preserving low-precision
#' fallback is used instead of arbitrarily clamping both parameters.
#'
#' @keywords internal
predictCoefficients <- function(model, ages, n = NULL) {
  if (!inherits(model, "cnormBetaBinomial")) {
    stop("Wrong object. Please provide object from class 'cnormBetaBinomial'.")
  }

  # Standardize new ages
  ages_std <- (ages - attr(model$result, "age_mean")) / attr(model$result, "age_sd")

  # Design matrices including intercept
  X_new <- bb_design_matrix(ages_std, model$mu)
  Z_new <- bb_design_matrix(ages_std, model$sigma)

  predicted_mu_std <- drop(X_new %*% model$beta_est)
  log_sigma_std <- pmin(pmax(drop(Z_new %*% model$gamma_est), -20), 20)
  predicted_sigma_std <- exp(log_sigma_std)

  # Unstandardize predictions
  predicted_mu <- predicted_mu_std * attr(model$result, "score_sd") +
    attr(model$result, "score_mean")
  predicted_sigma <- predicted_sigma_std * attr(model$result, "score_sd")

  if (is.null(n)) {
    n <- attr(model$result, "max")
  }

  # Method of moments; guard the mean against the boundaries of the support
  m <- pmin(pmax(predicted_mu, 1e-6), n - 1e-6)
  v <- pmax(predicted_sigma^2, 1e-12)

  m2 <- m * m
  m3 <- m2 * m
  denom <- n * v - n * m + m2

  a <- (m2 * n - m3 - m * v) / denom
  b <- a * ((n - m) / m)

  # Mean-preserving fallback for degenerate cases (e.g. variance at or above
  # the admissible bound): use a small precision phi with a = p*phi,
  # b = (1-p)*phi, which keeps E(X) = n*p intact.
  bad <- !is.finite(a) | !is.finite(b) | a <= 0 | b <= 0
  if (any(bad)) {
    p <- m[bad] / n
    phi <- 1e-3
    a[bad] <- pmax(p * phi, 1e-6)
    b[bad] <- pmax((1 - p) * phi, 1e-6)
  }

  data.frame(
    age = ages,
    mu = predicted_mu,
    sigma = predicted_sigma,
    a = a,
    b = b
  )
}

#' Predict alpha and beta parameters for a beta-binomial regression model
#'
#' This function generates predictions from a fitted beta-binomial regression
#' model (mode 2) for new age points.
#'
#' @param model An object of class "cnormBetaBinomial2".
#' @param ages A numeric vector of age points at which to make predictions.
#' @param n The maximum score to be achieved.
#'
#' @return A data frame with columns:
#'   \item{age}{The input age points}
#'   \item{a}{Predicted alpha values}
#'   \item{b}{Predicted beta values}
#'   \item{mu}{Predicted mean values}
#'   \item{sigma}{Predicted standard deviation values}
#'
#' @details
#' The function applies the same standardization used in model fitting and
#' the same clamping of the linear predictors (\eqn{\pm 20} on the log scale)
#' as the likelihood function, ensuring consistent behavior also under
#' (mild) extrapolation.
#'
#' @keywords internal
predictCoefficients2 <- function(model, ages, n = NULL) {
  if (!inherits(model, "cnormBetaBinomial2")) {
    stop("Wrong object. Please provide object from class 'cnormBetaBinomial2'.")
  }

  # Standardize new ages
  ages_std <- (ages - attr(model$result, "age_mean")) / attr(model$result, "age_sd")

  # Design matrices including intercept
  X_new <- bb_design_matrix(ages_std, model$alpha_degree)
  Z_new <- bb_design_matrix(ages_std, model$beta_degree)

  # Clamp consistently with log_likelihood2
  log_alpha <- pmin(pmax(drop(X_new %*% model$alpha_est), -20), 20)
  log_beta <- pmin(pmax(drop(Z_new %*% model$beta_est), -20), 20)

  alpha <- exp(log_alpha)
  beta <- exp(log_beta)

  if (is.null(n)) {
    n <- attr(model$result, "max")
  }

  # Mean and standard deviation of the beta-binomial distribution
  ab <- alpha + beta
  mu <- n * alpha / ab
  v <- (n * alpha * beta * (ab + n)) / (ab^2 * (ab + 1))

  data.frame(
    age = ages,
    a = alpha,
    b = beta,
    mu = mu,
    sigma = sqrt(v)
  )
}

#' Compute Parameters of a Beta Binomial Distribution
#'
#' This function calculates the \eqn{\alpha} (a) and \eqn{\beta} (b)
#' parameters of a beta binomial distribution, along with the mean (m) and
#' standard deviation (sd) based on the input vector \code{x} and the maximum
#' number \code{n}, using the method of moments.
#'
#' The beta-binomial distribution is a discrete probability distribution that
#' models the number of successes in a fixed number of trials, where the
#' probability of success varies from trial to trial. This variability in
#' success probability is modeled by a beta distribution. Such a calculation
#' is particularly relevant in scenarios where there is heterogeneity in
#' success probabilities across trials, which is common in real-world
#' situations, as for example the number of correct solutions in a
#' psychometric test, where the test has a fixed number of items.
#'
#' @param x A numeric vector of non-negative integers representing observed
#'   counts.
#' @param n The maximum number or the maximum possible value of \code{x}.
#'   If not specified, uses max(x) instead.
#'
#' @return A numeric vector containing the calculated parameters in the
#'   following order: alpha (a), beta (b), mean (m), standard deviation (sd),
#'   and the maximum number (n). If the method of moments is degenerate
#'   (mean at the boundary of the support or a non-positive moment
#'   denominator), alpha and beta are returned as \code{NA} with a warning.
#'   Non-positive estimates (indicating dispersion incompatible with a
#'   beta-binomial distribution) are returned as-is with a warning.
#'
#' @export
betaCoefficients <- function(x, n = NULL) {
  if (length(x) < 2 || any(!is.finite(x))) {
    stop("'x' must contain at least two finite values.")
  }
  if (is.null(n)) {
    n <- max(x)
  }
  if (max(x) > n) {
    stop("'x' contains values larger than n = ", n, ".")
  }

  m <- mean(x)
  s <- sd(x)
  v <- s^2

  m2 <- m * m
  m3 <- m2 * m
  denom <- n * v - n * m + m2

  if (m <= 0 || m >= n || !is.finite(denom) || denom == 0) {
    warning("Method of moments is degenerate (mean at the boundary of the ",
            "support or zero moment denominator). Returning NA for alpha ",
            "and beta.")
    return(c(NA_real_, NA_real_, m, s, n))
  }

  a <- (m2 * n - m3 - m * v) / denom
  b <- a * ((n - m) / m)

  if (a <= 0 || b <= 0) {
    warning("Method of moments produced non-positive parameter estimates. ",
            "The dispersion of 'x' is incompatible with a beta-binomial ",
            "distribution.")
  }

  c(a, b, m, s, n)
}


# --------------------------------------------------------------------------
# Norm tables and prediction of norm scores
# --------------------------------------------------------------------------

#' Calculate Cumulative Probabilities, Density, Percentiles, and Z-Scores for
#' Beta-Binomial Distribution
#'
#' This function generates a norm table for specific ages based on the beta
#' binomial regression model. In case a confidence coefficient (CI, default
#' .9) and the reliability is specified, confidence intervals are computed
#' for the true score estimates, including a correction for regression to
#' the mean (Eid & Schmidt, 2012, p. 272).
#'
#' @param model The model, fitted with \code{\link{cnorm.betabinomial}}.
#' @param ages A numeric vector of age points at which to compute the tables.
#' @param n The number of items resp. the maximum score.
#' @param m An optional stop criterion in table generation. Positive integer
#'   lower than n. Please note: The probabilities are always computed on the
#'   full support 0 to n; m only truncates the output table.
#' @param range The range of the norm scores in standard deviations.
#'   Default is 3. Thus, scores in the range of +/- 3 standard deviations
#'   are considered.
#' @param CI Confidence coefficient, ranging from 0 to 1, default .9.
#' @param reliability Reliability coefficient, ranging between 0 to 1.
#'
#' @return A list of data frames with columns: x, Px, Pcum, Percentile, z,
#'   norm score and possibly confidence intervals.
#' @export
normTable.betabinomial <- function(model,
                                   ages,
                                   n = NULL,
                                   m = NULL,
                                   range = 3,
                                   CI = .9,
                                   reliability = NULL) {
  if (!isBeta(model)) {
    stop("Wrong object. Please provide object from class 'cnormBetaBinomial' or 'cnormBetaBinomial2'.")
  }

  if (is.null(CI) || is.na(CI)) {
    reliability <- NULL
  } else if (CI > .99999 || CI < .00001) {
    stop("Confidence coefficient (CI) out of range. Please specify value between 0 and 1.")
  }

  rel <- FALSE
  if (!is.null(reliability)) {
    if (reliability > .9999 || reliability < .0001) {
      stop("Reliability coefficient out of range. Please specify value between 0 and 1.")
    } else {
      se <- qnorm(1 - ((1 - CI) / 2)) * sqrt(reliability * (1 - reliability))
      rel <- TRUE
    }
  }

  if (is.null(n)) {
    n <- attr(model$result, "max")
  }

  if (is.null(m) || m > n) {
    m <- n
  }

  if (inherits(model, "cnormBetaBinomial")) {
    predictions <- predictCoefficients(model, ages, n)
  } else {
    predictions <- predictCoefficients2(model, ages, n)
  }

  a <- predictions$a
  b <- predictions$b

  mScale <- attr(model$result, "scaleMean")
  sdScale <- attr(model$result, "scaleSD")

  result <- vector("list", length(a))

  for (k in seq_along(a)) {
    # IMPORTANT: probabilities are computed on the full support 0:n;
    # truncation at m happens afterwards without renormalization
    dist <- bb_distribution(a[k], b[k], n)

    # Clip via pmin/pmax instead of logical subscript assignment.
    z <- pmin(pmax(qnorm(dist$perc), -range), range)

    norm <- rep(NA_real_, length(z))
    if (!is.na(mScale) && !is.na(sdScale)) {
      norm <- mScale + sdScale * z
    }

    df <- data.frame(
      x = dist$x,
      Px = dist$Px,
      Pcum = dist$cum,
      Percentile = dist$perc * 100,
      z = z,
      norm = norm
    )

    if (rel) {
      zPredicted <- reliability * z
      df$lowerCI <- (zPredicted - se) * sdScale + mScale
      df$upperCI <- (zPredicted + se) * sdScale + mScale
      df$lowerCI_PR <- pnorm(zPredicted - se) * 100
      df$upperCI_PR <- pnorm(zPredicted + se) * 100
    }

    # Truncate output table at m (no renormalization!)
    result[[k]] <- df[df$x <= m, , drop = FALSE]
  }

  names(result) <- ages
  result
}

#' Predict Norm Scores from Raw Scores
#'
#' This function calculates norm scores based on raw scores, age, and a
#' fitted cnormBetaBinomial model.
#'
#' @param object A fitted model object of class 'cnormBetaBinomial' or
#'   'cnormBetaBinomial2'.
#' @param ... Additional arguments passed to the prediction method:
#'   \itemize{
#'      \item age A numeric vector of ages, same length as score.
#'      \item score A numeric vector of raw scores.
#'      \item range The range of the norm scores in standard deviations.
#'        Default is 3. Thus, scores in the range of +/- 3 standard
#'        deviations are considered.
#'    }
#'
#' @return A numeric vector of norm scores (or percentiles, if no norm scale
#'   was specified in the model). Missing, non-integer or out-of-range raw
#'   scores yield \code{NA}.
#'
#' @details
#' The function first predicts the alpha and beta parameters of the
#' beta-binomial distribution for each unique age using the provided model.
#' It then calculates the mid-p cumulative probability for each raw score
#' given these parameters and converts these probabilities to the norm scale
#' specified in the model. The distribution is computed only once per unique
#' age value, which considerably speeds up predictions for grouped data.
#'
#' @examples
#' \dontrun{
#' # Assuming you have a fitted model named 'bb_model':
#' model <- cnorm.betabinomial(ppvt$age, ppvt$raw)
#' raw <- c(100, 121, 97, 180)
#' ages <- c(7, 8, 9, 10)
#' norm_scores <- predict(model, ages, raw)
#' }
#'
#' @export
#' @family predict
predict.cnormBetaBinomial <- function(object, ...) {
  model <- object
  args <- list(...)

  if ("age" %in% names(args)) {
    age <- args$age
  } else if (length(args) > 0) {
    age <- args[[1]]
  } else {
    age <- NULL
  }
  if ("score" %in% names(args)) {
    score <- args$score
  } else if (length(args) > 1) {
    score <- args[[2]]
  } else {
    score <- NULL
  }
  if ("range" %in% names(args)) {
    range <- args$range
  } else if (length(args) > 2) {
    range <- args[[3]]
  } else {
    range <- 3
  }

  if (!isBeta(model)) {
    stop("Wrong object. Please provide object from class 'cnormBetaBinomial' or 'cnormBetaBinomial2'.")
  }

  if (is.null(age) || is.null(score)) {
    stop("Both 'age' and 'score' must be provided.")
  }

  if (length(age) != length(score)) {
    stop("The lengths of 'age' and 'score' must be the same.")
  }

  n <- attr(model$result, "max")

  # Compute distribution parameters only once per unique age
  unique_ages <- unique(age)
  idx <- match(age, unique_ages)

  if (inherits(model, "cnormBetaBinomial")) {
    predictions <- predictCoefficients(model, unique_ages, n)
  } else {
    predictions <- predictCoefficients2(model, unique_ages, n)
  }

  # Validate scores
  valid <- is.finite(score) & score >= 0 & score <= n & score == floor(score)
  if (any(!valid)) {
    warning("Some raw scores are missing, non-integer or outside the range 0 to ",
            n, ". Returning NA for these cases.")
  }

  z_scores <- rep(NA_real_, length(age))

  for (u in seq_along(unique_ages)) {
    rows <- which(idx == u & valid)
    if (length(rows) == 0) next

    dist <- bb_distribution(predictions$a[u], predictions$b[u], n)
    z_scores[rows] <- qnorm(dist$perc[score[rows] + 1])
  }

  z_scores <- pmin(pmax(z_scores, -range), range)

  mScale <- attr(model$result, "scaleMean")
  sdScale <- attr(model$result, "scaleSD")

  if (!is.na(mScale) && !is.na(sdScale)) {
    return(mScale + sdScale * z_scores)
  } else {
    return(pnorm(z_scores) * 100)
  }
}

#' Predict Norm Scores from Raw Scores
#'
#' This function calculates norm scores based on raw scores, age, and a
#' fitted cnormBetaBinomial2 model. See
#' \code{\link{predict.cnormBetaBinomial}} for details.
#'
#' @param object A fitted model object of class 'cnormBetaBinomial' or
#'   'cnormBetaBinomial2'.
#' @param ... Additional arguments passed to the prediction method:
#'   \itemize{
#'      \item age A numeric vector of ages, same length as score.
#'      \item score A numeric vector of raw scores.
#'      \item range The range of the norm scores in standard deviations.
#'        Default is 3.
#'    }
#'
#' @return A numeric vector of norm scores.
#'
#' @export
#' @family predict
predict.cnormBetaBinomial2 <- predict.cnormBetaBinomial


# --------------------------------------------------------------------------
# Plotting
# --------------------------------------------------------------------------

#' Plot cnormBetaBinomial Model with Data and Percentile Lines
#'
#' This function creates a visualization of a fitted cnormBetaBinomial model,
#' including the original data points, the manifest percentiles (diamond-shaped
#' dots per age group) and the model-implied percentile lines.
#'
#' By default (\code{discrete = TRUE}), the percentile lines represent the
#' exact quantiles of the fitted discrete beta-binomial distribution and are
#' therefore step functions. Setting \code{discrete = FALSE} draws smooth
#' lines based on the quantiles of the underlying beta (mixing) distribution.
#' Note that this continuous approximation omits the binomial stage of the
#' variance and therefore displays less spread than the fitted model actually
#' implies, particularly in the outer percentiles. Use it only for visual
#' comparability with continuous modelling approaches.
#'
#' Cases with non-finite values (NA, NaN, Inf) in age, score or weights are
#' removed prior to plotting, mirroring the behavior of the fitting functions.
#'
#' @param x A fitted model object of class "cnormBetaBinomial" or
#'   "cnormBetaBinomial2".
#' @param ... Additional arguments passed to the plot method.
#'   \itemize{
#'      \item age A vector of the age data.
#'      \item score A vector of the score data.
#'      \item weights An optional numeric vector of weights for each observation.
#'      \item percentiles An optional vector with the percentiles to plot.
#'      \item points Logical indicating whether to plot the data points.
#'        Default is TRUE.
#'      \item discrete Logical indicating whether to plot the exact discrete
#'        beta-binomial quantiles (TRUE, default) or a smooth continuous
#'        approximation via the underlying beta distribution (FALSE).
#'    }
#'
#' @return A ggplot object.
#'
#' @family plot
#' @export
plot.cnormBetaBinomial <- function(x, ...) {
  model <- x
  args <- list(...)

  if ("age" %in% names(args)) {
    age <- args$age
  } else if (length(args) > 0) {
    age <- args[[1]]
  } else {
    age <- NULL
  }
  if ("score" %in% names(args)) {
    score <- args$score
  } else if (length(args) > 1) {
    score <- args[[2]]
  } else {
    score <- NULL
  }
  weights <- if ("weights" %in% names(args)) args$weights else NULL
  percentiles <- if ("percentiles" %in% names(args)) args$percentiles
  else c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
  points <- if ("points" %in% names(args)) args$points else TRUE
  discrete <- if ("discrete" %in% names(args)) args$discrete else TRUE

  if (is.null(age) || is.null(score)) {
    stop("Please provide 'age' and 'score' vectors.")
  }

  if (!isBeta(model)) {
    stop("Wrong object. Please provide object from class 'cnormBetaBinomial' or 'cnormBetaBinomial2'.")
  }

  if (length(age) != length(score)) {
    stop("Length of 'age' and 'score' must be the same.")
  }

  if (!is.null(weights) && length(weights) != length(age)) {
    stop("Length of 'weights' must match length of 'age' and 'score'.")
  }

  # Remove non-finite cases (NA, NaN, Inf), mirroring the fitting functions.
  # Without this, weighted.quantile() propagates NA into the manifest
  # percentiles and the dots silently disappear from the plot.
  ok <- is.finite(age) & is.finite(score)
  if (!is.null(weights)) {
    ok <- ok & is.finite(weights)
  }
  if (!all(ok)) {
    message("Vector(s) contained non-finite values (NA, NaN, Inf). ",
            "These cases will be removed for plotting.")
    age <- age[ok]
    score <- score[ok]
    if (!is.null(weights)) {
      weights <- weights[ok]
    }
  }

  # Generate prediction points
  n_points <- 100
  data <- data.frame(age = age, score = score)
  data$w <- if (is.null(weights)) rep(1, nrow(data)) else weights

  age_range <- range(age)
  pred_ages <- seq(age_range[1], age_range[2], length.out = n_points)
  n_max <- attr(model$result, "max")

  # Get predictions
  if (inherits(model, "cnormBetaBinomial")) {
    preds <- predictCoefficients(model, pred_ages, n_max)
  } else {
    preds <- predictCoefficients2(model, pred_ages, n_max)
  }

  if (discrete) {
    # Exact quantiles of the fitted discrete beta-binomial distribution
    percentile_matrix <- vapply(seq_along(pred_ages), function(j) {
      dist <- bb_distribution(preds$a[j], preds$b[j], n_max)
      if (anyNA(dist$cum)) {
        return(rep(NA_real_, length(percentiles)))
      }
      vapply(percentiles,
             function(p) as.numeric(dist$x[which.max(dist$cum >= p)]),
             numeric(1))
    }, numeric(length(percentiles)))
    percentile_values <- t(percentile_matrix)   # rows: ages, cols: percentiles
  } else {
    # Continuous approximation via the underlying beta (mixing) distribution;
    # omits the binomial-stage variance and compresses outer percentiles
    percentile_values <- sapply(percentiles, function(p) {
      qbeta(p, shape1 = preds$a, shape2 = preds$b) * n_max
    })
  }

  # Wide-format prediction data, consistent with plot.cnormShash
  percentile_data <- as.data.frame(percentile_values)
  colnames(percentile_data) <- paste0("P", percentiles * 100)

  plot_data <- data.frame(
    age = pred_ages,
    mu = preds$mu,
    sigma = preds$sigma,
    percentile_data,
    check.names = FALSE
  )

  # Create the plot
  p <- ggplot()

  if (points) {
    p <- p + geom_point(
      data = data,
      aes(x = age, y = score),
      alpha = 0.2,
      size = 0.6
    )
  }

  # Grouping for manifest percentiles
  if (length(age) / length(unique(age)) > 50 && min(table(data$age)) > 30) {
    data$group <- age
  } else {
    data$group <- getGroups(age)
  }

  # Limit to max 30 groups for better visibility
  if (length(unique(data$group)) > 30) {
    data$group <- getGroups(age, n = 30)
  }

  # Get actual percentiles
  NAMES <- paste("PR", percentiles * 100, sep = "")
  percentile.actual <- as.data.frame(do.call("rbind", lapply(split(data, data$group), function(df) {
    c(age = mean(df$age),
      weighted.quantile(df$score, probs = percentiles, weights = df$w))
  })))
  colnames(percentile.actual) <- c("age", NAMES)
  manifest_data <- percentile.actual

  # Add percentile lines and points
  # Discrete quantiles are piecewise constant in age: render them with
  # geom_step (vertical risers) instead of geom_line, which would draw
  # sheared ramps between adjacent prediction ages.
  for (i in seq_along(percentiles)) {
    if (discrete) {
      p <- p + geom_step(
        data = plot_data,
        aes(
          x = .data$age,
          y = .data[[paste0("P", percentiles[i] * 100)]],
          color = !!NAMES[i]
        ),
        direction = "mid",
        linewidth = 0.6
      )
    } else {
      p <- p + geom_line(
        data = plot_data,
        aes(
          x = .data$age,
          y = .data[[paste0("P", percentiles[i] * 100)]],
          color = !!NAMES[i]
        ),
        linewidth = 0.6
      )
    }
    p <- p + geom_point(
      data = manifest_data,
      aes(
        x = .data$age,
        y = .data[[NAMES[i]]],
        color = !!NAMES[i]
      ),
      size = 2,
      shape = 18
    )
  }
  # Customize the plot
  p <- p +
    theme_minimal() +
    labs(
      title = "Percentile Plot (Beta-Binomial Model)",
      x = "Age",
      y = "Score",
      color = "Percentile"
    ) +
    scale_y_continuous(limits = c(0, n_max)) +
    scale_color_manual(
      values = setNames(rainbow(length(percentiles)), NAMES),
      breaks = NAMES,
      labels = paste0(percentiles * 100, "%")
    ) +
    guides(color = guide_legend(override.aes = list(
      linetype = rep("solid", length(NAMES)),
      shape = rep(18, length(NAMES))
    ))) +
    theme(
      plot.title = element_text(hjust = 0.5, size = 16, face = "bold"),
      plot.subtitle = element_text(hjust = 0.5, size = 12),
      axis.title = element_text(size = 12, face = "bold"),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  p
}

#' Plot cnormBetaBinomial2 Model with Data and Percentile Lines
#'
#' This function creates a visualization of a fitted cnormBetaBinomial2
#' model, including the original data points, manifest percentiles and
#' specified percentile lines. See \code{\link{plot.cnormBetaBinomial}}.
#'
#' @param x A fitted model object of class "cnormBetaBinomial" or
#'   "cnormBetaBinomial2".
#' @param ... Additional arguments passed to the plot method.
#'   \itemize{
#'      \item age A vector of the age data.
#'      \item score A vector of the score data.
#'      \item weights An optional numeric vector of weights for each observation.
#'      \item percentiles An optional vector with the percentiles to plot.
#'      \item points Logical indicating whether to plot the data points.
#'        Default is TRUE.
#'      \item discrete Logical indicating whether to plot the exact discrete
#'        beta-binomial quantiles (TRUE, default) or a smooth continuous
#'        approximation via the underlying beta distribution (FALSE).
#'    }
#'
#' @return A ggplot object.
#'
#' @family plot
#' @export
plot.cnormBetaBinomial2 <- plot.cnormBetaBinomial


# --------------------------------------------------------------------------
# Diagnostics and summary
# --------------------------------------------------------------------------

#' Diagnostic Information for Beta-Binomial Model
#'
#' This function provides diagnostic information for a fitted beta-binomial
#' model from the cnorm.betabinomial function. It returns various metrics
#' related to model convergence, fit, and complexity. In case age and raw
#' scores are provided, the function as well computes R2, RMSE and bias for
#' the norm scores based on the manifest and predicted norm scores.
#'
#' @param model An object of class "cnormBetaBinomial" or
#'   "cnormBetaBinomial2", typically the result of a call to
#'   cnorm.betabinomial().
#' @param age An optional vector with age values.
#' @param score An optional vector with raw values.
#' @param weights An optional vector with weights.
#'
#' @return A list containing diagnostic information, including convergence,
#'   number of evaluations, log-likelihood, AIC, BIC, the maximum absolute
#'   analytic gradient at the solution (first-order optimality), parameter
#'   estimates with standard errors, z and p values, and - if age and score
#'   are provided - R2, RMSE and bias of the norm scores.
#'
#' @details
#' The AIC and BIC are calculated as:
#' AIC = 2k - 2ln(L)
#' BIC = ln(n)k - 2ln(L)
#' where k is the number of parameters, L is the maximum likelihood, and n
#' is the number of observations.
#'
#' Please note that fit indices of mode 1 and mode 2 models are not directly
#' comparable, since the two modes maximize different likelihoods (normal
#' on standardized scores vs. discrete beta-binomial).
#'
#' @examples
#' \dontrun{
#' model <- cnorm.betabinomial(ppvt$age, ppvt$raw)
#' diag_info <- diagnostics.betabinomial(model)
#' print(diag_info)
#' summary(diag_info)
#'
#' if(diag_info$converged) {
#'   cat("Model converged successfully.\n")
#' } else {
#'   cat("Warning: Model did not converge.\n")
#' }
#'
#' cat("AIC:", diag_info$AIC, "\n")
#' cat("BIC:", diag_info$BIC, "\n")
#' }
#'
#' @export
diagnostics.betabinomial <- function(model,
                                     age = NULL,
                                     score = NULL,
                                     weights = NULL) {
  if (!isBeta(model)) {
    stop("Wrong object. Please provide object from class 'cnormBetaBinomial' or 'cnormBetaBinomial2'.")
  }

  opt_results <- model$result

  if (inherits(model, "cnormBetaBinomial")) {
    type <- "cnormBetaBinomial"
    n_params <- length(model$beta_est) + length(model$gamma_est)
    param_names <- c(paste0("beta_", 0:(length(model$beta_est) - 1)),
                     paste0("gamma_", 0:(length(model$gamma_est) - 1)))
    estimates <- c(model$beta_est, model$gamma_est)
  } else {
    type <- "cnormBetaBinomial2"
    n_params <- length(model$alpha_est) + length(model$beta_est)
    param_names <- c(paste0("alpha_", 0:(length(model$alpha_est) - 1)),
                     paste0("beta_", 0:(length(model$beta_est) - 1)))
    estimates <- c(model$alpha_est, model$beta_est)
  }

  n_obs <- attr(model$result, "N")
  convergence <- opt_results$convergence == 0

  # First-order optimality: max absolute analytic gradient at the solution
  # (stored by the fitting functions)
  max_gradient <- NA
  if (is.numeric(opt_results$gradient) && length(opt_results$gradient) > 0) {
    max_gradient <- max(abs(opt_results$gradient))
  }

  # Robust standard errors (Hessian may be singular or indefinite)
  se <- tryCatch({
    d <- diag(solve(opt_results$hessian))
    d[d < 0] <- NA_real_
    sqrt(d)
  }, error = function(e) {
    rep(NA_real_, length(estimates))
  })
  z_values <- estimates / se
  p_values <- 2 * (1 - pnorm(abs(z_values)))

  # Log-likelihood, AIC, and BIC
  log_likelihood <- -opt_results$value
  AIC <- 2 * n_params - 2 * log_likelihood
  BIC <- log(n_obs) * n_params - 2 * log_likelihood

  # Norm score recovery statistics, if data are provided
  R2 <- NA
  rmse <- NA
  bias <- NA
  if (!is.null(age) && !is.null(score)) {
    if (length(age) / length(unique(age)) > 50 && min(table(age)) > 30) {
      data <- data.frame(group = age, raw = score)
      data <- rankByGroup(
        data = data,
        raw = "raw",
        group = "group",
        weights = weights,
        scale = c(attr(model$result, "scaleMean"),
                  attr(model$result, "scaleSD"))
      )
      norm_scores <- predict(model, data$group, data$raw)
    } else {
      data <- data.frame(age = age, raw = score)
      data$groups <- getGroups(age)
      width <- (max(age) - min(age)) / length(unique(data$groups))
      data <- rankBySlidingWindow(
        data,
        age = "age",
        raw = "raw",
        width = width,
        weights = weights,
        scale = c(attr(model$result, "scaleMean"),
                  attr(model$result, "scaleSD"))
      )
      norm_scores <- predict(model, data$age, data$raw)
    }

    norm_manifest <- data$normValue
    R2 <- cor(norm_scores, norm_manifest, use = "pairwise.complete.obs")^2
    rmse <- sqrt(mean((norm_scores - norm_manifest)^2, na.rm = TRUE))
    bias <- mean(norm_scores - norm_manifest, na.rm = TRUE)
  }

  list(
    type = type,
    converged = convergence,
    n_evaluations = opt_results$counts["function"],
    n_gradient = opt_results$counts["gradient"],
    final_value = opt_results$value,
    message = opt_results$message,
    AIC = AIC,
    BIC = BIC,
    log_likelihood = log_likelihood,
    max_gradient = max_gradient,
    n_params = n_params,
    n_obs = n_obs,
    param_estimates = setNames(estimates, param_names),
    param_se = setNames(se, param_names),
    z_values = setNames(z_values, param_names),
    p_values = setNames(p_values, param_names),
    R2 = R2,
    rmse = rmse,
    bias = bias
  )
}

#' Summarize a Beta-Binomial Continuous Norming Model
#'
#' This function provides a summary of a fitted beta-binomial continuous
#' norming model, including model fit statistics, convergence information,
#' and parameter estimates.
#'
#' @param object An object of class "cnormBetaBinomial" or
#'   "cnormBetaBinomial2", typically the result of a call to
#'   \code{\link{cnorm.betabinomial}}.
#' @param ... Additional arguments passed to the summary method:
#'   \itemize{
#'      \item age An optional numeric vector of age values corresponding to
#'        the raw scores. If provided along with \code{score}, additional fit
#'        statistics (R-squared, RMSE, bias) will be calculated.
#'      \item score An optional numeric vector of raw scores. Must be
#'        provided if \code{age} is given.
#'      \item weights An optional numeric vector of weights for each
#'        observation.
#'    }
#'
#' @return Invisibly returns a list containing detailed diagnostic
#'   information about the model. The function primarily produces printed
#'   output summarizing the model.
#'
#' @details
#' The summary includes:
#' \itemize{
#'   \item Basic model information (type, number of observations, number of parameters)
#'   \item Model fit statistics (log-likelihood, AIC, BIC)
#'   \item R-squared, RMSE, and bias (if age and raw scores are provided)
#'         in comparison to manifest norm scores
#'   \item Convergence information, including the maximum absolute gradient
#'         at the solution (first-order optimality)
#'   \item Parameter estimates with standard errors, z-values, and p-values
#' }
#'
#' @examples
#' \dontrun{
#' model <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228)
#' summary(model)
#'
#' # Including R-squared, RMSE, and bias in the summary:
#' summary(model, age = ppvt$age, score = ppvt$raw)
#' }
#' @seealso \code{\link{cnorm.betabinomial}}, \code{\link{diagnostics.betabinomial}}
#'
#' @export
summary.cnormBetaBinomial <- function(object, ...) {
  args <- list(...)

  if ("age" %in% names(args)) {
    age <- args$age
  } else if (length(args) > 0) {
    age <- args[[1]]
  } else {
    age <- NULL
  }
  if ("score" %in% names(args)) {
    score <- args$score
  } else if (length(args) > 1) {
    score <- args[[2]]
  } else {
    score <- NULL
  }
  if ("weights" %in% names(args)) {
    weights <- args$weights
  } else if (length(args) > 2) {
    weights <- args[[3]]
  } else {
    weights <- NULL
  }

  diag <- diagnostics.betabinomial(object, age, score, weights)

  cat("Beta-Binomial Continuous Norming Model\n")
  cat("---------------------------------------\n")
  cat("Model type:", diag$type, "\n")
  cat("Number of observations:", diag$n_obs, "\n")
  cat("Number of parameters:", diag$n_params, "\n")
  cat("\n")

  cat("Model Fit:\n")
  cat("  Log-likelihood:", round(diag$log_likelihood, 2), "\n")
  cat("  AIC:", round(diag$AIC, 2), "\n")
  cat("  BIC:", round(diag$BIC, 2), "\n")
  if (!is.na(diag$R2)) {
    cat("  R-squared:", round(diag$R2, 4), "\n")
    cat("  RMSE:", round(diag$rmse, 4), "\n")
    cat("  Bias:", round(diag$bias, 4), "\n")
  }
  cat("\n")

  cat("Convergence:\n")
  cat("  Converged:", diag$converged, "\n")
  cat("  Function evaluations:", diag$n_evaluations, "\n")
  cat("  Gradient evaluations:", diag$n_gradient, "\n")
  if (!is.na(diag$max_gradient)) {
    cat("  Max gradient:", round(diag$max_gradient, 6), "\n")
  }
  if (!is.null(diag$message)) {
    cat("  Message:", diag$message, "\n")
  }
  cat("\n")

  cat("Parameter Estimates:\n")
  param_table <- data.frame(
    Estimate = diag$param_estimates,
    `Std. Error` = diag$param_se,
    `z value` = diag$z_values,
    `Pr(>|z|)` = diag$p_values,
    check.names = FALSE
  )
  print(param_table, digits = 4)

  invisible(diag)
}

#' Summarize a Beta-Binomial Continuous Norming Model
#'
#' Summary method for objects of class "cnormBetaBinomial2". See
#' \code{\link{summary.cnormBetaBinomial}} for details.
#'
#' @param object An object of class "cnormBetaBinomial" or
#'   "cnormBetaBinomial2", typically the result of a call to
#'   \code{\link{cnorm.betabinomial}}.
#' @param ... Additional arguments passed to the summary method:
#'   \itemize{
#'      \item age An optional numeric vector of age values.
#'      \item score An optional numeric vector of raw scores.
#'      \item weights An optional numeric vector of weights.
#'    }
#'
#' @return Invisibly returns a list containing detailed diagnostic
#'   information about the model.
#'
#' @seealso \code{\link{cnorm.betabinomial}}, \code{\link{diagnostics.betabinomial}}
#'
#' @export
summary.cnormBetaBinomial2 <- summary.cnormBetaBinomial


# --------------------------------------------------------------------------
# Automatic model selection
# --------------------------------------------------------------------------

#' Automatic model selection for beta-binomial continuous norming via BIC
#'
#' Selects polynomial degrees for the two components of a beta-binomial model
#' (\eqn{\alpha}/\eqn{\beta} when \code{mode = 2} or \eqn{\mu}/\eqn{\sigma}
#' when \code{mode = 1}) by minimizing BIC over the full grid of degree
#' combinations.
#'
#' Parallel execution is attempted by default. If the workers cannot access
#' the \pkg{cNORM} namespace (e.g. during \code{devtools::load_all()}), the
#' function transparently falls back to sequential execution.
#'
#' @param age,score Numeric vectors of predictor and response values.
#' @param n Maximum score. Defaults to \code{max(score)}.
#' @param weights Optional numeric vector of weights.
#' @param mode 1 for \eqn{\mu}/\eqn{\sigma}, 2 for direct
#'   \eqn{\alpha}/\eqn{\beta} (default).
#' @param max_alpha,max_beta Maximum polynomial degrees. Default 4.
#' @param min_alpha,min_beta Minimum polynomial degrees. Default 1.
#' @param control Optional control list passed to \code{\link[stats]{optim}}.
#' @param scale Norm scale (default \code{"T"}).
#' @param parallel Logical; attempt parallel execution. Default \code{TRUE}.
#' @param n_cores Number of cores. Defaults to all logical cores.
#' @param plot Logical; plot the selected model. Default \code{TRUE}.
#' @param verbose Logical; print progress. Default \code{TRUE}.
#'
#' @return The selected fitted model, with an additional element
#'   \code{selection} containing the evaluation table and selection details.
#'
#' @export
autoselect.betabinomial <- function(age,
                                    score,
                                    n         = NULL,
                                    weights   = NULL,
                                    mode      = 2,
                                    max_alpha = 4,
                                    max_beta  = 4,
                                    min_alpha = 1,
                                    min_beta  = 1,
                                    control   = NULL,
                                    scale     = "T",
                                    parallel  = TRUE,
                                    n_cores   = NULL,
                                    plot      = TRUE,
                                    verbose   = TRUE) {

  # ---- Input validation -------------------------------------------------
  if (length(age) != length(score))
    stop("Length of 'age' and 'score' must be the same.")
  if (!is.null(weights) && length(weights) != length(age))
    stop("Length of 'weights' must match length of 'age' and 'score'.")
  if (max_alpha < min_alpha) stop("'max_alpha' must be >= 'min_alpha'.")
  if (max_beta  < min_beta)  stop("'max_beta' must be >= 'min_beta'.")
  if (min_alpha < 1 || min_beta < 1) stop("Minimum degrees must be >= 1.")
  if (!(mode %in% c(1, 2))) stop("'mode' must be 1 or 2.")

  if (is.null(n)) {
    n <- max(score, na.rm = TRUE)
    if (verbose) message("n not specified; using max(score) = ", n)
  }

  say <- function(...) {
    if (verbose) { cat(..., sep = ""); utils::flush.console() }
  }

  # ---- Candidate grid ----------------------------------------------------
  grid <- expand.grid(alpha = min_alpha:max_alpha,
                      beta  = min_beta:max_beta,
                      KEEP.OUT.ATTRS = FALSE)
  pairs <- lapply(seq_len(nrow(grid)),
                  function(i) c(grid$alpha[i], grid$beta[i]))

  # ---- Parallel setup with dev-mode fallback -----------------------------
  use_parallel <- FALSE
  cl <- NULL
  if (isTRUE(parallel)) {
    avail <- tryCatch(parallel::detectCores(logical = TRUE),
                      error = function(e) 1L)
    if (is.null(n_cores)) n_cores <- avail
    n_cores <- max(1L, min(n_cores, length(pairs), avail))

    if (n_cores > 1L && length(pairs) > 1L) {
      cl <- tryCatch(parallel::makeCluster(n_cores),
                     error = function(e) NULL)
      if (!is.null(cl)) {
        # Probe workers: can they load cNORM (i.e. is the package installed)?
        worker_ok <- tryCatch({
          res <- parallel::clusterCall(cl, function() {
            requireNamespace("cNORM", quietly = TRUE) &&
              exists("cnorm.betabinomial",       envir = asNamespace("cNORM")) &&
              exists("diagnostics.betabinomial", envir = asNamespace("cNORM"))
          })
          all(vapply(res, isTRUE, logical(1)))
        }, error = function(e) FALSE)

        if (worker_ok) {
          use_parallel <- TRUE
          on.exit(try(parallel::stopCluster(cl), silent = TRUE), add = TRUE)
          parallel::clusterExport(
            cl,
            varlist = c("age", "score", "n", "weights",
                        "mode", "control", "scale"),
            envir = environment()
          )
          say(sprintf("Parallel mode: using %d cores.\n", n_cores))
        } else {
          try(parallel::stopCluster(cl), silent = TRUE); cl <- NULL
          say("Note: cNORM is not installed in the worker library path ",
              "(typical during devtools::load_all). ",
              "Falling back to sequential execution.\n")
        }
      }
    }
  }

  # ---- Worker function ---------------------------------------------------
  fit_worker <- function(p) {
    tryCatch({
      m <- suppressMessages(suppressWarnings(
        cNORM::cnorm.betabinomial(
          age = age, score = score, n = n, weights = weights,
          mode = mode, alpha = p[1], beta = p[2],
          control = control, scale = scale, plot = FALSE
        )
      ))
      d <- cNORM::diagnostics.betabinomial(m)
      list(alpha = p[1], beta = p[2], model = m,
           BIC = if (is.finite(d$BIC)) d$BIC else Inf,
           AIC = d$AIC, logLik = d$log_likelihood,
           converged = isTRUE(d$converged),
           status = if (!is.finite(d$BIC)) "error"
           else if (!isTRUE(d$converged)) "not_converged"
           else "ok",
           message = NA_character_)
    }, error = function(e) {
      list(alpha = p[1], beta = p[2], model = NULL,
           BIC = Inf, AIC = NA_real_, logLik = NA_real_,
           converged = FALSE, status = "error",
           message = conditionMessage(e))
    })
  }

  # ---- Evaluation --------------------------------------------------------
  report <- function(r) {
    tag <- switch(r$status,
                  ok            = "",
                  not_converged = "  (not strictly converged)",
                  error         = paste0("  [error: ", r$message, "]"))
    say(sprintf("  alpha = %d, beta = %d : BIC = %s%s\n",
                r$alpha, r$beta,
                formatC(r$BIC, digits = 3, format = "f"), tag))
  }

  say(sprintf("Evaluating %d model%s ...\n",
              length(pairs), if (length(pairs) == 1) "" else "s"))

  results <- list()
  if (use_parallel && length(pairs) > 1L) {
    # Process in chunks of size n_cores so output appears progressively
    chunks <- split(pairs, ceiling(seq_along(pairs) / n_cores))
    for (chunk in chunks) {
      chunk_res <- parallel::parLapply(cl, chunk, fit_worker)
      for (r in chunk_res) {
        results[[length(results) + 1L]] <- r
        report(r)
      }
    }
  } else {
    for (p in pairs) {
      r <- fit_worker(p)
      results[[length(results) + 1L]] <- r
      report(r)
    }
  }

  # ---- Select best model -------------------------------------------------
  bics <- vapply(results, `[[`, numeric(1), "BIC")
  if (all(!is.finite(bics)))
    stop("Selection failed: no model produced a finite BIC. ",
         "Inspect $selection$evaluated for per-fit messages.")
  current <- results[[which.min(bics)]]

  evaluated <- do.call(rbind, lapply(results, function(r)
    data.frame(alpha = r$alpha, beta = r$beta,
               BIC = r$BIC, AIC = r$AIC,
               logLik = r$logLik,
               converged = r$converged,
               status = r$status,
               message = r$message,
               stringsAsFactors = FALSE)))
  evaluated <- evaluated[order(evaluated$BIC), , drop = FALSE]
  rownames(evaluated) <- NULL

  say(sprintf("\nSelected model: alpha = %d, beta = %d (BIC = %.3f)\n",
              current$alpha, current$beta, current$BIC))

  final_model <- current$model
  if (is.null(final_model))
    stop("Selection failed: the best candidate did not produce a usable model.")

  final_model$selection <- list(
    evaluated = evaluated,
    selected  = list(alpha = current$alpha,
                     beta  = current$beta,
                     BIC   = current$BIC),
    mode      = mode
  )

  if (plot) print(plot(final_model, age = age, score = score, weights = weights))
  final_model
}

# =============================================================================
# cNORM — Comprehensive test suite (Updated for testthat 3e)
# =============================================================================
library(testthat)

# Use 3rd edition to match modern devtools standards
local_edition(3)

# ---------------------------------------------------------------------------
# 0. Shared test fixtures
# ---------------------------------------------------------------------------
set.seed(42)

.age     <- ppvt$age
.score   <- ppvt$raw
.n_items <- 228

.marginals <- data.frame(
  var   = c("sex",  "sex",  "migration", "migration"),
  level = c(1,      2,      0,           1),
  prop  = c(0.51,   0.49,   0.65,        0.35)
)

.elfe_cnorm  <- NULL
.ppvt_cnorm  <- NULL
.bb_model    <- NULL
.shash_model <- NULL

get_elfe <- function() {
  if (is.null(.elfe_cnorm)) {
    suppressMessages({
      .elfe_cnorm <<- cnorm(raw = elfe$raw, group = elfe$group, plot = FALSE)
    })
  }
  .elfe_cnorm
}

get_ppvt <- function() {
  if (is.null(.ppvt_cnorm)) {
    suppressMessages({
      .ppvt_cnorm <<- cnorm(raw = ppvt$raw, group = ppvt$group, plot = FALSE)
    })
  }
  .ppvt_cnorm
}

get_bb <- function() {
  if (is.null(.bb_model)) {
    suppressMessages({
      .bb_model <<- cnorm.betabinomial(.age, .score, n = .n_items, plot = FALSE)
    })
  }
  .bb_model
}

get_shash <- function() {
  if (is.null(.shash_model)) {
    suppressMessages({
      .shash_model <<- cnorm.shash(.age, .score, plot = FALSE)
    })
  }
  .shash_model
}

# =============================================================================
# 1. DATA PREPARATION
# =============================================================================

test_that("rankByGroup returns data.frame with expected columns", {
  skip_on_cran()
  suppressMessages({
    d <- rankByGroup(elfe, group = "group", raw = "raw")
  })
  expect_s3_class(d, "data.frame")
  expect_true(all(c("percentile", "normValue", "n", "m", "md", "sd") %in% colnames(d)))
})

test_that("rankByGroup percentiles lie strictly in (0, 1)", {
  skip_on_cran()
  suppressMessages({
    d <- rankByGroup(elfe, group = "group", raw = "raw")
  })
  expect_true(all(d$percentile > 0 & d$percentile < 1))
})

test_that("rankByGroup descend = TRUE inverts the ranking direction", {
  skip_on_cran()
  d_asc  <- suppressMessages(rankByGroup(elfe, group = "group", raw = "raw", descend = FALSE))
  d_desc <- suppressMessages(rankByGroup(elfe, group = "group", raw = "raw", descend = TRUE))
  expect_true(cor(d_asc$normValue, d_desc$normValue) < 0)
})

test_that("rankBySlidingWindow returns data.frame with expected columns", {
  skip_on_cran()
  d <- rankBySlidingWindow(ppvt, age = "age", raw = "raw", width = 1)
  expect_s3_class(d, "data.frame")
  expect_true(all(c("percentile", "normValue") %in% colnames(d)))
})

test_that("computePowers adds L and A columns", {
  skip_on_cran()
  d <- suppressMessages(rankByGroup(elfe, group = "group", raw = "raw"))
  d <- computePowers(d, k = 4, t = 3)
  expect_true("L4" %in% colnames(d))
  expect_true("A3" %in% colnames(d))
})

test_that("getGroups returns numeric vector of same length", {
  skip_on_cran()
  x <- rnorm(300, 50, 10)
  g <- getGroups(x, n = 5)
  expect_length(g, 300)
})

# =============================================================================
# 2. WEIGHTING AND RAKING
# =============================================================================

test_that("computeWeights returns a positive numeric vector", {
  skip_on_cran()
  w <- suppressMessages(computeWeights(ppvt, .marginals))
  expect_length(w, nrow(ppvt))
  expect_true(all(w > 0))
})

# =============================================================================
# 3. TAYLOR-POLYNOMIAL MODELLING
# =============================================================================

test_that("cnorm returns an object of class 'cnorm'", {
  skip_on_cran()
  expect_s3_class(get_elfe(), "cnorm")
})

test_that("regressionFunction returns a valid string", {
  skip_on_cran()
  f <- regressionFunction(get_elfe())
  expect_true(grepl("raw", f))
})

test_that("checkConsistency returns a logical value", {
  skip_on_cran()
  result <- checkConsistency(get_elfe(), minNorm = 25, maxNorm = 75, silent = TRUE)
  expect_true(is.logical(result))
})

# =============================================================================
# 4. PREDICTION
# =============================================================================

test_that("predictRaw respects clipping bounds", {
  skip_on_cran()
  m <- get_elfe()
  r_high <- predictRaw(95, 3, m, minRaw = 0, maxRaw = 28)
  expect_lte(r_high, 28)
})

test_that("normTable returns data.frame", {
  skip_on_cran()
  tab <- normTable(3, get_elfe(), minNorm = 25, maxNorm = 75)
  expect_s3_class(tab, "data.frame")
})

# =============================================================================
# 5. PARAMETRIC MODELS (Beta-Binomial & ShaSh)
#    Thoroughly tests beta binomial modelling
# =============================================================================

# ---- Helper: simulate beta-binomial data with age trend -------------------
simulate_bb <- function(N = 800, n_items = 40, seed = 42) {
  set.seed(seed)
  age <- runif(N, 6, 12)
  # increasing ability with age, moderate overdispersion
  a <- exp(0.2 + 0.25 * (age - 9))
  b <- exp(1.0 - 0.15 * (age - 9))
  p <- rbeta(N, a, b)
  score <- rbinom(N, size = n_items, prob = p)
  list(age = age, score = score, n = n_items)
}

sim <- simulate_bb()

# Fit once, reuse across tests (mode 2 = default path)
model <- suppressMessages(
  cnorm.betabinomial(sim$age, sim$score, n = sim$n, plot = FALSE)
)


test_that("cnorm.betabinomial (mode 2) fits and converges on simulated data", {
  expect_s3_class(model, "cnormBetaBinomial2")
  expect_equal(model$result$convergence, 0)
  expect_equal(attr(model$result, "max"), sim$n)
  expect_equal(attr(model$result, "N"), length(sim$score))

  d <- diagnostics.betabinomial(model)
  expect_true(d$converged)
  expect_true(is.finite(d$BIC))
})


test_that("log_likelihood2 with precomputed lchoose is identical", {
  X <- bb_design_matrix(standardize(sim$age), 3)
  Z <- bb_design_matrix(standardize(sim$age), 3)
  params <- c(model$alpha_est, model$beta_est)
  lch <- lchoose(sim$n, sim$score)

  expect_equal(
    log_likelihood2(params, X, Z, sim$score, sim$n),
    log_likelihood2(params, X, Z, sim$score, sim$n, lch = lch)
  )
})


test_that("bb_distribution is a proper mid-p distribution", {
  dist <- bb_distribution(a = 2.5, b = 4.0, n = 30)
  expect_equal(sum(dist$Px), 1, tolerance = 1e-12)
  expect_equal(dist$cum[31], 1)
  expect_true(all(diff(dist$perc) > 0))          # strictly increasing
  expect_true(all(dist$perc > 0 & dist$perc < 1))

  # invalid parameters yield NA, not errors
  bad <- bb_distribution(a = -1, b = 2, n = 30)
  expect_true(all(is.na(bad$Px)))
})


test_that("normTable.betabinomial: truncation via m does not renormalize", {
  ages <- c(7, 10)
  full  <- normTable.betabinomial(model, ages, CI = NULL)
  trunc <- normTable.betabinomial(model, ages, m = 20, CI = NULL)

  # truncated table must be an exact head of the full table (bugfix check)
  expect_equal(nrow(trunc[[1]]), 21)
  expect_equal(trunc[[1]]$Px,         full[[1]]$Px[1:21])
  expect_equal(trunc[[1]]$Pcum,       full[[1]]$Pcum[1:21])
  expect_equal(trunc[[1]]$Percentile, full[[1]]$Percentile[1:21])
})


test_that("predict is consistent with normTable and unique-age grouping", {
  ages  <- rep(c(7, 10), each = 3)
  raws  <- c(10, 20, 30, 10, 20, 30)
  preds <- predict(model, ages, raws)

  # cross-check against normTable norm scores
  tab <- normTable.betabinomial(model, c(7, 10), CI = NULL)
  expect_equal(preds[1:3], tab[["7"]]$norm[raws[1:3] + 1],  tolerance = 1e-10)
  expect_equal(preds[4:6], tab[["10"]]$norm[raws[4:6] + 1], tolerance = 1e-10)

  # monotonicity in raw score within age
  expect_true(all(diff(preds[1:3]) > 0))

  # out-of-range and non-integer scores yield NA with warning
  expect_warning(p2 <- predict(model, c(8, 8), c(sim$n + 5, 12.5)))
  expect_true(all(is.na(p2)))
})


test_that("mode 1 (cnorm.betabinomial1) remains functional (compatibility)", {
  m1 <- suppressMessages(
    cnorm.betabinomial(sim$age, sim$score, n = sim$n, mode = 1, plot = FALSE)
  )
  expect_s3_class(m1, "cnormBetaBinomial")

  p <- predict(m1, c(7, 9, 11), c(15, 20, 25))
  expect_true(all(is.finite(p)))

  # mean-preserving fallback in predictCoefficients: mu/(a+b) relation holds
  pc <- predictCoefficients(m1, c(7, 9, 11))
  expect_true(all(pc$a > 0 & pc$b > 0))
  expect_equal(sim$n * pc$a / (pc$a + pc$b), pc$mu, tolerance = 0.05)
})


test_that("betaCoefficients recovers parameters approximately", {
  set.seed(1)
  x <- rbinom(50000, 25, rbeta(50000, 3, 5))
  cf <- betaCoefficients(x, 25)
  expect_equal(cf[1], 3, tolerance = 0.25)
  expect_equal(cf[2], 5, tolerance = 0.35)
})

test_that("analytic beta-binomial gradient matches numerical gradient", {
  set.seed(1)
  n <- 60
  age <- runif(400, 6, 12); a_std <- (age - mean(age))/sd(age)
  y <- rbinom(400, n, rbeta(400, 4, 3))
  X <- cNORM:::bb_design_matrix(a_std, 3)
  Z <- cNORM:::bb_design_matrix(a_std, 3)
  p0 <- c(log(4), 0.2, -0.1, 0.05, log(3), -0.15, 0.1, 0)
  g_ana <- cNORM:::gradient_log_likelihood2(p0, X, Z, y, n)
  g_num <- numDeriv::grad(cNORM:::log_likelihood2, p0, X = X, Z = Z, y = y, n = n)
  expect_equal(g_ana, g_num, tolerance = 1e-6)
})



test_that("plot and summary run without error", {
  expect_s3_class(
    plot(model, age = sim$age, score = sim$score),
    "ggplot"
  )
  out <- capture.output(
    d <- summary(model, age = sim$age, score = sim$score)
  )
  expect_true(d$R2 > 0.9)   # norm score recovery on well-behaved data
})


test_that("cnorm.shash returns correct class", {
  skip_on_cran()
  expect_s3_class(get_shash(), "cnormShash")
})

test_that("predict.cnormShash returns finite scores", {
  skip_on_cran()
  pred <- predict(get_shash(), age = c(7, 10), score = c(100, 130))
  expect_length(pred, 2)
  expect_true(all(is.finite(pred)))
})

# =============================================================================
# 6. PLOTTING & COMPARISON
# =============================================================================

test_that("plotPercentiles returns a ggplot", {
  skip_on_cran()
  p <- plotPercentiles(get_elfe())
  expect_s3_class(p, "ggplot")
})

test_that("compare() returns ggplot", {
  skip_on_cran()
  suppressMessages({
    p <- compare(get_ppvt(), get_bb(), age = .age, score = .score)
  })
  expect_s3_class(p, "ggplot")
})

# =============================================================================
# 7. ERROR HANDLING
# =============================================================================

test_that("rankByGroup stops on missing columns", {
  skip_on_cran()
  expect_error(rankByGroup(elfe, group = "missing", raw = "raw"))
})

test_that("predict.cnormShash stops when score is missing", {
  skip_on_cran()
  expect_error(predict(get_shash(), age = c(7, 10)))
})

# =============================================================================
# 8. WEIGHTED MODELLING
# =============================================================================

test_that("bestModel works with case weights", {
  data <- prepareData(elfe)
  w <- runif(nrow(data), 0.5, 2)
  m <- bestModel(data, weights = w, plot = FALSE)
  expect_s3_class(m, "cnormModel")
  expect_false(anyNA(m$coefficients))

  # weights = FALSE must ignore existing weights without error
  m2 <- bestModel(data, weights = FALSE, plot = FALSE)
  expect_s3_class(m2, "cnormModel")
})

#' S3 method for printing model selection information
#'
#' After conducting the model fitting procedure on the data set, the best fitting
#' model has to be chosen. The print function shows the R2 and other information
#' on the different best fitting models with increasing number of predictors.
#' @param x The model from the 'bestModel' function or a cnorm object
#' @param ... additional parameters
#' @return A table with information criteria
#' @export
#'
#' @family model
print.cnorm <- function(x, ...) {  UseMethod("print") }

#' S3 method for printing the results and regression function of a cnorm model
#'
#' @param object A regression model or cnorm object
#' @param ... additional parameters
#' @return A report on the regression function, weights, R2 and RMSE
#' @export
#' @family model
summary.cnorm <- function(object, ...) {  UseMethod("summary") }



#' S3 function for plotting cnorm objects
#'
#' @param x the cnorm object
#' @param y the type of plot as a string, can be one of
#' 'raw' (1), 'norm' (2), 'curves' (3), 'percentiles' (4), 'series' (5), 'subset' (6),
#' or 'derivative' (7), either as a string or the according index
#' @param ... additional parameters for the specific plotting function
#'
#' @export
#'
#' @family plot
plot.cnorm <- function(x, y, ...) {  UseMethod("plot") }


print.cnorm <- printSubset
plot.cnorm <- plotCnorm
summary.cnorm <- modelSummary

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

#' Gauss-Hermite nodes and weights (probabilist form)
#'
#' Computes nodes and weights via the Golub-Welsch algorithm, rescaled so
#' that E[g(Z)] with Z ~ N(0,1) is approximated by sum(w * g(z)).
#' The weights sum to 1; no external dependencies.
#'
#' @param n Number of quadrature nodes.
#' @return A list with elements \code{z} (nodes on the standard normal scale)
#'   and \code{w} (probability weights summing to 1).
#' @keywords internal
#' @noRd
.gaussHermite <- function(n) {
  i <- seq_len(n - 1)
  J <- matrix(0, n, n)
  J[cbind(i, i + 1)] <- J[cbind(i + 1, i)] <- sqrt(i / 2)
  e <- eigen(J, symmetric = TRUE)
  list(z = sqrt(2) * e$values,     # nodes on N(0,1) scale
       w = e$vectors[1, ]^2)       # weights, sum to 1
}

#' Evaluate polynomial via Horner's method
#'
#' @param x Numeric vector of evaluation points.
#' @param coeff Coefficients in ascending order (c0, c1, ..., ck).
#' @keywords internal
#' @noRd
.evalPoly <- function(x, coeff) {
  y <- rep(coeff[length(coeff)], length(x))
  if (length(coeff) > 1) {
    for (j in (length(coeff) - 1):1) y <- y * x + coeff[j]
  }
  y
}

#' Central moments from a discrete/weighted representation
#'
#' @param x Values (quadrature evaluations or discrete support points).
#' @param w Probability weights (must sum to 1).
#' @return Named list: mean, variance, sd, skewness, kurtosis (excess).
#' @keywords internal
#' @noRd
.weightedMoments <- function(x, w) {
  m  <- sum(w * x)
  d  <- x - m
  v  <- sum(w * d^2)
  if (v <= .Machine$double.eps^0.5 * max(1, m^2)) {
    return(list(mean = m, variance = max(v, 0), sd = sqrt(max(v, 0)),
                skewness = NA_real_, kurtosis = NA_real_))
  }
  m3 <- sum(w * d^3)
  m4 <- sum(w * d^4)
  list(mean     = m,
       variance = v,
       sd       = sqrt(v),
       skewness = m3 / v^1.5,
       kurtosis = m4 / v^2 - 3)     # excess kurtosis; 0 for the normal
}

#' Assemble result data.frame
#' @keywords internal
#' @noRd
.momentsDataFrame <- function(age, momentsList, method) {
  df <- data.frame(
    age      = age,
    mean     = vapply(momentsList, `[[`, numeric(1), "mean"),
    sd       = vapply(momentsList, `[[`, numeric(1), "sd"),
    variance = vapply(momentsList, `[[`, numeric(1), "variance"),
    skewness = vapply(momentsList, `[[`, numeric(1), "skewness"),
    kurtosis = vapply(momentsList, `[[`, numeric(1), "kurtosis"),
    row.names = NULL
  )
  attr(df, "method") <- method
  df
}

#' Input check for the age vector
#' @keywords internal
#' @noRd
.checkAge <- function(age) {
  if (!is.numeric(age) || length(age) == 0 || any(!is.finite(age)))
    stop("'age' must be a numeric vector with finite values.", call. = FALSE)
  invisible(age)
}


# ---------------------------------------------------------------------------
# Generic
# ---------------------------------------------------------------------------

#' Model-Implied Distributional Moments at Specific Ages
#'
#' Computes the mean, standard deviation, variance, skewness and (excess)
#' kurtosis of the raw score distribution implied by a fitted cNORM model
#' at one or more ages (or, more generally, values of the explanatory
#' variable). The moments are model-implied population moments of the
#' conditional raw score distribution, censored at the bounds of the raw
#' score range \code{[minRaw, maxRaw]} for consistency across model families.
#'
#' @details
#' The computation strategy depends on the model family:
#' \describe{
#'   \item{Taylor polynomial (\code{cnorm})}{The bivariate regression function
#'     is collapsed at the specified age into a univariate polynomial in the
#'     norm score (location) variable. Moments are then obtained by
#'     Gauss-Hermite quadrature, which is mathematically exact for
#'     polynomial quantile functions (up to the censoring at
#'     \code{minRaw}/\code{maxRaw}).}
#'   \item{Beta-binomial (\code{cnormBetaBinomial},
#'     \code{cnormBetaBinomial2})}{Moments are computed exactly by summation
#'     over the discrete probability mass function on the support
#'     \code{0:n}, using the age-specific predicted \eqn{\alpha} and
#'     \eqn{\beta} parameters. This respects the discreteness of the
#'     distribution; no continuity approximation is involved.}
#'   \item{SHASH (\code{cnormShash})}{Moments are obtained by Gauss-Hermite
#'     quadrature of the quantile function \code{qshash} evaluated at the
#'     age-specific distribution parameters, censored at
#'     \code{minRaw}/\code{maxRaw}.}
#' }
#'
#' Kurtosis is reported as \emph{excess} kurtosis (0 for the normal
#' distribution).
#'
#' Note that the skewness and kurtosis of the \emph{censored} distribution
#' are reported. For well-fitting models whose raw score range covers the
#' probability mass of the conditional distribution, censoring effects are
#' negligible; for distributions with substantial floor or ceiling effects,
#' the censored moments are the substantively meaningful ones.
#'
#' @param model A model object of class \code{cnorm},
#'   \code{cnormBetaBinomial}, \code{cnormBetaBinomial2} or
#'   \code{cnormShash}.
#' @param age A numeric vector of ages (values of the explanatory variable)
#'   at which to compute the moments.
#' @param ... Additional parameters passed to the methods, e.g.
#'   \code{nNodes}.
#'
#' @return A \code{data.frame} with one row per age and the columns
#'   \code{age}, \code{mean}, \code{sd}, \code{variance}, \code{skewness}
#'   and \code{kurtosis} (excess). The computation method is stored in the
#'   attribute \code{"method"}.
#'
#' @examples
#' \dontrun{
#' # Taylor polynomial model
#' model <- cnorm(raw = elfe$raw, group = elfe$group)
#' predictMoments(model, age = c(2.25, 2.75, 3.25, 3.75, 4.25))
#'
#' # Beta-binomial model
#' bb <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
#' predictMoments(bb, age = seq(4, 16, by = 2))
#' }
#'
#' @references
#' Isserlis, L. (1918). On a formula for the product-moment coefficient of
#' any order of a normal frequency distribution in any number of variables.
#' Biometrika, 12(1/2), 134-139.
#'
#' Jones, M. C. & Pewsey, A. (2009). Sinh-arcsinh distributions.
#' Biometrika, 96(4), 761-780.
#'
#' @seealso \code{\link{normTable}}, \code{\link{predictNorm}},
#'   \code{\link{predictRaw}}
#' @export
predictMoments <- function(model, age, ...) {
  UseMethod("predictMoments")
}


# ---------------------------------------------------------------------------
# Default method
# ---------------------------------------------------------------------------

#' @rdname predictMoments
#' @export
predictMoments.default <- function(model, age, ...) {
  stop("predictMoments is not defined for objects of class '",
       paste(class(model), collapse = "', '"),
       "'. Supported classes: cnorm, cnormBetaBinomial, cnormBetaBinomial2, ",
       "cnormShash.", call. = FALSE)
}


# ---------------------------------------------------------------------------
# Taylor polynomial models
# ---------------------------------------------------------------------------

#' @rdname predictMoments
#' @param nNodes Number of Gauss-Hermite quadrature nodes (default 100).
#'   Only relevant for the Taylor and SHASH methods; ignored for
#'   beta-binomial models, which are computed exactly by summation.
#' @export
predictMoments.cnorm <- function(model, age, nNodes = 100, ...) {
  .checkAge(age)
  m <- model$model                        # inner regression model

  gh <- .gaussHermite(nNodes)
  L  <- m$scaleM + m$scaleSD * gh$z       # nodes on the norm score scale

  momentsList <- lapply(age, function(a) {
    cf <- calcPolyInLBase2(raw = 0, age = a,
                           coeff = m$coefficients, k = m$k)
    x  <- .evalPoly(L, cf)
    x  <- pmin(pmax(x, m$minRaw), m$maxRaw)   # censoring
    .weightedMoments(x, gh$w)
  })

  .momentsDataFrame(age, momentsList,
                    method = "Taylor polynomial / Gauss-Hermite quadrature")
}


# ---------------------------------------------------------------------------
# Beta-binomial models (exact, via pmf summation)
# ---------------------------------------------------------------------------

#' Retrieve age-specific alpha/beta parameters of a beta-binomial model
#'
#' NOTE: Adapt the accessor calls below to the actual internal predictor
#' functions of the development version (e.g. predictCoefficients /
#' predictCoefficients2). The contract of this helper: return a data.frame
#' with columns 'a' and 'b', one row per age.
#'
#' @keywords internal
#' @noRd
.betaBinomialParams <- function(model, age) {
  if (inherits(model, "cnormBetaBinomial2")) {
    pars <- predictCoefficients2(model, age)   # mu/sigma parameterization
  } else {
    pars <- predictCoefficients(model, age)    # direct alpha/beta
  }
  if (!all(c("a", "b") %in% names(pars)))
    stop("Internal parameter prediction did not return columns 'a' and 'b'.",
         call. = FALSE)
  pars
}

#' Exact moments of the beta-binomial distribution via pmf summation
#'
#' Numerically stable through log-space computation of the pmf. Exact on
#' the discrete support 0:n; censoring is implicit since the support
#' coincides with the raw score range.
#'
#' @keywords internal
#' @noRd
.betaBinomialMoments <- function(a, b, n) {
  x    <- 0:n
  logp <- lchoose(n, x) + lbeta(x + a, n - x + b) - lbeta(a, b)
  p    <- exp(logp)
  p    <- p / sum(p)                      # guard against numeric drift
  .weightedMoments(x, p)
}

#' @rdname predictMoments
#' @export
predictMoments.cnormBetaBinomial <- function(model, age, ...) {
  .checkAge(age)

  # Number of items; adapt accessor if stored differently
  n <- attr(model$result, "max")
  if (is.null(n))
    stop("Could not retrieve the number of items 'n' from the model object.",
         call. = FALSE)

  pars <- .betaBinomialParams(model, age)

  momentsList <- lapply(seq_along(age), function(i) {
    .betaBinomialMoments(pars$a[i], pars$b[i], n)
  })

  .momentsDataFrame(age, momentsList,
                    method = "Beta-binomial / exact pmf summation")
}

#' @rdname predictMoments
#' @export
predictMoments.cnormBetaBinomial2 <- predictMoments.cnormBetaBinomial


# ---------------------------------------------------------------------------
# SHASH models
# ---------------------------------------------------------------------------

#' @keywords internal
#' @noRd
.shashParams <- function(model, age) {
  pars <- predictCoefficients_shash(model, age)
  needed <- c("mu", "sigma", "epsilon", "delta")
  if (!all(needed %in% names(pars)))
    stop("predictCoefficients_shash did not return columns ",
         paste(needed, collapse = ", "), ".", call. = FALSE)
  pars
}

#' @rdname predictMoments
#' @export
predictMoments.cnormShash <- function(model, age, nNodes = 100, ...) {
  .checkAge(age)

  minRaw <- attr(model$result, "min")
  maxRaw <- attr(model$result, "max")
  if (is.null(minRaw) || is.null(maxRaw))
    stop("Could not retrieve minRaw/maxRaw from the model object.",
         call. = FALSE)

  gh   <- .gaussHermite(nNodes)
  p    <- pnorm(gh$z)
  pars <- .shashParams(model, age)

  momentsList <- lapply(seq_along(age), function(i) {
    x <- qshash(p,
                mu      = pars$mu[i],
                sigma   = pars$sigma[i],
                epsilon = pars$epsilon[i],
                delta   = pars$delta[i])
    # Censoring also neutralizes +/-Inf from p ~ 0 or 1 at extreme nodes
    x <- pmin(pmax(x, minRaw), maxRaw)
    .weightedMoments(x, gh$w)
  })

  .momentsDataFrame(age, momentsList,
                    method = "SHASH / Gauss-Hermite quadrature")
}

.onLoad <- function(libname, pkgname) {
  op <- options()
  op.cNORM <-
    list(
      cNORM.install.args = "",
      cNORM.name = "Wolfgang Lenhard, Alexandra Lenhard & Sebastian Gary",
      cNORM.desc.author = "Wolfgang Lenhard <wolfgang.lenhard@uni-wuerzburg.de> [aut, cre];
      Alexandra Lenhard <lenhard@psychometrica.de> [aut];  Sebastian Gary [aut]",
      cNORM.desc.license = "AGPL-3",
      cNORM.desc = list()
    )
  toset <- !(names(op.cNORM) %in% names(op))
  if (any(toset)) {
    options(op.cNORM[toset])
  }

  invisible()
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(
    "Good morning star-shine!\ncNORM is free software. Please report bugs: https://github.com/WLenhard/cNORM/issues"
  )
}


#' Prepare data for modeling in one step (convenience method)
#'
#' This is a convenience method to either load the inbuilt sample dataset, or
#' to provide a data frame with the variables "raw" (for the raw scores) and "group"
#' The function ranks the data within groups, computes norm values, powers of the norm
#' scores and interactions. Afterwards, you can use these preprocessed data to
#' determine the best fitting model.
#'
#' The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
#' \code{computePowers} and \code{prepareData} are usually not called directly, but accessed
#' through other functions like \code{cnorm}.
#'
#' @param data data.frame with a grouping variable named 'group' and a raw score variable
#' named 'raw'.
#' @param group grouping variable in the data, e. g. age groups, grades ...
#' Setting group = FALSE deactivates modeling in dependence of age. Use this in case you do want
#' conventional norm tables.
#' @param raw the raw scores
#' @param age the continuous explanatory variable; by default set to "group"
#' @param width if a width is provided, the function switches to rankBySlidingWindow to determine the
#' observed raw scores, otherwise, ranking is done by group (default)
#' @param weights Vector or variable name in the dataset with weights for each individual case. It can be used
#' to compensate for moderate imbalances due to insufficient norm data stratification. Weights should be numerical
#' and positive. Please use the 'computeWeights' function for this purpose.
#' @param scale type of norm scale, either T (default), IQ, z or percentile (= no
#' transformation); a double vector with the mean and standard deviation can as well,
#' be provided f. e. c(10, 3) for Wechsler scale index point
#' @param descend ranking order (default descent = FALSE): inverses the
#' ranking order with higher raw scores getting lower norm scores; relevant
#' for example when norming error scores, where lower scores mean higher
#' performance
#' @param k The power parameter, default = 4
#' @param t the age power parameter (default NULL). If not set, cNORM automatically uses k. The age power parameter
#' can be used to specify the k to produce rectangular matrices and specify the course of scores per  independently from k
#' @param silent set to TRUE to suppress messages
#' @return data frame including the norm scores, powers and interactions of the norm score and
#' grouping variable
#' @examples
#' \dontrun{
#' # conducts ranking and computation of powers and interactions with the 'elfe' dataset
#' data.elfe <- prepareData(elfe)
#'
#' # use vectors instead of data frame
#' data.elfe <- prepareData(raw=elfe$raw, group=elfe$group)
#'
#' # variable names can be specified as well, here with the BMI data included in the package
#' data.bmi <- prepareData(CDC, group = "group", raw = "bmi", age = "age")
#' }
#'
#' # modeling with only one group with the 'elfe' dataset as an example
#' # this results in conventional norming
#' data.elfe2 <- prepareData(data = elfe, group = FALSE)
#' m <- bestModel(data.elfe2)
#' @export
#' @family prepare
prepareData <- function(data = NULL,
                        group = "group",
                        raw = "raw",
                        age = "group",
                        k = 4,
                        t = NULL,
                        width = NA,
                        weights = NULL,
                        scale = "T",
                        descend = FALSE,
                        silent = FALSE) {

  # ---- Assemble working data.frame ------------------------------------------
  if (is.null(data)) {
    normData <- data.frame(raw = raw)
    raw <- "raw"
  } else {
    normData <- as.data.frame(data)
    if (is.numeric(raw) && length(raw) == nrow(normData)) {
      normData$raw <- raw
      raw <- "raw"
    }
  }
  n_rows <- nrow(normData)

  # ---- Grouping variable ----------------------------------------------------
  has_group <- TRUE
  if (is.logical(group)) {
    has_group <- FALSE                      # group = FALSE -> conventional norming
  } else if (is.numeric(group)) {
    if (length(group) != n_rows)
      stop("Length of the group vector (", length(group),
           ") does not match the number of cases (", n_rows, ").")
    normData$group <- group
    group <- "group"
  } else if (is.character(group)) {
    if (!(group %in% colnames(normData)))
      stop("Grouping variable '", group, "' does not exist in data object.")
  } else {
    stop("Invalid 'group' specification.")
  }

  # ---- Raw score variable ---------------------------------------------------
  if (is.numeric(raw)) {                    # only reachable if it did NOT match n_rows
    stop("Length of the raw score vector (", length(raw),
         ") does not match the number of cases (", n_rows, ").")
  } else if (is.character(raw)) {
    if (!(raw %in% colnames(normData)))
      stop("Raw score variable '", raw, "' does not exist in data object.")
  } else {
    stop("Invalid 'raw' specification.")
  }

  # ---- Age (continuous modelling) variable ----------------------------------
  has_age <- FALSE
  if (!has_group) {
    age   <- FALSE                          # conventional norming: no age modelling
    width <- NA
  } else if (is.numeric(age)) {
    if (length(age) != n_rows)
      stop("Length of the age vector (", length(age),
           ") does not match the number of cases (", n_rows, ").")
    normData$age <- age
    age <- "age"
    has_age <- TRUE
  } else if (is.character(age)) {
    if (age %in% colnames(normData)) {
      has_age <- TRUE
    } else {
      # age column absent -> use the grouping variable as the age proxy,
      # and disable the sliding window (no continuous age available)
      age   <- group
      width <- NA
      has_age <- TRUE
    }
  } else {
    stop("Invalid 'age' specification.")
  }

  # ---- Type checks (numeric requirements) -----------------------------------
  if (has_group && !is.numeric(normData[[group]]))
    warning("Grouping variable '", group, "' has to be numeric.")
  if (!is.numeric(normData[[raw]]))
    warning("Raw variable '", raw, "' has to be numeric.")
  if (has_age && !is.numeric(normData[[age]]))
    warning("Age variable '", age, "' has to be numeric.")

  # ---- Age / group range plausibility ---------------------------------------
  if (has_group && has_age && group != age &&
      is.numeric(normData[[group]]) && is.numeric(normData[[age]])) {
    rng_age <- range(normData[[age]],   na.rm = TRUE)
    rng_grp <- range(normData[[group]], na.rm = TRUE)
    if (rng_age[2] < rng_grp[1] || rng_age[1] > rng_grp[2]) {
      warning("The range of the age and group variable do not match. ",
              "Please specify a grouping variable whose values relate to the range of the age variable. ",
              "You can automatically generate a grouping variable by using the 'rankBySlidingWindow' ",
              "function and setting a desired number of groups with the 'nGroup' parameter.")
    }
  }

  # ---- Dispatch decision ----------------------------------------------------
  use_group_ranking <- is.null(width) || (length(width) == 1L && is.na(width))

  # ---- Remove cases with missing age in the GROUP-ranking path ---------------
  # rankByGroup() does not see 'age', yet computePowers() will model on it.
  # Keep a supplied weights *vector* aligned with the rows we drop.
  if (use_group_ranking && has_age &&
      (age %in% colnames(normData)) && anyNA(normData[[age]])) {
    age_ok <- !is.na(normData[[age]])
    if (!silent)
      message("Excluding ", sum(!age_ok), " case(s) with missing age before modelling.")
    if (!is.null(weights) && !is.character(weights) &&
        length(weights) == nrow(normData))
      weights <- weights[age_ok]
    normData <- normData[age_ok, , drop = FALSE]
  }

  # ---- Ranking --------------------------------------------------------------
  if (use_group_ranking) {
    normData <- rankByGroup(
      data    = normData,
      group   = group,
      raw     = raw,
      scale   = scale,
      descend = descend,
      weights = weights,
      silent  = silent
    )
  } else {
    normData <- rankBySlidingWindow(
      data    = normData,
      age     = age,
      raw     = raw,
      group   = group,        # passed for the 'group' attribute / plotting
      width   = width,
      weights = weights,
      scale   = scale,
      descend = descend,
      silent  = silent
    )
  }

  # ---- Powers & interactions ------------------------------------------------
  if (has_group) {
    normData <- computePowers(normData, k = k, t = t,
                              norm = "normValue", age = age, silent = silent)
  } else {
    normData <- computePowers(normData, k = k, t = t,
                              norm = "normValue", silent = silent)
  }

  return(normData)
}


#' Determine the norm scores of the participants in each subsample
#'
#' This is the initial step, usually done in all kinds of test norming projects,
#' after the scale is constructed and the norm sample is established. First,
#' the data is grouped according to a grouping variable and afterwards, the percentile
#' for each raw value is retrieved. The percentile can be used for the modeling
#' procedure, but in case, the samples to not deviate too much from normality,
#' T, IQ or z scores can be computed via a normal rank procedure based on the
#' inverse cumulative normal distribution. In case of bindings, we use the medium rank
#' and there are different methods for estimating the percentiles (default RankIt).
#'
#' @section Remarks on using covariates:
#' So far the inclusion of a binary covariate is experimental and far from optimized.
#' The according variable name has to be specified in the ranking procedure
#' and the modeling includes this in the further process. At the moment, during ranking
#' the data are split into the according cells group x covariate, which leads to small
#' sample sizes. Please take care to have enough cases in each combination. Additionally,
#' covariates can lead to unstable modeling solutions. The question, if it is really
#' reasonable to include covariates when norming a test is a decision beyond the pure data
#' modeling. Please use with care or alternatively split the dataset into the two groups
#' beforehand and model them separately.
#'
#' The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
#' \code{computePowers} and \code{prepareData} are usually not called directly, but accessed
#' through other functions like \code{cnorm}.
#'
#' @param data data.frame with norm sample data. If no data.frame is provided, the raw score
#' and group vectors are directly used
#' @param group name of the grouping variable (default 'group') or numeric vector, e. g. grade, setting
#' group to FALSE cancels grouping (data is treated as one group)
#' @param raw name of the raw value variable (default 'raw') or numeric vector
#' @param weights Vector or variable name in the dataset with weights for each individual case. It can be used
#' to compensate for moderate imbalances due to insufficient norm data stratification. Weights should be numerical
#' and positive.  Please use the 'computeWeights' function for this purpose.
#' @param method Ranking method in case of bindings, please provide an index,
#' choosing from the following methods: 1 = Blom (1958), 2 = Tukey (1949),
#' 3 = Van der Warden (1952), 4 = Rankit (default), 5 = Levenbach (1953),
#' 6 = Filliben (1975), 7 = Yu & Huang (2001)
#' @param scale type of norm scale, either T (default), IQ, z or percentile (= no
#' transformation); a double vector with the mean and standard deviation can as well,
#' be provided f. e. c(10, 3) for Wechsler scale index points
#' @param descend ranking order (default descent = FALSE): inverses the
#' ranking order with higher raw scores getting lower norm scores; relevant
#' for example when norming error scores, where lower scores mean higher
#' performance
#' @param descriptives If set to TRUE (default), information in n, mean, median and
#' standard deviation per group is added to each observation
#' @param na.rm remove values, where the percentiles could not be estimated,
#' most likely happens in the context of weighting
#' @param silent set to TRUE to suppress messages
#' @return the dataset with the percentiles and norm scales per group
#'
#' @examples
#' \dontrun{
#' # Transformation with default parameters: RankIt and converting to T scores
#' data.elfe <- rankByGroup(elfe, group = "group") # using a data frame with vector names
#' data.elfe2 <- rankByGroup(raw=elfe$raw, group=elfe$group) # use vectors for raw score and group
#'
#' # Transformation into Wechsler scores with Yu & Huang (2001) ranking procedure
#' data.elfe <- rankByGroup(raw = elfe$raw, group = elfe$group, method = 7, scale = c(10, 3))
#'
#' # cNORM can as well be used for conventional norming, in case no group is given
#' d <- rankByGroup(raw = elfe$raw)
#' d <- computePowers(d)
#' m <- bestModel(d)
#' rawTable(0, m) # please use an arbitrary value for age when generating the tables
#' }
#'
#' @seealso rankBySlidingWindow, computePowers, computeWeights, weighted.rank
#' @export
#' @family prepare
rankByGroup <- function(data = NULL,
                        group = "group",
                        raw = "raw",
                        weights = NULL,
                        method = 4,
                        scale = "T",
                        descend = FALSE,
                        descriptives = TRUE,
                        na.rm = TRUE,
                        silent = FALSE) {

  # ---- assemble working data.frame ------------------------------------------
  if (is.null(data)) {
    d <- data.frame(raw = raw)
    if (is.numeric(group)) {
      if (length(group) != nrow(d))
        stop("Length of the group vector (", length(group),
             ") does not match the number of cases (", nrow(d), ").")
      d$group <- group
      group <- "group"
    } else {
      group <- FALSE
    }
    raw <- "raw"
  } else {
    d <- as.data.frame(data)

    # Grouping variable: existence checked BEFORE assignment
    if (is.character(group)) {
      if (!(group %in% colnames(d)))
        stop("Grouping variable '", group, "' does not exist in the data object.")
      d$group <- d[[group]]
      group <- "group"
    } else if (is.numeric(group)) {
      if (length(group) != nrow(d))
        stop("Length of the group vector (", length(group),
             ") does not match the number of cases (", nrow(d), ").")
      d$group <- group
      group <- "group"
    } # else: group is logical FALSE -> no grouping

    # Raw score variable
    if (is.character(raw)) {
      if (!(raw %in% colnames(d)))
        stop("Raw value variable '", raw, "' does not exist in the data object.")
      d$raw <- d[[raw]]
      raw <- "raw"
    } else if (is.numeric(raw)) {
      if (length(raw) != nrow(d))
        stop("Length of the raw score vector (", length(raw),
             ") does not match the number of cases (", nrow(d), ").")
      d$raw <- raw
      raw <- "raw"
    }
  }

  # ---- weights --------------------------------------------------------------
  weighting <- NULL
  has_weights <- FALSE
  if (!is.null(weights)) {
    if (is.character(weights)) {
      if (!(weights %in% colnames(d))) {
        if (!silent) warning("Weighting variable '", weights,
                             "' does not exist in dataset. Proceeding without weighting.")
        weights <- NULL
      } else {
        weighting <- as.numeric(d[[weights]])
        has_weights <- TRUE
      }
    } else {
      if (length(weights) != nrow(d)) {
        if (!silent) warning("Length of weights vector has to match the number of cases. Proceeding without weighting.")
        weights <- NULL
      } else {
        d$weights <- as.numeric(weights)
        weighting <- as.numeric(weights)
        weights <- "weights"
        has_weights <- TRUE
      }
    }
  }

  # ---- drop missing values (raw / group / weight) ---------------------------
  group_has_na  <- if (is.logical(group)) FALSE else anyNA(d[["group"]])
  raw_has_na    <- anyNA(d[["raw"]])
  weight_has_na <- has_weights && anyNA(weighting)

  if (group_has_na || raw_has_na || weight_has_na) {
    if (!silent) message("Missing values found in grouping, raw score or weight variable. Excluding affected cases.")
    valid_idx <- !is.na(d[["raw"]])
    if (!is.logical(group)) valid_idx <- valid_idx & !is.na(d[["group"]])
    if (has_weights)        valid_idx <- valid_idx & !is.na(weighting)
    d <- d[valid_idx, , drop = FALSE]
    if (has_weights) weighting <- weighting[valid_idx]
  }

  # ---- validate weight positivity -------------------------------------------
  if (has_weights && any(weighting <= 0)) {
    if (!silent) warning("Weights must be positive, non-zero values. Proceeding without weighting.")
    has_weights <- FALSE; weighting <- NULL; weights <- NULL
  }

  # ---- plotting-position constants ------------------------------------------
  numerator   <- c(-3/8, -1/3, 0, -1/2, -0.3, -0.3175, -0.326)
  denominator <- c( 1/4,  1/3, 1,    0,  0.4,  0.365,   0.348)

  if (method < 1 || method > length(numerator)) {
    if (!silent) message("Method parameter out of range, setting to RankIt (4).")
    method <- 4
  }
  num <- numerator[method]
  den <- denominator[method]
  sign_mult <- if (descend) -1 else 1

  # ---- ranking --------------------------------------------------------------
  if (is.logical(group) && !group) {
    if (!silent) message("No grouping variable specified. Ranking without grouping.")
    n_cases <- nrow(d)

    x <- sign_mult * d[["raw"]]
    ranks <- if (has_weights) weighted.rank(x, weighting) else rank(x)
    d$percentile <- (ranks + num) / (n_cases + den)

    if (descriptives) {
      d$n  <- n_cases
      d$m  <- mean(d[["raw"]])
      d$md <- median(d[["raw"]])
      d$sd <- if (n_cases > 1) sd(d[["raw"]]) else NA_real_
    }

  } else {
    idx_split <- split(seq_len(nrow(d)), d[["group"]])
    ranks   <- numeric(nrow(d))
    n_group <- numeric(nrow(d))

    for (idx in idx_split) {
      x <- sign_mult * d[["raw"]][idx]
      ranks[idx] <- if (has_weights) weighted.rank(x, weighting[idx]) else rank(x)
      n_group[idx] <- length(idx)
    }
    d$percentile <- (ranks + num) / (n_group + den)

    if (descriptives) {
      raw_split <- split(d[["raw"]], d[["group"]])
      grp_m  <- vapply(raw_split, mean,   numeric(1))
      grp_md <- vapply(raw_split, median, numeric(1))
      grp_sd <- vapply(raw_split, function(z) if (length(z) > 1) sd(z) else NA_real_, numeric(1))
      match_idx <- match(as.character(d[["group"]]), names(raw_split))

      d$n  <- n_group
      d$m  <- grp_m[match_idx]
      d$md <- grp_md[match_idx]
      d$sd <- grp_sd[match_idx]
    }
  }

  # ---- norm scale -----------------------------------------------------------
  d$percentile <- clipPercentile(d$percentile)
  sc <- applyNormScale(d$percentile, scale)
  d$normValue <- sc$normValue

  attr(d, "group")     <- group
  attr(d, "age")       <- group
  attr(d, "raw")       <- raw
  attr(d, "scaleMean") <- sc$M
  attr(d, "scaleSD")   <- sc$SD
  attr(d, "descend")   <- descend
  attr(d, "normValue") <- "normValue"
  attr(d, "width")     <- NA
  attr(d, "weights")   <- weights

  if (na.rm) {
    naPerc <- sum(is.na(d$percentile))
    if (naPerc > 0) {
      if (!silent) message("Could not determine manifest percentile for ", naPerc, " cases. These will be dropped.")
      d <- d[!is.na(d$percentile), , drop = FALSE]
    }
  }

  if (descriptives && nrow(d) > 0 && min(d$n) < 30 && !silent) {
    warning("The dataset includes cases whose percentile depends on fewer than 30 cases. Consider redividing the cases over the grouping variable or reducing 'k'.")
  }

  return(d)
}

#' Determine the norm scores of the participants by sliding window
#'
#' The function retrieves all individuals in the predefined age range (x +/- width/2)
#' around each case and ranks that individual based on this individually drawn sample.
#' This function can be directly used with a continuous age variable in order to avoid
#' grouping. When collecting data on the basis of a continuous age variable, cases
#' located far from the mean age of the group receive distorted percentiles when building
#' discrete groups and generating percentiles with the traditional approach. The distortion
#' increases with distance from the group mean and this effect can be avoided by the
#' sliding window. Nonetheless, please ensure, that the optional grouping variable in fact
#' represents the correct mean age of the respective age groups, as this variable is
#' later on used for displaying the manifest data in the percentile plots.
#'
#' In case of bindings, the function uses the medium rank and applies the algorithms
#' already described in the \code{\link{rankByGroup}} function. At the upper and lower end of the
#' data sample, the sliding stops and the sample is drawn from the interval min + width and
#' max - width, respectively.
#'
#' @section Remarks on using covariates:
#' So far the inclusion of a binary covariate is experimental and far from optimized.
#' The according variable name has to be specified in the ranking procedure
#' and the modeling includes this in the further process. At the moment, during ranking
#' the data are split into the according degrees of the covariate and the ranking is done
#' separately. This may lead to small sample sizes. Please take care to have enough cases in each combination. Additionally,
#' covariates can lead to unstable modeling solutions. The question, if it is really
#' reasonable to include covariates when norming a test is a decision beyond the pure data
#' modeling. Please use with care or alternatively split the dataset into the two groups
#' beforehand and model them separately.
#'
#' The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
#' \code{computePowers} and \code{prepareData} are usually not called directly, but accessed
#' through other functions like \code{cnorm}.
#'
#' @param data data.frame with norm sample data
#' @param age the continuous age variable. Setting 'age' to FALSE inhibits computation of
#' powers of age and the interactions
#' @param raw name of the raw value variable (default 'raw')
#' @param width the width of the sliding window
#' @param weights Vector or variable name in the dataset with weights for each individual case. It can be used
#' to compensate for moderate imbalances due to insufficient norm data stratification. Weights should be numerical
#' and positive. It can be resource intense when applied to the sliding window. Please use the 'computeWeights' function for this purpose.
#' @param method Ranking method in case of bindings, please provide an index,
#' choosing from the following methods: 1 = Blom (1958), 2 = Tukey (1949),
#' 3 = Van der Warden (1952), 4 = Rankit (default), 5 = Levenbach (1953),
#' 6 = Filliben (1975), 7 = Yu & Huang (2001)
#' @param scale type of norm scale, either T (default), IQ, z or percentile (= no
#' transformation); a double vector with the mean and standard deviation can as well,
#' be provided f. e. c(10, 3) for Wechsler scale index points
#' @param descend ranking order (default descent = FALSE): inverses the
#' ranking order with higher raw scores getting lower norm scores; relevant
#' for example when norming error scores, where lower scores mean higher
#' performance
#' @param descriptives If set to TRUE (default), information in n, mean, median and
#' standard deviation per group is added to each observation
#' @param nGroup If set to a positive value, a grouping variable is created with the desired number of
#' equi distant groups, named by the group mean age of each group. It creates the
#' column 'group' in the data.frame and in case, there is already one with that name,
#' overwrites it.
#' @param group Optional parameter for providing the name of the grouping variable (if present; overwritten
#' if ngroups is used)
#' @param na.rm remove values, where the percentiles could not be estimated,
#' most likely happens in the context of weighting
#' @param silent set to TRUE to suppress messages
#' @return the dataset with the individual percentiles and norm scores
#'
#' @examples
#' \dontrun{
#' # Transformation using a sliding window
#' data.elfe2 <- rankBySlidingWindow(elfe, raw = "raw", age = "group", width = 0.5)
#'
#' # Comparing this to the traditional approach should give us exactly the same
#' # values, since the sample dataset only has a grouping variable for age
#' data.elfe <- rankByGroup(elfe, group = "group")
#' mean(data.elfe$normValue - data.elfe2$normValue)
#' }
#' @seealso rankByGroup, computePowers, computeWeights, weighted.rank, weighted.quantile
#' @export
#' @family prepare
rankBySlidingWindow <- function(data = NULL,
                                age = "age",
                                raw = "raw",
                                weights = NULL,
                                width,
                                method = 4,
                                scale = "T",
                                descend = FALSE,
                                descriptives = TRUE,
                                nGroup = 0,
                                group = NA,
                                na.rm = TRUE,
                                silent = FALSE) {

  if (missing(width) || !is.numeric(width) || length(width) != 1 || is.na(width) || width <= 0)
    stop("Please provide a single positive numeric value for 'width'.")

  # ---- assemble working data.frame ------------------------------------------
  if (is.null(data)) {
    d <- data.frame(raw = raw, age = age)
    raw <- "raw"; age <- "age"
  } else {
    d <- as.data.frame(data)
    if (is.numeric(raw) && length(raw) == nrow(d)) { d$raw <- raw; raw <- "raw" }
    if (is.numeric(age) && length(age) == nrow(d)) { d$age <- age; age <- "age" }
  }

  if (!(age %in% colnames(d))) stop("Age variable '", age, "' does not exist in data object.")
  if (!(raw %in% colnames(d))) stop("Raw value variable '", raw, "' does not exist in data object.")

  # ---- weights --------------------------------------------------------------
  weighting <- NULL
  has_weights <- FALSE
  if (!is.null(weights)) {
    if (is.character(weights)) {
      if (!(weights %in% colnames(d))) {
        if (!silent) warning("Weighting variable '", weights, "' does not exist. Proceeding without weighting.")
        weights <- NULL
      } else {
        weighting <- as.numeric(d[[weights]]); has_weights <- TRUE
      }
    } else {
      if (length(weights) != nrow(d)) {
        if (!silent) warning("Length of weights vector does not match the number of cases. Proceeding without weighting.")
        weights <- NULL
      } else {
        d$weights <- as.numeric(weights)
        weighting <- as.numeric(weights)
        weights <- "weights"; has_weights <- TRUE
      }
    }
  }

  # ---- drop missing values --------------------------------------------------
  weight_has_na <- has_weights && anyNA(weighting)
  if (anyNA(d[[raw]]) || anyNA(d[[age]]) || weight_has_na) {
    if (!silent) message("Missing values found in raw score, age or weight variable. Excluding affected cases.")
    valid_idx <- !is.na(d[[raw]]) & !is.na(d[[age]])
    if (has_weights) valid_idx <- valid_idx & !is.na(weighting)
    d <- d[valid_idx, , drop = FALSE]
    if (has_weights) weighting <- weighting[valid_idx]
  }

  if (has_weights && any(weighting <= 0)) {
    if (!silent) warning("Weights must be positive, non-zero values. Proceeding without weighting.")
    has_weights <- FALSE; weighting <- NULL; weights <- NULL
  }

  # ---- plotting-position constants ------------------------------------------
  numerator   <- c(-3/8, -1/3, 0, -1/2, -0.3, -0.3175, -0.326)
  denominator <- c( 1/4,  1/3, 1,    0,  0.4,  0.365,   0.348)

  if (method < 1 || method > length(numerator)) {
    if (!silent) message("Method parameter out of range, setting to RankIt (4).")
    method <- 4
  }
  num <- numerator[method]
  den <- denominator[method]

  n <- nrow(d)
  if (n == 0) stop("No valid cases left after removing missing values.")

  # ---- sort by age, keep original order -------------------------------------
  d$orig_id <- seq_len(n)
  ord <- order(d[[age]])
  d <- d[ord, , drop = FALSE]

  a_vec <- d[[age]]
  r_vec <- d[[raw]]
  w_vec <- if (has_weights) d[["weights"]] else NULL

  # descend handled once, outside the loop
  r_vec_eval <- if (descend) -r_vec else r_vec

  MIN.AGE <- a_vec[1L]
  MAX.AGE <- a_vec[n]

  minAge <- a_vec - (width / 2)
  maxAge <- a_vec + (width / 2)

  idx_low  <- minAge < MIN.AGE
  idx_high <- (maxAge > MAX.AGE) & !idx_low

  if (any(idx_low)) {
    minAge[idx_low] <- MIN.AGE
    maxAge[idx_low] <- MIN.AGE + width
  }
  if (any(idx_high)) {
    minAge[idx_high] <- MAX.AGE - width
    maxAge[idx_high] <- MAX.AGE
  }

  left_idx  <- findInterval(minAge - 1e-12, a_vec) + 1L
  right_idx <- findInterval(maxAge + 1e-12, a_vec)

  # ---- group identical windows into contiguous runs -------------------------
  # left_idx/right_idx are monotone non-decreasing in i, so identical windows
  # are contiguous: a run-length id groups them without any overflow-prone key.
  if (n == 1L) {
    run_id <- 1L
  } else {
    changed <- (diff(left_idx) != 0L) | (diff(right_idx) != 0L)
    run_id  <- cumsum(c(TRUE, changed))
  }
  win_groups <- split(seq_len(n), run_id)

  percentile_vec <- numeric(n)
  if (descriptives) {
    n_vec  <- integer(n)
    m_vec  <- numeric(n)
    md_vec <- numeric(n)
    sd_vec <- numeric(n)
  }

  for (members in win_groups) {
    l <- left_idx[members[1L]]
    r <- right_idx[members[1L]]
    nObs <- r - l + 1L
    vals <- r_vec_eval[l:r]

    if (length(members) == 1L) {
      # single case -> cheap mid-rank, no sorting
      t_raw <- r_vec_eval[members]
      if (has_weights) {
        ww <- w_vec[l:r]
        W_less  <- sum(ww[vals <  t_raw])
        W_equal <- sum(ww[vals == t_raw])
        rank_val <- ((W_less + W_equal / 2) / sum(ww)) * nObs + 0.5
      } else {
        n_less  <- sum(vals <  t_raw)
        n_equal <- sum(vals == t_raw)
        rank_val <- n_less + (n_equal + 1) / 2
      }
      percentile_vec[members] <- (rank_val + num) / (nObs + den)
    } else {
      # several cases share this window -> rank the window once, then index
      if (has_weights) {
        ranks_window <- weighted.rank(vals, w_vec[l:r])
      } else {
        ranks_window <- rank(vals)                 # ties.method = "average"
      }
      pos <- members - l + 1L
      percentile_vec[members] <- (ranks_window[pos] + num) / (nObs + den)
    }

    if (descriptives) {
      orig_raw <- r_vec[l:r]                        # original (non-descended)
      n_vec[members]  <- nObs
      m_vec[members]  <- mean(orig_raw)
      md_vec[members] <- median(orig_raw)
      sd_vec[members] <- if (nObs > 1L) sd(orig_raw) else NA_real_
    }
  }

  d$percentile <- clipPercentile(percentile_vec)
  if (descriptives) {
    d$n  <- n_vec
    d$m  <- m_vec
    d$md <- md_vec
    d$sd <- sd_vec
  }

  # ---- restore original order -----------------------------------------------
  d <- d[order(d$orig_id), , drop = FALSE]
  d$orig_id <- NULL

  # ---- norm scale -----------------------------------------------------------
  sc <- applyNormScale(d$percentile, scale)
  d$normValue <- sc$normValue

  # ---- grouping variable for display ----------------------------------------
  group_name <- "group"
  if (nGroup > 0) {
    group_cut <- as.factor(as.numeric(cut(d[[age]], nGroup)))
    d$group   <- ave(d[[age]], group_cut, FUN = mean)
  } else if (is.character(group) && length(group) == 1 && group %in% colnames(d)) {
    group_name <- group
  }

  attr(d, "age")       <- age
  attr(d, "raw")       <- raw
  attr(d, "scaleMean") <- sc$M
  attr(d, "scaleSD")   <- sc$SD
  attr(d, "descend")   <- descend
  attr(d, "width")     <- width
  attr(d, "normValue") <- "normValue"
  attr(d, "group")     <- group_name
  attr(d, "weights")   <- weights

  if (na.rm) {
    naPerc <- sum(is.na(d$percentile))
    if (naPerc > 0) {
      if (!silent) message("Could not determine manifest percentile for ", naPerc, " cases. These will be dropped.")
      d <- d[!is.na(d$percentile), , drop = FALSE]
    }
  }

  if (descriptives && nrow(d) > 0 && min(d$n) < 30 && !silent) {
    warning("The dataset includes cases whose percentile depends on fewer than 30 cases. Increasing the width parameter might help.")
  }

  return(d)
}

#' Clip percentiles into the open interval (0, 1), NA-safe
#' @keywords internal
#' @noRd
clipPercentile <- function(p) {
  lo <- !is.na(p) & p <= 0
  hi <- !is.na(p) & p >= 1
  p[lo] <- 1e-12
  p[hi] <- 1 - 1e-12
  p
}

#' Map percentiles onto the requested norm scale
#' @return list(normValue, M, SD)
#' @keywords internal
#' @noRd
applyNormScale <- function(percentile, scale) {
  if (is.numeric(scale) && length(scale) == 2) {
    return(list(normValue = qnorm(percentile, scale[1], scale[2]),
                M = scale[1], SD = scale[2]))
  }
  if (is.character(scale) && length(scale) == 1) {
    if (scale == "T")          return(list(normValue = qnorm(percentile, 50, 10),  M = 50,  SD = 10))
    if (scale == "IQ")         return(list(normValue = qnorm(percentile, 100, 15), M = 100, SD = 15))
    if (scale == "z")          return(list(normValue = qnorm(percentile, 0, 1),    M = 0,   SD = 1))
    if (scale == "percentile") return(list(normValue = percentile,                 M = NA,  SD = NA))
  }
  warning("Unknown 'scale' specification; defaulting to T scores (M = 50, SD = 10).")
  list(normValue = qnorm(percentile, 50, 10), M = 50, SD = 10)
}


#' Compute powers of the explanatory variable a as well as of the person
#' location l (data preparation)
#'
#' The function computes powers of the norm variable e. g. T scores (location, L),
#' an explanatory variable, e. g. age or grade of a data frame (age, A) and the
#' interactions of both (L X A). The k variable indicates the degree up to which
#' powers and interactions are build. These predictors can be used later on in the
#' \code{\link{bestModel}} function to model the norm sample. Higher values of k
#' allow for modeling the norm sample closer, but might lead to over-fit. In general
#' k = 3 or k = 4 (default) is sufficient to model human performance data. For example,
#' k = 2 results in the variables L1, L2, A1, A2, and their interactions L1A1, L2A1, L1A2
#' and L2A2 (but k = 2 is usually not sufficient for the modeling). Please note, that
#' you do not need to use a normal rank transformed scale like T or IQ; you can
#' use the percentiles for the 'normValue' as well.
#'
#' The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
#' \code{computePowers} and \code{prepareData} are usually not called directly, but accessed
#' through other functions like \code{cnorm}.
#'
#' @param data data.frame with the norm data
#' @param k degree of the location polynomial (1..6)
#' @param norm name of the norm variable in the data.frame (T scores, IQ, percentiles, ...).
#'   If `NULL`, the `"normValue"` attribute of `data` is used.
#' @param age explanatory variable (e.g. age or grade). May be a column name, a numeric
#'   vector of `nrow(data)`, `FALSE` to disable age handling, or `NULL` to fall back
#'   to the `"age"` attribute of `data`.
#' @param t age power parameter (1..6). If `NULL`, falls back to `k`.
#' @param silent set to TRUE to suppress messages
#' @return data.frame with the powers and interactions of location and explanatory
#'   variable / age
#' @seealso bestModel
#' @examples
#' \dontrun{
#' # Dataset with grade levels as grouping
#' data.elfe <- rankByGroup(elfe)
#' data.elfe <- computePowers(data.elfe)
#'
#' # Dataset with continuous age variable and k = 5
#' data.ppvt <- rankByGroup(ppvt)
#' data.ppvt <- computePowers(data.ppvt, age = "age", k = 5)
#' }
#' @export
#' @family prepare
computePowers <- function(data,
                          k = 5,
                          norm = NULL,
                          age = NULL,
                          t = 3,
                          silent = FALSE) {
  d <- as.data.frame(data)

  if (is.null(norm)) {
    norm <- attr(d, "normValue")
  }
  if (is.null(norm) || !(norm %in% colnames(d))) {
    stop("ERROR: Norm variable '",
         norm,
         "' does not exist in data object.")
  }
  if (!is.numeric(d[[norm]])) {
    warning("Norm score variable '", norm, "' has to be numeric.")
  }

  if (is.null(age)) {
    age <- attr(d, "age")
  }

  if (is.numeric(age) && length(age) == nrow(d)) {
    d$age <- age
    age   <- "age"
  }

  useAge <- TRUE
  if (is.null(age) ||
      (is.logical(age) && !isTRUE(age))) {
    useAge <- FALSE
    age    <- NULL
  }

  if (useAge) {
    if (!(age %in% colnames(d))) {
      stop("ERROR: Explanatory variable '",
           age,
           "' does not exist in data object.")
    }
    if (!is.numeric(d[[age]])) {
      warning("Age variable '", age, "' has to be numeric.")
    }
  }

  if (k < 1 || k > 6) {
    if (!silent)
      message("Parameter k out of range, setting to 5.")
    k <- 5
  }

  if (is.null(t)) {
    t <- k
  }
  if (t < 1 || t > 6) {
    if (!silent)
      message("Parameter t out of range, setting to 3.")
    t <- 3
  }

  L1 <- as.numeric(d[[norm]])

  # Powers of L
  for (i in seq_len(k)) {
    d[[paste0("L", i)]] <- L1^i
  }

  if (useAge) {
    A1 <- as.numeric(d[[age]])

    # Powers of A
    for (j in seq_len(t)) {
      d[[paste0("A", j)]] <- A1^j
    }

    # Interactions L_i * A_j
    for (i in seq_len(k)) {
      for (j in seq_len(t)) {
        d[[paste0("L", i, "A", j)]] <- L1^i * A1^j
      }
    }
  }

  attr(d, "age")        <- age
  attr(d, "normValue")  <- norm
  attr(d, "k")          <- k
  attr(d, "t")          <- t
  attr(d, "useAge")     <- useAge

  if (useAge && !silent) {
    cat("Powers of location: k = ", k, "\n", sep = "")
    cat("Powers of age:      t = ", t, "\n", sep = "")

    raw_name <- attr(d, "raw")
    if (!is.null(raw_name) && raw_name %in% colnames(d)) {
      A1 <- as.numeric(d[[age]])
      r2 <- summary.lm(lm(as.numeric(d[[raw_name]]) ~ poly(A1, t, raw = TRUE)))$r.squared

      if (r2 < 0.05 && t > 2) {
        warning(
          "Multiple R^2 between the explanatory variable and the raw score is low ",
          "(R^2 = ",
          round(r2, 4),
          "). The continuous norming model may be unstable. ",
          "Consider reducing t (e.g. to 1 for a linear age effect) and/or the number ",
          "of age groups."
        )
      } else {
        cat(
          "Multiple R^2 between raw score and explanatory variable: R^2 = ",
          round(r2, 4),
          "\n\n",
          sep = ""
        )
      }
    }
  }

  return(d)
}

# Required packages
library(shiny)
library(shinycssloaders)
library(cNORM)
library(DT)
library(markdown)
library(foreign)
library(readxl)

data(elfe, package = "cNORM")
data(ppvt, package = "cNORM")
data(CDC,  package = "cNORM")

# Helper: split user-entered numbers separated by space/comma
parseNums <- function(s) {
  if (is.null(s) || !nzchar(trimws(s))) return(NULL)
  as.numeric(unlist(strsplit(s, "[,\\s]+", perl = TRUE)))
}

shinyServer(function(input, output, session) {

  session$onSessionEnded(function() stopApp())

  # ---- Data loading ----------------------------------------------------------
  currentFile <- reactive({
    if (is.null(input$file)) {
      ex <- input$Example
      if (is.null(ex) || ex == "") return(NULL)
      return(switch(ex,
                    "elfe" = as.data.frame(elfe),
                    "ppvt" = as.data.frame(ppvt),
                    "CDC"  = as.data.frame(CDC),
                    NULL))
    }
    path <- input$file$datapath
    ext  <- tolower(tools::file_ext(path))
    switch(ext,
           "csv"  = as.data.frame(read.csv2(path)),
           "sav"  = as.data.frame(foreign::read.spss(path, to.data.frame = TRUE)),
           "xlsx" = as.data.frame(readxl::read_xlsx(path)),
           "xls"  = as.data.frame(readxl::read_xls(path)),
           "rda"  = { e <- new.env(); load(path, e); get(ls(e)[1], envir = e) },
           NULL)
  })

  variableNames <- reactive({
    req(currentFile())
    names(currentFile())
  })

  # ---- Variable selectors ----------------------------------------------------
  output$GroupingVariable <- renderUI({
    nm  <- variableNames()
    sel <- if ("group" %in% nm) "group" else if ("Group" %in% nm) "Group" else NULL
    selectInput("InputGroupingVariable", "Grouping variable", choices = nm, selected = sel)
  })

  output$RawValues <- renderUI({
    nm  <- variableNames()
    sel <- if ("raw" %in% nm) "raw" else if ("Raw" %in% nm) "Raw" else NULL
    selectInput("InputRawValues", "Raw-score variable", choices = nm, selected = sel)
  })

  # ---- Intro / data preview --------------------------------------------------
  output$introduction <- renderUI({
    if (is.null(currentFile())) includeHTML("www/introduction.html") else NULL
  })

  output$table <- DT::renderDT({
    req(currentFile())
    currentFile()
  }, options = list(pageLength = 10, scrollX = TRUE))

  # ---- Combined preparation + modeling --------------------------------------
  # Holds the current model + prepared data; NULL means "no model yet".
  cnormResult <- reactiveVal(NULL)

  # Whenever a new dataset is loaded (file or example), drop the old model.
  observeEvent(currentFile(), {
    cnormResult(NULL)
  }, ignoreNULL = FALSE)

  # Also clear if the user changes a key model setting after running once,
  # so stale plots can't be shown against new parameters.
  observeEvent(list(input$InputGroupingVariable,
                    input$InputRawValues,
                    input$Scale,
                    input$NumberOfPowers,
                    input$NumberOfPowersAge,
                    input$InputNumberOfTerms), {
                      cnormResult(NULL)
                    }, ignoreInit = TRUE)

  # Build the model when the user clicks "Run model".
  observeEvent(input$RunModel, {
    req(currentFile(), input$InputGroupingVariable, input$InputRawValues)

    df    <- currentFile()
    grp   <- input$InputGroupingVariable
    raw   <- input$InputRawValues
    scale <- input$Scale
    k     <- as.numeric(input$NumberOfPowers)
    t     <- as.numeric(input$NumberOfPowersAge)
    nT    <- input$InputNumberOfTerms

    ranked   <- cNORM::rankByGroup(df, group = grp, raw = raw, scale = scale)
    prepared <- cNORM::computePowers(ranked, k = k, t = t, norm = "normValue")

    model <- if (is.null(nT) || is.na(nT) || nT < 1) {
      cNORM::bestModel(prepared, raw = raw, k = k, t = t,
                       terms = 0, plot = FALSE)
    } else {
      cNORM::bestModel(prepared, raw = raw, k = k, t = t,
                       terms = as.integer(nT), plot = FALSE)
    }

    cnormResult(list(data = prepared, model = model,
                     group = grp, raw = raw, scale = scale))
  })

  preparedData   <- reactive({ req(cnormResult()); cnormResult()$data })
  bestModel      <- reactive({ req(cnormResult()); cnormResult()$model })
  chosenGrouping <- reactive({ req(cnormResult()); cnormResult()$group })
  chosenRaw      <- reactive({ req(cnormResult()); cnormResult()$raw })
  chosenScale    <- reactive({ req(cnormResult()); cnormResult()$scale })
  cnormObj       <- reactive({
    req(cnormResult())
    cNORM::buildCnormObject(preparedData(), bestModel())
  })

  ageRange <- reactive({
    req(currentFile(), chosenGrouping())
    range(currentFile()[[chosenGrouping()]], na.rm = TRUE)
  })

  # ---- Model report ----------------------------------------------------------
  output$ModelReport <- renderPrint({
    req(bestModel())
    cat(bestModel()$report, sep = "\n")
  })

  output$ModelReportUI <- renderUI({
    if (is.null(cnormResult())) {
      tags$div(class = "alert alert-info",
               "No model yet. Choose your variables and click ",
               tags$b("Run model"), ".")
    } else {
      verbatimTextOutput("ModelReport")
    }
  })

  output$ConsistencyCheck <- renderText({
    req(bestModel())
    if (cNORM::checkConsistency(bestModel()))
      "⚠ Model may be inconsistent. Check the percentile plot for intersecting curves and try a different number of terms."
    else
      "✓ No consistency violations within the data range."
  })

  output$RangeCheck <- renderText({
    req(bestModel())
    cNORM::rangeCheck(bestModel())
  })

  # ---- Modeling diagnostics --------------------------------------------------
  output$modelPlot <- renderPlot({
    rng <- ageRange()
    cNORM::plotPercentiles(cnormObj(), raw = chosenRaw(), group = chosenGrouping(),
                           minAge = rng[1], maxAge = rng[2])
  })

  chosenTypeOfPlotSubset <- reactive({
    m <- c("R2 by Number of Predictors",
           "Mallow's (log) Cp by R2",
           "Bayesian Information Criterion (BIC) by R2",
           "RMSE by Number of Predictors")
    match(input$chosenTypePlotSubset, m) - 1
  })

  output$PlotWL <- renderPlot({
    cNORM::plotSubset(bestModel(), type = chosenTypeOfPlotSubset())
  })

  output$PrintSubset <- DT::renderDT({
    cNORM::printSubset(bestModel())
  }, options = list(pageLength = 10, scrollX = TRUE))

  # ---- Visualization ---------------------------------------------------------
  output$PlotPercentiles <- renderPlot({
    rng <- ageRange()
    pct <- parseNums(input$PercentilesForPercentiles) / 100
    if (is.null(pct))
      cNORM::plotPercentiles(cnormObj(), raw = chosenRaw(), group = chosenGrouping(),
                             minAge = rng[1], maxAge = rng[2])
    else
      cNORM::plotPercentiles(cnormObj(), raw = chosenRaw(), group = chosenGrouping(),
                             minAge = rng[1], maxAge = rng[2], percentiles = pct)
  })

  output$Series <- renderPlot({
    cNORM::plotPercentileSeries(cnormObj(), start = input$terms, end = input$terms)
  })

  output$NormCurves <- renderPlot({
    rng <- ageRange()
    pct <- parseNums(input$PercentilesForNormCurves) / 100
    if (is.null(pct)) pct <- c(.02276, .1587, .5, .8413, .97724)
    norms <- switch(chosenScale(),
                    "T"  = qnorm(pct, 50, 10),
                    "z"  = qnorm(pct),
                    "IQ" = qnorm(pct, 100, 15))
    cNORM::plotNormCurves(bestModel(), minAge = rng[1], maxAge = rng[2],
                          normList = round(norms, 2))
  })

  output$PlotDensity <- renderPlot({
    grps <- parseNums(input$densities)
    if (is.null(grps)) cNORM::plotDensity(bestModel())
    else cNORM::plotDensity(bestModel(), group = grps)
  })

  output$PlotDerivatives <- renderPlot({
    rng <- ageRange()
    cNORM::plotDerivative(bestModel(), minAge = rng[1], maxAge = rng[2])
  })

  output$PlotNormScores <- renderPlot({
    input$normScoreButton
    type <- if (isTRUE(input$differences)) 1 else 0
    cNORM::plotNorm(cnormObj(), group = isTRUE(input$grouping), type = type)
  }) |> bindEvent(input$normScoreButton)

  output$PlotRawScores <- renderPlot({
    input$rawScoreButton
    type <- if (isTRUE(input$differences1)) 1 else 0
    if (isTRUE(input$grouping1))
      cNORM::plotRaw(cnormObj(), group = TRUE, type = type)
    else
      cNORM::plotRaw(cnormObj(), type = type)
  }) |> bindEvent(input$rawScoreButton)

  # ---- Cross validation ------------------------------------------------------
  cvResult <- eventReactive(input$CrossValidation, {
    cNORM::cnorm.cv(preparedData(),
                    repetitions = input$RepetitionsCV,
                    norms = input$NormsCV,
                    min = 1, max = input$MaxTermsCV,
                    group = chosenGrouping(),
                    raw   = chosenRaw(),
                    age   = chosenGrouping())
  })
  output$PlotCV  <- renderPlot({ cvResult() })
  output$TableCV <- DT::renderDT({ cvResult() })

  # ---- Predictions -----------------------------------------------------------
  output$InputNormValue <- renderUI({
    req(bestModel())
    tagList(
      numericInput("NormValueInputAge", "Age",  value = NULL,
                   min = bestModel()$minA1,  max = bestModel()$maxA1),
      numericInput("NormValueInputRaw", "Raw score", value = NULL,
                   min = bestModel()$minRaw, max = bestModel()$maxRaw),
      actionButton("CalcNormValue", "Compute norm score")
    )
  })

  output$NormValue <- renderText({
    input$CalcNormValue
    isolate({
      req(input$NormValueInputAge, input$NormValueInputRaw)
      m <- bestModel()
      v <- cNORM::predictNorm(input$NormValueInputRaw, input$NormValueInputAge,
                              model = m, minNorm = m$minL1, maxNorm = m$maxL1)
      paste("Predicted norm score:", round(v, 2))
    })
  }) |> bindEvent(input$CalcNormValue)

  output$InputRawValue <- renderUI({
    req(bestModel())
    tagList(
      numericInput("RawValueInputAge", "Age", value = NULL,
                   min = bestModel()$minA1, max = bestModel()$maxA1),
      numericInput("RawValueInputNorm", "Norm score", value = NULL,
                   min = bestModel()$minL1, max = bestModel()$maxL1),
      actionButton("CalcRawValue", "Compute raw score")
    )
  })

  output$RawValue <- renderText({
    input$CalcRawValue
    isolate({
      req(input$RawValueInputAge, input$RawValueInputNorm)
      m <- bestModel()
      v <- cNORM::predictRaw(input$RawValueInputNorm, input$RawValueInputAge,
                             coefficients = m$coefficients,
                             minRaw = m$minRaw, maxRaw = m$maxRaw)
      paste("Predicted raw score:", round(v, 2))
    })
  }) |> bindEvent(input$CalcRawValue)

  output$InputNormTable <- renderUI({
    req(bestModel())
    m <- bestModel()
    tagList(
      numericInput("NormTableInput",         "Age",            value = NULL,
                   min = 0, max = m$maxA1),
      numericInput("NormTableInputStart",    "Norm start",     value = m$scaleM - 2.5 * m$scaleSD),
      numericInput("NormTableInputEnd",      "Norm end",       value = m$scaleM + 2.5 * m$scaleSD),
      numericInput("NormTableInputStepping", "Step width",     value = NULL),
      numericInput("NormTableCI",            "Confidence",     value = 0.9),
      numericInput("NormTableRel",           "Reliability",    value = NULL),
      actionButton("CalcNormTables", "Generate norm table"),
      downloadButton("DownloadNormTable", "Download")
    )
  })

  normTable <- eventReactive(input$CalcNormTables, {
    req(input$NormTableInput, input$NormTableInputStart,
        input$NormTableInputEnd, input$NormTableInputStepping)
    m <- bestModel()
    tab <- cNORM::normTable(input$NormTableInput, m,
                            minNorm = input$NormTableInputStart,
                            maxNorm = input$NormTableInputEnd,
                            minRaw  = m$minRaw, maxRaw = m$maxRaw,
                            step    = input$NormTableInputStepping,
                            CI          = if (is.na(input$NormTableCI))  NULL else input$NormTableCI,
                            reliability = if (is.na(input$NormTableRel)) NULL else input$NormTableRel,
                            pretty = TRUE)
    tab$raw        <- round(tab$raw, 2)
    tab$percentile <- round(tab$percentile, 2)
    tab
  })

  output$NormTable         <- DT::renderDT({ normTable() })
  output$DownloadNormTable <- downloadHandler(
    filename = function() "NormTable.csv",
    content  = function(file) write.csv(normTable(), file, row.names = FALSE)
  )

  output$InputRawTable <- renderUI({
    req(bestModel())
    m <- bestModel()
    tagList(
      numericInput("RawTableInput",         "Age",         value = NULL,
                   min = 0, max = m$maxA1),
      numericInput("RawTableInputStart",    "Raw start",   value = m$minRaw),
      numericInput("RawTableInputEnd",      "Raw end",     value = m$maxRaw),
      numericInput("RawTableInputStepping", "Step width",  value = NULL),
      numericInput("RawTableCI",            "Confidence",  value = 0.9),
      numericInput("RawTableRel",           "Reliability", value = NULL),
      actionButton("CalcRawTables", "Generate raw table"),
      downloadButton("DownloadRawTable", "Download")
    )
  })

  rawTable <- eventReactive(input$CalcRawTables, {
    req(input$RawTableInput, input$RawTableInputStart,
        input$RawTableInputEnd, input$RawTableInputStepping)
    m <- bestModel()
    tab <- cNORM::rawTable(input$RawTableInput, m,
                           minNorm = m$minL1, maxNorm = m$maxL1,
                           minRaw  = input$RawTableInputStart,
                           maxRaw  = input$RawTableInputEnd,
                           step    = input$RawTableInputStepping,
                           CI          = if (is.na(input$RawTableCI))  NULL else input$RawTableCI,
                           reliability = if (is.na(input$RawTableRel)) NULL else input$RawTableRel,
                           pretty = TRUE)
    tab$norm       <- round(tab$norm, 3)
    tab$percentile <- round(tab$percentile, 2)
    tab
  })

  output$RawTable         <- DT::renderDT({ rawTable() })
  output$DownloadRawTable <- downloadHandler(
    filename = function() "RawTable.csv",
    content  = function(file) write.csv(rawTable(), file, row.names = FALSE)
  )

  # ---- Model export ----------------------------------------------------------
  output$downloadModel <- downloadHandler(
    filename = function() "cnorm_model.RData",
    content  = function(file) {
      model.cnorm <- bestModel()
      data.cnorm  <- preparedData()
      save(model.cnorm, data.cnorm, file = file)
    }
  )
})

#' Sentence completion test from ELFE 1-6
#'
#' A dataset containing the raw data of 1400 students from grade 2 to 5 in the sentence
#' comprehension test from ELFE 1-6 (Lenhard & Schneider, 2006). In this test, students
#' are presented lists of sentences with one gap. The student has to fill in the correct
#' solution by selecting from a list of 5 alternatives per sentence. The alternatives
#' include verbs, adjectives, nouns, pronouns and conjunctives. Each item stems from
#' the same word type. The text is speeded, with a time cutoff of 180 seconds. The
#' variables are as follows:
#'
#' @format A data frame with 1400 rows and 3 variables:
#' \describe{
#'   \item{personID}{ID of the student}
#'   \item{group}{grade level, with x.5 indicating the end of the school year and x.0 indicating the middle of the school year}
#'   \item{raw}{the raw score of the student, spanning values from 0 to 28}
#' }
#' @source \url{https://www.psychometrica.de/elfe2.html}
#' @references Lenhard, W. & Schneider, W.(2006). Ein Leseverstaendnistest fuer Erst- bis Sechstklaesser. Goettingen/Germany: Hogrefe.
#' @docType data
#' @keywords datasets
#' @concept reading comprehension
#' @name elfe
#' @examples
#' \dontrun{
#'   # prepare data, retrieve model and plot percentiles
#'   model <- cnorm(group = elfe$group, raw = elfe$raw)
#' }
#' @format A data frame with 1400 rows and 3 columns
"elfe"

#' Vocabulary development from 2.5 to 17
#'
#' A dataset based on an unstratified sample of PPVT4 data (German adaption). The PPVT4 consists of blocks of items with
#' 12 items each. Each item consists of 4 pictures. The test taker is given a word orally and he or she has to point out
#' the picture matching the oral word. Bottom and ceiling blocks of items are determined according to age and performance. For
#' instance, when a student knows less than 4 word from a block of 12 items, the testing stops. The sample is not identical
#' with the norm sample and includes doublets of cases in order to align the sample size per age group. It is
#' primarily intended for running the cNORM analyses with regard to modeling and stratification.
#'
#' @format A data frame with 4542 rows and 6 variables:
#' \describe{
#'   \item{age}{the chronological age of the child}
#'   \item{sex}{the sex of the test taker, 1=male, 2=female}
#'   \item{migration}{migration status of the family, 0=no, 1=yes}
#'   \item{region}{factor specifying the region, the data were collected; grouped into south, north, east and west}
#'   \item{raw}{the raw score of the student, spanning values from 0 to 228}
#'   \item{group}{age group of the child, determined by the getGroups()-function with 12 equidistant age groups}
#' }
#' @source \url{https://www.psychometrica.de/ppvt4.html}
#' @references Lenhard, A., Lenhard, W., Segerer, R. & Suggate, S. (2015). Peabody Picture Vocabulary Test - Revision IV (Deutsche Adaption). Frankfurt a. M./Germany: Pearson Assessment.
#' @docType data
#' @keywords datasets
#' @concept vocabulary acquisition development receptive
#' @name ppvt
#' @examples
#' \dontrun{
#' # Example with continuous age variable, ranked with sliding window
#' model.ppvt.sliding <- cnorm(age=ppvt$age, raw=ppvt$raw, width=1)
#'
#' # Example with age groups; you might first want to experiment with
#' # the granularity of the groups via the 'getGroups()' function
#' model.ppvt.group <- cnorm(group=ppvt$group, raw=ppvt$raw) # with predefined groups
#' model.ppvt.group <- cnorm(group=getGroups(ppvt$age, n=15, equidistant = T),
#'                           raw=ppvt$raw) # groups built 'on the fly'
#'
#'
#' # plot information function
#' plot(model.ppvt.group, "subset")
#'
#' # check model consistency
#' checkConsistency(model.ppvt.group)
#'
#' # plot percentiles
#' plot(model.ppvt.group, "percentiles")
#' }
#' @format A data frame with 5600 rows and 9 columns
"ppvt"

#' BMI growth curves from age 2 to 25
#'
#' By the courtesy of the Center of Disease Control (CDC), cNORM includes human growth data for children and adolescents
#' age 2 to 25 that can be used to model trajectories of the body mass index and to estimate percentiles for clinical
#' definitions of under- and overweight. The data stems from the NHANES surveys in the US and was published in 2012
#' as public domain. The data was cleaned by removing missing values and it includes the following variables from or
#' based on the original dataset.
#'
#' @format A data frame with 45053 rows and 7 variables:
#' \describe{
#'   \item{age}{continuous age in years, based on the month variable}
#'   \item{group}{age group; chronological age in years at the time of examination}
#'   \item{month}{chronological age in month at the time of examination}
#'   \item{sex}{sex of the participant, 1 = male, 2 = female}
#'   \item{height}{height of the participants in cm}
#'   \item{weight}{weight of the participants in kg}
#'   \item{bmi}{the body mass index, computed by (weight in kg)/(height in m)^2}
#' }
#' @docType data
#' @keywords datasets
#' @concept Body Mass Index growth curves weight height
#' @references Center for Disease Control and Prevention (2012).
#'   National Health and Nutrition Examination Survey: Questionnaires,
#'   datasets and related documentation. U.S. Department of Health and Human Services
#'   (original source not available anymore).
#' @name CDC
#' @format A data frame with 45035 rows and 7 columns
"CDC"

#' Weighting of cases through iterative proportional fitting (Raking)
#'
#' Computes and standardizes weights via raking to compensate for non-stratified
#' samples. It is based on the implementation in the survey R package. It reduces
#' data collection biases in the norm data by the means of post stratification,
#' thus reducing the effect of unbalanced data in percentile estimation and norm
#' data modeling.
#'
#' This function computes standardized raking weights to overcome biases in norm
#' samples. It generates weights by drawing on the information of population
#' shares (e. g. for sex, ethnic group, region ...) and subsequently reduces the
#' influence of over-represented groups or increases underrepresented cases. The
#' returned weights are either raw or standardized and scaled to be larger than 0.
#'
#' Raking in general has a number of advantages over post stratification and it
#' additionally allows cNORM to draw on larger datasets, since less cases have
#' to be removed during stratification. To use this function, additionally to the
#' data, a data frame with stratification variables has to be specified. The data
#' frame should include a row with (a) the variable name, (b) the level of the
#' variable and (c) the according population proportion.
#'
#' @param data data.frame with norm sample data.
#' @param population.margins A data.frame including three columns, specifying the
#' variable name in the original dataset used for data stratification, the factor
#' level of the variable and the according population share. Please ensure the
#' original data does not include factor levels not present in the
#' population.margins. Additionally, summing up the shares of the different
#' levels of a variable should result in a value near 1.0. The first column must
#' specify the name of the stratification variable, the second the level and
#' the third the proportion.
#' @param standardized If TRUE (default), the raking weights are scaled to
#' weights/min(weights)
#' @return a vector with the standardized weights
#' @examples
#' \dontrun{
#' # cNORM features a dataset on vocabulary development (ppvt)
#' # that includes variables like sex or migration. In order
#' # to weight the data, we have to specify the population shares.
#' # According to census, the population includes 52% boys
#' # (factor level 1 in the ppvt dataset) and 70% / 30% of persons
#' # without / with a history of migration (= 0 / 1 in the dataset).
#' # First we set up the population margins with all shares of the
#' # different levels:
#'
#' margins <- data.frame(variables = c("sex", "sex",
#'                                     "migration", "migration"),
#'                       levels = c(1, 2, 0, 1),
#'                       share = c(.52, .48, .7, .3))
#' head(margins)
#'
#' # Now we use the population margins to generate weights
#' # through raking
#'
#' weights <- computeWeights(ppvt, margins)
#'
#'
#' # There are as many different weights as combinations of
#' # factor levels, thus only four in this specific case
#'
#' unique(weights)
#'
#'
#' # To include the weights in the cNORM modelling, we have
#' # to pass them as weights. They are then used to set up
#' # weighted quantiles and as weights in the regression.
#'
#' model <- cnorm(raw = ppvt$raw,
#'                group = ppvt$group,
#'                weights = weights)
#' }
#' @export
computeWeights <- function(data, population.margins, standardized = TRUE) {
  data <- as.data.frame(data)
  names(population.margins) <- c("sv", "level", "proportion")
  sv <- unique(population.margins[, 1])

  # Check if all stratification variables are present in the dataset
  missing.variables <- sv[!sv %in% names(data)]
  if (length(missing.variables) > 0) {
    stop(
      paste0(
        "Weighting aborted due to missing variable(s): ",
        paste(missing.variables, collapse = ", "),
        " not present in the dataset."
      )
    )
  }

  # Joint distribution of the sample; drop empty cross-classification cells,
  # as they are inert in the IPF iteration, are never mapped onto a data row,
  # and would otherwise distort the standardization (min of weights).
  props <- as.data.frame(prop.table(xtabs(
    formula(paste0("~", paste(sv, collapse = " + "))), data = data
  )))
  props <- props[props$Freq > 0, , drop = FALSE]
  rownames(props) <- NULL
  props.adjusted <- props

  # Check that the levels of every stratification variable match between the
  # data and the population margins (in both directions).
  for (lev in seq_along(sv)) {
    data_levels   <- unique(as.character(data[[sv[lev]]]))
    margin_levels <- unique(as.character(
      population.margins$level[population.margins$sv == sv[lev]]))

    if (!setequal(data_levels, margin_levels)) {
      stop(
        "Levels of the data and the population margins in variable '",
        sv[lev],
        "' do not match.\n  In data but not in margins: ",
        paste(setdiff(data_levels, margin_levels), collapse = ", "),
        "\n  In margins but not in data: ",
        paste(setdiff(margin_levels, data_levels), collapse = ", "),
        "\n  Please make sure every level of every variable is contained in",
        " both the data set and the population margins."
      )
    }
  }

  # Check if the sum of proportions per variable lies between 0.95 and 1.05
  for (marg in seq_along(sv)) {
    marg_sum <- abs(sum(population.margins$proportion[population.margins$sv == sv[marg]]) - 1)
    if (marg_sum > 0.05) {
      warning(
        paste(
          "Sum of proportions of variable",
          sv[marg],
          "is not within [0.95;1.05].\nPlease make sure the proportions for every single stratification variable sum up to almost 1.00"
        )
      )
    }
  }

  # Iterative proportional fitting (raking) -------------------------------------
  weights <- rep(1, length.out = nrow(props))

  converged <- FALSE
  for (repetitions in 1:100) {
    weights_start <- weights   # snapshot before the full cycle over all variables

    for (i in seq_along(sv)) {
      sv.sum <- aggregate(formula(paste0("Freq ~ ", sv[i])),
                          data = props.adjusted, FUN = sum)
      tmp <- population.margins[population.margins$sv == sv[i], ]

      for (j in seq_len(nrow(sv.sum))) {
        current_level <- as.character(sv.sum[j, 1])

        # Find matching row in population margins
        match_idx <- which(as.character(tmp$level) == current_level)
        if (length(match_idx) == 0) {
          warning(
            paste(
              "No matching level found for",
              current_level,
              "in variable",
              sv[i],
              "- check your population.margins data"
            )
          )
          next
        }

        # Rows of the joint distribution belonging to this level
        indizes <- which(as.character(props[, sv[i]]) == current_level)
        if (length(indizes) == 0) {
          warning(
            paste(
              "No matching observations found for level",
              current_level,
              "in variable",
              sv[i]
            )
          )
          next
        }

        weights[indizes] <- weights[indizes] *
          tmp$proportion[match_idx] / sv.sum[j, 2]

        props.adjusted$Freq <- props$Freq * weights
      }
    }

    # Convergence is evaluated after a complete pass through ALL variables,
    # not after a single level adjustment.
    if (sum(abs(weights - weights_start)) < 1e-6) {
      converged <- TRUE
      message("Raking converged after ", repetitions, " iterations.")
      break
    }
  }

  if (!converged)
    warning("Raking did not converge within 100 iterations. ",
            "Using the raking weights is not recommended.")

  checkWeights(weights)

  if (standardized)
    weights <- standardizeRakingWeights(weights)

  # Map the cell weights back onto the individual observations -------------------
  stratification.weights <- rep(1, length.out = nrow(data))

  if (length(sv) > 1) {
    # Multiple stratification variables: match by the full level combination
    data_combined <- do.call("paste", c(lapply(sv, function(v)
      as.character(data[[v]])), sep = "|"))
    props_combined <- do.call("paste", c(lapply(sv, function(v)
      as.character(props[[v]])), sep = "|"))

    for (i in seq_along(weights)) {
      stratification.weights[data_combined == props_combined[i]] <- weights[i]
    }
  } else {
    # Single stratification variable
    sv_name <- sv[1]
    for (i in seq_len(nrow(props))) {
      level_val <- as.character(props[[sv_name]][i])
      stratification.weights[as.character(data[[sv_name]]) == level_val] <- weights[i]
    }
  }

  return(stratification.weights)
}


#' Check, if NA or values <= 0 occur and issue warning
#' @param weights Raking weights
#' @keywords internal
checkWeights <- function(weights) {
  if (anyNA(weights)) {
    warning("Undefined value occurred during raking. Using the raking weights is not recommended.")
  } else if (any(weights <= 0)) {
    warning("Negative values or zeros occurred during raking. Using the raking weights is not recommended.")
  }
}


#' Function for standardizing raking weights
#'
#' Raking weights get divided by the smallest weight. Thereby, all weights
#' become larger or equal to 1 without changing the ratio of the weights
#' to each other.
#' @param weights Raking weights computed by computeWeights()
#' @return the standardized weights
#' @keywords internal
standardizeRakingWeights <- function(weights) {
  weights_min <- min(weights)
  if (weights_min <= 0) {
    warning("Smallest raking weight is zero or below. Weights cannot be standardized.\nNon-standardized weights are returned.")
    return(weights)
  }
  weights / weights_min
}

library(shiny)
library(cNORM)
library(DT)
library(ggplot2)

# Hilfsoperator: NULL-coalescing
`%||%` <- function(a, b) if (is.null(a)) b else a

shinyServer(function(input, output, session) {

  # Reactive values to store data and model
  rv <- reactiveValues(
    data = NULL,
    model = NULL,
    normTables = NULL
  )

  # ============================================================================
  # DATA INPUT TAB
  # ============================================================================

  # Load data based on selection
  observeEvent(input$Example, {
    req(input$Example)

    if (input$Example == "elfe") {
      rv$data <- elfe
    } else if (input$Example == "ppvt") {
      rv$data <- ppvt
    } else if (input$Example == "CDC") {
      rv$data <- CDC
    }

    # Invalidate model when new data is loaded
    rv$model <- NULL
    rv$normTables <- NULL
  })

  # Load data from file
  observeEvent(input$file, {
    req(input$file)

    tryCatch({
      ext <- tools::file_ext(input$file$name)

      if (ext == "csv") {
        rv$data <- read.csv(input$file$datapath)
      } else if (ext %in% c("xlsx", "xls")) {
        rv$data <- readxl::read_excel(input$file$datapath)
      } else if (ext == "rda" || ext == "RData") {
        load(input$file$datapath)
        rv$data <- get(ls()[1])
      } else if (ext == "sav") {
        rv$data <- haven::read_sav(input$file$datapath)
      }

      # Invalidate model when new data is loaded
      rv$model <- NULL
      rv$normTables <- NULL

      showNotification("Data loaded successfully!", type = "message")
    }, error = function(e) {
      showNotification(paste("Error loading file:", e$message), type = "error")
    })
  })

  # Display data preview
  output$dataPreview <- DT::renderDT({
    req(rv$data)

    DT::datatable(
      rv$data,
      options = list(
        pageLength = 15,
        scrollX = TRUE,
        dom = 'Bfrtip'
      ),
      rownames = FALSE
    )
  })

  # ============================================================================
  # MODELING TAB - Dynamic UI
  # ============================================================================

  # Age variable selector
  output$ageVariable <- renderUI({
    req(rv$data)

    selectInput(
      "age_var",
      "Age Variable:",
      choices = names(rv$data),
      selected = names(rv$data)[1]
    )
  })

  # Score variable selector
  output$scoreVariable <- renderUI({
    req(rv$data)

    selectInput(
      "score_var",
      "Raw Score Variable:",
      choices = names(rv$data),
      selected = if (ncol(rv$data) > 1) names(rv$data)[2] else names(rv$data)[1]
    )
  })

  # Weight variable selector
  output$weightVariable <- renderUI({
    req(rv$data)

    selectInput(
      "weight_var",
      "Weight Variable (optional):",
      choices = c("None" = "", names(rv$data)),
      selected = ""
    )
  })

  # ============================================================================
  # MODELING TAB - Compute Model
  # ============================================================================

  observeEvent(input$computeModel, {
    req(rv$data, input$age_var, input$score_var)

    # Show progress
    withProgress(message = 'Fitting model...', value = 0, {

      tryCatch({
        # Extract variables
        age <- rv$data[[input$age_var]]
        score <- rv$data[[input$score_var]]

        # Handle weights
        weights <- NULL
        if (!is.null(input$weight_var) && input$weight_var != "") {
          weights <- rv$data[[input$weight_var]]
        }

        incProgress(0.2, detail = "Preparing data...")

        # Fit model based on distribution type
        if (input$distributionType == "shash") {

          # Determine delta parameter
          delta_deg <- if (input$fix_delta) NULL else input$delta_degree
          delta_val <- if (input$fix_delta) input$delta_value else 1

          incProgress(0.3, detail = "Fitting SHASH model...")

          rv$model <- cnorm.shash(
            age = age,
            score = score,
            weights = weights,
            mu_degree = input$mu_degree,
            sigma_degree = input$sigma_degree,
            epsilon_degree = input$epsilon_degree,
            delta_degree = delta_deg,
            delta = delta_val,
            scale = input$scale,
            plot = FALSE
          )

        } else if (input$distributionType == "betabinomial") {

          incProgress(0.3, detail = "Fitting Beta-Binomial model...")

          rv$model <- cnorm.betabinomial(
            age = age,
            score = score,
            weights = weights,
            alpha = input$alpha_degree,
            beta = input$beta_degree,
            scale = input$scale,
            plot = FALSE
          )
        }

        # Reset existing norm tables when a new model is fit
        rv$normTables <- NULL

        incProgress(1, detail = "Model fitted!")

        showNotification("Model computed successfully!", type = "message")

      }, error = function(e) {
        showNotification(
          paste("Error fitting model:", e$message),
          type = "error",
          duration = 10
        )
      })
    })
  })

  # ============================================================================
  # MODELING TAB - Outputs
  # ============================================================================

  # Model summary
  output$modelSummary <- renderPrint({
    req(rv$model)
    summary(rv$model,
            age = rv$data[[input$age_var]],
            score = rv$data[[input$score_var]])
  })

  # Percentile plot
  output$percentilePlot <- renderPlot({
    req(rv$model, rv$data, input$age_var, input$score_var)

    weights <- NULL
    if (!is.null(input$weight_var) && input$weight_var != "") {
      weights <- rv$data[[input$weight_var]]
    }

    plot(rv$model,
         age = rv$data[[input$age_var]],
         score = rv$data[[input$score_var]],
         weights = weights,
         points = TRUE)
  })

  # Download model
  output$downloadModel <- downloadHandler(
    filename = function() {
      paste0("cnorm_model_", input$distributionType, "_", Sys.Date(), ".rds")
    },
    content = function(file) {
      saveRDS(rv$model, file)
    }
  )

  # ============================================================================
  # NORM TABLES TAB
  # ============================================================================

  # Auto-populate score range when data/score variable changes
  observe({
    req(rv$data, input$score_var)

    score <- rv$data[[input$score_var]]

    updateNumericInput(
      session,
      "norm_start",
      value = floor(min(score, na.rm = TRUE))
    )

    updateNumericInput(
      session,
      "norm_end",
      value = ceiling(max(score, na.rm = TRUE))
    )
  })

  observeEvent(input$generateTables, {
    req(rv$model, input$norm_ages)

    withProgress(message = 'Generating norm tables...', value = 0, {

      tryCatch({
        # Parse ages
        ages_text <- gsub(",", " ", input$norm_ages)
        ages <- as.numeric(unlist(strsplit(ages_text, "\\s+")))
        ages <- ages[!is.na(ages)]

        if (length(ages) == 0) {
          showNotification("Please enter valid ages", type = "error")
          return(NULL)
        }

        incProgress(0.3, detail = "Computing percentiles...")

        # CI / Reliability
        ci_val  <- if (isTRUE(input$include_ci)) input$ci_level    else NULL
        rel_val <- if (isTRUE(input$include_ci)) input$reliability else NULL

        # Generate tables
        if (input$distributionType == "shash") {

          # Determine score range
          start_score <- input$norm_start
          end_score   <- input$norm_end

          if (is.null(start_score) || is.na(start_score)) {
            start_score <- attr(rv$model$result, "min") %||%
              floor(min(rv$data[[input$score_var]], na.rm = TRUE))
          }
          if (is.null(end_score) || is.na(end_score)) {
            end_score <- attr(rv$model$result, "max") %||%
              ceiling(max(rv$data[[input$score_var]], na.rm = TRUE))
          }

          rv$normTables <- normTable.shash(
            rv$model,
            ages = ages,
            start = start_score,
            end = end_score,
            step = input$norm_step,
            CI = ci_val,
            reliability = rel_val
          )

        } else if (input$distributionType == "betabinomial") {

          # Beta-Binomial: keine start/end/step Argumente!
          # Tabelle wird automatisch ueber den gesamten Itembereich (0:n) erzeugt.
          rv$normTables <- normTable.betabinomial(
            rv$model,
            ages = ages,
            CI = ci_val,
            reliability = rel_val
          )

          # Optional: Tabelle nachtraeglich auf gewuenschten Bereich filtern
          start_score <- input$norm_start
          end_score   <- input$norm_end

          if ((!is.null(start_score) && !is.na(start_score)) ||
              (!is.null(end_score)   && !is.na(end_score))) {

            rv$normTables <- lapply(rv$normTables, function(df) {
              # Spalte mit Rohwerten ermitteln
              sc <- if ("raw"   %in% names(df)) "raw"
              else if ("x"     %in% names(df)) "x"
              else if ("score" %in% names(df)) "score"
              else names(df)[1]

              keep <- rep(TRUE, nrow(df))
              if (!is.null(start_score) && !is.na(start_score))
                keep <- keep & df[[sc]] >= start_score
              if (!is.null(end_score) && !is.na(end_score))
                keep <- keep & df[[sc]] <= end_score
              df[keep, , drop = FALSE]
            })
          }
        }

        incProgress(1, detail = "Done!")

        showNotification("Norm tables generated!", type = "message")

      }, error = function(e) {
        showNotification(
          paste("Error generating tables:", e$message),
          type = "error",
          duration = 10
        )
      })
    })
  })

  # Display norm tables
  output$normTablesUI <- renderUI({
    req(rv$normTables)

    table_tabs <- lapply(names(rv$normTables), function(age_name) {
      tabPanel(
        paste("Age", age_name),
        tags$br(),
        DT::DTOutput(paste0("normTable_", age_name))
      )
    })

    do.call(tabsetPanel, table_tabs)
  })

  # Render individual norm tables
  observe({
    req(rv$normTables)

    lapply(names(rv$normTables), function(age_name) {
      output[[paste0("normTable_", age_name)]] <- DT::renderDT({

        df <- rv$normTables[[age_name]]

        # Numerische Spalten, die gerundet werden sollen
        candidate_cols <- c("Px", "Pcum", "Percentile", "z", "norm",
                            "lowerCI", "upperCI",
                            "lowerCI_PR", "upperCI_PR",
                            "lower", "upper")
        cols_to_round <- intersect(candidate_cols, names(df))

        dt <- DT::datatable(
          df,
          options = list(
            pageLength = 25,
            scrollX = TRUE,
            dom = 'Bfrtip'
          ),
          rownames = FALSE
        )

        if (length(cols_to_round) > 0) {
          dt <- DT::formatRound(dt, columns = cols_to_round, digits = 2)
        }
        dt
      })
    })
  })

  # Download norm tables
  output$downloadTables <- downloadHandler(
    filename = function() {
      paste0("norm_tables_", input$distributionType, "_", Sys.Date(), ".csv")
    },
    content = function(file) {
      req(rv$normTables)

      # Combine all tables with age column
      combined <- do.call(rbind, lapply(names(rv$normTables), function(age_name) {
        df <- rv$normTables[[age_name]]
        df$age <- suppressWarnings(as.numeric(age_name))
        df
      }))

      write.csv(combined, file, row.names = FALSE)
    }
  )
})

#' Fit a Sinh-Arcsinh (shash) Regression Model for Continuous Norming
#'
#' This function fits a Sinh-Arcsinh (shash; Jones & Pewsey, 2009) regression model for continuous norm
#' score modeling, where the distribution parameters vary smoothly as polynomial functions of age or other
#' predictors. The shash distribution is well-suited for psychometric data as it can flexibly model
#' skewness and tail weight independently, making it ideal for handling floor effects, ceiling effects,
#' and varying degrees of individual differences across age groups. In a simulation study (Lenhard et
#' al, 2019), the shash model demonstrated superior performance compared to other parametric approaches
#' from the Box Cox family of functions. In contrast to Box Cox, Sinh-Arcsinh can model distributions
#' including zero and negative values.
#'
#'
#' @param age A numeric vector of predictor values (typically age, but can be any continuous predictor).
#' @param score A numeric vector of response values (raw test scores). Must be the same length as age.
#'   The value range is unrestricted and it can include zeros and negative values.
#' @param weights An optional numeric vector of weights for each observation.
#'   Useful for incorporating sampling weights. If NULL (default), all observations are weighted equally.
#' @param mu_degree Integer specifying the degree of the polynomial for modeling the location parameter mu(age).
#'   Default is 3. Higher degrees allow more flexible modeling of how the central tendency changes with age,
#'   but may lead to overfitting with small samples. Common choices:
#'   \itemize{
#'     \item 1: Linear change with age
#'     \item 2: Quadratic change (allows one inflection point)
#'     \item 3: Cubic change (allows two inflection points, suitable for most developmental curves)
#'     \item 4+: Higher-order changes (use cautiously, mainly for large samples)
#'   }
#' @param sigma_degree Integer specifying the degree of the polynomial for modeling the scale parameter sigma(age).
#'   Default is 2. This controls how the variability (spread) of scores changes with age.
#'   Lower degrees are often sufficient as variability typically changes more smoothly than location.
#'
#' @param epsilon_degree Integer specifying the degree of the polynomial for modeling the skewness parameter epsilon(age).
#'   Default is 2. This controls how the asymmetry of the distribution changes with age.
#'
#' @param delta_degree Integer specifying the degree of the polynomial for modeling the tail weight parameter delta(age).
#'   Default is 1. The tail weight can be fixed as well in case of numerical instability. In that case, set
#'   \code{delta_degree} to NULL and specify a value for \code{delta} instead. Recommendation: Keep
#'   \code{delta_degree} low to avoid overfitting.
#'
#' @param delta Fixed tail weight parameter (must be > 0). Default is 1. This parameter controls the
#'   heaviness of the distribution tails and is kept constant across all ages when \code{delta_degree}
#'   is set to NULL. Common values:
#'   \itemize{
#'     \item delta = 1: Normal-like tail behavior (baseline)
#'     \item delta < 1: Heavier tails, higher kurtosis (more extreme scores than normal distribution)
#'     \item delta > 1: Lighter tails, lower kurtosis (fewer extreme scores than normal distribution)
#'   }
#'
#' @param control An optional list of control parameters passed to the \code{optim} function for
#'   maximum likelihood estimation. Any parameters not specified are filled with sensible defaults
#'   based on the model complexity. Common parameters to adjust:
#'   \itemize{
#'     \item \code{factr}: Controls precision of optimization (default: 1e-8)
#'     \item \code{maxit}: Maximum number of iterations (default: n_parameters * 200)
#'     \item \code{lmm}: Memory limit for L-BFGS-B (default: min(n_parameters, 20))
#'   }
#'   Increase \code{maxit} or decrease \code{factr} if optimization fails to converge.
#'
#' @param scale Character string or numeric vector specifying the type of norm scale for output.
#'   This affects the scaling of derived norm scores but does not influence model fitting:
#'   \itemize{
#'     \item "T": T-scores (mean = 50, SD = 10) - default
#'     \item "IQ": IQ-like scores (mean = 100, SD = 15)
#'     \item "z": z-scores (mean = 0, SD = 1)
#'     \item c(M, SD): Custom scale with specified mean M and standard deviation SD
#'   }
#'
#' @param plot Logical indicating whether to automatically display a diagnostic plot of the fitted model.
#'   Default is TRUE.
#'
#' @return An object of class "cnormShash" containing the fitted model results. This is a list with components:
#'   \item{mu_est}{Numeric vector of estimated coefficients for the location parameter mu(age).
#'     The first coefficient is the intercept, subsequent coefficients correspond to polynomial terms.}
#'   \item{sigma_est}{Numeric vector of estimated coefficients for the scale parameter log(sigma(age)).
#'     Note: These are coefficients for log(sigma) to ensure sigma > 0.}
#'   \item{epsilon_est}{Numeric vector of estimated coefficients for the skewness parameter epsilon(age).}
#'   \item{delta}{The fixed tail weight parameter value used in fitting (relevant if \code{delta_degree} is NULL).}
#'   \item{delta_est}{Numeric vector of estimated coefficients for log(delta(age)) -
#'      in case a degree has been set, otherwise NULL.}
#'   \item{se}{Numeric vector of standard errors for all estimated coefficients (if Hessian computation succeeds).}
#'   \item{mu_degree, sigma_degree, epsilon_degree}{The polynomial degrees used for each parameter.}
#'   \item{delta_degree}{The polynomial degree used for delta, or NULL if delta was fixed.}
#'   \item{result}{Complete output from the \code{optim} function, including convergence information,
#'     log-likelihood value, and other optimization details.}
#'
#' @details
#' This implementation uses the Jones & Pewsey (2009) parameterization of the Sinh-Arcsinh distribution.
#' Parameters are estimated using maximum likelihood via the L-BFGS-B algorithm with analytic
#' gradients. In case optimization fails, try reducing model complexity by reducing polynomial
#' degrees or fixing the delta parameter.
#'
#' \subsection{The Sinh-Arcsinh Distribution}{
#' The shash distribution is defined by the transformation:
#' \deqn{X = \mu + \sigma \cdot \sinh\left(\frac{\text{arcsinh}(Y) - \epsilon}{\delta}\right)}
#' where Y is a standard normal variable, Y ~ N(0,1).
#'
#' This transformation provides:
#' \itemize{
#'   \item mu: Location parameter (similar to mean)
#'   \item sigma: Scale parameter (similar to standard deviation)
#'   \item epsilon: Skewness parameter (epsilon = 0 for symmetry)
#'   \item delta: Tail weight parameter (delta = 1 for normal-like tails)
#' }
#' }
#'
#' \subsection{Model Selection}{
#' Choose polynomial degrees based on:
#' \itemize{
#'   \item Sample size (higher degrees need more data)
#'   \item Theoretical expectations about developmental trajectories
#'   \item Model comparison criteria (AIC, BIC)
#'   \item Visual inspection of fitted curves
#' }
#'
#' For most applications, mu_degree = 3, sigma_degree = 2, epsilon_degree = 2, delta_degree = 1 provides
#' a good balance of flexibility and parsimony.
#' }
#'
#' @note
#' \itemize{
#'   \item The function requires the input data to have sufficient variability. Very small datasets
#'     or datasets with little age spread may cause convergence problems.
#'   \item Polynomial models can exhibit edge effects at the boundaries of the age range.
#'     Predictions outside the observed age range should be made cautiously.
#'   \item If convergence fails, try: (1) reducing polynomial degrees, (2) adjusting the delta parameter,
#'     (3) providing custom control parameters, or (4) checking for data quality issues.
#'   \item If parameters end up at their box constraints, the reported standard errors and
#'     Wald tests for these parameters are not valid; a warning is issued in this case.
#' }
#'
#' @seealso
#' \code{\link{plot}} for plotting fitted models,
#' \code{\link{predict}} for generating predictions,
#' \code{\link{cnorm.betabinomial2}} for discrete beta-binomial alternative
#'
#' @examples
#' \dontrun{
#' # Basic usage with default settings
#' model <- cnorm.shash(age = children$age, score = children$raw_score)
#'
#' # Custom polynomial degrees for complex developmental pattern
#' model_complex <- cnorm.shash(
#'   age = adolescents$age,
#'   score = adolescents$vocabulary_score,
#'   mu_degree = 4,         # Complex mean trajectory
#'   sigma_degree = 3,      # Changing variability pattern
#'   epsilon_degree = 2,    # Skewness shifts
#'   delta_degree = NULL,   # set to NULL to activate fixed delta
#'   delta = 0.8            # Slightly heavy tails
#' )
#'
#' # With sampling weights
#' model_weighted <- cnorm.shash(
#'   age = survey$age,
#'   score = survey$score,
#'   weights = survey$sample_weight
#' )
#'
#' # Custom optimization control for difficult convergence
#' model_robust <- cnorm.shash(
#'   age = mixed$age,
#'   score = mixed$score,
#'   control = list(factr = 1e-6, maxit = 2000),
#'   delta = 1.5
#' )
#'
#' # Compare model fit
#' compare(model, model_complex)
#' }
#'
#' @author Wolfgang Lenhard
#' @references
#' Jones, M. C., & Pewsey, A. (2009). Sinh-arcsinh distributions. *Biometrika*, 96(4), 761-780.
#'
#' Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A
#' simulation study of parametric and semi-parametric approaches. *PLoS ONE*, 14(9), e0222279.
#' https://doi.org/10.1371/journal.pone.0222279
#'
#' @export
cnorm.shash <- function(age,
                        score,
                        weights = NULL,
                        mu_degree = 3,
                        sigma_degree = 2,
                        epsilon_degree = 2,
                        delta_degree = 1,
                        delta = 1,
                        control = NULL,
                        scale = "T",
                        plot = TRUE) {
  # Input validation
  if (length(age) != length(score)) {
    stop("Length of 'age' and 'score' must be the same.")
  }

  if (!is.null(weights) && length(age) != length(weights)) {
    stop("Length of 'weights' must match length of 'age' and 'score'.")
  }

  # ensure positivity of delta
  if (delta <= 0) {
    stop("Delta parameter must be positive.")
  }

  # validate 'scale' early with explicit errors. Accepts integer vectors as well.
  if (is.numeric(scale) && length(scale) == 2) {
    scaleM <- scale[1]
    scaleSD <- scale[2]
  } else if (is.character(scale) && length(scale) == 1) {
    scaleM <- switch(scale,
                     "T" = 50,
                     "IQ" = 100,
                     "z" = 0,
                     stop("Unknown scale '", scale,
                          "'. Use 'T', 'IQ', 'z', or a numeric vector c(M, SD)."))
    scaleSD <- switch(scale, "T" = 10, "IQ" = 15, "z" = 1)
  } else {
    stop("'scale' must be 'T', 'IQ', 'z', or a numeric vector c(M, SD).")
  }

  # Prepare vectors
  vectors_to_check <- list(age = age, score = score)
  if (!is.null(weights)) {
    vectors_to_check$weights <- weights
  }

  # Check if filtering needed
  needs_filtering <- any(sapply(vectors_to_check, function(x)
    any(!is.finite(x))))

  if (needs_filtering) {
    message("Vector(s) contained non-finite values (NA, NaN, Inf). These cases will be removed.")
    tmp <- do.call(filter_complete, c(vectors_to_check, verbose = FALSE))
    age <- tmp[[1]]
    score <- tmp[[2]]
    if (!is.null(weights))
      weights <- tmp[[3]]
  }

  # Standardize age
  age_std <- standardize(age)

  # Create design matrices
  X_mu <- cbind(1, poly(age_std, degree = mu_degree, raw = TRUE))
  X_sigma <- cbind(1, poly(age_std, degree = sigma_degree, raw = TRUE))
  X_epsilon <- cbind(1, poly(age_std, degree = epsilon_degree, raw = TRUE))

  # In case, delta polynomials are used as well
  if (!is.null(delta_degree)) {
    X_delta <- cbind(1, poly(age_std, degree = delta_degree, raw = TRUE))
    use_varying_delta <- TRUE
  } else {
    X_delta <- NULL
    use_varying_delta <- FALSE
  }

  # Calculate initial parameter values
  initial_mu <- mean(score)
  initial_sigma <- sd(score)
  initial_epsilon <- 0  # Start with symmetric distribution

  # Guard against degenerate input (zero/undefined variance)
  if (!is.finite(initial_sigma) || initial_sigma <= 0) {
    stop("'score' has zero or undefined variance. The model cannot be fitted.")
  }

  # Initial parameter vectors
  initial_params <- c(
    c(initial_mu, rep(0, mu_degree)),
    # mu parameters
    c(log(initial_sigma), rep(0, sigma_degree)),
    # sigma parameters
    c(initial_epsilon, rep(0, epsilon_degree))   # epsilon parameters
  )

  if (use_varying_delta) {
    initial_params <- c(initial_params, c(log(delta), rep(0, delta_degree))) # delta parameters
  }

  n_param <- length(initial_params)
  control_defaults <- list(factr = 1e-8,
                           maxit = n_param * 200,
                           lmm = min(n_param, 20))
  control <- if (is.null(control))
    control_defaults
  else
    utils::modifyList(control_defaults, control)

  # Parameter bounds
  n_params <- length(initial_params)
  lower_bounds <- rep(-Inf, n_params)
  upper_bounds <- rep(Inf, n_params)

  # Bounds for sigma parameters (keep log_sigma reasonable)
  sigma_start <- ncol(X_mu) + 1
  sigma_end <- sigma_start + ncol(X_sigma) - 1
  lower_bounds[sigma_start:sigma_end] <- -10
  upper_bounds[sigma_start:sigma_end] <- 10

  # Bounds for epsilon parameters
  epsilon_start <- sigma_end + 1
  epsilon_end   <- epsilon_start + ncol(X_epsilon) - 1
  lower_bounds[epsilon_start:epsilon_end] <- -10
  upper_bounds[epsilon_start:epsilon_end] <- 10

  if (use_varying_delta) {
    delta_start <- epsilon_end + 1
    delta_end   <- delta_start + ncol(X_delta) - 1
    lower_bounds[delta_start:delta_end] <- -2
    upper_bounds[delta_start:delta_end] <- 2
  }

  # Optimization (with analytic gradient)
  result <- tryCatch({
    optim(
      initial_params,
      log_likelihood_shash,
      gr = gradient_shash,
      X_mu = X_mu,
      X_sigma = X_sigma,
      X_epsilon = X_epsilon,
      X_delta = X_delta,
      y = score,
      weights = weights,
      fixed_delta = if (!use_varying_delta)
        delta
      else
        NULL,
      method = "L-BFGS-B",
      lower = lower_bounds,
      upper = upper_bounds,
      hessian = TRUE,
      control = control
    )
  }, error = function(e) {
    message("First optimization attempt failed. Trying with different parameters...")

    # Try with different, more robust initial values
    initial_params[1] <- median(score)

    # mad(score) can be 0 with strong floor/ceiling effects
    # (> 50% identical scores); fall back to sd in that case.
    s_robust <- mad(score)
    if (!is.finite(s_robust) || s_robust <= 0)
      s_robust <- sd(score)
    initial_params[ncol(X_mu) + 1] <- log(s_robust)

    # guard against missing list entries when mutating a
    # user-supplied control list (NULL * 10 == numeric(0)).
    control$factr <- if (is.null(control$factr)) 1e-7 else control$factr * 10
    control$maxit <- if (is.null(control$maxit)) 1000 else control$maxit * 2

    optim(
      initial_params,
      log_likelihood_shash,
      gr = gradient_shash,
      X_mu = X_mu,
      X_sigma = X_sigma,
      X_epsilon = X_epsilon,
      X_delta = X_delta,
      y = score,
      weights = weights,
      fixed_delta = if (!use_varying_delta)
        delta
      else
        NULL,
      method = "L-BFGS-B",
      lower = lower_bounds,
      upper = upper_bounds,
      hessian = TRUE,
      control = control
    )
  })

  # Check convergence
  if (result$convergence != 0) {
    warning(
      "Optimization did not converge (code: ",
      result$convergence,
      "). Consider adjusting control parameters or degrees of the polynomials. Check percentile curves for plausibility."
    )
  }

  # Warn when parameters sit at their box constraints; SEs and Wald
  # tests are invalid there.
  at_bound <- is.finite(lower_bounds) &
    (result$par <= lower_bounds + 1e-6 | result$par >= upper_bounds - 1e-6)
  if (any(at_bound)) {
    warning(
      "Parameter(s) at index ",
      paste(which(at_bound), collapse = ", "),
      " reached their bounds. Standard errors and Wald tests for these ",
      "parameters are not valid. Consider reducing model complexity."
    )
  }

  # Extract parameter estimates
  n_mu <- ncol(X_mu)
  n_sigma <- ncol(X_sigma)
  n_epsilon <- ncol(X_epsilon)

  mu_est <- result$par[1:n_mu]
  sigma_est <- result$par[(n_mu + 1):(n_mu + n_sigma)]
  epsilon_est <- result$par[(n_mu + n_sigma + 1):(n_mu + n_sigma + n_epsilon)]

  if (use_varying_delta) {
    n_delta <- ncol(X_delta)
    delta_est <- result$par[(n_mu + n_sigma + n_epsilon + 1):(n_mu + n_sigma + n_epsilon + n_delta)]
  } else {
    delta_est <- NULL
  }

  # Guarded SE computation. solve() on an indefinite Hessian could otherwise
  # either error or silently produced NaN via sqrt of negative diagonals.
  se <- tryCatch({
    h_inv <- solve(result$hessian)
    d <- diag(h_inv)
    if (any(d < 0, na.rm = TRUE)) {
      warning("Hessian is not positive definite; standard errors are set to NA where invalid.")
      d[d < 0] <- NA
    }
    sqrt(d)
  }, error = function(e) {
    warning("Could not compute standard errors: Hessian matrix issue")
    rep(NA_real_, length(result$par))
  })

  # Store attributes
  attr(result, "age_mean") <- mean(age)
  attr(result, "age_sd") <- sd(age)
  attr(result, "ageMin") <- min(age)
  attr(result, "ageMax") <- max(age)
  attr(result, "score_mean") <- mean(score)
  attr(result, "score_sd") <- sd(score)
  attr(result, "max") <- max(score)
  attr(result, "min") <- min(score)
  attr(result, "N") <- length(score)
  attr(result, "scaleMean") <- scaleM
  attr(result, "scaleSD") <- scaleSD
  attr(result, "delta") <- delta

  # Create model object
  model <- list(
    mu_est = mu_est,
    sigma_est = sigma_est,
    epsilon_est = epsilon_est,
    delta_est = delta_est,
    delta = delta,
    se = se,
    mu_degree = mu_degree,
    sigma_degree = sigma_degree,
    epsilon_degree = epsilon_degree,
    delta_degree = delta_degree,
    result = result
  )

  class(model) <- "cnormShash"

  if (plot) {
    p <- plot.cnormShash(model, age, score, weights = weights)
    print(p)
  }

  return(model)
}



#' Sinh-Arcsinh (shash) Distribution
#'
#' Density, distribution function, quantile function and random generation
#' for the Sinh-Arcsinh distribution with location parameter \code{mu},
#' scale parameter \code{sigma}, skewness parameter \code{epsilon}, and
#' tail weight parameter \code{delta}.
#'
#' @name shash
#' @aliases dshash pshash qshash rshash
#'
#' @param x,q vector of quantiles
#' @param p vector of probabilities
#' @param n number of observations. If \code{length(n) > 1}, the length is
#'   taken to be the number required.
#' @param mu location parameter (default: 0)
#' @param sigma scale parameter (must be > 0, default: 1)
#' @param epsilon skewness parameter (default: 0, symmetric distribution)
#' @param delta tail weight parameter (must be > 0, default: 1 for normal-like
#'   tails; \code{delta < 1} gives heavier tails, \code{delta > 1} gives
#'   lighter tails)
#' @param log,log.p logical; if TRUE, probabilities p are given as log(p)
#' @param lower.tail logical; if TRUE (default), probabilities are P[X <= x]
#'   otherwise, P[X > x]
#'
#' @details
#' The Sinh-Arcsinh distribution (Jones & Pewsey, 2009) is defined by the transformation:
#' \deqn{X = \mu + \sigma \cdot \sinh\left(\frac{\text{asinh}(Z) - \epsilon}{\delta}\right)}
#' where \eqn{Z \sim N(0,1)} is a standard normal variable.
#'
#' The four parameters control:
#' \itemize{
#'   \item \code{mu}: Location (similar to mean)
#'   \item \code{sigma}: Scale (similar to standard deviation)
#'   \item \code{epsilon}: Skewness (\code{epsilon = 0} gives symmetry)
#'   \item \code{delta}: Tail weight (\code{delta = 1} gives normal-like tails,
#'         \code{delta < 1} gives heavier tails, \code{delta > 1} gives
#'         lighter tails)
#' }
#'
#' @return
#' \code{dshash} gives the density, \code{pshash} gives the distribution
#' function, \code{qshash} gives the quantile function, and \code{rshash}
#' generates random deviates.
#'
#' The length of the result is determined by \code{n} for \code{rshash}, and
#' is the maximum of the lengths of the numerical arguments for the other functions.
#'
#' @references
#' Jones, M. C., & Pewsey, A. (2009). Sinh-arcsinh distributions.
#' \emph{Biometrika}, 96(4), 761-780. \doi{10.1093/biomet/asp053}
#'
#' @seealso
#' \link[stats]{Normal} for the normal distribution.
#'
#' @examples
#' # Generate random samples
#' x <- rshash(1000, mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2)
#'
#' # Density
#' plot(density(x))
#' curve(dshash(x, mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2),
#'       add = TRUE, col = "red")
#'
#' # Cumulative probability
#' pshash(0, mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2)
#'
#' # Quantiles
#' qshash(c(0.025, 0.5, 0.975), mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2)
#'
#' # Compare with normal distribution (epsilon = 0, delta = 1)
#' oldpar <- par(mfrow = c(2, 2))
#' x_vals <- seq(-4, 4, length.out = 200)
#' plot(x_vals, dshash(x_vals), type = "l", main = "Symmetric (like normal)")
#' plot(x_vals, dshash(x_vals, epsilon = 1), type = "l", main = "Right skewed")
#' plot(x_vals, dshash(x_vals, delta = 0.5), type = "l", main = "Heavy tails")
#' plot(x_vals, dshash(x_vals, delta = 2), type = "l", main = "Light tails")
#' par(oldpar)
#'
#' @export
#' @rdname shash
dshash <- function(x,
                   mu = 0,
                   sigma = 1,
                   epsilon = 0,
                   delta = 1,
                   log = FALSE) {
  z <- (x - mu) / sigma

  # Use builtin asinh() -- the manual log(z + sqrt(z^2 + 1)) suffers
  # catastrophic cancellation for large negative z and is slower.
  arg <- delta * asinh(z) + epsilon
  abs_arg <- pmin.int(abs(arg), 350)

  # log(cosh(a)) = |a| + log1p(exp(-2|a|)) - log(2), stable for all a
  log_cosh <- abs_arg + log1p(exp(-2 * abs_arg)) - 0.6931471805599453  # log(2)

  # sinh(arg)^2 == sinh(|arg|)^2
  S <- sinh(abs_arg)

  logdens <- log(delta) - log(sigma) - 0.918938533204673 -  # log(sqrt(2*pi))
    0.5 * log1p(z * z) - 0.5 * S * S + log_cosh

  # Safety net: any residual numerical NaN from extreme (but valid) inputs
  # corresponds to density 0. Genuine NaN/NA inputs are propagated.
  nan_idx <- is.nan(logdens) & !is.nan(x + mu + sigma + epsilon + delta)
  if (any(nan_idx)) {
    logdens[nan_idx] <- -Inf
  }

  if (log)
    return(logdens)
  else
    return(exp(logdens))
}

#' @export
#' @rdname shash
pshash <- function(q,
                   mu = 0,
                   sigma = 1,
                   epsilon = 0,
                   delta = 1,
                   lower.tail = TRUE,
                   log.p = FALSE) {
  z <- (q - mu) / sigma
  u <- sinh(delta * asinh(z) + epsilon)

  p <- pnorm(u, lower.tail = lower.tail, log.p = log.p)
  return(p)
}

#' @export
#' @rdname shash
qshash <- function(p,
                   mu = 0,
                   sigma = 1,
                   epsilon = 0,
                   delta = 1,
                   lower.tail = TRUE,
                   log.p = FALSE) {
  # Delegate lower.tail/log.p to qnorm() instead of computing
  # exp(p) and 1 - p manually, which loses precision in the tails.
  u <- qnorm(p, lower.tail = lower.tail, log.p = log.p)
  z <- sinh((asinh(u) - epsilon) / delta)
  x <- mu + sigma * z

  return(x)
}

#' @export
#' @rdname shash
rshash <- function(n,
                   mu = 0,
                   sigma = 1,
                   epsilon = 0,
                   delta = 1) {
  u <- rnorm(n)
  z <- sinh((asinh(u) - epsilon) / delta)
  x <- mu + sigma * z

  return(x)
}

#' Calculate the negative log-likelihood for a shash regression model
#'
#' This function computes the negative log-likelihood for a Sinh-Arcsinh regression model
#' where the location, scale, skewness, and (optionally) tail weight parameters
#' are modeled as functions of predictors.
#'
#' @param params A numeric vector containing all model parameters
#' @param X_mu Design matrix for location parameter
#' @param X_sigma Design matrix for scale parameter
#' @param X_epsilon Design matrix for skewness parameter
#' @param X_delta Design matrix for tail weight parameter (or NULL for fixed delta)
#' @param fixed_delta If not NULL, the delta parameter is fixed to this value
#' @param y Response vector
#' @param weights Observation weights
#'
#' @return The negative log-likelihood of the model
#'
#' @keywords internal
log_likelihood_shash <- function(params,
                                 X_mu,
                                 X_sigma,
                                 X_epsilon,
                                 X_delta = NULL,
                                 y,
                                 weights = NULL,
                                 fixed_delta = NULL) {
  n_mu <- ncol(X_mu)
  n_sigma <- ncol(X_sigma)
  n_epsilon <- ncol(X_epsilon)
  n_delta <- if (!is.null(X_delta))
    ncol(X_delta)
  else
    0

  # Extract coefficients
  mu_coef <- params[seq_len(n_mu)]
  sigma_coef <- params[n_mu + seq_len(n_sigma)]
  epsilon_coef <- params[n_mu + n_sigma + seq_len(n_epsilon)]

  # Compute parameters with efficient matrix operations
  mu <- drop(X_mu %*% mu_coef)
  log_sigma <- pmin.int(pmax.int(drop(X_sigma %*% sigma_coef), -10), 10)
  sigma <- exp(log_sigma)
  epsilon <- pmin.int(pmax.int(drop(X_epsilon %*% epsilon_coef), -10), 10)

  if (!is.null(X_delta)) {
    delta_coef <- params[n_mu + n_sigma + n_epsilon + seq_len(n_delta)]
    log_delta <- pmin.int(pmax.int(drop(X_delta %*% delta_coef), -2), 2)
    delta <- exp(log_delta)
  } else {
    delta <- fixed_delta
  }

  if (is.null(weights)) {
    weights <- 1  # Will broadcast in multiplication
  }

  log_densities <- dshash(y, mu, sigma, epsilon, delta, log = TRUE)
  bad <- !is.finite(log_densities)
  if (any(bad)) {
    if (anyNA(log_densities))
      return(1e10)
    log_densities[bad] <- -1e6
  }

  loglik <- sum(weights * log_densities)

  if (!is.finite(loglik))
    return(1e10)

  return(-loglik)
}

#' Analytic gradient of the negative log-likelihood for a shash regression model
#'
#' Computes the exact gradient of \code{\link{log_likelihood_shash}} with
#' respect to all regression coefficients. Supplying this gradient to
#' \code{\link[stats]{optim}} (method \code{"L-BFGS-B"}) avoids costly and
#' noisy finite-difference approximations, typically yielding a speedup
#' and more reliable convergence.
#'
#' @details
#' With \eqn{z = (y - \mu)/\sigma} and \eqn{A = \delta \, \mathrm{asinh}(z) + \epsilon},
#' the per-observation log-density is
#' \deqn{\ell = \log\delta - \log\sigma - \tfrac{1}{2}\log(2\pi)
#'   - \tfrac{1}{2}\log(1+z^2) - \tfrac{1}{2}\sinh^2(A) + \log\cosh(A).}
#' The gradient uses the chain rule through the linear predictors:
#' \deqn{\partial\ell/\partial A = \tanh(A) - \sinh(A)\cosh(A)}
#' \deqn{\partial\ell/\partial z = -z/(1+z^2) + (\partial\ell/\partial A)\,\delta/\sqrt{1+z^2}}
#' \deqn{\partial\ell/\partial\mu = -(\partial\ell/\partial z)/\sigma}
#' \deqn{\partial\ell/\partial\log\sigma = -1 - z\,(\partial\ell/\partial z)}
#' \deqn{\partial\ell/\partial\epsilon = \partial\ell/\partial A}
#' \deqn{\partial\ell/\partial\log\delta = 1 + (\partial\ell/\partial A)\,\delta\,\mathrm{asinh}(z)}
#' Coefficient gradients follow as \eqn{-X^\top (w \cdot \partial\ell/\partial\eta)}
#' for the negative log-likelihood. Contributions of observations whose linear
#' predictor is clamped (see \code{log_likelihood_shash}) are zeroed, which is
#' the exact subgradient of the clamped objective. In extreme regions where
#' \eqn{\sinh(A)\cosh(A)} would overflow, the gradient magnitude is capped at
#' a large finite value; these regions coincide with the penalty branch of the
#' objective, so the capped direction remains correct.
#'
#' @inheritParams log_likelihood_shash
#'
#' @return Numeric vector: gradient of the negative log-likelihood with
#'   respect to \code{params}.
#'
#' @keywords internal
gradient_shash <- function(params,
                           X_mu,
                           X_sigma,
                           X_epsilon,
                           X_delta = NULL,
                           y,
                           weights = NULL,
                           fixed_delta = NULL) {
  n_mu <- ncol(X_mu)
  n_sigma <- ncol(X_sigma)
  n_epsilon <- ncol(X_epsilon)
  n_delta <- if (!is.null(X_delta))
    ncol(X_delta)
  else
    0

  # Extract coefficients
  mu_coef <- params[seq_len(n_mu)]
  sigma_coef <- params[n_mu + seq_len(n_sigma)]
  epsilon_coef <- params[n_mu + n_sigma + seq_len(n_epsilon)]

  # Linear predictors with clamping identical to log_likelihood_shash;
  # indicator vectors record where the clamp is inactive (gradient nonzero)
  mu <- drop(X_mu %*% mu_coef)

  eta_sigma <- drop(X_sigma %*% sigma_coef)
  sigma_free <- (eta_sigma > -10) & (eta_sigma < 10)
  sigma <- exp(pmin.int(pmax.int(eta_sigma, -10), 10))

  eta_epsilon <- drop(X_epsilon %*% epsilon_coef)
  epsilon_free <- (eta_epsilon > -10) & (eta_epsilon < 10)
  epsilon <- pmin.int(pmax.int(eta_epsilon, -10), 10)

  if (!is.null(X_delta)) {
    delta_coef <- params[n_mu + n_sigma + n_epsilon + seq_len(n_delta)]
    eta_delta <- drop(X_delta %*% delta_coef)
    delta_free <- (eta_delta > -2) & (eta_delta < 2)
    delta <- exp(pmin.int(pmax.int(eta_delta, -2), 2))
  } else {
    delta <- fixed_delta
  }

  if (is.null(weights)) {
    weights <- 1
  }

  # Core quantities
  z <- (y - mu) / sigma
  asinh_z <- asinh(z)
  A <- delta * asinh_z + epsilon

  # Overflow guard: sinh(a)*cosh(a) = 0.5*sinh(2a) overflows for |a| > ~355.
  # Cap at 350 (product ~ 2.5e303, still finite), then bound the magnitude.
  A_c <- pmin.int(pmax.int(A, -350), 350)
  S <- sinh(A_c)
  C <- cosh(A_c)

  gA <- tanh(A) - S * C
  gA <- pmin.int(pmax.int(gA, -1e10), 1e10)

  r <- 1 / sqrt(1 + z * z)            # d asinh(z) / dz
  gz <- -z * r * r + gA * delta * r   # dl/dz

  d_mu   <- -gz / sigma               # dl/dmu
  d_lsig <- -1 - gz * z               # dl/d(log sigma)
  d_eps  <- gA                        # dl/depsilon

  # Zero out contributions from numerically failed observations
  # (consistent with the constant penalty branch of the objective)
  ok <- is.finite(d_mu) & is.finite(d_lsig) & is.finite(d_eps)

  if (!is.null(X_delta)) {
    d_ldel <- 1 + gA * delta * asinh_z  # dl/d(log delta)
    ok <- ok & is.finite(d_ldel)
    d_ldel[!ok] <- 0
    d_ldel <- d_ldel * delta_free
  }

  d_mu[!ok] <- 0
  d_lsig[!ok] <- 0
  d_eps[!ok] <- 0

  # Apply clamp subgradients
  d_lsig <- d_lsig * sigma_free
  d_eps <- d_eps * epsilon_free

  # Gradient of the NEGATIVE weighted log-likelihood w.r.t. coefficients
  w <- weights
  grad <- c(
    -drop(crossprod(X_mu, w * d_mu)),
    -drop(crossprod(X_sigma, w * d_lsig)),
    -drop(crossprod(X_epsilon, w * d_eps)),
    if (!is.null(X_delta)) -drop(crossprod(X_delta, w * d_ldel)) else NULL
  )

  # Final safety: L-BFGS-B cannot handle non-finite gradients
  grad[!is.finite(grad)] <- 0

  return(grad)
}


#' Predict parameters for a shash regression model
#'
#' @param model An object of class "cnormShash"
#' @param ages A numeric vector of age points for prediction
#'
#' @return A data frame with predicted parameters and statistics
#'
#' @keywords internal
predictCoefficients_shash <- function(model, ages) {
  if (!isSHASH(model)) {
    stop("Wrong object. Please provide object from class 'cnormShash'.")
  }

  # Standardize new ages
  ages_std <- (ages - attr(model$result, "age_mean")) / attr(model$result, "age_sd")

  # Create design matrices
  X_mu_new <- cbind(1, poly(ages_std, degree = model$mu_degree, raw = TRUE))
  X_sigma_new <- cbind(1, poly(ages_std, degree = model$sigma_degree, raw = TRUE))
  X_epsilon_new <- cbind(1, poly(ages_std, degree = model$epsilon_degree, raw = TRUE))

  if (!is.null(model$delta_degree)) {
    X_delta_new <- cbind(1, poly(ages_std, degree = model$delta_degree, raw = TRUE))
  }

  # Predict parameters
  mu <- X_mu_new %*% model$mu_est
  log_sigma <- X_sigma_new %*% model$sigma_est
  sigma <- exp(log_sigma)
  epsilon <- X_epsilon_new %*% model$epsilon_est

  if (!is.null(model$delta_degree)) {
    log_delta <- X_delta_new %*% model$delta_est
    delta <- exp(log_delta)
  } else {
    delta <- rep(model$delta, length(ages))
  }

  predicted <- data.frame(
    age = ages,
    mu = as.vector(mu),
    sigma = as.vector(sigma),
    epsilon = as.vector(epsilon),
    delta = as.vector(delta)
  )

  return(predicted)
}

#' Plot Sinh-Arcsinh Model with Data and Percentile Lines
#'
#' @param x A fitted model object of class "cnormShash"
#' @param ... Additional arguments including age, score, weights, percentiles, points
#'
#' @return A ggplot object
#'
#' @export
plot.cnormShash <- function(x, ...) {
  model <- x
  args <- list(...)

  if ("age" %in% names(args)) {
    age <- args$age
  } else {
    if (length(args) > 0)
      age <- args[[1]]
    else
      age <- NULL
  }
  if ("score" %in% names(args)) {
    score <- args$score
  } else {
    if (length(args) > 1)
      score <- args[[2]]
    else
      score <- NULL
  }
  if ("weights" %in% names(args)) {
    weights <- args$weights
  } else {
    weights <- NULL
  }
  if ("percentiles" %in% names(args)) {
    percentiles <- args$percentiles
  } else {
    percentiles <- c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975)
  }
  if ("points" %in% names(args)) {
    points <- args$points
  } else {
    points <- TRUE
  }

  if (is.null(age) || is.null(score))
    stop("Please provide 'age' and 'score' vectors.")

  if (!isSHASH(model)) {
    stop("Wrong object. Please provide object from class 'cnormShash'.")
  }

  if (length(age) != length(score)) {
    stop("Length of 'age' and 'score' must be the same.")
  }

  if (!is.null(weights) && length(weights) != length(age)) {
    stop("Length of 'weights' must match length of 'age' and 'score'.")
  }

  # Generate prediction points
  n_points <- 100
  data <- data.frame(age = age, score = score)
  if (!is.null(weights)) {
    data$w <- weights
  } else {
    data$w <- rep(1, nrow(data))
  }

  age_range <- range(age)
  pred_ages <- seq(age_range[1], age_range[2], length.out = n_points)

  # Get predictions
  preds <- predictCoefficients_shash(model, pred_ages)

  # Calculate percentile lines using shash quantiles
  percentile_lines <- lapply(percentiles, function(p) {
    qshash(
      p,
      mu = preds$mu,
      sigma = preds$sigma,
      epsilon = preds$epsilon,
      delta = preds$delta
    )
  })

  percentile_data <- do.call(cbind, percentile_lines)
  colnames(percentile_data) <- paste0("P", percentiles * 100)

  plot_data <- data.frame(
    age = pred_ages,
    mu = preds$mu,
    sigma = preds$sigma,
    epsilon = preds$epsilon,
    percentile_data
  )

  # Create the plot
  p <- ggplot()

  if (points)
    p <- p + geom_point(
      data = data,
      aes(x = age, y = score),
      alpha = 0.2,
      size = 0.6
    )

  # Calculate manifest percentiles
  if (length(age) / length(unique(age)) > 50 &&
      min(table(data$age)) > 30) {
    data$group <- age
  } else {
    data$group <- getGroups(age)
  }

  # Limit to max 30 groups for better visibility
  if (length(unique(data$group)) > 30) {
    data$group <- getGroups(age, n = 30)
  }

  # Get actual percentiles
  NAMES <- paste("PR", percentiles * 100, sep = "")
  percentile.actual <- as.data.frame(do.call("rbind", lapply(split(data, data$group), function(df) {
    c(age = mean(df$age),
      weighted.quantile(df$score, probs = percentiles, weights = df$w))
  })))
  colnames(percentile.actual) <- c("age", NAMES)
  manifest_data <- percentile.actual

  # Add percentile lines and points
  for (i in seq_along(percentiles)) {
    p <- p +
      geom_line(
        data = plot_data,
        aes(
          x = .data$age,
          y = .data[[paste0("P", percentiles[i] * 100)]],
          color = !!NAMES[i]
        ),
        linewidth = 0.6
      ) +
      geom_point(
        data = manifest_data,
        aes(
          x = .data$age,
          y = .data[[NAMES[i]]],
          color = !!NAMES[i]
        ),
        size = 2,
        shape = 18
      )
  }

  # Customize the plot
  p <- p +
    theme_minimal() +
    labs(
      title = "Percentile Plot (Sinh-Arcsinh Model)",
      x = "Age",
      y = "Score",
      color = "Percentile"
    ) +
    scale_color_manual(
      values = setNames(rainbow(length(percentiles)), NAMES),
      breaks = NAMES,
      labels = paste0(percentiles * 100, "%")
    ) +
    guides(color = guide_legend(override.aes = list(
      linetype = rep("solid", length(NAMES)),
      shape = rep(18, length(NAMES))
    )))

  p <- p +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      plot.subtitle = element_text(hjust = 0.5, size = 12),
      axis.title = element_text(size = 12, face = "bold"),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  return(p)
}

#' Print method for Sinh-Arcsinh objects
#'
#' @param x A cnormShash object
#' @param ... Additional arguments
#' @export
print.cnormShash <- function(x, ...) {
  cat("SinH-ArcSinH Regression Model for Continuous Norming\n")
  cat("====================================================\n\n")

  cat("Model Parameters:\n")
  cat("- Location (mu) polynomial degree:", x$mu_degree, "\n")
  cat("- Scale (sigma) polynomial degree:", x$sigma_degree, "\n")
  cat("- Skewness (epsilon) polynomial degree:",
      x$epsilon_degree,
      "\n")
  if (!is.null(x$delta_degree)) {
    cat("- Tail weight (delta): Polynomial degree",
        x$delta_degree,
        "\n\n")
  } else {
    cat("- Tail weight (delta): Fixed at", x$delta, "\n\n")
  }

  cat("Sample size:", attr(x$result, "N"), "\n")
  cat("Age range:",
      attr(x$result, "ageMin"),
      "to",
      attr(x$result, "ageMax"),
      "\n")
  cat("Score range:",
      attr(x$result, "min"),
      "to",
      attr(x$result, "max"),
      "\n\n")

  cat("Optimization:\n")
  cat("- Convergence:",
      ifelse(x$result$convergence == 0, "Successful", "Failed"),
      "\n")
  cat("- Log-likelihood:", -x$result$value, "\n")
  cat("- AIC:", 2 * length(x$result$par) + 2 * x$result$value, "\n")
}

#' Automatic model selection for SinH-ArcSinH continuous norming via BIC
#'
#' Selects the polynomial degrees for the four components of a
#' \code{\link{cnorm.shash}} model (\eqn{\mu}, \eqn{\sigma}, \eqn{\epsilon},
#' \eqn{\delta}) by minimizing BIC over a full grid.
#'
#' A degree of \code{0} for \eqn{\delta} is interpreted as a **fixed**
#' \eqn{\delta} (passed to \code{\link{cnorm.shash}} as \code{delta_degree = NULL})
#' with value taken from the \code{delta} argument. This allows the grid
#' search to choose between a fixed and a varying tail-weight model.
#'
#' Parallel execution is attempted by default. If the workers cannot access
#' the \pkg{cNORM} namespace, the function transparently falls
#' back to sequential execution.
#'
#' @param age,score Numeric vectors of predictor and response values.
#' @param weights Optional numeric vector of observation weights.
#' @param max_mu,max_sigma,max_epsilon,max_delta Maximum polynomial degrees for
#'   the four shash parameters. Defaults: \code{4, 3, 3, 1}.
#' @param min_mu,min_sigma,min_epsilon Minimum polynomial degrees (default 1).
#' @param min_delta Minimum degree for \eqn{\delta}. \code{0} (default) means
#'   the fixed-delta variant is included in the search.
#' @param delta Value of \eqn{\delta} used whenever the candidate uses
#'   \code{delta_degree = 0} (i.e. fixed delta). Default 1.
#' @param control Optional control list passed to \code{\link[stats]{optim}}.
#' @param scale Norm scale (default \code{"T"}).
#' @param parallel Logical; attempt parallel execution. Default \code{TRUE}.
#' @param n_cores Number of cores. Defaults to all logical cores.
#' @param plot Logical; plot the selected model. Default \code{TRUE}.
#' @param verbose Logical; print progress. Default \code{TRUE}.
#'
#' @return The selected fitted \code{cnormShash} model with an additional
#'   element \code{$selection} containing:
#'   \itemize{
#'     \item \code{evaluated}: data frame of every combination tried, sorted by BIC.
#'     \item \code{selected}: list with the chosen degrees and BIC.
#'   }
#'
#' @examples
#' \dontrun{
#' m <- autoselect.shash(elfe$group, elfe$raw)
#' m$selection$evaluated
#' summary(m, age = elfe$group, score = elfe$raw)
#' }
#'
#' @seealso \code{\link{cnorm.shash}}, \code{\link{autoselect.betabinomial}}
#' @export
autoselect.shash <- function(age,
                             score,
                             weights      = NULL,
                             max_mu       = 4,
                             max_sigma    = 3,
                             max_epsilon  = 3,
                             max_delta    = 1,
                             min_mu       = 1,
                             min_sigma    = 1,
                             min_epsilon  = 1,
                             min_delta    = 0,
                             delta        = 1,
                             control      = NULL,
                             scale        = "T",
                             parallel     = TRUE,
                             n_cores      = NULL,
                             plot         = TRUE,
                             verbose      = TRUE) {

  # ---- Input validation -------------------------------------------------
  if (length(age) != length(score))
    stop("Length of 'age' and 'score' must be the same.")
  if (!is.null(weights) && length(weights) != length(age))
    stop("Length of 'weights' must match length of 'age' and 'score'.")
  if (max_mu      < min_mu)      stop("'max_mu' must be >= 'min_mu'.")
  if (max_sigma   < min_sigma)   stop("'max_sigma' must be >= 'min_sigma'.")
  if (max_epsilon < min_epsilon) stop("'max_epsilon' must be >= 'min_epsilon'.")
  if (max_delta   < min_delta)   stop("'max_delta' must be >= 'min_delta'.")
  if (min_mu < 1 || min_sigma < 1 || min_epsilon < 1)
    stop("Minimum degrees for mu, sigma, epsilon must be >= 1.")
  if (min_delta < 0)
    stop("'min_delta' must be >= 0 (0 means fixed delta).")
  if (delta <= 0) stop("'delta' must be > 0.")

  say <- function(...) {
    if (verbose) { cat(..., sep = ""); utils::flush.console() }
  }

  # ---- Build the 4-D grid ----------------------------------------------
  grid <- expand.grid(mu      = min_mu:max_mu,
                      sigma   = min_sigma:max_sigma,
                      epsilon = min_epsilon:max_epsilon,
                      delta   = min_delta:max_delta,
                      KEEP.OUT.ATTRS = FALSE)
  pairs <- lapply(seq_len(nrow(grid)),
                  function(i) c(grid$mu[i], grid$sigma[i],
                                grid$epsilon[i], grid$delta[i]))
  say(sprintf(
    "Search grid: %d candidate models (mu %d:%d, sigma %d:%d, epsilon %d:%d, delta %d:%d).\n",
    length(pairs),
    min_mu, max_mu, min_sigma, max_sigma,
    min_epsilon, max_epsilon, min_delta, max_delta
  ))

  # ---- Parallel setup with dev-mode fallback ---------------------------
  use_parallel <- FALSE
  cl <- NULL
  if (isTRUE(parallel)) {
    avail <- tryCatch(parallel::detectCores(logical = TRUE),
                      error = function(e) 1L)
    if (is.null(n_cores)) n_cores <- avail
    n_cores <- max(1L, min(n_cores, length(pairs), avail))

    if (n_cores > 1L && length(pairs) > 1L) {
      cl <- tryCatch(parallel::makeCluster(n_cores),
                     error = function(e) NULL)
      if (!is.null(cl)) {
        worker_ok <- tryCatch({
          res <- parallel::clusterCall(cl, function() {
            requireNamespace("cNORM", quietly = TRUE) &&
              exists("cnorm.shash",      envir = asNamespace("cNORM")) &&
              exists("diagnostics.shash", envir = asNamespace("cNORM"))
          })
          all(vapply(res, isTRUE, logical(1)))
        }, error = function(e) FALSE)

        if (worker_ok) {
          use_parallel <- TRUE
          on.exit(try(parallel::stopCluster(cl), silent = TRUE), add = TRUE)
          parallel::clusterExport(
            cl,
            varlist = c("age", "score", "weights",
                        "delta", "control", "scale"),
            envir = environment()
          )
          say(sprintf("Parallel mode: using %d cores.\n", n_cores))
        } else {
          try(parallel::stopCluster(cl), silent = TRUE); cl <- NULL
          say("Note: cNORM is not installed in the worker library path ",
              "(typical during devtools::load_all). ",
              "Falling back to sequential execution.\n")
        }
      }
    }
  }

  # ---- Worker --------------------------------------------------------
  fit_worker <- function(p) {
    # p = c(mu_deg, sigma_deg, epsilon_deg, delta_deg)
    # delta_deg == 0  -> pass delta_degree = NULL (fixed delta)
    dd <- if (p[4] == 0L) NULL else p[4]
    tryCatch({
      m <- suppressMessages(suppressWarnings(
        cNORM::cnorm.shash(
          age = age, score = score, weights = weights,
          mu_degree      = p[1],
          sigma_degree   = p[2],
          epsilon_degree = p[3],
          delta_degree   = dd,
          delta          = delta,
          control        = control,
          scale          = scale,
          plot           = FALSE
        )
      ))
      d <- cNORM::diagnostics.shash(m)
      list(mu = p[1], sigma = p[2], epsilon = p[3], delta = p[4],
           model = m,
           BIC = if (is.finite(d$BIC)) d$BIC else Inf,
           AIC = d$AIC, logLik = d$log_likelihood,
           converged = isTRUE(d$converged),
           status  = if (!is.finite(d$BIC)) "error"
           else if (!isTRUE(d$converged)) "not_converged"
           else "ok",
           message = NA_character_)
    }, error = function(e) {
      list(mu = p[1], sigma = p[2], epsilon = p[3], delta = p[4],
           model = NULL, BIC = Inf, AIC = NA_real_, logLik = NA_real_,
           converged = FALSE, status = "error",
           message = conditionMessage(e))
    })
  }

  # ---- Cache + reporting -----------------------------------------------
  cache <- new.env(hash = TRUE, parent = emptyenv())
  ck <- function(p) sprintf("%d_%d_%d_%d", p[1], p[2], p[3], p[4])

  fmt_delta <- function(d) if (d == 0L) "fixed" else as.character(d)

  report <- function(r) {
    tag <- switch(r$status,
                  ok            = "",
                  not_converged = "  (not strictly converged)",
                  error         = paste0("  [error: ", r$message, "]"))
    say(sprintf("  mu=%d, sigma=%d, epsilon=%d, delta=%s : BIC = %s%s\n",
                r$mu, r$sigma, r$epsilon, fmt_delta(r$delta),
                formatC(r$BIC, digits = 3, format = "f"), tag))
  }

  evaluate_pairs <- function(pairs) {
    todo <- pairs[!vapply(pairs,
                          function(p) exists(ck(p), envir = cache),
                          logical(1))]
    if (length(todo) == 0L) return(invisible(NULL))

    say(sprintf("Evaluating %d model%s ...\n",
                length(todo), if (length(todo) == 1) "" else "s"))

    if (use_parallel && length(todo) > 1L) {
      chunks <- split(todo, ceiling(seq_along(todo) / n_cores))
      for (chunk in chunks) {
        results <- parallel::parLapply(cl, chunk, fit_worker)
        for (r in results) {
          assign(ck(c(r$mu, r$sigma, r$epsilon, r$delta)), r, envir = cache)
          report(r)
        }
      }
    } else {
      for (p in todo) {
        r <- fit_worker(p)
        assign(ck(c(r$mu, r$sigma, r$epsilon, r$delta)), r, envir = cache)
        report(r)
      }
    }
    invisible(NULL)
  }

  # ---- Run the search --------------------------------------------------
  evaluate_pairs(pairs)
  all_res <- mget(ls(envir = cache), envir = cache)
  bics    <- vapply(all_res, `[[`, numeric(1), "BIC")
  if (all(!is.finite(bics)))
    stop("Selection failed: no model produced a finite BIC. ",
         "Inspect $selection$evaluated for per-fit messages.")
  current <- all_res[[which.min(bics)]]

  # ---- Compile results --------------------------------------------------
  evaluated <- do.call(rbind, lapply(
    all_res,
    function(r) data.frame(mu = r$mu, sigma = r$sigma,
                           epsilon = r$epsilon, delta = r$delta,
                           BIC = r$BIC, AIC = r$AIC,
                           logLik = r$logLik,
                           converged = r$converged,
                           status = r$status,
                           message = r$message,
                           stringsAsFactors = FALSE)))
  evaluated <- evaluated[order(evaluated$BIC), , drop = FALSE]
  rownames(evaluated) <- NULL

  say(sprintf("\nSelected model: mu=%d, sigma=%d, epsilon=%d, delta=%s (BIC = %.3f)\n",
              current$mu, current$sigma, current$epsilon,
              fmt_delta(current$delta), current$BIC))

  final_model <- current$model
  if (is.null(final_model))
    stop("Selection failed: the best candidate did not produce a usable model.")

  final_model$selection <- list(
    evaluated = evaluated,
    selected  = list(mu_degree      = current$mu,
                     sigma_degree   = current$sigma,
                     epsilon_degree = current$epsilon,
                     delta_degree   = if (current$delta == 0L) NULL else current$delta,
                     delta          = delta,
                     BIC            = current$BIC)
  )

  if (plot) print(plot(final_model, age = age, score = score, weights = weights))
  return(final_model)
}

#' Calculate Norm Tables for Sinh-Arcsinh Distribution
#'
#' Generates norm tables for specific ages based on a fitted Sinh-Arcsinh (shash) regression model.
#' Computes interval probabilities, percentiles, z-scores, and norm scores for a specified range of raw scores.
#' Optionally includes confidence intervals when reliability is provided.
#'
#' @param model Fitted shash model object of class "cnormShash"
#' @param ages Numeric vector of age points for norm table generation
#' @param start Minimum raw score value for the norm table
#' @param end Maximum raw score value for the norm table
#' @param step Step size between consecutive raw scores (default: 1)
#' @param CI Confidence coefficient (0-1, default: 0.9) for confidence intervals
#' @param reliability Reliability coefficient (0-1) for true score confidence intervals
#'
#' @return List of data frames (one per age) containing:
#'   \item{x}{Raw scores}
#'   \item{Px}{Approximate interval probabilities (density times step width).
#'     These are consistent with \code{Pcum} but do not necessarily sum to 1
#'     over the tabulated score range.}
#'   \item{Pcum}{Cumulative probabilities}
#'   \item{Percentile}{Percentile ranks (0-100)}
#'   \item{z}{Standardized z-scores}
#'   \item{norm}{Norm scores in specified scale}
#'   \item{lowerCI, upperCI}{Confidence intervals (if reliability provided)}
#'   \item{lowerCI_PR, upperCI_PR}{CI as percentile ranks (if reliability provided)}
#'
#' @details
#' For continuous shash distributions, densities are converted to interval probabilities
#' (density times step width) and cumulative probabilities. When reliability is specified,
#' confidence intervals include correction for regression to the mean.
#'
#' @examples
#' \dontrun{
#' # Basic norm table
#' model <- cnorm.shash(age, score)
#' tables <- normTable.shash(model, ages = c(7, 8, 9), start = 0, end = 50)
#'
#' # With confidence intervals and finer granularity
#' tables_ci <- normTable.shash(model, ages = c(8, 9), start = 10, end = 40,
#'                              step = 0.5, CI = 0.95, reliability = 0.85)
#' }
#'
#' @export
normTable.shash <- function(model,
                            ages,
                            start = NULL,
                            end = NULL,
                            step = 1,
                            CI = .9,
                            reliability = NULL) {
  # Input validation
  if (!isSHASH(model)) {
    stop("Wrong object. Please provide object from class 'cnormShash'.")
  }

  if (is.null(start)) {
    start <- attr(model$result, "min")
  }

  if (is.null(end)) {
    end <- attr(model$result, "max")
  }

  if (is.null(step)) {
    step <- 1
  }

  if (start >= end) {
    stop("Start value must be less than end value.")
  }

  if (step <= 0) {
    stop("Step size must be positive.")
  }

  if (is.null(CI) || is.na(CI)) {
    reliability <- NULL
  } else if (CI > .99999 || CI < .00001) {
    stop("Confidence coefficient (CI) out of range. Please specify value between 0 and 1.")
  }

  # Setup reliability and confidence intervals
  rel <- FALSE
  if (!is.null(reliability)) {
    if (reliability > .9999 || reliability < .0001) {
      stop("Reliability coefficient out of range. Please specify value between 0 and 1.")
    } else {
      se <- qnorm(1 - ((1 - CI) / 2)) * sqrt(reliability * (1 - reliability))
      rel <- TRUE
    }
  }

  # Get predictions for all ages
  predictions <- predictCoefficients_shash(model, ages)

  # Create score sequence
  x <- seq(from = start, to = end, by = step)

  # Get scale information
  mScale <- attr(model$result, "scaleMean")
  sdScale <- attr(model$result, "scaleSD")

  result <- list()

  # Generate norm table for each age
  for (k in seq_along(ages)) {
    mu_k <- predictions$mu[k]
    sigma_k <- predictions$sigma[k]
    epsilon_k <- predictions$epsilon[k]
    delta_k <- predictions$delta[k]

    Px <- dshash(
      x,
      mu = mu_k,
      sigma = sigma_k,
      epsilon = epsilon_k,
      delta = delta_k
    ) * step

    # Calculate cumulative probabilities using the continuous CDF
    cum <- pshash(
      x,
      mu = mu_k,
      sigma = sigma_k,
      epsilon = epsilon_k,
      delta = delta_k
    )

    # Calculate percentiles
    perc <- cum

    # Clamp extreme cumulative probabilities instead of mapping
    # non-finite z-scores to 0 (which assigned the MEAN norm score to
    # floor/ceiling raw scores).
    z <- qnorm(pmin(pmax(cum, 1e-12), 1 - 1e-12))

    # Calculate norm scores
    norm <- rep(NA, length(z))
    if (!is.na(mScale) && !is.na(sdScale)) {
      norm <- mScale + sdScale * z
    }

    # Create data frame
    df <- data.frame(
      x = x,
      Px = Px,
      Pcum = cum,
      Percentile = perc * 100,
      z = z,
      norm = norm
    )

    # Add confidence intervals if reliability is specified
    if (rel) {
      # True score estimates (regression to the mean correction)
      zPredicted <- reliability * z

      # Confidence intervals
      df$lowerCI <- (zPredicted - se) * sdScale + mScale
      df$upperCI <- (zPredicted + se) * sdScale + mScale
      df$lowerCI_PR <- pnorm(zPredicted - se) * 100
      df$upperCI_PR <- pnorm(zPredicted + se) * 100

      # Ensure percentile ranks are within bounds
      df$lowerCI_PR <- pmax(0, pmin(100, df$lowerCI_PR))
      df$upperCI_PR <- pmax(0, pmin(100, df$upperCI_PR))
    }

    result[[k]] <- df
  }

  # Name the list elements with ages
  names(result) <- as.character(ages)

  return(result)
}

#' Summarize a Sinh-Arcsinh Continuous Norming Model
#'
#' This function provides a summary of a fitted Sinh-Arcsinh (shash) continuous norming model,
#' including model fit statistics, convergence information, and parameter estimates.
#'
#' @param object An object of class "cnormShash", typically the result of a call to
#'   \code{\link{cnorm.shash}}.
#' @param ... Additional arguments passed to the summary method:
#'   \itemize{
#'      \item age An optional numeric vector of age values corresponding to the raw scores.
#'        If provided along with \code{score}, additional fit statistics (R-squared, RMSE, bias)
#'        will be calculated.
#'      \item score An optional numeric vector of raw scores. Must be provided if \code{age} is given.
#'      \item weights An optional numeric vector of weights for each observation.
#'    }
#'
#' @return Invisibly returns a list containing detailed diagnostic information about the model.
#'   The function primarily produces printed output summarizing the model.
#'
#' @details
#' The summary includes:
#' \itemize{
#'   \item Basic model information (polynomial degrees, delta parameter, number of observations)
#'   \item Model fit statistics (log-likelihood, AIC, BIC)
#'   \item R-squared, RMSE, and bias (if age and raw scores are provided)
#'   \item Convergence information
#'   \item Parameter estimates with standard errors, z-values, and p-values
#' }
#'
#' @examples
#' \dontrun{
#' model <- cnorm.shash(children$age, children$score)
#' summary(model)
#'
#' # Including R-squared, RMSE, and bias in the summary:
#' summary(model, age = children$age, score = children$score)
#' }
#'
#' @seealso \code{\link{cnorm.shash}}, \code{\link{diagnostics.shash}}
#' @export
summary.cnormShash <- function(object, ...) {
  args <- list(...)

  if ("age" %in% names(args)) {
    age <- args$age
  } else {
    if (length(args) > 0)
      age <- args[[1]]
    else
      age <- NULL
  }
  if ("score" %in% names(args)) {
    score <- args$score
  } else {
    if (length(args) > 1)
      score <- args[[2]]
    else
      score <- NULL
  }
  if ("weights" %in% names(args)) {
    weights <- args$weights
  } else {
    if (length(args) > 2)
      weights <- args[[3]]
    else
      weights <- NULL
  }

  diag <- diagnostics.shash(object, age, score, weights)

  cat("SinH-ArcSinH Continuous Norming Model\n")
  cat("-------------------------------------\n")
  cat("Polynomial degrees:\n")
  cat("  Location (mu):", diag$mu_degree, "\n")
  cat("  Scale (sigma):", diag$sigma_degree, "\n")
  cat("  Skewness (epsilon):", diag$epsilon_degree, "\n")
  if (!is.null(diag$delta_degree)) {
    cat("  Tail weight (delta):", diag$delta_degree, "\n")
  } else {
    cat("  Tail weight (delta): Fixed at", diag$delta, "\n")
  }
  cat("Number of observations:", diag$n_obs, "\n")
  cat("Number of parameters:", diag$n_params, "\n")
  cat("\n")

  cat("Model Fit:\n")
  cat("  Log-likelihood:", round(diag$log_likelihood, 2), "\n")
  cat("  AIC:", round(diag$AIC, 2), "\n")
  cat("  BIC:", round(diag$BIC, 2), "\n")
  if (!is.na(diag$R2)) {
    cat("  R-squared:", round(diag$R2, 4), "\n")
    cat("  RMSE:", round(diag$rmse, 4), "\n")
    cat("  Bias:", round(diag$bias, 4), "\n")
  }
  cat("\n")

  cat("Convergence:\n")
  cat("  Converged:", diag$converged, "\n")
  cat("  Function evaluations:", diag$n_evaluations, "\n")
  cat("  Max Hessian eigenvalue:", round(diag$max_gradient, 6), "\n")
  cat("  Message:", diag$message, "\n")
  cat("\n")

  cat("Parameter Estimates:\n")
  cat("Location (mu) parameters:\n")
  mu_table <- data.frame(
    Estimate = diag$mu_estimates,
    `Std. Error` = diag$mu_se,
    `z value` = diag$mu_z_values,
    `Pr(>|z|)` = diag$mu_p_values
  )
  rownames(mu_table) <- paste0("mu_", 0:(length(diag$mu_estimates) - 1))
  print(mu_table, digits = 4)
  cat("\n")

  cat("Scale (sigma) parameters:\n")
  sigma_table <- data.frame(
    Estimate = diag$sigma_estimates,
    `Std. Error` = diag$sigma_se,
    `z value` = diag$sigma_z_values,
    `Pr(>|z|)` = diag$sigma_p_values
  )
  rownames(sigma_table) <- paste0("log(sigma)_", 0:(length(diag$sigma_estimates) -
                                                      1))
  print(sigma_table, digits = 4)
  cat("\n")

  cat("Skewness (epsilon) parameters:\n")
  epsilon_table <- data.frame(
    Estimate = diag$epsilon_estimates,
    `Std. Error` = diag$epsilon_se,
    `z value` = diag$epsilon_z_values,
    `Pr(>|z|)` = diag$epsilon_p_values
  )
  rownames(epsilon_table) <- paste0("epsilon_", 0:(length(diag$epsilon_estimates) -
                                                     1))
  print(epsilon_table, digits = 4)

  if (!is.null(diag$delta_estimates)) {
    cat("\n")
    cat("Tail weight (delta) parameters:\n")
    delta_table <- data.frame(
      Estimate = diag$delta_estimates,
      `Std. Error` = diag$delta_se,
      `z value` = diag$delta_z_values,
      `Pr(>|z|)` = diag$delta_p_values
    )
    rownames(delta_table) <- paste0("log(delta)_", 0:(length(diag$delta_estimates) -
                                                        1))
    print(delta_table, digits = 4)
  }

  invisible(diag)
}

#' Diagnostic Statistics for Sinh-Arcsinh Continuous Norming Model
#'
#' This function computes detailed diagnostic statistics for a fitted shash model,
#' including fit statistics, parameter estimates, and convergence information.
#'
#' @param object An object of class "cnormShash"
#' @param age An optional numeric vector of age values for computing fit statistics
#' @param score An optional numeric vector of raw scores for computing fit statistics
#' @param weights An optional numeric vector of observation weights
#'
#' @return A list containing comprehensive diagnostic information
#'
#' @keywords internal
diagnostics.shash <- function(object,
                              age = NULL,
                              score = NULL,
                              weights = NULL) {
  if (!isSHASH(object)) {
    stop("Wrong object. Please provide object from class 'cnormShash'.")
  }

  # Basic model information
  n_obs <- attr(object$result, "N")
  n_params <- length(object$result$par)
  log_likelihood <- -object$result$value
  AIC <- 2 * n_params - 2 * log_likelihood
  BIC <- n_params * log(n_obs) - 2 * log_likelihood

  # Parameter estimates and standard errors
  mu_estimates <- object$mu_est
  sigma_estimates <- object$sigma_est
  epsilon_estimates <- object$epsilon_est
  delta_estimates <- object$delta_est

  se <- object$se
  if (is.null(se) || any(is.na(se))) {
    mu_se <- rep(NA, length(mu_estimates))
    sigma_se <- rep(NA, length(sigma_estimates))
    epsilon_se <- rep(NA, length(epsilon_estimates))
    delta_se <- if (!is.null(delta_estimates))
      rep(NA, length(delta_estimates))
    else
      NULL
  } else {
    n_mu <- length(mu_estimates)
    n_sigma <- length(sigma_estimates)
    n_epsilon <- length(epsilon_estimates)

    mu_se <- se[1:n_mu]
    sigma_se <- se[(n_mu + 1):(n_mu + n_sigma)]
    epsilon_se <- se[(n_mu + n_sigma + 1):(n_mu + n_sigma + n_epsilon)]

    if (!is.null(delta_estimates)) {
      n_delta <- length(delta_estimates)
      delta_se <- se[(n_mu + n_sigma + n_epsilon + 1):(n_mu + n_sigma + n_epsilon + n_delta)]
    } else {
      delta_se <- NULL
    }
  }

  # Calculate z-values and p-values
  mu_z_values <- ifelse(is.na(mu_se) |
                          mu_se == 0, NA, mu_estimates / mu_se)
  sigma_z_values <- ifelse(is.na(sigma_se) |
                             sigma_se == 0, NA, sigma_estimates / sigma_se)
  epsilon_z_values <- ifelse(is.na(epsilon_se) |
                               epsilon_se == 0,
                             NA,
                             epsilon_estimates / epsilon_se)

  mu_p_values <- ifelse(is.na(mu_z_values), NA, 2 * (1 - pnorm(abs(mu_z_values))))
  sigma_p_values <- ifelse(is.na(sigma_z_values), NA, 2 * (1 - pnorm(abs(sigma_z_values))))
  epsilon_p_values <- ifelse(is.na(epsilon_z_values), NA, 2 * (1 - pnorm(abs(epsilon_z_values))))

  if (!is.null(delta_estimates)) {
    delta_z_values <- ifelse(is.na(delta_se) |
                               delta_se == 0,
                             NA,
                             delta_estimates / delta_se)
    delta_p_values <- ifelse(is.na(delta_z_values), NA, 2 * (1 - pnorm(abs(delta_z_values))))
  } else {
    delta_z_values <- NULL
    delta_p_values <- NULL
  }

  # Convergence information
  converged <- object$result$convergence == 0
  n_evaluations <- object$result$counts["function"]
  message <- switch(
    as.character(object$result$convergence),
    "0" = "Successful convergence",
    "1" = "Maximum iterations reached",
    "10" = "Degeneracy in Nelder-Mead simplex",
    "51" = "Warning from L-BFGS-B",
    "52" = "Error from L-BFGS-B",
    paste("Convergence code:", object$result$convergence)
  )

  # Largest Hessian eigenvalue (curvature diagnostic), if available
  max_gradient <- NA
  if (!is.null(object$result$hessian)) {
    tryCatch({
      eigenvals <- eigen(object$result$hessian, only.values = TRUE)$values
      max_gradient <- max(abs(eigenvals), na.rm = TRUE)
    }, error = function(e) {
      max_gradient <- NA
    })
  }

  # Calculate R-squared if age and score data are provided
  R2 <- NA
  rmse <- NA
  bias <- NA
  if (!is.null(age) && !is.null(score)) {
    if (length(age) / length(unique(age)) > 50 &&
        min(table(age)) > 30) {
      data <- data.frame(group = age, raw = score)
      data <- rankByGroup(
        data = data,
        raw = "raw",
        group = "group",
        weights = weights,
        scale = c(
          attr(object$result, "scaleMean"),
          attr(object$result, "scaleSD")
        )
      )
      norm_scores <- predict(object, data$group, data$raw)
    } else{
      data <- data.frame(age = age, raw = score)
      data$groups <- getGroups(age)
      width <- (max(age) - min(age)) / length(unique(data$groups))
      data <- rankBySlidingWindow(
        data,
        age = "age",
        raw = "raw",
        width = width,
        weights = weights,
        scale = c(
          attr(object$result, "scaleMean"),
          attr(object$result, "scaleSD")
        )
      )
      norm_scores <- predict(object, data$age, data$raw)
    }

    norm_manifest <- data$normValue
    R2 <- cor(norm_scores, norm_manifest, use = "pairwise.complete.obs")^2
    rmse <- sqrt(mean((norm_scores - norm_manifest)^2))
    bias <- mean(norm_scores - norm_manifest)
  } else {
    message <- "No age and raw scores provided. Cannot calculate R2, RMSE, and bias."
  }

  # Return comprehensive diagnostic information
  list(
    # Basic model info
    mu_degree = object$mu_degree,
    sigma_degree = object$sigma_degree,
    epsilon_degree = object$epsilon_degree,
    delta = object$delta,
    delta_degree = object$delta_degree,
    n_obs = n_obs,
    n_params = n_params,

    # Fit statistics
    log_likelihood = log_likelihood,
    AIC = AIC,
    BIC = BIC,
    R2 = R2,
    rmse = rmse,
    bias = bias,

    # Parameter estimates
    mu_estimates = mu_estimates,
    sigma_estimates = sigma_estimates,
    epsilon_estimates = epsilon_estimates,
    delta_estimates = delta_estimates,

    # Standard errors
    mu_se = mu_se,
    sigma_se = sigma_se,
    epsilon_se = epsilon_se,
    delta_se = delta_se,

    # Test statistics
    mu_z_values = mu_z_values,
    sigma_z_values = sigma_z_values,
    epsilon_z_values = epsilon_z_values,
    delta_z_values = delta_z_values,

    mu_p_values = mu_p_values,
    sigma_p_values = sigma_p_values,
    epsilon_p_values = epsilon_p_values,
    delta_p_values = delta_p_values,

    # Convergence info
    converged = converged,
    n_evaluations = n_evaluations,
    max_gradient = max_gradient,
    message = message
  )
}

#' Predict Norm Scores from Raw Scores
#'
#' This function calculates norm scores based on raw scores, age, and a fitted cnormShash model.
#'
#' @param object A fitted model object of class 'cnormShash'.
#' @param ... Additional arguments passed to the prediction method:
#'   \itemize{
#'      \item age A numeric vector of ages, same length as score.
#'      \item score A numeric vector of raw scores.
#'      \item range The range of the norm scores in standard deviations. Default is 3.
#'        Thus, scores in the range of +/- 3 standard deviations are considered.
#'    }
#'
#' @return A numeric vector of norm scores.
#'
#' @details
#' The function predicts the Sinh-Arcsinh (shash) distribution parameters (mu, sigma, epsilon, delta)
#' for each age using the provided model. It then calculates the cumulative probability for each
#' raw score given these parameters using the continuous shash distribution. Finally,
#' it converts these probabilities to the norm scale specified in the model. Raw scores
#' at the extreme floor or ceiling of the distribution are mapped to the boundary of the
#' norm score range (mean +/- \code{range} standard deviations), not to the mean.
#'
#' @examples
#' \dontrun{
#' # Assuming you have a fitted model named 'shash_model':
#' model <- cnorm.shash(children$age, children$score)
#' raw_scores <- c(25.5, 30.2, 18.7, 45.3)
#' ages <- c(7, 8, 9, 10)
#' norm_scores <- predict(model, ages, raw_scores)
#' }
#'
#' @export
#' @family predict
predict.cnormShash <- function(object, ...) {
  model <- object
  args <- list(...)

  if ("age" %in% names(args)) {
    age <- args$age
  } else {
    if (length(args) > 0)
      age <- args[[1]]
    else
      age <- NULL
  }
  if ("score" %in% names(args)) {
    score <- args$score
  } else {
    if (length(args) > 1)
      score <- args[[2]]
    else
      score <- NULL
  }
  if ("range" %in% names(args)) {
    range <- args$range
  } else {
    if (length(args) > 2)
      range <- args[[3]]
    else
      range <- 3
  }

  if (!isSHASH(model)) {
    stop("Wrong object. Please provide object from class 'cnormShash'.")
  }

  if (is.null(age) || is.null(score)) {
    stop("Both 'age' and 'score' must be provided.")
  }

  if (length(age) != length(score)) {
    stop("The lengths of 'age' and 'score' must be the same.")
  }

  # Get predicted distribution parameters for each age
  predictions <- predictCoefficients_shash(model, age)

  mu <- predictions$mu
  sigma <- predictions$sigma
  epsilon <- predictions$epsilon
  delta <- predictions$delta

  # Calculate cumulative probabilities (percentiles) for each score
  percentiles <- pshash(
    score,
    mu = mu,
    sigma = sigma,
    epsilon = epsilon,
    delta = delta
  )

  percentiles_c <- pmin(pmax(percentiles, 1e-12), 1 - 1e-12)
  z_scores <- qnorm(percentiles_c)

  # Apply range constraints
  z_scores <- pmin(pmax(z_scores, -range), range)

  # Get scale information
  mScale <- attr(model$result, "scaleMean")
  sdScale <- attr(model$result, "scaleSD")

  if (!is.na(mScale) && !is.na(sdScale)) {
    # Scale z-scores to specified norm scale
    norm_scores <- mScale + sdScale * z_scores
    return(norm_scores)
  } else {
    # Return percentile ranks if no scale specified
    return(percentiles * 100)
  }
}

library(shiny)
library(shinycssloaders)
library(cNORM)
library(DT)
library(markdown)
library(foreign)
library(readxl)

shinyUI(fluidPage(
  title = "cNORM - Shiny",

  tabsetPanel(

    # ---------------------------------------------------------------- DATA ----
    tabPanel("Data",
             sidebarLayout(
               sidebarPanel(
                 img(src = "logo.png", align = "right", style = "max-width:100px;"),
                 h3("Data input"),
                 selectInput("Example", "Example dataset:",
                             choices = c("(none)" = "", "elfe", "ppvt", "CDC"),
                             selected = ""),
                 fileInput("file", "Or upload your own file:",
                           accept = c(".csv", ".xlsx", ".xls", ".rda", ".sav")),
                 helpText("Accepted: CSV (semicolon), Excel, SPSS, R data.",
                          "Uploaded files take precedence over the example.")
               ),
               mainPanel(
                 uiOutput("introduction"),
                 DT::DTOutput("table")
               )
             )
    ),

    # ------------------------------------------------------------ MODELING ----
    tabPanel("Modeling",
             sidebarLayout(
               sidebarPanel(
                 h3("Model settings"),
                 uiOutput("GroupingVariable"),
                 uiOutput("RawValues"),
                 selectInput("Scale", "Norm scale:",
                             choices = c("T (M=50, SD=10)"   = "T",
                                         "IQ (M=100, SD=15)" = "IQ",
                                         "z (M=0, SD=1)"     = "z"),
                             selected = "T"),
                 selectInput("NumberOfPowers",    "Power degree of location (k):",
                             choices = 1:6, selected = 5),
                 selectInput("NumberOfPowersAge", "Power degree of explanatory variable (t), e. g. age:",
                             choices = 1:6, selected = 3),
                 numericInput("InputNumberOfTerms",
                              "Number of terms (leave blank for auto selection):",
                              value = NA, min = 1, step = 1),
                 actionButton("RunModel", "Run model", class = "btn-primary"),
                 downloadButton("downloadModel", "Download model"),
                 hr(),
                 helpText(
                   "k controls the polynomial complexity for the location term, ",
                   "t for the explanatory variable (typically age). ",
                   "Leaving the number of terms blank autoselects the highest consistent model."
                 )
               ),
               mainPanel(
                 h4("Model summary"),
                 withSpinner(uiOutput("ModelReportUI"), type = 5),
                 verbatimTextOutput("ConsistencyCheck"),
                 verbatimTextOutput("RangeCheck"),
                 h4("Percentile plot"),
                 withSpinner(plotOutput("modelPlot", height = "500px"), type = 5),
                 h4("Model fit by number of terms"),
                 selectInput("chosenTypePlotSubset", NULL,
                             choices = c("R2 by Number of Predictors",
                                         "Mallow's (log) Cp by R2",
                                         "Bayesian Information Criterion (BIC) by R2",
                                         "RMSE by Number of Predictors"),
                             selected = "RMSE by Number of Predictors"),
                 withSpinner(plotOutput("PlotWL", height = "500px"), type = 5),
                 DT::DTOutput("PrintSubset")
               )
             )
    ),

    # ------------------------------------------------------- VISUALIZATION ----
    navbarMenu("Visualization",

               tabPanel("Percentiles", sidebarLayout(
                 sidebarPanel(
                   h3("Percentile curves"),
                   helpText("Dots: observed percentiles. Lines: model. ",
                            "Intersecting curves suggest an inconsistent model — try adjusting ",
                            "the number of terms in the Modeling tab."),
                   textInput("PercentilesForPercentiles",
                             "Custom percentiles (comma/space separated):", value = "2.5, 10, 25, 50, 75, 90, 97.5")
                 ),
                 mainPanel(plotOutput("PlotPercentiles", height = "700px"))
               )),

               tabPanel("Series", sidebarLayout(
                 sidebarPanel(
                   h3("Percentile series"),
                   helpText("Choose the number of terms with the slider to compare models. ",
                            "Avoid intersecting or wavy curves."),
                   sliderInput("terms", "Number of terms:",
                               min = 1, max = 24, value = 5)
                 ),
                 mainPanel(plotOutput("Series", height = "700px"))
               )),

               tabPanel("Norm curves", sidebarLayout(
                 sidebarPanel(
                   h3("Norm curves"),
                   textInput("PercentilesForNormCurves",
                             "Percentiles (comma/space separated):", value = "2.5, 10, 25, 50, 75, 90, 97.5"),
                   helpText("z = -2, -1, 0, 1, 2 corresponds to ",
                            "2.276, 15.87, 50.00, 84.13, 97.724")
                 ),
                 mainPanel(plotOutput("NormCurves", height = "700px"))
               )),

               tabPanel("Density", sidebarLayout(
                 sidebarPanel(
                   h3("Density"),
                   helpText("Probability density of raw scores by group. Helps detect ",
                            "non-normality or model violations."),
                   textInput("densities", "Groups (comma/space separated):", "")
                 ),
                 mainPanel(withSpinner(plotOutput("PlotDensity", height = "700px"), type = 5))
               )),

               tabPanel("Derivative", sidebarLayout(
                 sidebarPanel(
                   h3("First derivative"),
                   helpText("Negative regions indicate violations of the bijective ",
                            "mapping between raw and norm scores.")
                 ),
                 mainPanel(withSpinner(plotOutput("PlotDerivatives", height = "600px"), type = 5))
               )),

               tabPanel("Norm scores", sidebarLayout(
                 sidebarPanel(
                   h3("Predicted vs. observed norm scores"),
                   checkboxInput("grouping",   "Plot by group", FALSE),
                   checkboxInput("differences","Plot differences", FALSE),
                   actionButton("normScoreButton", "Update")
                 ),
                 mainPanel(withSpinner(plotOutput("PlotNormScores", height = "700px"), type = 5))
               )),

               tabPanel("Raw scores", sidebarLayout(
                 sidebarPanel(
                   h3("Predicted vs. observed raw scores"),
                   checkboxInput("grouping1",   "Plot by group", FALSE),
                   checkboxInput("differences1","Plot differences", FALSE),
                   actionButton("rawScoreButton", "Update")
                 ),
                 mainPanel(withSpinner(plotOutput("PlotRawScores", height = "700px"), type = 5))
               ))
    ),

    # ---------------------------------------------------- CROSS-VALIDATION ----
    tabPanel("Cross-validation",
             sidebarLayout(
               sidebarPanel(
                 h3("Cross-validation"),
                 helpText("Stratified 80/20 splits, repeated. Compares RMSE and norm-score ",
                          "reliability across model sizes. Computationally heavy."),
                 sliderInput("MaxTermsCV",    "Max number of terms:", min = 1, max = 24, value = 10),
                 sliderInput("RepetitionsCV", "Repetitions:",         min = 1, max = 20, value = 1),
                 checkboxInput("NormsCV", "Evaluate norm scores", TRUE),
                 actionButton("CrossValidation", "Run cross-validation", class = "btn-primary")
               ),
               mainPanel(
                 withSpinner(plotOutput("PlotCV", height = "800px"), type = 5),
                 DT::DTOutput("TableCV")
               )
             )
    ),

    # ---------------------------------------------------------- PREDICTION ---
    navbarMenu("Prediction",
               tabPanel("Norm score",  sidebarLayout(
                 sidebarPanel(uiOutput("InputNormValue")),
                 mainPanel(verbatimTextOutput("NormValue")))),
               tabPanel("Raw score",   sidebarLayout(
                 sidebarPanel(uiOutput("InputRawValue")),
                 mainPanel(verbatimTextOutput("RawValue")))),
               tabPanel("Norm table",  sidebarLayout(
                 sidebarPanel(uiOutput("InputNormTable")),
                 mainPanel(DT::DTOutput("NormTable")))),
               tabPanel("Raw table",   sidebarLayout(
                 sidebarPanel(uiOutput("InputRawTable")),
                 mainPanel(DT::DTOutput("RawTable"))))
    )
  )
))

## ----setup, include = FALSE-------------------------------
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
options(width = 60)

library(cNORM)
library(ggplot2)

## ----message = FALSE--------------------------------------
library(cNORM)
# Assign data to object norm.data
norm.data <- ppvt
head(norm.data)

## ---------------------------------------------------------
# Generate population marginals
marginals <- data.frame(var = c("sex", "sex", "migration", "migration"),
                             level = c(1,2,0,1),
                             prop = c(0.51, 0.49, 0.65, 0.35))
head(marginals)

## ----message=FALSE----------------------------------------
weights <- computeWeights(data = norm.data, population.margins = marginals)

## ----message = FALSE, results = 'hide', warning=FALSE, fig0, fig.height = 4, fig.width = 7----
norm.model <- cnorm(raw = norm.data$raw, group = norm.data$group,
					weights = weights, terms=6)

## ---------------------------------------------------------
summary(norm.model)

## ----message = FALSE, results = 'hide', warning=FALSE, fig1, fig.height = 4, fig.width = 7----
plot(norm.model, "subset")
plot(norm.model, "norm")

#' cNORM: Continuous Norming
#'
#' The package provides methods for generating regression based continuous standard
#' scores, as f. e. for psychometric test development, biometrics (e. g. physiological
#' growth curves), and screenings in the medical domain. It includes a distribution free approach
#' on the basis of Taylor polynomials and parametric modelling with beta binomial distributions.
#' Both approaches can generate robust norm data models and alleviate the computation of norm scores
#' and norm tables.
#'
#' Conventional methods for producing test norm score tables are often plagued with
#' "jumps" or "gaps" (i.e., discontinuities) in norm tables and low confidence for
#' assessing extreme scores.  The continuous norming method introduced by A. Lenhard et
#' al. (2016, <doi:10.1177/1073191116656437>; 2019, <doi:10.1371/journal.pone.0222279>)
#' addresses these problems and also has the added advantage of not requiring assumptions
#' about the distribution of the raw data: The norm scores are established from raw data
#' by modeling the latter ones as a function  of both percentile scores and an explanatory
#' variable (e.g., age). The method minimizes bias arising from sampling and measurement
#' error, while handling marked deviations from normality - such as are commonplace in
#' clinical samples. For pre-requisites and use cases of the beta binomial modelling, please
#' consult the vignette 'Beta Binomial'.
#'
#' Conducting the analysis consists of four steps and cNORM offers all according functions
#' for preparing data, conducting the  regression, selecting the best model and generating
#' norm tables (according functions in brackets):
#' \enumerate{
#'   \item Data preparation (\code{\link{rankByGroup}}, \code{\link{rankBySlidingWindow}},
#'   \code{\link{computePowers}})
#'   \item Establishing the regression model and selecting the parameters (\code{\link{bestModel}},
#'   \code{\link{printSubset}}, \code{\link{plotSubset}}, \code{\link{regressionFunction}},
#'   \code{\link{derive}})
#'   \item Validating the model (\code{\link{checkConsistency}}, \code{\link{plotPercentiles}},
#'   \code{\link{plotPercentileSeries}}, \code{\link{plotRaw}}, \code{\link{plotNorm}}, \code{\link{derivationTable}},
#'   \code{\link{plotDerivative}})
#'   \item Generating norm tables and predicting scores (\code{\link{predictNorm}},
#'   \code{\link{predictRaw}}, \code{\link{normTable}}, \code{\link{getNormCurve}},
#'   \code{\link{plotNormCurves}})
#' }
#'
#' The function \link{cnorm}
#' For an easy start, you can use the graphical user interface by typing \code{cNORM.GUI()} on the console.
#' Example datasets with large cohorts are available for demonstration purposes ('elfe',
#' 'ppvt', and 'CDC' sample data from the references). Use
#' \code{model <- cnorm(raw = elfe$raw, group = elfe$group)} to get a first impression.
#' Use  \code{vignette(cNORM-Demo)} for a walk through on
#' conducting  the modeling and \url{https://www.psychometrica.de/cNorm_en.html} for a
#' comprehensive tutorial.
#'
#' @references
#' \enumerate{
#'   \item Center for Disease Control and Prevention (2012).
#'   National Health and Nutrition Examination Survey: Questionnaires,
#'   datasets and related documentation. U.S. Department of Health and Human Services
#'   (original source not available anymore).
#'   \item Gary, S., Lenhard, W., Lenhard, A., & Herzberg, D. (2023). A tutorial on automatic post-stratification and weighting in conventional and regression-based norming of psychometric tests. Behavior Research Methods. https://doi.org/10.3758/s13428-023-02207-0
#'   \item Gary, S., Lenhard, A., Lenhard, W., & Herzberg, D. S. (2023). Reducing the Bias of Norm Scores in Non-Representative Samples: Weighting as an Adjunct to Continuous Norming Methods. Assessment, 30(8), 2491–2509. https://doi.org/10.1177/10731911231153832
#'   \item Lenhard, W., & Lenhard, A. (2021). Improvement of Norm Score Quality via Regression-Based
#'   Continuous Norming. Educational and Psychological Measurement, 81(2), 229–261.
#'   doi: 10.1177/0013164420928457
#'   \item Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests:
#'   A simulation study of parametric and semi-parametric approaches.
#'   PLoS ONE, 14(9),  e0222279. doi: 10.1371/journal.pone.0222279
#'   \item Lenhard, A., Lenhard, W., Suggate, S. & Segerer, R. (2016). A continuous solution to
#'   the norming problem. Assessment, Online first, 1-14. doi: 10.1177/1073191116656437
#'   \item Lenhard, A., Lenhard, W., Segerer, R. & Suggate, S. (2015). Peabody Picture Vocabulary
#'   Test - Revision IV (German Adaption). Frankfurt a. M.: Pearson Assessment.
#'   \item Lenhard, W. & Schneider, W. (2006). ELFE 1-6 - Ein Leseverstaendnistest fuer Erst- bis
#'   Sechstklässler. Goettingen: Hogrefe.
#' }
#' @author Wolfgang Lenhard, Alexandra Lenhard and Sebastian Gary
#' @keywords Psychometrics Biometrics Test Development Regression Based Norming
#' @docType PACKAGE
#' @name cNORM
#' @seealso cNORM.GUI
#' @examples
#' \dontrun{
#' # Model internal 'elfe' dataset with the default k = 4 regression on T scores
#' results <- cnorm(raw = elfe$raw, group = elfe$group)
#'
#' # Show model fit of models with progressing number of predictors
#' print(results)
#' plot(results, "subset")
#'
#' # Plot manifest and predicted values, plot series of percentile charts
#' plot(results, "raw")
#' plot(results, "series", start = 3, end = 9)
#'
#'
#' # Additional tests: Check model assumptions
#' checkConsistency(results)
#' plot(results, "derivative")
#'
#' # Generate norm tables; predict values, here: grade 3.75 from T score 25
#' # to 75 and within the raw value range of this specific test (0 to 28)
#' normTable <- normTable(3.75, results, minNorm=25, maxNorm=75, step=0.5)
#' rawTable <- rawTable(3.75, results, minRaw = 0, maxRaw = 28, minNorm=25,
#'                      maxNorm=75)
#'
#' # Predict a specific norm score
#' score <- predictNorm(raw = 21, A = 3.75,
#'                           model = results, minNorm=25, maxNorm=75)
#' }
NULL


#' Launcher for the graphical user interface of cNORM for distribution free
#' continuous norming
#'
#' @param launch.browser Default TRUE; automatically open browser for GUI
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Launch graphical user interface
#' cNORM.GUI()
#' }
cNORM.GUI <- function(launch.browser = TRUE) {
  packageList <- c("shiny",
                   "shinycssloaders",
                   "foreign",
                   "readxl",
                   "markdown",
                   "rmarkdown")

  # Check which packages are missing
  missing <- packageList[!sapply(packageList, requireNamespace, quietly = TRUE)]

  if (length(missing) > 0) {
    cat("Additional packages are needed to start the user interface:\n")
    cat(paste("-", missing, collapse = "\n"), "\n")
    cat("\nWould you like to try to install them now?\n")
    installChoice <- menu(c("yes", "no"))

    if (installChoice == 1) {
      utils::install.packages(missing)
    } else {
      stop("Required packages are missing. Unable to start the GUI")
    }
  }

  appDir <- system.file("shiny", "app1", package = "cNORM")
  shiny::runApp(appDir, display.mode = "normal", launch.browser = launch.browser)
}

#' Launch the cNORM Parametric Modeling Shiny Application
#'
#' @param launch.browser Logical, whether to launch browser automatically (default: TRUE)
#' @examples
#' \dontrun{
#' # Launch graphical user interface
#' cNORM.GUI2()
#' }
#' @export
cNORM.GUI2 <- function(launch.browser = TRUE) {
  # Required packages for the parametric modeling GUI
  packageList <- c("shiny",
                   "shinycssloaders",
                   "DT",
                   "ggplot2",
                   "foreign",
                   "readxl",
                   "haven")

  # Check which packages are missing
  missing <- packageList[!sapply(packageList, requireNamespace, quietly = TRUE)]

  if (length(missing) > 0) {
    cat("Additional packages are needed to start the parametric modeling interface:\n")
    cat(paste("-", missing, collapse = "\n"), "\n")
    cat("\nWould you like to install them now?\n")
    installChoice <- utils::menu(c("Yes", "No"))

    if (installChoice == 1) {
      cat("\nInstalling packages...\n")
      utils::install.packages(missing)

      # Verify installation succeeded
      still_missing <- missing[!sapply(missing, requireNamespace, quietly = TRUE)]
      if (length(still_missing) > 0) {
        stop(
          "Failed to install: ",
          paste(still_missing, collapse = ", "),
          "\nPlease install manually and try again."
        )
      }
      cat("Installation complete!\n\n")
    } else {
      stop("Required packages are missing. Unable to start the GUI.")
    }
  }

  # Locate the app directory
  appDir <- system.file("shiny", "app2", package = "cNORM")

  if (appDir == "") {
    stop("Could not find app directory. Try re-installing `cNORM`.",
         call. = FALSE)
  }

  # Launch the app
  message("Launching cNORM Parametric Modeling GUI...")
  shiny::runApp(appDir, display.mode = "normal", launch.browser = launch.browser)
}

#' Continuous Norming
#'
#' Conducts continuous norming in one step and returns an object including ranked
#' raw data and the continuous norming model. Please consult the function
#' description of 'rankByGroup', 'rankBySlidingWindow' and 'bestModel' for specifics
#' of the steps in the data preparation and modeling process. In addition to the
#' raw scores, either provide
#' \itemize{
#'  \item{a numeric vector for the grouping information (group)}
#'  \item{a numeric age vector and the width of the sliding window (age, width)}
#' }
#' for the ranking of the raw scores. You can
#' adjust the grade of smoothing of the regression model by setting the k and terms
#' parameter. In general, increasing k to more than 4 and the number of terms lead
#' to a higher fit, while lower values lead to more smoothing. The power parameter
#' for the age trajectory can be specified independently by 't'. If both parameters
#' are missing, cnorm uses k = 5 and t = 3 by default.
#'
#' @param raw Numeric vector of raw scores
#' @param group Numeric vector of grouping variable, e.g. grade. If no group
#'   or age variable is provided, conventional norming is applied.
#' @param age Numeric vector with chronological age. If used without `group`,
#'   please additionally specify `width`.
#' @param width Size of the sliding window in case an age vector is used.
#' @param scale Type of norm scale, either "T" (default), "IQ", "z" or
#'   "percentile" (= no transformation); a numeric vector with mean and SD can
#'   also be provided, e.g. `c(10, 3)` for Wechsler scale index points.
#' @param method Ranking method in case of ties; an integer index from 1
#'   (Blom 1958) through 7 (Yu & Huang 2001). Default is 4 (Rankit).
#' @param descend If TRUE, inverts the ranking order so that higher raw scores
#'   receive lower norm scores (e.g. for error scores).
#' @param weights Optional numeric vector of case weights for post-stratification.
#' @param terms If > 0, fix the model to this number of terms.
#' @param R2 Stopping criterion (adjusted R-squared) for model selection.
#' @param k Power degree for the location dimension (max 6).
#' @param t Power degree for the age dimension (max 6).
#' @param plot If TRUE (default), display percentile plot and report.
#' @param extensive If TRUE (default), screen models for monotonic consistency.
#' @param averaging If TRUE (default FALSE), apply BIC-weighted model averaging
#'   across the consistency-screened candidate models instead of selecting a
#'   single model. Requires \code{extensive = TRUE} and age-based norming.
#' @param minDip Tolerance for monotonicity check. Allow small violations
#'   (default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
#'   for strict checking.
#' @return cnorm object including the ranked raw data and the regression model.
#' @seealso rankByGroup, rankBySlidingWindow, computePowers, bestModel
#' @examples
#' \dontrun{
#' # Conventional norming
#' cnorm(raw = elfe$raw)
#'
#' # Continuous norming via group
#' m1 <- cnorm(raw = elfe$raw, group = elfe$group)
#'
#' # Continuous norming via continuous age + sliding window
#' m2 <- cnorm(raw = ppvt$raw, age = ppvt$age, width = 1)
#'
#' # Norm tables with confidence intervals
#' normTable(c(2.125, 2.375, 2.625), m1, CI = .9, reliability = .95)
#' rawTable (c(2.125, 2.375, 2.625), m1, CI = .9, reliability = .95)
#' }
#' @export
#' @references
#' \enumerate{
#'   \item Gary, S. & Lenhard, W. (2021). In norming we trust. Diagnostica.
#'   \item Gary, S., Lenhard, W. & Lenhard, A. (2021). Modelling Norm Scores with the cNORM Package in R. Psych, 3(3), 501-521. https://doi.org/10.3390/psych3030033
#'   \item Gary, S., Lenhard, W., Lenhard, A., & Herzberg, D. (2023). A tutorial on automatic post-stratification and weighting in conventional and regression-based norming of psychometric tests. Behavior Research Methods. https://doi.org/10.3758/s13428-023-02207-0
#'   \item Gary, S., Lenhard, A., Lenhard, W., & Herzberg, D. S. (2023). Reducing the Bias of Norm Scores in Non-Representative Samples: Weighting as an Adjunct to Continuous Norming Methods. Assessment, 30(8), 2491–2509. https://doi.org/10.1177/10731911231153832
#'   \item Lenhard, A., Lenhard, W., Suggate, S. & Segerer, R. (2016). A continuous solution to the norming problem. Assessment, Online first, 1-14. doi:10.1177/1073191116656437
#'   \item Lenhard, A., Lenhard, W., Gary, S. (2018). Continuous Norming (cNORM). The Comprehensive R Network, Package cNORM, available: https://CRAN.R-project.org/package=cNORM
#'   \item Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. PLoS ONE, 14(9),  e0222279. doi:10.1371/journal.pone.0222279
#'   \item Lenhard, W., & Lenhard, A. (2020). Improvement of Norm Score Quality via Regression-Based Continuous Norming. Educational and Psychological Measurement(Online First), 1-33. https://doi.org/10.1177/0013164420928457
#' }
cnorm <- function(raw = NULL,
                  group = NULL,
                  age = NULL,
                  width = NA,
                  weights = NULL,
                  scale = "T",
                  method = 4,
                  descend = FALSE,
                  k = NULL,
                  t = NULL,
                  terms = 0,
                  R2 = NULL,
                  plot = TRUE,
                  extensive = TRUE,
                  averaging = FALSE,
                  minDip = .01) {

  # ---- input validation -----------------------------------------------------
  if (is.null(raw) || !is.numeric(raw))
    stop("Please provide a numeric vector for the raw scores.")
  if (!is.null(group) && !is.numeric(group))
    stop("`group` must be a numeric vector.")
  if (!is.null(age) && !is.numeric(age))
    stop("`age` must be a numeric vector.")
  if (!is.null(weights) && !is.numeric(weights))
    stop("`weights` must be a numeric vector.")
  if (!is.null(group) && length(group) != length(raw))
    stop("`raw` and `group` must have the same length.")
  if (!is.null(age) && length(age) != length(raw))
    stop("`raw` and `age` must have the same length.")
  if (!is.null(weights) && length(weights) != length(raw))
    stop("`weights` must have the same length as `raw`.")
  if (!is.null(weights) && any(weights <= 0, na.rm = TRUE))
    stop("`weights` must be positive, non-zero values (see computeWeights).")

  silent <- !plot

  if (!is.null(group) && !is.null(age) && !silent) {
    message("Both 'group' and 'age' supplied: ranking within 'group' and ",
            "modelling on the continuous 'age' variable.")
  }

  # ---- default smoothing parameters -----------------------------------------
  if (is.null(k) && is.null(t)) {
    k <- 5; t <- 3
  } else if (is.null(t)) {
    t <- k
  } else if (is.null(k)) {
    k <- t
  }

  # ---- assemble & clean (single source of NA removal, weights kept aligned) --
  df_in <- data.frame(raw = raw)
  if (!is.null(group))   df_in$group   <- group
  if (!is.null(age))     df_in$age     <- age
  if (!is.null(weights)) df_in$weights <- weights
  df_in <- df_in[stats::complete.cases(df_in), , drop = FALSE]
  if (nrow(df_in) == 0L)
    stop("After removing missing values no observations remain.")

  weights_name <- if (!is.null(weights)) "weights" else NULL

  conventional <- is.null(group) && is.null(age)
  use_window   <- is.null(group) && !is.null(age) && !is.na(width)

  # ============================ CONVENTIONAL =================================
  if (conventional) {
    data <- rankByGroup(data = df_in, group = FALSE, raw = "raw",
                        scale = scale, weights = weights_name,
                        descend = descend, method = method, silent = silent)
    data <- computePowers(data, k = k, t = t, silent = silent)

    model <- bestModel(data, k = k, t = t, terms = terms, R2 = R2,
                       weights = data$weights, plot = FALSE,
                       extensive = extensive, averaging=averaging, minDip = minDip)

    result <- structure(list(data = data, model = model), class = "cnorm")
    if (plot) {
      cat(model$report, sep = "\n")
      print(rawTable(0, result))
      plotPercentiles(result)
    }
    return(result)
  }

  # ============================ SLIDING WINDOW ==============================
  if (use_window) {
    if (!silent) message("Ranking data with sliding window ...")

    # Synthesise a grouping variable for display / manifest percentiles only;
    # the model itself is fitted on the continuous age.
    if (is.null(df_in$group)) df_in$group <- getGroups(df_in$age)

    data <- rankBySlidingWindow(data = df_in, age = "age", raw = "raw",
                                group = "group", width = width, scale = scale,
                                weights = weights_name, descend = descend,
                                method = method, silent = silent)
    data <- data[!is.na(data$normValue), , drop = FALSE]
    data <- computePowers(data, k = k, t = t, age = "age", silent = silent)

    # ============================ BY GROUP ====================================
  } else {
    group_was_synthesised <- is.null(group)

    if (group_was_synthesised) {
      if (length(df_in$age) / length(unique(df_in$age)) > 50 &&
          min(table(df_in$age)) > 30) {
        if (!silent) message("Width missing. Using age directly as grouping variable.")
        df_in$group <- df_in$age
      } else {
        if (!silent) message("Width missing. Discretising age via getGroups().")
        df_in$group <- getGroups(df_in$age)
      }
    }

    data <- rankByGroup(data = df_in, group = "group", raw = "raw",
                        scale = scale, weights = weights_name,
                        descend = descend, method = method, silent = silent)
    data <- data[!is.na(data$normValue), , drop = FALSE]

    # Model on the continuous age only when the user supplied a *real* group
    # together with a separate age vector; otherwise the group already IS the
    # (discretised) age axis. 'age' rides along as a column of `data`, so it
    # stays row-aligned even if cases were dropped during ranking.
    model_on_age <- (!group_was_synthesised) && !is.null(df_in$age)
    if (model_on_age) {
      data <- computePowers(data, k = k, t = t, age = "age", silent = silent)
    } else {
      data <- computePowers(data, k = k, t = t, silent = silent)
    }
  }

  # ---- model, assemble, report ----------------------------------------------
  model <- bestModel(data, k = k, t = t, terms = terms, R2 = R2,
                     weights = data$weights, plot = FALSE,
                     extensive = extensive, averaging=averaging, minDip = minDip)

  result <- structure(list(data = data, model = model), class = "cnorm")
  if (plot) {
    cat(model$report, sep = "\n")
    plotPercentiles(result)
  }
  return(result)
}

#' Swiftly compute Taylor regression models for distribution free continuous norming
#'
#' Conducts distribution free continuous norming and aims to find a fitting model. Raw data are modelled as a Taylor polynomial
#' of powers of age and location and their interactions. In addition to the
#' raw scores, either provide a numeric vector for the grouping information (group)
#' for the ranking of the raw scores. You can adjust the grade of smoothing of the regression model by setting the k, t and terms
#' parameter. In general, increasing k and t leads to a higher fit, while lower values lead to more smoothing. If both parameters
#' are missing, taylorSwift uses k = 5 and t = 3 by default.
#'
#' @param raw Numeric vector of raw scores
#' @param group Numeric vector of grouping variable, e. g. grade. If no group
#' or age variable is provided, conventional norming is applied
#' @param age Numeric vector with chronological age, please additionally specify
#' width of window
#' @param width Size of the sliding window in case an age vector is used
#' @param scale type of norm scale, either T (default), IQ, z or percentile (= no
#' transformation); a double vector with the mean and standard deviation can as
#' well, be provided f. e. c(10, 3) for Wechsler scale index points
#' @param method Ranking method in case of bindings, please provide an index,
#' choosing from the following methods: 1 = Blom (1958), 2 = Tukey (1949),
#' 3 = Van der Warden (1952), 4 = Rankit (default), 5 = Levenbach (1953),
#' 6 = Filliben (1975), 7 = Yu & Huang (2001)
#' @param descend ranking order (default descent = FALSE): inverses the
#' ranking order with higher raw scores getting lower norm scores; relevant
#' for example when norming error scores, where lower scores mean higher
#' performance
#' @param weights Vector or variable name in the dataset with weights for each
#' individual case. It can be used to compensate for moderate imbalances due to
#' insufficient norm data stratification. Weights should be numerical and positive.
#' @param terms Selection criterion for model building. The best fitting model with
#' this number of terms is used
#' @param R2 Adjusted R square as a stopping criterion for the model building
#' (default R2 = 0.99)
#' @param k The power constant. Higher values result in more detailed approximations
#' but have the danger of over-fit (max = 6). If not set, it uses t and if both
#' parameters are NULL, k is set to 5.
#' @param t The age power parameter (max = 6). If not set, it uses k and if both
#' parameters are NULL, k is set to 3, since age trajectories are most often well
#' captured by cubic polynomials.
#' @param plot Default TRUE; plots the regression model and prints report
#' @param extensive If TRUE, screen models for consistency and - if possible, exclude inconsistent ones
#' @param averaging If TRUE (default FALSE), apply BIC-weighted model averaging
#'   across the consistency-screened candidate models instead of selecting a
#'   single model. Requires \code{extensive = TRUE} and age-based norming.
#' @param minDip Tolerance for monotonicity check. Allow small violations
#'   (default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
#'   for strict checking.
#'
#' @return cnorm object including the ranked raw data and the regression model
#' @seealso rankByGroup, rankBySlidingWindow, computePowers, bestModel
#' @examples
#' \dontrun{
#' # Using this function with the example dataset 'ppvt'
#' # You can use the 'getGroups()' function to set up grouping variable in case,
#' # you have a continuous age variable.
#' model <- taylorSwift(raw = ppvt$raw, group = ppvt$group)
#'
#' # return norm tables including 90% confidence intervals for a
#' # test with a reliability of r = .85; table are set to mean of quartal
#' # in grade 3 (children completed 2 years of schooling)
#' normTable(c(5, 15), model, CI = .90, reliability = .95)
#'
#' # ... or instead of raw scores for norm scores, the other way round
#' rawTable(c(8, 12), model, CI = .90, reliability = .95)
#' }
#' @export
#' @references
#' \enumerate{
#'   \item Gary, S. & Lenhard, W. (2021). In norming we trust. Diagnostica.
#'   \item Gary, S., Lenhard, W. & Lenhard, A. (2021). Modelling Norm Scores with the cNORM Package in R. Psych, 3(3), 501-521. https://doi.org/10.3390/psych3030033
#'   \item Lenhard, A., Lenhard, W., Suggate, S. & Segerer, R. (2016). A continuous solution to the norming problem. Assessment, Online first, 1-14. doi:10.1177/1073191116656437
#'   \item Lenhard, A., Lenhard, W., Gary, S. (2018). Continuous Norming (cNORM). The Comprehensive R Network, Package cNORM, available: https://CRAN.R-project.org/package=cNORM
#'   \item Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. PLoS ONE, 14(9),  e0222279. doi:10.1371/journal.pone.0222279
#'   \item Lenhard, W., & Lenhard, A. (2020). Improvement of Norm Score Quality via Regression-Based Continuous Norming. Educational and Psychological Measurement(Online First), 1-33. https://doi.org/10.1177/0013164420928457
#'
#' }
taylorSwift <- cnorm

#' Standardize a numeric vector
#'
#' This function standardizes a numeric vector by subtracting the mean
#' and dividing by the standard deviation. The resulting vector will have
#' a mean of 0 and a standard deviation of 1.
#'
#' @param x A numeric vector to be standardized.
#'
#' @return A numeric vector of the same length as x, containing the standardized values.
#'
#' @examples
#' \dontrun{
#' data <- c(1, 2, 3, 4, 5)
#' standardized_data <- standardize(data)
#' print(standardized_data)
#' }
#'
#' @export
standardize <- function(x) {
  (x - mean(x)) / sd(x)
}


#' Weighted rank estimation
#'
#' Conducts weighted ranking on the basis of sums of weights per unique raw score.
#' Please provide a vector with raw values and an additional vector with the weight
#' of each observation. In case the weight vector is NULL, a normal ranking is done.
#' The vectors may not include NAs and the weights should be positive non-zero
#' values. With equal weights, the result reproduces the mid-ranks of base R's
#' \code{rank()} (\code{ties.method = "average"}).
#'
#' @param x A numerical vector
#' @param weights A numerical vector with weights; should have the same length as x
#' @return the weighted absolute ranks
#' @export
weighted.rank <- function(x, weights = NULL) {
  if (is.null(weights)) {
    return(rank(x))
  }

  # Input validation
  if (length(x) != length(weights)) {
    stop("x and weights must have the same length.")
  }
  if (anyNA(x) || anyNA(weights)) {
    stop("x and weights must not contain NAs.")
  }
  if (any(weights <= 0)) {
    stop("weights must be positive non-zero values.")
  }

  # 1. Aggregate weights by unique x value.
  #    factor() orders the levels numerically, so the cumulative ("strictly
  #    less than") logic below is valid, and the final mapping is done through
  #    the integer factor codes - this avoids the numeric -> character ->
  #    numeric round-trip and the floating-point match() failures it can cause.
  f          <- factor(x)
  sum_weights <- tapply(weights, f, sum)
  W_eq        <- as.numeric(sum_weights)

  # 2. Cumulative weight strictly below each unique value
  W_less <- cumsum(c(0, W_eq[-length(W_eq)]))

  # 3. Empirical mid-rank probability for each unique value
  S        <- sum(weights)
  mid_prob <- (W_less + (W_eq / 2)) / S

  # 4. Map the probability onto the continuous rank scale [1, N];
  #    reproduces base R rank() (ties.method = "average") for equal weights.
  N            <- length(x)
  unique_ranks <- (mid_prob * N) + 0.5

  # 5. Map the unique ranks back onto the original layout of x via factor codes
  final_ranks <- unique_ranks[as.integer(f)]

  return(final_ranks)
}

#' Weighted quantile estimator
#'
#' Computes weighted quantiles (code from Andrey Akinshin (2023) "Weighted quantile estimators" arXiv:2304.07265 [stat.ME]
#' Code made available via the CC BY-NC-SA 4.0 license) on the basis of either the weighted Harrell-Davis
#' quantile estimator or an adaption of the type 7 quantile estimator of the generic quantile function in
#' the base package. Please provide a vector with raw values, the probabilities for the quantiles and an
#' additional vector with the weight of each observation. In case the weight vector is NULL, a normal
#' quantile estimation is done. The vectors may not include NAs and the weights should be positive non-zero
#' values. Please draw on the computeWeights() function for retrieving weights in post stratification.
#'
#' @param x A numerical vector
#' @param probs Numerical vector of quantiles
#' @param type Type of estimator, can either be "inflation", "Harrell-Davis" using a beta function to
#' approximate the weighted percentiles (Harrell & Davis, 1982) or "Type7" (default; Hyndman & Fan, 1996), an adaption
#' of the generic quantile function in R, including weighting. The inflation procedure is essentially
#' a numerical, non-parametric solution that gives the same results as Harrel-Davis. It requires less
#' ressources with small datasets and always finds a solution (e. g. 1000 cases with
#' weights between 1 and 10). If it becomes too resource intense, it switches to Harrell-Davis automatically.
#' Harrel-Davis and Type7 code is based on the work of Akinshin (2023).
#' @param weights A numerical vector with weights; should have the same length as x
#' @references
#' \enumerate{
#'   \item Harrell, F.E. & Davis, C.E. (1982). A new distribution-free quantile estimator. Biometrika, 69(3), 635-640.
#'   \item Hyndman, R. J. & Fan, Y. (1996). Sample quantiles in statistical packages, American Statistician 50, 361–365.
#'   \item Akinshin, A. (2023). Weighted quantile estimators arXiv:2304.07265 [stat.ME]
#' }
#' @seealso weighted.quantile.inflation, weighted.quantile.harrell.davis, weighted.quantile.type7
#' @return the weighted quantiles
#' @export
weighted.quantile <- function(x,
                              probs,
                              weights = NULL,
                              type = "Harrell-Davis") {
  if (is.null(weights)) {
    return(quantile(x, probs))
  } else if (type == "Harrell-Davis") {
    return(weighted.quantile.harrell.davis(x, probs, weights))
  } else if (type == "inflation") {
    return(weighted.quantile.inflation(x, probs, weights))
  } else{
    return(weighted.quantile.type7(x, probs, weights))
  }
}

#' Weighted quantile estimator through case inflation
#'
#' Applies weighted ranking numerically by inflating cases according to weight. This function
#' will be resource intensive, if inflated cases get too high and in this cases, it switches
#' to the parametric Harrell-Davis estimator.
#'
#' @param x A numerical vector
#' @param probs Numerical vector of quantiles
#' @param weights A numerical vector with weights; should have the same length as x.
#' @param degree power parameter for case inflation (default = 3, equaling factor 1000)
#' If no weights are provided (NULL), it falls back to the base quantile function, type 7
#' @param cutoff stop criterion for the sum of standardized weights to switch to Harrell-Davis,
#' default = 1000000
#' @return the quantiles
#' @export
weighted.quantile.inflation <- function(x,
                                        probs,
                                        weights = NULL,
                                        degree = 3,
                                        cutoff = 10000000) {
  if (is.null(weights)) {
    return(quantile(x, probs = probs))
  }

  data <- x
  w <- round(weights / min(weights) * (10^degree), digits = 0)

  # switch to Harrell Davis, if computations gets to resource intensive
  if (sum(w) > cutoff) {
    print("Inflation too intense - switching to parametric Harrel-Davis estimator")
    return(weighted.quantile.harrell.davis(x, probs, weights))
  } else{
    # expand
    for (i in 1:length(x)) {
      data <- c(data, rep(data[i], w[i]))
    }
    return(quantile(data, probs = probs))
  }
}



#' Weighted Harrell-Davis quantile estimator
#'
#' Computes weighted quantiles; code from Andrey Akinshin (2023) "Weighted quantile estimators" arXiv:2304.07265 [stat.ME]
#' Code made available via the CC BY-NC-SA 4.0 license
#'
#' @param x A numerical vector
#' @param probs Numerical vector of quantiles
#' @param weights A numerical vector with weights; should have the same length as x.
#' If no weights are provided (NULL), it falls back to the base quantile function, type 7
#'
#' @return the quantiles
#' @export
weighted.quantile.harrell.davis <- function(x, probs, weights = NULL) {
  if (is.null(weights)) {
    return(quantile(x, probs = probs))
  }

  cdf.gen <- function(n, p)
    return(function(cdf.probs) {
      pbeta(cdf.probs, (n + 1) * p, (n + 1) * (1 - p))
    })
  wquantile.generic(x, probs, cdf.gen, weights)
}


#' Weighted type7 quantile estimator
#'
#' Computes weighted quantiles; code from Andrey Akinshin (2023) "Weighted quantile estimators" arXiv:2304.07265 [stat.ME]
#' Code made available via the CC BY-NC-SA 4.0 license
#'
#' @param x A numerical vector
#' @param probs Numerical vector of quantiles
#' @param weights A numerical vector with weights; should have the same length as x.
#' If no weights are provided (NULL), it falls back to the base quantile function, type 7
#'
#' @return the quantiles
#' @export
weighted.quantile.type7 <- function(x, probs, weights = NULL) {
  if (is.null(weights)) {
    return(quantile(x, probs = probs))
  }

  cdf.gen <- function(n, p)
    return(function(cdf.probs) {
      h <- p * (n - 1) + 1
      u <- pmax((h - 1) / n, pmin(h / n, cdf.probs))
      u * n - h + 1
    })
  wquantile.generic(x, probs, cdf.gen, weights)
}

# Weighted generic quantile estimator
#
# Code made available via the CC BY-NC-SA 4.0 license from code from
# Andrey Akinshin (2023) "Weighted quantile estimators" arXiv:2304.07265 [stat.ME]
wquantile.generic <- function(x, probs, cdf.gen, weights = NULL) {
  if (is.null(weights)) {
    return(quantile(x, probs = probs))
  }

  n <- length(x)
  if (any(is.na(weights)))
    weights <- rep(1 / n, n)
  nw <- sum(weights) / max(weights)

  indexes <- order(x)
  x <- x[indexes]
  weights <- weights[indexes]

  weights <- weights / sum(weights)
  cdf.probs <- cumsum(c(0, weights))

  sapply(probs, function(p) {
    cdf <- cdf.gen(nw, p)
    q <- cdf(cdf.probs)
    w <- tail(q, -1) - head(q, -1)
    sum(w * x)
  })
}



#' Determine groups and group means
#'
#' Splits a continuous variable into \code{n} groups (either equidistant or
#' with equal number of observations) and returns the group mean for every
#' observation.
#'
#' @param x Numeric vector to be split.
#' @param n Desired number of groups. If \code{NULL}, the function chooses
#'   roughly one group per 100 observations, constrained to 3 \eqn{\le} n \eqn{\le} 20.
#' @param equidistant If \code{TRUE}, intervals have equal width; otherwise
#'   (default) each interval contains roughly the same number of observations.
#'
#' @return Numeric vector of the same length as \code{x} with the group mean
#'   assigned to each observation.
#' @export
#'
#' @examples
#' \dontrun{
#' x <- rnorm(1000, mean = 50, sd = 10)
#' m <- getGroups(x, n = 10)
#' }
getGroups <- function(x, n = NULL, equidistant = FALSE) {

  if (!is.numeric(x))
    stop("'x' must be a numeric vector.")

  # Auto rule: ~one group per 100 obs, clamped to [3, 20]; ensure integer.
  if (is.null(n)) {
    n <- floor(length(x) / 100)
    n <- max(3L, min(20L, n))
  } else {
    if (!is.numeric(n) || length(n) != 1L || n < 2)
      stop("'n' must be a single integer >= 2.")
    n <- as.integer(n)
  }

  # Build group codes
  if (equidistant) {
    groups <- as.integer(cut(x, breaks = n, include.lowest = TRUE))
  } else {
    breaks <- stats::quantile(
      x,
      probs   = seq(0, 1, length.out = n + 1L),
      na.rm   = TRUE,
      names   = FALSE,
      type    = 7
    )
    # Collapse duplicate breaks (heavy ties).  Warn if effective n < requested.
    ubreaks <- unique(breaks)
    if (length(ubreaks) < length(breaks)) {
      warning(sprintf(
        "Requested %d groups, but only %d distinct quantile breakpoints could be formed (ties in 'x'). Effective number of groups: %d.",
        n, length(ubreaks), length(ubreaks) - 1L
      ))
    }
    groups <- findInterval(x, ubreaks, all.inside = TRUE)
  }

  # Group means, broadcast back to each observation.
  ave(x, groups, FUN = function(z) mean(z, na.rm = TRUE))
}


#' Simulate mean per age
#'
#' @param age the age variable
#'
#' @return return predicted means
#'
#' @examples
#' \dontrun{
#' x <- simMean(a)
#' }
simMean <- function(age) {
  return(1.5 * age - 0.05 * age^2 + 0.0001 * age^4)
}

#' Simulate sd per age
#'
#' @param age the age variable
#'
#' @return return predicted sd
#'
#' @examples
#' \dontrun{
#' x <- simSD(a)
#' }
simSD <- function(age) {
  return((1.5 * age - 0.05 * age^2 + 0.0001 * age^4) * 0.2 + 1)
}

#' Simulate raw test scores based on Rasch model
#'
#' For testing purposes only:
#' The function simulates raw test scores based on a virtual Rasch based test with n results per
#' age group, an evenly distributed age variable, items.n test items with a simulated difficulty and
#' standard deviation. The development trajectories over age group are modeled by a curve linear
#' function of age, with at first fast progression, which slows down over age, and a slightly increasing
#' standard deviation in order to model a scissor effects. The item difficulties can be accessed via $theta
#' and the raw data via $data of the returned object.
#'
#' @param data data.frame from previous simulations for recomputation (overrides n, minAge, maxAge)
#' @param n The sample size per age group
#' @param minAge The minimum age (default 1)
#' @param maxAge The maximum age (default 7)
#' @param items.n The number of items of the test
#' @param items.m The mean difficulty of the items
#' @param items.sd The standard deviation of the item difficulty
#' @param Theta irt scales difficulty parameters, either "random" for drawing a random sample,
#' "even" for evenly distributed or a set of predefined values, which then overrides the item.n
#' parameters
#' @param width The width of the window size for the continuous age per group; +- 1/2 width around group
#' center
#' on items.m and item.sd; if set to FALSE, the distribution is not drawn randomly but normally nonetheless
#' @export
#' @return a list containing the simulated data and thetas
#' \describe{
#'   \item{data}{the data.frame with only age, group and raw}
#'   \item{sim}{the complete simulated data with item level results}
#'   \item{theta}{the difficulty of the items}
#' }
#'
#' @examples
#' \dontrun{
#' # simulate data for a rather easy test (m = -1.0)
#' sim <- simulateRasch(n=150, minAge=1,
#'                      maxAge=7, items.n = 30, items.m = -1.0,
#'                      items.sd = 1, Theta = "random", width = 1.0)
#'
#' # Show item difficulties
#' mean(sim$theta)
#' sd(sim$theta)
#' hist(sim$theta)
#'
#' # Plot raw scores
#' boxplot(raw~group, data=sim$data)
#'
#' # Model data
#' data <- prepareData(sim$data, age="age")
#' model <- bestModel(data, k = 4)
#' printSubset(model)
#' plotSubset(model, type=0)
#' }
simulateRasch <- function(data = NULL,
                          n = 100,
                          minAge = 1,
                          maxAge = 7,
                          items.n = 21,
                          items.m = 0,
                          items.sd = 1,
                          Theta = "random",
                          width = 1) {
  # draw sample
  if (is.null(data)) {
    groups <- seq.int(from = minAge, to = maxAge)
    latent <- vector(mode = "numeric", length = 0)
    age <- vector(mode = "numeric", length = 0)
    group <- vector(mode = "numeric", length = 0)
    i <- 1

    while (i <= length(groups)) {
      group <- c(group, rep(groups[i], times = n))
      age <- c(age, runif(
        n,
        min = groups[i] - (width / 2),
        max = groups[i] + (width / 2)
      ))

      i <- i + 1
    }

    data <- data.frame(age, group, mean = simMean(age), sd = simSD(age))
    data$z <- rnorm(nrow(data))
    data$latent <- data$z * data$sd + data$mean

    meanL <- mean(data$latent)
    sdL <- sd(data$latent)

    # compute standardized value over complete sample
    data$zOverall <- (data$latent - meanL) / sdL
  }

  # generate item difficulties either randomly or evenly distributed
  if (is.character(Theta)) {
    if (Theta == "random") {
      theta <- rnorm(items.n, items.m, items.sd)
    } else if (Theta == "even") {
      p <- seq(
        from = 0.5 / items.n,
        to = (items.n - 0.5) / items.n,
        length.out = items.n
      )
      theta <- qnorm(p, items.m, items.sd)
    }
  } else {
    theta <- Theta
    items.n <- length(Theta)
  }


  # compute propabilities
  i <- 1
  while (i <= items.n) {
    # prob <- exp(data$zOverall - theta[i])/(1 + exp(data$zOverall - theta[i]))
    prob <- pnorm(data$zOverall - theta[i]) # use real cumulative normal instead of logit
    name <- paste0("prob", i)
    data <- cbind(data, prob)
    names(data)[[ncol(data)]] <- name

    i <- i + 1
  }

  data <- transform(data, expected = rowSums(data[, 8:items.n + 7]))

  i <- 1
  # compute propabilities
  while (i <= items.n) {
    rand <- runif(length(data$latent))
    item <- round((sign(data[, i + 7] - rand) + 1) / 2)
    name <- paste0("item", i)
    data <- cbind(data, item)
    names(data)[[ncol(data)]] <- name

    i <- i + 1
  }
  sim <- transform(data, raw = rowSums(data[, (9 + items.n):ncol(data)]))
  dat <- sim[, c(1, 2, 5, ncol(sim))]
  return(list(
    data = dat,
    sim = sim,
    theta = theta
  ))
}

#' Build cnorm object from data and bestModel model object
#'
#' Helper function to build a cnorm object from a data object and
#' a model object from the bestModel function for compatibility reasons.
#'
#'
#' @param data A data object from 'prepareData', or from 'rankByGroup' and
#'             'computePower'
#' @param model Object obtained from the bestModel function
#'
#' @return A cnorm object
#'
#' @examples
#' \dontrun{
#'   data <- prepareData(elfe)
#'   model <- bestModel(data, k = 4)
#'   model.cnorm <- buildCnormObject(data, model)
#' }
#'
#' @export
buildCnormObject <- function(data, model) {
  result <- list(data = data, model = model)
  class(result) <- "cnorm"
  return(result)
}

#' Prepare design matrix for regression
#'
#' @description
#' This function prepares the design matrix, including powers and interaction terms of location and age.
#'
#' @param location Numeric vector of location values (norm scores).
#' @param age Numeric vector of age values.
#' @param k Integer. Maximum power for location terms. Default is 5.
#' @param t Integer. Maximum power for age terms. Default is 3.
#'
#' @return A matrix with columns for powers of location and age, and their interactions.
#'
#' @details
#' The function creates a matrix with columns for powers of location (up to k),
#' powers of age (up to t), and their interactions. The resulting matrix has
#' (k+1)*(t+1)-1 columns (excluding the intercept term).
#'
#' @note
#' It is recommended to keep k and t below 8 to avoid numerical instability.
#'
#' @keywords internal
#'
#' @noRd
prepare_matrix <- function(location, age, k = 5, t = 3) {
  if (!is.numeric(location) ||
      !is.numeric(age) || length(location) != length(age))
    stop("location and age must be numeric vectors of the same length")

  if (!is.numeric(k) || !is.numeric(t) || k < 0 || t < 0 ||
      k != round(k) || t != round(t))
    stop("k and t must be non-negative integers")

  if (k > 8 || t > 8)
    warning("k and t should not exceed 8")

  n <- length(location)

  # Precompute all powers once — n × (k+1) and n × (t+1)
  loc_pow <- outer(location, 0:k, `^`)
  age_pow <- outer(age, 0:t, `^`)

  # All (i, j) pairs excluding (0, 0)
  idx <- expand.grid(i = 0:k, j = 0:t)
  idx <- idx[!(idx$i == 0L & idx$j == 0L), ]

  # FIX: wrap in matrix() — mapply simplifies to a vector when n = 1,
  # which has no colnames dimension
  mat <- matrix(mapply(function(i, j)
    loc_pow[, i + 1L] * age_pow[, j + 1L], idx$i, idx$j), nrow = n)

  colnames(mat) <- ifelse(
    idx$i > 0L & idx$j > 0L,
    paste0("L", idx$i, "A", idx$j),
    ifelse(idx$i > 0L, paste0("L", idx$i), paste0("A", idx$j))
  )

  attr(mat, "k") <- k
  attr(mat, "t") <- t

  mat
}

# Check if object is of class cnorm
#' @keywords internal
isTaylor <- function(model) {
  return(inherits(model, "cnorm") || inherits(model, "cnormTemp"))
}

# Check if model is of class cnormBetaBinomial or cnormBetaBinomial2
#' @keywords internal
isBeta <- function(model) {
  return(inherits(model, "cnormBetaBinomial2") ||
           inherits(model, "cnormBetaBinomial"))
}

# Check if model is of class cnormShash
#' @keywords internal
isSHASH <- function(model) {
  return(inherits(model, "cnormShash") || inherits(model, "cnormShaSh"))
}

# Check if model is parametric (SHASH or BetaBinomial)
#' @keywords internal
isParametric <- function(model) {
  return(isSHASH(model) || isBeta(model))
}

# filter incomplete cases
#' @keywords internal
filter_complete <- function(...) {
  v <- list(...)
  keep <- Reduce(`&`, lapply(v, is.finite))
  lapply(v, `[`, keep)
}

#' Plot manifest and fitted raw scores
#'
#' The function plots the raw data against the fitted scores from
#' the regression model per group. This helps to inspect the precision
#' of the modeling process. The scores should not deviate too far from
#' regression line.
#' @param model The regression model from the 'cnorm' function
#' @param group Should the fit be displayed by group?
#' @param type Type of display: 0 = plot manifest against fitted values, 1 = plot
#' manifest against difference values
#' @examples
#' \dontrun{
#'   # Compute model with example dataset and plot results
#'   result <- cnorm(raw = elfe$raw, group = elfe$group)
#'   plotRaw(result)
#' }
#' @import ggplot2
#' @export
#' @family plot
plotRaw <- function(model, group = FALSE, type = 0) {
  if (isParametric(model)) {
    stop("This function is not applicable for parametric models.")
  }

  if (!isTaylor(model)) {
    stop("Please provide a cnorm object.")
  }

  d <- model$data
  model <- model$model

  d$fitted <- model$fitted.values
  d$diff <- d$fitted - d$raw
  mse <- round(model$rmse, digits = 4)
  r <- round(cor(d$fitted, d$raw, use = "pairwise.complete.obs"), digits = 4)
  d <- as.data.frame(d)

  if (group) {
    if ("group" %in% colnames(d)) {
      d$group <- as.factor(d$group)
    } else {
      d$group <- as.factor(getGroups(d$age))
    }
  }

  if (type == 0) {
    p <- ggplot(d, aes(x = .data$raw, y = .data$fitted)) +
      geom_point(alpha = 0.2, color = "#0033AA") +
      geom_abline(
        intercept = 0,
        slope = 1,
        color = "red",
        linetype = "dashed"
      ) +
      labs(
        title = if (isTRUE(group))
          "Observed vs. Fitted Raw Scores by Group"
        else
          "Observed vs. Fitted Raw Scores",
        subtitle = paste("r =", r, ", RMSE =", mse),
        x = "Observed Score",
        y = "Fitted Scores"
      )
  } else {
    p <- ggplot(d, aes(x = .data$raw, y = .data$diff)) +
      geom_point(alpha = 0.2, color = "#0033AA") +
      geom_hline(yintercept = 0,
                 color = "red",
                 linetype = "dashed") +
      labs(
        title = if (isTRUE(group))
          "Observed Raw Scores vs. Difference Scores by Group"
        else
          "Observed Raw Scores vs. Difference Scores",
        subtitle = paste("r =", r, ", RMSE =", mse),
        x = "Observed Score",
        y = "Difference Scores"
      )
  }

  if (group) {
    p <- p + facet_wrap( ~ group)
  }

  p <- p + theme_minimal() +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      plot.subtitle = element_text(hjust = 0.5, size = 12),
      axis.title = element_text(size = 12, face = "bold"),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  return(p)
}

#' @title Plot manifest and fitted norm scores
#'
#' @description
#' This function plots the manifest norm score against the fitted norm score from
#' the inverse regression model per group. This helps to inspect the precision
#' of the modeling process. The scores should not deviate too far from
#' the regression line. Applicable for Taylor polynomial, beta-binomial, and shash models.
#'
#' @param model The regression model, usually from the 'cnorm', 'cnorm.betabinomial', or 'cnorm.shash' function
#' @param age In case of parametric models, please provide the age vector
#' @param score In case of parametric models, please provide the score vector
#' @param width In case of parametric models, please provide the width for the sliding window.
#'              If null, the function tries to determine a sensible setting.
#' @param weights Vector or variable name in the dataset with weights for each
#' individual case. If NULL, no weights are used.
#' @param group On optional grouping variable, use empty string for no group, the variable name
#'              for Taylor polynomial models or a vector with the groups for parametric models
#' @param minNorm lower bound of fitted norm scores
#' @param maxNorm upper bound of fitted norm scores
#' @param type Type of display: 0 = plot manifest against fitted values, 1 = plot
#' manifest against difference values
#'
#' @return A ggplot object representing the norm scores plot.
#'
#' @examples
#' \dontrun{
#' # Load example data set, compute model and plot results
#'
#' # Taylor polynomial model
#' model <- cnorm(raw = elfe$raw, group = elfe$group)
#' plot(model, "norm")
#'
#' # Beta binomial models; maximum number of items in elfe is n = 28
#' model.bb <- cnorm.betabinomial(elfe$group, elfe$raw, n = 28)
#' plotNorm(model.bb, age = elfe$group, score = elfe$raw)
#' }
#'
#' @import ggplot2
#' @export
#' @family plot
plotNorm <- function(model,
                     age = NULL,
                     score = NULL,
                     width = NULL,
                     weights = NULL,
                     group = FALSE,
                     minNorm = NULL,
                     maxNorm = NULL,
                     type = 0) {
  is_taylor <- isTaylor(model)
  if (is_taylor) {
    data <- model$data
    model <- model$model

    if (is.null(minNorm)) {
      minNorm <- model$minL1
    }

    if (is.null(maxNorm)) {
      maxNorm <- model$maxL1
    }

    d <- data
    raw <- data[[model$raw]]
    if (attr(data, "useAge"))
      age <- data[[model$age]]
    else
      age <- rep(0, length = nrow(data))

    d$fitted <- predictNorm(raw, age, model, minNorm = minNorm, maxNorm = maxNorm)

    if (group) {
      if ("group" %in% colnames(d)) {
        d$group <- as.factor(d$group)
      } else {
        d$group <- as.factor(getGroups(d$age))
      }
    }

  } else if (isParametric(model)) {
    if (is.null(age) || is.null(score)) {
      stop(
        "Please provide age and score vectors for beta-binomial or shash models and the width for the sliding window."
      )
    }

    # Extract model scale ensuring manifest norm scores match fitted norm scores
    scaleMean <- attr(model$result, "scaleMean")
    scaleSD <- attr(model$result, "scaleSD")
    model_scale <- c(scaleMean, scaleSD)

    d <- data.frame(age = age, score = score)
    if (is.null(width)) {
      if (length(age) / length(unique(age)) < 50)
        stop("Please provide a width for the sliding window.")

      d$group <- d$age
      if (is.null(weights))
        d <- rankByGroup(data = d,
                         group = "age",
                         raw = "score",
                         scale = model_scale)
      else
        d <- rankByGroup(
          data = d,
          group = "age",
          raw = "score",
          weights = weights,
          scale = model_scale
        )
    } else{
      if (is.null(weights))
        d <- rankBySlidingWindow(
          data = d,
          age = "age",
          raw = "score",
          width = width,
          scale = model_scale
        )
      else
        d <- rankBySlidingWindow(
          data = d,
          age = "age",
          raw = "score",
          weights = weights,
          width = width,
          scale = model_scale
        )
    }

    # Generic S3 dispatch so predict.cnormShash OR predict.cnormBetaBinomial is correctly used
    d$fitted <- predict(model, d$age, d$score)

  } else {
    stop("Please provide an object of type cnorm, cnormBetaBinomial, cnormBetaBinomial2, or cnormShash.")
  }

  if (!"normValue" %in% colnames(d)) {
    stop(
      "The 'normValue' column is missing from the data. Please ensure it's present for both cnorm and parametric models."
    )
  }

  d$diff <- d$fitted - d$normValue
  d <- d[!is.na(d$fitted) & !is.na(d$diff), ]

  rmse <- round(sqrt(mean(d$diff^2)), digits = 4)
  r <- round(cor(d$fitted, d$normValue, use = "pairwise.complete.obs"),
             digits = 4)

  if (type == 0) {
    if (is_taylor) {
      title <- if (isTRUE(group) ||
                   (is.character(group) &&
                    nzchar(group)))
        paste("Observed vs. Fitted Norm Scores by", group)
      else
        "Observed vs. Fitted Norm Scores"
    } else{
      title <- if (is.numeric(group))
        paste("Observed vs. Fitted Norm Scores by group")
      else
        "Observed vs. Fitted Norm Scores"
    }

    p <- ggplot(d, aes(x = .data$normValue, y = .data$fitted)) +
      geom_point(alpha = 0.2, color = "#0033AA") +
      geom_abline(
        intercept = 0,
        slope = 1,
        color = "red",
        linetype = "dashed"
      ) +
      labs(
        title = title,
        subtitle = paste("r =", r, ", RMSE =", rmse),
        x = "Observed Scores",
        y = "Fitted Scores"
      )
  } else {
    if (is_taylor) {
      title <- if (group != "" &&
                   !is.null(group))
        paste("Observed Norm Scores vs. Difference Scores by", group)
      else
        "Observed Norm Scores vs. Difference Scores"
    } else{
      title <- if (is.numeric(group))
        paste("Observed Norm Scores vs. Difference Scores by group")
      else
        "Observed Norm Scores vs. Difference Scores"
    }

    p <- ggplot(d, aes(x = .data$normValue, y = .data$diff)) +
      geom_point(alpha = 0.5, color = "#0033AA") +
      geom_hline(yintercept = 0,
                 color = "red",
                 linetype = "dashed") +
      labs(
        title = title,
        subtitle = paste("r =", r, ", RMSE =", rmse),
        x = "Observed Scores",
        y = "Difference"
      )
  }

  if (isTRUE(group) || (is.character(group) && nzchar(group))) {
    p <- p + facet_wrap( ~ group)
  }

  p <- p + theme_minimal() +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      plot.subtitle = element_text(hjust = 0.5, size = 12),
      axis.title = element_text(size = 12, face = "bold"),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  return(p)
}

#' @import ggplot2
#' @export
#' @family plot
#'
#' @title Plot norm curves
#'
#' @description
#' This function plots the norm curves based on the regression model. It supports both
#' Taylor polynomial models, beta-binomial models, and shash models.
#'
#' @param model The model from the bestModel function, a cnorm object, or a cnormBetaBinomial / cnormBetaBinomial2 / cnormShash object.
#' @param normList Vector with norm scores to display. If NULL, default values are used.
#' @param minAge Age to start with checking. If NULL, it's automatically determined from the model.
#' @param maxAge Upper end of the age check. If NULL, it's automatically determined from the model.
#' @param step Stepping parameter for the age check, usually 1 or 0.1; lower scores indicate higher precision.
#' @param minRaw Lower end of the raw score range, used for clipping implausible results. If NULL, it's automatically determined from the model.
#' @param maxRaw Upper end of the raw score range, used for clipping implausible results. If NULL, it's automatically determined from the model.
#'
#' @details
#' Please check the function for inconsistent curves: The different curves should not intersect.
#' Violations of this assumption are a strong indication of violations of model assumptions in
#' modeling the relationship between raw and norm scores.
#'
#' Common reasons for inconsistencies include:
#' 1. Vertical extrapolation: Choosing extreme norm scores (e.g., scores <= -3 or >= 3).
#' 2. Horizontal extrapolation: Using the model scores outside the original dataset.
#' 3. The data cannot be modeled with the current approach, or you need another power parameter (k) or R2 for the model.
#'
#' @return A ggplot object representing the norm curves.
#'
#' @seealso \code{\link{checkConsistency}}, \code{\link{plotDerivative}}, \code{\link{plotPercentiles}}
#'
#' @examples
#' \dontrun{
#' # For Taylor continuous norming model
#' m <- cnorm(raw = ppvt$raw, group = ppvt$group)
#' plotNormCurves(m, minAge=2, maxAge=5)
#'
#' # For beta-binomial model
#' bb_model <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
#' plotNormCurves(bb_model)
#' }
plotNormCurves <- function(model,
                           normList = NULL,
                           minAge = NULL,
                           maxAge = NULL,
                           step = 0.1,
                           minRaw = NULL,
                           maxRaw = NULL) {
  if (isTaylor(model)) {
    model <- model$model
  }

  parametric <- isParametric(model)
  is_beta_binomial <- isBeta(model)
  is_shash <- isSHASH(model)

  if (!parametric && !model$useAge) {
    stop("Age or group variable explicitly set to FALSE in dataset. No plotting available.")
  }

  # Get scale information
  if (parametric) {
    scaleMean <- attr(model$result, "scaleMean")
    scaleSD <- attr(model$result, "scaleSD")
  } else {
    scaleMean <- model$scaleM
    scaleSD <- model$scaleSD
  }

  if (is.null(normList)) {
    normList <- c(-2, -1, 0, 1, 2) * scaleSD + scaleMean
  }

  if (is.null(minAge)) {
    minAge <- if (parametric)
      attr(model$result, "age_mean") - 2 * attr(model$result, "age_sd")
    else
      model$minA1
  }

  if (is.null(maxAge)) {
    maxAge <- if (parametric)
      attr(model$result, "age_mean") + 2 * attr(model$result, "age_sd")
    else
      model$maxA1
  }

  if (is.null(minRaw)) {
    minRaw <- if (is_beta_binomial)
      0
    else if (is_shash)
      attr(model$result, "min")
    else
      model$minRaw
  }

  if (is.null(maxRaw)) {
    maxRaw <- if (parametric)
      attr(model$result, "max")
    else
      model$maxRaw
  }

  frame_list <- lapply(normList, function(norm) {
    if (parametric) {
      ages <- seq(minAge, maxAge, by = step)
      p_val <- pnorm((norm - scaleMean) / scaleSD)

      if (is_beta_binomial) {
        # Loop for backwards compat with un-vectorized underlying coefficient variants
        raws <- sapply(ages, function(a) {
          if (inherits(model, "cnormBetaBinomial")) {
            pred <- predictCoefficients(model, a)
          } else {
            pred <- predictCoefficients2(model, a, attr(model$result, "max"))
          }
          qbeta(p_val, pred$a, pred$b) * attr(model$result, "max")
        })
      } else if (is_shash) {
        # Predict parameters vectorized and calculate exact inverse quantiles
        preds <- predictCoefficients_shash(model, ages)
        raws <- qshash(p_val, mu = preds$mu, sigma = preds$sigma, epsilon = preds$epsilon, delta = preds$delta)
      }

      data.frame(n = norm, raw = raws, age = ages)
    } else {
      normCurve <- getNormCurve(
        norm,
        model,
        minAge = minAge,
        maxAge = maxAge,
        step = step,
        minRaw = minRaw,
        maxRaw = maxRaw
      )
      data.frame(n = norm,
                 raw = normCurve$raw,
                 age = normCurve$age)
    }
  })
  valueList <- do.call(rbind, frame_list)

  # Create rainbow color palette
  n_colors <- length(unique(valueList$n))
  color_palette <- rainbow(n_colors)

  # Create ggplot
  p <- ggplot(valueList, aes(
    x = .data$age,
    y = .data$raw,
    color = factor(.data$n)
  )) +
    geom_line(linewidth = 1) +
    scale_color_manual(
      name = "Norm Score",
      values = color_palette,
      labels = paste("Norm", normList)
    ) +
    labs(title = "Norm Curves", x = "Explanatory Variable (Age)", y = "Raw Score") +
    theme_minimal() +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      axis.title = element_text(size = 12, face = "bold"),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  return(p)
}

#' Cumulative norm distribution plot for conventional-norming models
#'
#' Used internally when `plotPercentiles()` is called on a cnorm object that
#' was built without age/group (i.e. `useAge == FALSE`). Plots raw score on
#' the x-axis against the model-implied percentile on the y-axis, with the
#' empirical cumulative distribution of the manifest sample overlaid.
#'
#' @param x A cnorm object.
#' @param minRaw,maxRaw Plot range on the raw-score axis. Defaults to the
#'   model's stored raw range.
#' @param title,subtitle Plot annotations; defaults match `plotPercentiles`.
#' @noRd
plotCumulative <- function(x,
                           minRaw   = NULL,
                           maxRaw   = NULL,
                           title    = NULL,
                           subtitle = NULL) {
  if (!isTaylor(x))
    stop("plotCumulative requires a cnorm object.")

  data    <- x$data
  m       <- x$model
  raw_var <- m$raw
  scaleM  <- m$scaleM
  scaleSD <- m$scaleSD

  if (is.null(minRaw))
    minRaw <- m$minRaw
  if (is.null(maxRaw))
    maxRaw <- m$maxRaw

  curve_df <- rawTable(0, x, minRaw, maxRaw, pretty = F)
  raw_obs <- data[[raw_var]]
  raw_obs <- raw_obs[!is.na(raw_obs)]
  n_obs   <- length(raw_obs)

  unique_raw <- sort(unique(raw_obs))
  manifest_df <- data.frame(raw        = unique_raw,
                            percentile = vapply(unique_raw, function(r) {
                              (sum(raw_obs < r) + 0.5 * sum(raw_obs == r)) / n_obs * 100
                            }, numeric(1)))

  if (is.null(title)) {
    title <- "Cumulative Norm Distribution (Conventional Norming)"
    subtitle <- bquote(paste("Model: ", .(m$ideal.model), ", R"^2, " = ", .(round(
      m$subsets$adjr2[[m$ideal.model]], digits = 4
    ))))
  }

  p <- ggplot() +
    geom_point(
      data = manifest_df,
      aes(x = .data$raw, y = .data$percentile),
      colour = "black",
      alpha = 0.6,
      size = 1.4
    ) +
    geom_line(
      data = curve_df,
      aes(x = .data$raw, y = .data$percentile),
      colour = "#1f77b4",
      linewidth = 0.9
    ) +
    labs(
      title    = title,
      subtitle = subtitle,
      x        = paste0("Raw Score (", raw_var, ")"),
      y        = "Percentile"
    ) +
    scale_y_continuous(limits = c(0, 100), breaks = seq(0, 100, by = 10)) +
    theme_minimal() +
    theme(
      plot.title       = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      plot.subtitle    = element_text(hjust = 0.5, size = 12),
      axis.title       = element_text(size = 12, face = "bold"),
      axis.title.x     = element_text(margin = margin(t = 10)),
      axis.title.y     = element_text(margin = margin(r = 10)),
      axis.text        = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )


  print(p)
  invisible(p)
}

#' Plot norm curves against actual percentiles
#'
#' The function plots the norm curves based on the regression model against
#' the actual percentiles from the raw data. As in 'plotNormCurves',
#' please check for inconsistent curves, especially intersections.
#' Violations of this assumption are a strong indication for problems
#' in modeling the relationship between raw and norm scores.
#' In general, extrapolation (point 1 and 2) can carefully be done to a
#' certain degree outside the original sample, but it should in general
#' be handled with caution.
#' The original percentiles are displayed as distinct points in the according
#' color, the model based projection of percentiles are drawn as lines.
#' Please note, that the estimation of the percentiles of the raw data is done with
#' the quantile function with the default settings.
#' In case, you get 'jagged' or disorganized percentile curve, try to reduce the 'k'
#' and/or 't' parameter in modeling.
#'
#' @param model The Taylor polynomial regression model object from the cNORM
#' @param minRaw Lower bound of the raw score (default = 0)
#' @param maxRaw Upper bound of the raw score
#' @param minAge Variable to restrict the lower bound of the plot to a specific age
#' @param maxAge Variable to restrict the upper bound of the plot to a specific age
#' @param raw The name of the raw variable
#' @param group The name of the grouping variable; the distinct groups are automatically
#' determined
#' @param percentiles Vector with percentile scores, ranging from 0 to 1 (exclusive)
#' @param scale The norm scale, either 'T', 'IQ', 'z', 'percentile' or
#' self defined with a double vector with the mean and standard deviation,
#' f. e. c(10, 3) for Wechsler scale index points; if NULL, scale information from the
#' data preparation is used (default)
#' @param title custom title for plot
#' @param subtitle custom title for plot
#' @param points Logical indicating whether to plot the data points. Default is TRUE.
#' @seealso plotNormCurves, plotPercentileSeries
#' @examples
#' \dontrun{
#'   # Load example data set, compute model and plot results
#'   result <- cnorm(raw = elfe$raw, group = elfe$group)
#'   plotPercentiles(result)
#' }
#' @export
#' @family plot
plotPercentiles <- function(model,
                            minRaw = NULL,
                            maxRaw = NULL,
                            minAge = NULL,
                            maxAge = NULL,
                            raw = NULL,
                            group = NULL,
                            percentiles = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975),
                            scale = NULL,
                            title = NULL,
                            subtitle = NULL,
                            points = FALSE) {
  if (isParametric(model)) {
    stop(
      "This function is not applicable for parametric models (Beta Binomial or Sinh-Arcsinh). ",
      "Please use 'plot(model, age, raw)' instead."
    )
  }

  if (isTaylor(model)) {
    data <- model$data
    m    <- model$model
  } else {
    stop("Please provide a cnorm object.")
  }

  if (!isTRUE(m$useAge)) {
    return(
      plotCumulative(
        model,
        minRaw      = minRaw,
        maxRaw      = maxRaw,
        title       = title,
        subtitle    = subtitle
      )
    )
  }


  if (is.null(group)) {
    group <- attr(data, "group")
  }

  age <- NULL
  if (is.null(data[[group]])) {
    age <- data[, attributes(data)$age]
    data$group <- getGroups(data[, attributes(data)$age])
    data$age <- data[, attributes(data)$age]
    group <- "group"
  }

  if (is.null(minAge)) {
    minAge <- m$minA1
  }

  if (is.null(maxAge)) {
    maxAge <- m$maxA1
  }

  if (is.null(minRaw)) {
    minRaw <- m$minRaw
  }

  if (is.null(maxRaw)) {
    maxRaw <- m$maxRaw
  }

  if (is.null(raw)) {
    raw <- m$raw
  }

  if (!(raw %in% colnames(data))) {
    stop(paste(
      c(
        "ERROR: Raw score variable '",
        raw,
        "' does not exist in data object."
      ),
      collapse = ""
    ))
  }

  if (!(group %in% colnames(data))) {
    stop(paste(
      c(
        "ERROR: Grouping variable '",
        group,
        "' does not exist in data object."
      ),
      collapse = ""
    ))
  }

  if (typeof(group) == "logical" && !group) {
    stop("The plotPercentiles-function does not work without a grouping variable.")
  }


  # compute norm scores from percentile vector
  if (is.null(scale)) {
    # fetch scale information from model
    T <- qnorm(percentiles, m$scaleM, m$scaleSD)
  } else if ((typeof(scale) == "double" && length(scale) == 2)) {
    T <- qnorm(percentiles, scale[1], scale[2])
  } else if (scale == "IQ") {
    T <- qnorm(percentiles, 100, 15)
  } else if (scale == "z") {
    T <- qnorm(percentiles)
  } else if (scale == "T") {
    T <- qnorm(percentiles, 50, 10)
  } else {
    # no transformation
    T <- percentiles
  }

  # generate variable names
  NAMES <- paste("PR", percentiles * 100, sep = "")
  NAMESP <- paste("PredPR", percentiles * 100, sep = "")

  # build function for xyplot and aggregate actual percentiles per group
  xyFunction <- paste(
    paste(NAMES, collapse = " + "),
    paste(NAMESP, collapse = " + "),
    sep = " + ",
    collapse = " + "
  )
  xyFunction <- paste(xyFunction, group, sep = " ~ ")

  w <- attributes(data)$weights
  data[, group] <- round(data[, group], digits = 3)
  AGEP <- unique(data[, group])

  # get actual percentiles
  if (!is.null(attr(data, "descend")) && attr(data, "descend")) {
    percentile.actual <- as.data.frame(do.call("rbind", lapply(split(data, data[, group]), function(df) {
      weighted.quantile(df[, raw],
                        probs = 1 - percentiles,
                        weights = df$w)
    })))
  } else{
    percentile.actual <- as.data.frame(do.call("rbind", lapply(split(data, data[, group]), function(df) {
      weighted.quantile(df[, raw], probs = percentiles, weights = df$w)
    })))
  }
  percentile.actual$group <- as.numeric(rownames(percentile.actual))
  colnames(percentile.actual) <- c(NAMES, c(group))
  rownames(percentile.actual) <- AGEP

  # build finer grained grouping variable for prediction and fit predicted percentiles
  share <- seq(from = minAge,
               to = maxAge,
               length.out = 100)
  AGEP <- c(AGEP, share)
  percentile.fitted <- data.frame(matrix(NA, nrow = length(AGEP), ncol = length(T)))

  norm_rep <- rep(T, times = length(AGEP))   # T1,T2,...,Tk, T1,T2,...,Tk, ...
  age_rep  <- rep(AGEP, each  = length(T))       # A1,A1,...,A1, A2,A2,...,A2, ...

  preds <- predictRaw(norm_rep,
                      age_rep,
                      m$coefficients,
                      minRaw = minRaw,
                      maxRaw = maxRaw)

  percentile.fitted <- as.data.frame(matrix(
    preds,
    nrow = length(AGEP),
    ncol = length(T),
    byrow = TRUE
  ))
  percentile.fitted$group <- AGEP
  percentile.fitted <- percentile.fitted[!duplicated(percentile.fitted$group), ]
  colnames(percentile.fitted)  <- c(NAMESP, group)
  rownames(percentile.fitted)  <- percentile.fitted$group

  # Merge actual and predicted scores and plot them show lines
  # for predicted scores and dots for actual scores
  percentile <- merge(percentile.actual,
                      percentile.fitted,
                      by = group,
                      all = TRUE)

  END <- .8
  COL1 <- rainbow(length(percentiles), end = END)
  COL2 <- c(rainbow(length(percentiles), end = END), rainbow(length(percentiles), end = END))


  if (is.null(title)) {
    title <- "Observed and Predicted Percentile Curves"
    subtitle <- bquote(paste("Model: ", .(m$ideal.model), ", R"^2, "=", .(round(
      m$subsets$adjr2[[m$ideal.model]], digits = 4
    ))))
  }

  # Prepare data for ggplot
  plot_data <- data.frame(
    group = rep(percentile$group, 2 * length(percentiles)),
    value = c(as.matrix(percentile[, NAMES]), as.matrix(percentile[, NAMESP])),
    type = rep(
      c("Observed", "Predicted"),
      each = nrow(percentile) * length(percentiles)
    ),
    percentile = factor(rep(rep(
      NAMES, each = nrow(percentile)
    ), 2), levels = NAMES)
  )

  plot_data_predicted <- plot_data[plot_data$type == "Predicted", ]
  plot_data_observed <- plot_data[plot_data$type == "Observed", ]

  # Create the ggplot - ALIGNED ORDER: Points first, then lines
  p <- ggplot()

  # Add raw scores FIRST if points is TRUE (matches plot.cnormShash)
  if (points) {
    if (is.null(age)) {
      p <- p + geom_point(
        data = data,
        aes(x = .data[[group]], y = .data[[raw]]),
        color = "black",
        alpha = 0.2,
        size = 0.6
      )  # Matched size
    } else{
      p <- p + geom_point(
        data = data,
        aes(x = .data$age, y = .data[[raw]]),
        color = "black",
        alpha = 0.2,
        size = 0.6
      )  # Matched size
    }
  }

  # Then add percentile lines and observed points
  p <- p +
    geom_line(
      data = plot_data_predicted,
      aes(
        x = .data$group,
        y = .data$value,
        color = .data$percentile
      ),
      linewidth = 0.6
    ) +  # Matched linewidth
    geom_point(
      data = plot_data_observed,
      aes(
        x = .data$group,
        y = .data$value,
        color = .data$percentile
      ),
      na.rm = TRUE,
      size = 2,
      shape = 18
    ) +
    labs(
      title = title,
      subtitle = subtitle,
      x = paste0("Explanatory Variable (", group, ")"),
      y = paste0("Raw Score (", raw, ")"),
      color = "Percentile"
    ) +  # Added legend title
    scale_color_manual(
      values = setNames(COL1, NAMES),
      labels = paste0(percentiles * 100, "%")  # Matched label format
    ) +
    guides(color = guide_legend(override.aes = list(
      linetype = rep("solid", length(NAMES)),
      shape = rep(18, length(NAMES))
    )))  # Matched legend override

  # Apply consistent theme
  p <- p + theme_minimal() +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      plot.subtitle = element_text(hjust = 0.5, size = 12),
      axis.title = element_text(size = 12, face = "bold"),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )
  print(p)
  invisible(p)
}



#' Plot the density function per group by raw score
#'
#' This function plots density curves based on the regression model against the raw scores.
#' It supports both traditional continuous norming models and beta-binomial models.
#' The function allows for customization of the plot range and groups to be displayed.
#'
#' @param model The model from the bestModel function, a cnorm object, a cnormBetaBinomial, a cnormBetaBinomial2 or
#'    cnormShash object.
#' @param minRaw Lower bound of the raw score. If NULL, it's automatically determined based on the model type.
#' @param maxRaw Upper bound of the raw score. If NULL, it's automatically determined based on the model type.
#' @param minNorm Lower bound of the norm score. If NULL, it's automatically determined based on the model type.
#' @param maxNorm Upper bound of the norm score. If NULL, it's automatically determined based on the model type.
#' @param group Numeric vector specifying the age groups to plot. If NULL, groups are automatically selected.
#'
#' @return A ggplot object representing the density functions.
#'
#' @details
#' The function generates density curves for specified age groups, allowing for easy comparison of score distributions
#' across different ages.
#'
#' For beta-binomial models, the density is based on the probability mass function, while for
#' traditional models, it uses a normal distribution based on the norm scores.
#'
#' @note
#' Please check for inconsistent curves, especially those showing implausible shapes
#' such as violations of biuniqueness in the cnorm models.
#'
#' @seealso \code{\link{plotNormCurves}}, \code{\link{plotPercentiles}}
#'
#' @examples
#' \dontrun{
#' # For traditional continuous norming model
#' result <- cnorm(raw = elfe$raw, group = elfe$group)
#' plotDensity(result, group = c(2, 4, 6))
#'
#' # For beta-binomial model
#' bb_model <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
#' plotDensity(bb_model)
#' }
#'
#' @import ggplot2
#' @export
#' @family plot
plotDensity <- function(model,
                        minRaw = NULL,
                        maxRaw = NULL,
                        minNorm = NULL,
                        maxNorm = NULL,
                        group = NULL) {
  if (isTaylor(model)) {
    model <- model$model
  }

  is_beta_binomial <- isBeta(model)
  is_shash <- isSHASH(model)

  if (is.null(minNorm)) {
    minNorm <- if (is_beta_binomial || is_shash)
      -3
    else
      model$minL1
  }

  if (is.null(maxNorm)) {
    maxNorm <- if (is_beta_binomial || is_shash)
      3
    else
      model$maxL1
  }

  if (is.null(minRaw)) {
    minRaw <- if (is_beta_binomial)
      0
    else if (is_shash)
      attr(model$result, "min")
    else
      model$minRaw
  }
  if (is.null(maxRaw)) {
    maxRaw <- if (is_beta_binomial || is_shash)
      attr(model$result, "max")
    else
      model$maxRaw
  }

  if (is.null(group)) {
    if (is_beta_binomial || is_shash) {
      age_min <- attr(model$result, "ageMin")
      age_max <- attr(model$result, "ageMax")
      group <- round(seq(
        from = age_min,
        to = age_max,
        length.out = 4
      ), digits = 3)
    } else if (model$useAge) {
      group <- round(seq(
        from = model$minA1,
        to = model$maxA1,
        length.out = 4
      ),
      digits = 3)
    } else {
      group <- c(1)
    }
  }

  step <- (maxNorm - minNorm) / 100

  # Calculate step size uniquely required for continuous density approximation for shash
  step_shash <- NULL
  if (is_shash) {
    step_shash <- (maxRaw - minRaw) / 100
    if (step_shash <= 0) step_shash <- 1
  }

  matrix_list <- lapply(group, function(g) {
    if (is_beta_binomial) {
      norm <- normTable.betabinomial(model, ages = g, n = attr(model$result, "max"))[[1]]
      norm$group <- rep(g, length.out = nrow(norm))
      colnames(norm)[colnames(norm) == "x"] <- "raw"
      colnames(norm)[colnames(norm) == "norm"] <- "norm1"
      colnames(norm)[colnames(norm) == "z"] <- "norm"
    } else if (is_shash) {
      norm <- normTable.shash(model, ages = g, start = minRaw, end = maxRaw, step = step_shash)[[1]]
      norm$group <- rep(g, length.out = nrow(norm))
      colnames(norm)[colnames(norm) == "x"] <- "raw"
      colnames(norm)[colnames(norm) == "norm"] <- "norm1"
      colnames(norm)[colnames(norm) == "z"] <- "norm"
    } else {
      norm <- normTable(
        g,
        model = model,
        minNorm = minNorm,
        maxNorm = maxNorm,
        minRaw = minRaw,
        maxRaw = maxRaw,
        step = step,
        pretty = FALSE
      )
      norm$group <- rep(g, length.out = nrow(norm))
    }
    return(norm)
  })

  matrix <- do.call(rbind, matrix_list)
  matrix <- matrix[matrix$norm > minNorm & matrix$norm < maxNorm, ]
  matrix <- matrix[matrix$raw > minRaw & matrix$raw < maxRaw, ]

  if (is_beta_binomial) {
    matrix$density <- matrix$Px
  } else if (is_shash) {
    # Re-scale back to pure density utilizing defined step_shash
    matrix$density <- matrix$Px / step_shash
  } else {
    matrix$density <- dnorm(matrix$norm,
                            mean = model$scaleM,
                            sd = model$scaleSD)
  }

  # Create ggplot
  title <- ""
  if (is_beta_binomial) {
    title <- "Density Functions (Beta-Binomial)"
  } else if (is_shash) {
    title <- "Density Functions (SHASH)"
  } else {
    title <- "Density Functions (Taylor Polynomial)"
  }

  matrix <- matrix[complete.cases(matrix), ]
  p <- ggplot(matrix, aes(
    x = .data$raw,
    y = .data$density,
    color = factor(group)
  )) +
    geom_line(linewidth = 1, na.rm = TRUE) +
    scale_color_viridis_d(name = "Group",
                          labels = paste("Group", group),
                          option = "plasma") +
    labs(title = title, x = "Raw Score", y = "Density") +
    theme_minimal() +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      axis.title = element_text(size = 12, face = "bold"),
      axis.text = element_text(size = 10),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  return(p)
}


#' Generates a series of plots with percentile curves for different models
#'
#' This function makes use of 'plotPercentiles' to generate a series of plots
#' for models with an increasing number of terms. It draws on the information
#' provided by the model object to determine the bounds of the modeling (age
#' and standard score range). It can be used as an additional model check to
#' determine the best fitting model. Please have a look at the
#' 'plotPercentiles' function for further information.
#'
#' Each model of the series is refitted on the complete norm sample (applying
#' case weights, if the norm data were post stratified). Models are identified
#' by their actual number of terms, which - after consistency screening in
#' \code{bestModel} - is not necessarily identical to the row number of the
#' model selection table. The subtitle of each plot reports the number of
#' terms, the adjusted R2 and, if available, the result of the consistency
#' check.
#'
#' @param model The Taylor polynomial regression model object or a cnorm object
#' @param start Number of terms to start with (default 1)
#' @param end Number of terms to end with; defaults to the largest available
#'   model
#' @param group The name of the grouping variable; the distinct groups are
#'   automatically determined
#' @param percentiles Vector with percentile scores, ranging from 0 to 1
#'   (exclusive)
#' @param filename Prefix of the filename. If specified, the plots are saved as
#'   png files in the directory of the workspace, instead of only displaying
#'   them. The number of terms is appended to the prefix.
#' @seealso plotPercentiles
#' @return A named list of plots (names indicate the number of terms),
#'   returned invisibly
#' @export
#'
#' @examples
#' \dontrun{
#'   # Load example data set, compute model and plot results
#'   result <- cnorm(raw = elfe$raw, group = elfe$group)
#'   plotPercentileSeries(result, start = 4, end = 6)
#' }
#'
#' @family plot
plotPercentileSeries <- function(model,
                                 start = 1,
                                 end = NULL,
                                 group = NULL,
                                 percentiles = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975),
                                 filename = NULL) {
  if (isParametric(model)) {
    stop("This function is not applicable for parametric models (Beta Binomial ",
         "or Sinh-Arcsinh). Please use the plotDensity function instead.")
  }

  if (isTaylor(model)) {
    d <- model$data
    model <- model$model
  } else {
    stop("Please provide a cnorm object.")
  }

  if (!isTRUE(attr(d, "useAge"))) {
    stop("Age or group variable explicitly set to FALSE in dataset. ",
         "No plotting available.")
  }

  subsets <- model$subsets
  outmat  <- subsets$outmat
  nTerms  <- rowSums(outmat == "*")   # actual model sizes (row index is
  # NOT reliable after screening)
  maxTerms <- max(nTerms)

  # sanitize requested range
  if (is.null(end) || end > maxTerms) end <- maxTerms
  if (start < 1) start <- 1
  if (start > end) start <- end

  # select the rows whose term count falls into the requested range;
  # after screening there is one model per size, otherwise take the first
  rows <- which(nTerms >= start & nTerms <= end)
  rows <- rows[!duplicated(nTerms[rows])]
  if (length(rows) == 0L) {
    stop("No models with ", start, " to ", end, " terms available.")
  }

  # case weights (post stratification), analogous to bestModel
  w <- if (!is.null(attr(d, "weights")) && !is.null(d$weights))
    d$weights else NULL

  # static model information, assembled once
  minR <- min(d[[model$raw]])
  maxR <- max(d[[model$raw]])

  fields <- list(
    ideal.model = model$ideal.model,
    cutoff  = model$cutoff,
    useAge  = model$useAge,
    minA1   = model$minA1,
    maxA1   = model$maxA1,
    minL1   = model$minL1,
    maxL1   = model$maxL1,
    minRaw  = minR,
    maxRaw  = maxR,
    raw     = model$raw,
    scaleSD = attr(d, "scaleSD"),
    scaleM  = attr(d, "scaleM"),
    descend = attr(d, "descend"),
    group   = attr(d, "group"),
    age     = attr(d, "age"),
    k       = attr(d, "k"),
    A       = attr(d, "A")
  )

  termNames <- colnames(outmat)
  l <- vector("list", length(rows))
  names(l) <- as.character(nTerms[rows])

  for (idx in seq_along(rows)) {
    row <- rows[idx]
    size <- nTerms[row]
    message("Plotting model with ", size, " terms ...")

    # refit the candidate model on the complete sample
    selected <- termNames[outmat[row, ] == "*"]
    f <- stats::reformulate(selected, response = model$raw)
    bestformula <- if (is.null(w)) stats::lm(f, data = d)
    else stats::lm(f, data = d, weights = w)

    # attach the static model information
    bestformula[names(fields)] <- fields

    result <- list(data = d, model = bestformula)
    class(result) <- "cnormTemp"

    # subtitle: actual size, fit and consistency information
    r2 <- round(subsets$adjr2[row], digits = 4)
    consInfo <- if (!is.null(subsets$consistent) && !is.na(subsets$consistent[row])) {
      if (subsets$consistent[row]) ", consistent" else ", inconsistent"
    } else {
      ""
    }

    l[[idx]] <- plotPercentiles(
      result,
      minAge = model$minA1,
      maxAge = model$maxA1,
      minRaw = minR,
      maxRaw = maxR,
      percentiles = percentiles,
      scale = NULL,
      group = group,
      title = "Observed and Predicted Percentiles",
      subtitle = bquote(paste("Model with ", .(size), " terms, ",
                              R^2, " = ", .(r2), .(consInfo)))
    )

    if (!is.null(filename)) {
      ggsave(
        filename = paste0(filename, size, ".png"),
        plot = l[[idx]],
        device = "png",
        width = 10,
        height = 7,
        dpi = 300
      )
    }
  }

  invisible(l)
}


#' #' Evaluate information criteria for regression model
#'
#' This function plots various information criteria and model fit statistics
#' against the number of terms or the adjusted R-squared, depending on the type
#' of plot selected. It helps in model selection by visualizing different
#' aspects of model performance. Models which did not pass the consistency
#' check are depicted with an empty circle; the automatically selected model is
#' highlighted in red. If BIC-weighted model averaging was applied
#' (\code{averaging = TRUE}), a caption indicates that the final coefficients
#' are a weighted combination of the consistent candidate models shown.
#'
#' @param model The regression model from the bestModel function or a cnorm object.
#' @param type Integer specifying the type of plot to generate:
#'   \itemize{
#'     \item 0: Adjusted R2 by number of terms (default)
#'     \item 1: Log-transformed Mallows's Cp by adjusted R2
#'     \item 2: Bayesian Information Criterion (BIC) by adjusted R2
#'     \item 3: Root Mean Square Error (RMSE) by number of terms
#'     \item 4: Residual Sum of Squares (RSS) by number of terms
#'     \item 5: F-test statistic for consecutive models by number of terms
#'     \item 6: p-value for model tests by number of terms
#'   }
#'
#' @return A ggplot object representing the selected information criterion plot.
#'
#' @details
#' The function generates different plots to help in model selection:
#'
#' - For types 1 and 2 (Mallows's Cp and BIC), look for the "elbow" in the curve
#'   where the information criterion begins to drop. This often indicates a good
#'   balance between model fit and complexity.
#' - For type 0 (Adjusted R2), higher values indicate better fit, but be cautious
#'   of overfitting with values approaching 1.
#' - For types 3 and 4 (RMSE and RSS), lower values indicate better fit.
#' - For type 5 (F-test), higher values suggest significant improvement with
#'   added terms.
#' - For type 6 (p-values), values below the significance level (typically 0.05)
#'   suggest significant improvement with added terms.
#'
#' The F-tests and p-values compare each model with the preceding (smaller) one,
#' with degrees of freedom based on the actual difference in the number of
#' parameters. After consistency screening, consecutive models may differ by
#' more than one term.
#'
#' @note
#' It's important to balance statistical measures with practical considerations
#' and to visually inspect the model fit using functions like
#' \code{plotPercentiles}.
#'
#' @seealso \code{\link{bestModel}}, \code{\link{plotPercentiles}}, \code{\link{printSubset}}
#'
#' @examples
#' \dontrun{
#' # Compute model with example data and plot information function
#' cnorm.model <- cnorm(raw = elfe$raw, group = elfe$group)
#' plotSubset(cnorm.model)
#'
#' # Plot BIC against adjusted R-squared
#' plotSubset(cnorm.model, type = 2)
#'
#' # Plot RMSE against number of terms
#' plotSubset(cnorm.model, type = 3)
#' }
#'
#' @import ggplot2
#' @export
#' @family plot
plotSubset <- function(model, type = 0) {
  if (isParametric(model)) {
    stop("This function is not applicable for parametric models ",
         "(Beta Binomial or Sinh-Arcsinh).")
  }

  if (isTaylor(model)) {
    model <- model$model
  }

  if (is.null(model$subsets)) {
    stop("The model object does not contain model selection information ('subsets').")
  }

  if (!(type %in% 0:6)) {
    warning("Unknown plot type; using type = 0 (adjusted R2).")
    type <- 0
  }

  subsets <- model$subsets
  nModels <- length(subsets$rss)
  n       <- length(model$fitted.values)

  # actual model complexity per row (row index != number of terms
  # after consistency screening!)
  nTerms  <- rowSums(subsets$outmat == "*")
  nParams <- nTerms + 1L

  # F-tests between consecutive models, df from actual parameter counts
  Fvals <- rep(NA_real_, nModels)
  pvals <- rep(NA_real_, nModels)
  if (nModels > 1L) {
    df1 <- diff(nParams)
    df1[df1 < 1L] <- NA                      # guard non-nested comparisons
    df2 <- n - nParams[-1L]
    Fs  <- (-diff(subsets$rss) / df1) / (subsets$rss[-1L] / df2)
    Fvals[-1L] <- Fs
    pvals[-1L] <- stats::pf(Fs, df1, df2, lower.tail = FALSE)
  }

  # consistency flags: NA (unscreened, e.g. custom predictors) shown as filled
  consistent <- subsets$consistent
  if (is.null(consistent)) consistent <- rep(TRUE, nModels)
  consistent[is.na(consistent)] <- TRUE

  cutoff <- if (!is.null(model$cutoff)) model$cutoff else .99
  sel    <- if (!is.null(model$ideal.model)) model$ideal.model else NA_integer_

  dat <- data.frame(
    adjr2 = subsets$adjr2,
    bic   = subsets$bic,
    cp    = subsets$cp,
    RSS   = subsets$rss,
    RMSE  = sqrt(subsets$rss / n),
    Fstat = Fvals,
    pval  = pvals,
    nr    = nTerms,
    consistency = factor(ifelse(consistent, "consistent", "inconsistent"),
                         levels = c("consistent", "inconsistent"))
  )

  # ---- plot configuration per type -----------------------------------------
  xlab_terms <- "Number of terms"
  xlab_r2    <- expression(paste("Adjusted ", R^2))

  cfg <- switch(type + 1L,
                list(x = "nr",    y = "adjr2",                                     # 0
                     title = expression(paste("Information Function: Adjusted ", R^2)),
                     xlab = xlab_terms,
                     ylab = expression(paste("Adjusted ", R^2))),
                list(x = "adjr2", y = "cp",                                        # 1
                     title = "Information Function: Mallows's Cp",
                     xlab = xlab_r2,
                     ylab = "Mallows's Cp"),
                list(x = "adjr2", y = "bic",                                       # 2
                     title = "Information Function: BIC",
                     xlab = xlab_r2,
                     ylab = "Bayesian Information Criterion (BIC)"),
                list(x = "nr",    y = "RMSE",                                      # 3
                     title = "Information Function: RMSE",
                     xlab = xlab_terms,
                     ylab = "Root Mean Square Error (Raw Score)"),
                list(x = "nr",    y = "RSS",                                       # 4
                     title = "Information Function: RSS",
                     xlab = xlab_terms,
                     ylab = "Residual Sum of Squares (RSS)"),
                list(x = "nr",    y = "Fstat",                                     # 5
                     title = "Information Function: F-test Statistics",
                     xlab = xlab_terms,
                     ylab = "F-test Statistics for Consecutive Models"),
                list(x = "nr",    y = "pval",                                      # 6
                     title = "Information Function: p-values",
                     xlab = xlab_terms,
                     ylab = expression(paste("p-values for Tests on ", R^2,
                                             " adj. of Consecutive Models")))
  )

  theme_custom <- theme_minimal() +
    theme(
      plot.title   = element_text(face = "bold", size = 16, hjust = 0.5),
      axis.title   = element_text(face = "bold", size = 12),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text    = element_text(size = 10),
      legend.title = element_blank(),
      legend.text  = element_text(size = 10),
      # show the shape legend only if inconsistent models are present
      legend.position = if (any(dat$consistency == "inconsistent"))
        "bottom" else "none",
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  # ---- base plot -------------------------------------------------------------
  plt <- ggplot(dat, aes(x = .data[[cfg$x]], y = .data[[cfg$y]])) +
    theme_custom +
    geom_line(color = "#1f77b4", linewidth = .75, na.rm = TRUE) +
    geom_point(aes(shape = .data$consistency),
               color = "#1f77b4", size = 2.5, na.rm = TRUE) +
    # named values: robust even when only one level is present
    scale_shape_manual(values = c(consistent = 16, inconsistent = 1),
                       drop = FALSE) +
    labs(title = cfg$title, x = cfg$xlab, y = cfg$ylab, shape = NULL)

  # highlight the selected model
  if (!is.na(sel) && sel >= 1L && sel <= nModels) {
    plt <- plt +
      geom_point(data = dat[sel, , drop = FALSE],
                 aes(shape = .data$consistency),
                 color = "#3322AA", size = 2.5, stroke = 1.1, na.rm = TRUE)
  }

  # ---- type-specific decorations ----------------------------------------------
  if (type == 0) {
    plt <- plt +
      geom_hline(yintercept = cutoff, linetype = "dashed",
                 linewidth = .8, color = "#d62728")
  } else if (type == 1) {
    # Mallows's Cp can legitimately be <= 0; log scale only when valid
    if (all(dat$cp > 0, na.rm = TRUE)) {
      plt <- plt + scale_y_log10() +
        labs(y = "log-transformed Mallows's Cp")
    } else {
      message("Mallows's Cp contains non-positive values; ",
              "using a linear scale instead of log10.")
    }
  } else if (type == 6) {
    plt <- plt +
      geom_hline(yintercept = 0.05, linetype = "dashed",
                 linewidth = 1, color = "#d62728") +
      # clip the view without silently removing observations (unlike ylim)
      coord_cartesian(ylim = c(-0.005, 0.11))
  }

  # indicate model averaging
  if (isTRUE(model$averaged)) {
    plt <- plt + labs(caption = paste0(
      "Final coefficients: BIC-weighted average over ",
      length(model$averagingWeights), " consistent candidate models."))
  }

  return(plt)
}

#'
#' @title Plot first order derivative of regression model
#'
#' @description
#' This function plots the scores obtained via the first order derivative of the regression model
#' in dependence of the norm score.
#'
#' @param model The model from the bestModel function, a cnorm object.
#' @param minAge Minimum age to start checking. If NULL, it's automatically determined from the model.
#' @param maxAge Maximum age for checking. If NULL, it's automatically determined from the model.
#' @param minNorm Lower end of the norm score range. If NULL, it's automatically determined from the model.
#' @param maxNorm Upper end of the norm score range. If NULL, it's automatically determined from the model.
#' @param stepAge Stepping parameter for the age check, usually 1 or 0.1; lower values indicate higher precision.
#' @param stepNorm Stepping parameter for norm scores.
#' @param order Degree of the derivative (default = 1).
#'
#' @details
#' The results indicate the progression of the norm scores within each age group. The regression-based
#' modeling approach relies on the assumption of a linear progression of the norm scores. Negative scores
#' in the first order derivative indicate a violation of this assumption. Scores near zero are typical
#' for bottom and ceiling effects in the raw data.
#'
#' The regression models usually converge within the range of the original values. In case of vertical
#' and horizontal extrapolation, with increasing distance to the original data, the risk of assumption
#' violation increases as well.
#'
#' @note
#' This function is currently incompatible with reversed raw score scales ('descent' option).
#'
#' @return A ggplot object representing the derivative of the regression function.
#'
#' @seealso \code{\link{checkConsistency}}, \code{\link{bestModel}}, \code{\link{derive}}
#'
#' @examples
#' \dontrun{
#' # For traditional continuous norming model
#' result <- cnorm(raw = elfe$raw, group = elfe$group)
#' plotDerivative(result, minAge=2, maxAge=5, stepAge=.2, minNorm=25, maxNorm=75, stepNorm=1)
#' }
#'
#' @import ggplot2
#' @export
#' @family plot
plotDerivative <- function(model,
                           minAge = NULL,
                           maxAge = NULL,
                           minNorm = NULL,
                           maxNorm = NULL,
                           stepAge = NULL,
                           stepNorm = NULL,
                           order = 1) {
  if (isTaylor(model)) {
    model <- model$model
  } else if (isParametric(model)) {
    stop(
      "This function is not applicable for parametric models (Beta Binomial or Sinh-Arcsinh). Please use the plotDensity function instead."
    )
  }

  if (!model$useAge) {
    stop("Age or group variable explicitly set to FALSE in dataset. No plotting available.")
  }

  if (is.null(minAge)) {
    minAge <- model$minA1
  }

  if (is.null(maxAge)) {
    maxAge <- model$maxA1
  }

  if (is.null(minNorm)) {
    minNorm <- model$minL1
  }

  if (is.null(maxNorm)) {
    maxNorm <- model$maxL1
  }

  if (is.null(stepAge)) {
    stepAge <- (maxAge - minAge) / 100
  }

  if (is.null(stepNorm)) {
    stepNorm <- (maxNorm - minNorm) / 100
  }

  if (order <= 0)
    stop("Order of derivative must be a positive integer.")

  rowS <- seq(minNorm, maxNorm, by = stepNorm)
  colS <- seq(minAge, maxAge, by = stepAge)

  coeff <- derive(model, order)
  if (length(coeff) == 0) {
    stop("Derivative of order ", order, " not available for this model.")
  }

  cat(
    paste0(
      rangeCheck(model, minAge, maxAge, minNorm, maxNorm),
      " Coefficients from the ",
      order,
      " order derivative function:\n\n"
    )
  )
  print(coeff)


  dev2 <- expand.grid(X = rowS, Y = colS)
  dev2$Z <- mapply(function(norm, age)
    predictRaw(norm, age, coeff), dev2$X, dev2$Y)

  ordinal <- if (order <= 3)
    c("st", "nd", "rd")[order]
  else
    "th"
  desc <- paste0(order, ordinal, " Order Derivative")
  custom_palette <- c(
    "#FF0000",
    "#FF4000",
    "#FF8000",
    "#FFBF00",
    "#FFFF00",
    "#80FF00",
    "#00FF00",
    "#00FF80",
    "#00FFFF",
    "#0080FF",
    "#0000FF",
    "#4B0082",
    "#8B00FF"
  )
  theme_custom <- theme_minimal() +
    theme(
      plot.title = element_text(
        face = "bold",
        size = 14,
        hjust = 0.5
      ),
      axis.title = element_text(face = "bold", size = 12),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 8),
      legend.position = "right",
      legend.text = element_text(size = 8),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  p <- ggplot(dev2, aes(x = .data$Y, y = .data$X, z = .data$Z)) +
    geom_tile(aes(fill = .data$Z)) +
    geom_contour(color = "white", alpha = 0.5) +
    scale_fill_gradientn(colors = custom_palette) +
    labs(
      title = "Slope of the Regression Function",
      x = "Explanatory Variable (Age)",
      y = paste("Norm Score - ", desc),
      fill = "Derivative"
    ) +
    theme_custom +
    theme(legend.position = "right")

  if (min(dev2$Z) < 0 && max(dev2$Z) > 0)
    p <- p + geom_contour(
      aes(z = .data$Z),
      color = "black",
      linewidth = 0.5,
      breaks = 0,
      linetype = "dashed"
    )

  return(p)
}

#' General convenience plotting function
#'
#' @param x a cnorm object
#' @param y the type of plot as a string, can be one of
#' 'raw' (1), 'norm' (2), 'curves' (3), 'percentiles' (4), 'series' (5), 'subset' (6),
#' or 'derivative' (7), either as a string or the according index
#' @param ... additional parameters for the specific plotting function
#'
#' @export
plotCnorm <- function(x, y, ...) {
  if (!isTaylor(x)) {
    message("Please provide a cnorm object as x.")
    return(invisible(NULL))
  }
  if (!is.character(y) && !is.numeric(y)) {
    message("y must be a plot-type string or integer index.")
    return(invisible(NULL))
  }
  if (y == "raw"         || y == 1)
    plotRaw(x, ...)
  else if (y == "norm"        || y == 2)
    plotNorm(x, ...)
  else if (y == "curves"      || y == 3)
    plotNormCurves(x, ...)
  else if (y == "percentiles" || y == 4)
    plotPercentiles(x, ...)
  else if (y == "density"     || y == 5)
    plotDensity(x, ...)
  else if (y == "series"      ||
           y == 6)
    plotPercentileSeries(x, ...)
  else if (y == "subset"      || y == 7)
    plotSubset(x, ...)
  else if (y == "derivative"  || y == 8)
    plotDerivative(x, ...)
  else
    stop("Unkown plot type")
}

#' Compare Two Norm Models Visually
#'
#' This function creates a visualization comparing two norm models by displaying
#' their percentile curves. The first model is shown with solid lines, the second
#' with dashed lines. If age and score vectors are provided, manifest percentiles
#' are displayed as dots. The function works with regular cnorm models, beta-binomial
#' models, and shash models, allowing comparison between different model types.
#'
#' For beta-binomial models, the exact quantiles of the discrete beta-binomial
#' distribution are displayed by default as step functions (\code{discrete = TRUE}).
#' Setting \code{discrete = FALSE} draws smooth lines based on the quantiles of
#' the underlying beta (mixing) distribution instead. Note that this continuous
#' approximation omits the binomial stage of the variance and therefore displays
#' less spread than the fitted model actually implies, particularly in the outer
#' percentiles. The parameter has no effect on Taylor polynomial or shash models.
#'
#' @param model1 First model object (distribution free, beta-binomial, or shash)
#' @param model2 Second model object (distribution free, beta-binomial, or shash)
#' @param age Optional vector with manifest age or group values
#' @param score Optional vector with manifest raw score values
#' @param weights Optional vector with manifest weights
#' @param percentiles Vector with percentile scores, ranging from 0 to 1 (exclusive)
#' @param title Custom title for plot (optional)
#' @param subtitle Custom subtitle for plot (optional)
#' @param discrete Logical indicating whether beta-binomial models are displayed
#'   with their exact discrete quantiles as step functions (TRUE, default) or
#'   with a smooth continuous approximation via the underlying beta
#'   distribution (FALSE). Ignored for other model types.
#'
#' @return A ggplot object showing the comparison of both models
#'
#' @examples
#' \dontrun{
#' # Compare different types of models
#' model1 <- cnorm(group = elfe$group, raw = elfe$raw)
#' model2 <- cnorm.betabinomial(elfe$group, elfe$raw)
#' model3 <- cnorm.shash(elfe$group, elfe$raw)
#'
#' # Compare traditional cnorm with shash
#' compare(model1, model3, age = elfe$group, score = elfe$raw)
#'
#' # Compare beta-binomial with shash
#' compare(model2, model3, age = elfe$group, score = elfe$raw)
#' }
#'
#' @export
#' @family plot
compare <- function(model1,
                    model2,
                    percentiles = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975),
                    age = NULL,
                    score = NULL,
                    weights = NULL,
                    title = NULL,
                    subtitle = NULL,
                    discrete = TRUE) {
  # retrieve score from model if score is null and one of the
  # models is a cnorm object
  if (is.null(score) && isTaylor(model1)) {
    score <- model1$data[[attributes(model1$data)$raw]]
    age <- model1$data[[attributes(model1$data)$age]]
  }

  if (is.null(score) && isTaylor(model2)) {
    score <- model2$data[[attributes(model2$data)$raw]]
    age <- model2$data[[attributes(model2$data)$age]]
  }

  # Function to get predictions for beta-binomial models
  get_bb_predictions <- function(model, pred_ages) {
    if (inherits(model, "cnormBetaBinomial")) {
      preds <- predictCoefficients(model, pred_ages)
    } else {
      preds <- predictCoefficients2(model, pred_ages)
    }

    n_max <- attr(model$result, "max")
    pred_matrix <- matrix(NA_real_,
                          nrow = length(pred_ages),
                          ncol = length(percentiles))

    if (discrete) {
      # Exact quantiles of the discrete beta-binomial distribution;
      # one pmf evaluation per age, all percentiles read from it
      for (j in seq_along(pred_ages)) {
        dist <- bb_distribution(preds$a[j], preds$b[j], n_max)
        if (!anyNA(dist$cum)) {
          pred_matrix[j, ] <- vapply(percentiles, function(p) {
            as.numeric(dist$x[which.max(dist$cum >= p)])
          }, numeric(1))
        }
      }
    } else {
      # Continuous approximation via the underlying beta (mixing)
      # distribution; omits the binomial-stage variance
      for (i in seq_along(percentiles)) {
        pred_matrix[, i] <- qbeta(percentiles[i],
                                  shape1 = preds$a,
                                  shape2 = preds$b) * n_max
      }
    }

    pred_data <- data.frame(age = pred_ages, pred_matrix)
    names(pred_data)[-1] <- paste0("P", percentiles * 100)
    return(pred_data)
  }

  # Function to get predictions for shash models
  get_shash_predictions <- function(model, pred_ages) {
    preds <- predictCoefficients_shash(model, pred_ages)

    pred_matrix <- matrix(NA,
                          nrow = length(pred_ages),
                          ncol = length(percentiles))
    for (i in seq_along(percentiles)) {
      pred_matrix[, i] <- qshash(
        percentiles[i],
        mu = preds$mu,
        sigma = preds$sigma,
        epsilon = preds$epsilon,
        delta = preds$delta
      )
    }

    pred_data <- data.frame(age = pred_ages, pred_matrix)
    names(pred_data)[-1] <- paste0("P", percentiles * 100)
    return(pred_data)
  }

  # Function to get predictions for cnorm models
  get_cnorm_predictions <- function(model, pred_ages) {
    m <- model$model
    T <- qnorm(percentiles, m$scaleM, m$scaleSD)

    pred_matrix <- matrix(NA,
                          nrow = length(pred_ages),
                          ncol = length(percentiles))
    for (i in 1:length(pred_ages)) {
      pred_matrix[i, ] <- predictRaw(T, pred_ages[i], m$coefficients)
    }

    pred_data <- data.frame(age = pred_ages, pred_matrix)
    names(pred_data)[-1] <- paste0("P", percentiles * 100)
    return(pred_data)
  }

  # Determine age range
  get_age_range <- function(model) {
    if (isParametric(model)) {
      return(c(
        attr(model$result, "ageMin"),
        attr(model$result, "ageMax")
      ))
    } else {
      m <- model$model
      return(c(m$minA1, m$maxA1))
    }
  }

  # Get age ranges for both models
  range1 <- get_age_range(model1)
  range2 <- get_age_range(model2)

  # Create common age sequence
  pred_ages <- seq(min(range1[1], range2[1]), max(range1[2], range2[2]), length.out = 100)

  # Get predictions for both models; remember which models are displayed
  # as step functions (discrete beta-binomial quantiles)
  step1 <- isBeta(model1) && discrete
  step2 <- isBeta(model2) && discrete

  plot_data1 <- if (isBeta(model1)) {
    get_bb_predictions(model1, pred_ages)
  } else if (isSHASH(model1)) {
    get_shash_predictions(model1, pred_ages)
  } else {
    get_cnorm_predictions(model1, pred_ages)
  }

  plot_data2 <- if (isBeta(model2)) {
    get_bb_predictions(model2, pred_ages)
  } else if (isSHASH(model2)) {
    get_shash_predictions(model2, pred_ages)
  } else {
    get_cnorm_predictions(model2, pred_ages)
  }

  # Prepare data for plotting (reshape to long format using base R)
  plot_data_long <- data.frame(
    age = numeric(),
    value = numeric(),
    percentile = character(),
    model = character()
  )

  # Reshape data for model 1
  for (i in 2:ncol(plot_data1)) {
    plot_data_long <- rbind(
      plot_data_long,
      data.frame(
        age = plot_data1$age,
        value = plot_data1[[i]],
        percentile = names(plot_data1)[i],
        model = "Model 1"
      )
    )
  }

  # Reshape data for model 2
  for (i in 2:ncol(plot_data2)) {
    plot_data_long <- rbind(
      plot_data_long,
      data.frame(
        age = plot_data2$age,
        value = plot_data2[[i]],
        percentile = names(plot_data2)[i],
        model = "Model 2"
      )
    )
  }

  if (!is.null(score)) {
    plot_data_long$value[plot_data_long$value < min(score)] <- min(score)
    plot_data_long$value[plot_data_long$value > max(score)] <- max(score)
  }

  # Set factor levels for correct ordering
  plot_data_long$percentile <- factor(plot_data_long$percentile, levels = paste0("P", percentiles * 100))

  # Set default title if none provided
  if (is.null(title)) {
    title <- "Visual Model Comparison"
  }

  if (is.null(subtitle)) {
    subtitle <- "Model 1: solid lines, Model 2: dashed lines"
  }

  # Layer helper: piecewise-constant discrete quantiles are rendered with
  # geom_step (vertical risers, direction "mid"), continuous curves with
  # geom_line
  model_layer <- function(dat, lty, use_step) {
    if (use_step) {
      geom_step(
        data = dat,
        aes(x = .data$age, y = .data$value, color = .data$percentile),
        direction = "mid",
        linetype = lty,
        linewidth = 0.6
      )
    } else {
      geom_line(
        data = dat,
        aes(x = .data$age, y = .data$value, color = .data$percentile),
        linetype = lty,
        linewidth = 0.6
      )
    }
  }

  # Create plot
  p <- ggplot() +
    model_layer(plot_data_long[plot_data_long$model == "Model 1", ],
                "solid", step1) +
    model_layer(plot_data_long[plot_data_long$model == "Model 2", ],
                "dashed", step2) +
    scale_color_manual(values = rainbow(length(percentiles)),
                       labels = paste0(percentiles * 100, "%")) +
    labs(
      title = title,
      subtitle = subtitle,
      x = "Age",
      y = "Score",
      color = "Percentile"
    ) +
    theme_minimal() +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        size = 16,
        face = "bold"
      ),
      plot.subtitle = element_text(hjust = 0.5, size = 12),
      axis.title = element_text(size = 12, face = "bold"),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.position = "right",
      legend.title = element_text(size = 12, face = "bold"),
      legend.text = element_text(size = 10),
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  # Information criteria
  if (isTaylor(model1)) {
    ideal.model <- model1$model$ideal.model
    rss <- model1$model$subsets$rss[ideal.model]
    n <- nrow(model1$data)

    sigma2 <- rss / (nrow(model1$data) - ideal.model - 1)  # residual variance
    loglik <- -0.5 * n * (log(2 * pi) + log(sigma2) + 1)

    AIC1 <- -2 * loglik + 2 * ideal.model
    BIC1 <- model1$model$subsets$bic[ideal.model]
  } else{
    n_obs <- attr(model1$result, "N")
    n_params <- length(model1$result$par)
    log_likelihood <- -model1$result$value
    AIC1 <- 2 * n_params - 2 * log_likelihood
    BIC1 <- n_params * log(n_obs) - 2 * log_likelihood
  }

  if (isTaylor(model2)) {
    ideal.model <- model2$model$ideal.model
    rss <- model2$model$subsets$rss[ideal.model]
    n <- nrow(model2$data)

    sigma2 <- rss / (nrow(model2$data) - ideal.model - 1)  # residual variance
    loglik <- -0.5 * n * (log(2 * pi) + log(sigma2) + 1)

    AIC2 <- -2 * loglik + 2 * ideal.model
    BIC2 <- model2$model$subsets$bic[ideal.model]
  } else{
    n_obs <- attr(model2$result, "N")
    n_params <- length(model2$result$par)
    log_likelihood <- -model2$result$value
    AIC2 <- 2 * n_params - 2 * log_likelihood
    BIC2 <- n_params * log(n_obs) - 2 * log_likelihood
  }

  if (!is.null(score) & !is.null(age)) {
    # Prepare data for manifest percentiles and fit statistics
    data <- data.frame(age = age, score = score)
    if (!is.null(weights)) {
      data$w <- weights
    } else{
      data$w <- rep(1, length(age))
    }

    # Calculate groups for manifest percentiles
    if (length(age) / length(unique(age)) > 50 &&
        min(table(data$age)) > 30) {
      data$group <- age
    } else {
      data$group <- getGroups(age)
    }

    # Calculate manifest percentiles
    percentile.actual <- as.data.frame(do.call("rbind", lapply(split(
      data, data$group
    ), function(df) {
      c(age = mean(df$age),
        weighted.quantile(df$score, probs = percentiles, weights = df$w))
    })))
    colnames(percentile.actual) <- c("age", paste0("P", percentiles * 100))

    # Reshape manifest data
    manifest_data_long <- data.frame(age = numeric(),
                                     value = numeric(),
                                     percentile = character())

    for (i in 2:ncol(percentile.actual)) {
      manifest_data_long <- rbind(
        manifest_data_long,
        data.frame(
          age = percentile.actual$age,
          value = percentile.actual[[i]],
          percentile = names(percentile.actual)[i]
        )
      )
    }

    manifest_data_long$percentile <- factor(manifest_data_long$percentile,
                                            levels = paste0("P", percentiles * 100))

    # Add manifest percentiles to plot
    p <- p + geom_point(
      data = manifest_data_long,
      aes(
        x = .data$age,
        y = .data$value,
        color = .data$percentile
      ),
      size = 2,
      shape = 18
    )

    # Calculate fit statistics
    if (is.null(weights)) {
      data <- rankByGroup(data, raw = "score", group = "group")
    } else{
      data <- rankByGroup(data,
                          raw = "score",
                          group = "group",
                          weights = "w")
    }
    data$normValue <- 10 * (data$normValue - attributes(data)$scaleMean) / attributes(data)$scaleSD

    # Get predictions for both models
    if (isTaylor(model1)) {
      data$fitted1 <- predictNorm(
        data$score,
        data$age,
        model1,
        minNorm = model1$model$minL1,
        maxNorm = model1$model$maxL1
      )
      data$fitted1 <- 10 * (data$fitted1 - attributes(model1$data)$scaleMean) / attributes(model1$data)$scaleSD
    } else if (isParametric(model1)) {
      data$fitted1 <- predict(model1, data$age, data$score)
      scaleMean <- attr(model1$result, "scaleMean")
      scaleSD <- attr(model1$result, "scaleSD")
      data$fitted1 <- 10 * (data$fitted1 - scaleMean) / scaleSD
    }

    if (isTaylor(model2)) {
      data$fitted2 <- predictNorm(
        data$score,
        data$age,
        model2,
        minNorm = model2$model$minL1,
        maxNorm = model2$model$maxL1
      )
      data$fitted2 <- 10 * (data$fitted2 - attributes(model2$data)$scaleMean) / attributes(model2$data)$scaleSD
    } else if (isParametric(model2)) {
      data$fitted2 <- predict(model2, data$age, data$score)
      scaleMean <- attr(model2$result, "scaleMean")
      scaleSD <- attr(model2$result, "scaleSD")
      data$fitted2 <- 10 * (data$fitted2 - scaleMean) / scaleSD
    }

    # Calculate fit statistics
    R2a <- cor(data$fitted1, data$normValue, use = "pairwise.complete.obs")^2
    R2b <- cor(data$fitted2, data$normValue, use = "pairwise.complete.obs")^2

    bias1 <- mean(data$fitted1 - data$normValue, na.rm = TRUE)
    bias2 <- mean(data$fitted2 - data$normValue, na.rm = TRUE)

    RMSE1 <- sqrt(mean((data$fitted1 - data$normValue)^2, na.rm = TRUE))
    RMSE2 <- sqrt(mean((data$fitted2 - data$normValue)^2, na.rm = TRUE))

    MAD1 <- mean(abs(data$fitted1 - data$normValue), na.rm = TRUE)
    MAD2 <- mean(abs(data$fitted2 - data$normValue), na.rm = TRUE)


    # Create and print summary table
    fit_table <- data.frame(
      Metric = c("R2", "Bias", "RMSE", "MAD", "AIC", "BIC"),
      Model1 = c(R2a, bias1, RMSE1, MAD1, AIC1, BIC1),
      Model2 = c(R2b, bias2, RMSE2, MAD2, AIC2, BIC2),
      Difference = c(
        R2b - R2a,
        bias2 - bias1,
        RMSE2 - RMSE1,
        MAD2 - MAD1,
        AIC2 - AIC1,
        BIC2 - BIC1
      )
    )

    # Round values
    fit_table[, 2:4] <- round(fit_table[, 2:4], 4)

    cat("\nModel Comparison Summary:\n")
    cat("------------------------\n")
    print(format(fit_table, justify = "right"), row.names = FALSE)
    cat("\nNote: Difference = Model2 - Model1\n")
    cat("      Fit indices are based on the manifest and fitted norm scores of both models.\n")
    cat("      Scale metrics are T scores (scaleSD = 10)\n")
    cat("      AIC and BIC should only be used when comparing models of the same type.\n")
  } else{
    # Create and print summary table
    fit_table <- data.frame(
      Metric = c("AIC", "BIC"),
      Model1 = c(AIC1, BIC1),
      Model2 = c(AIC2, BIC2),
      Difference = c(AIC2 - AIC1, BIC2 - BIC1)
    )

    cat("\nModel Comparison Summary:\n")
    cat("------------------------\n")
    print(format(fit_table, justify = "right"), row.names = FALSE)
    cat("\nNote: Difference = Model2 - Model1\n")
    cat("      AIC and BIC should only be used when comparing models of the same type.\n")
  }

  return(p)
}

## ----setup, include = FALSE-------------------------------
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
library(cNORM)
library(ggplot2)


## ----echo=FALSE, fig.align='center', fig.width=10---------
knitr::include_graphics("plos.png")

## ----eval=FALSE-------------------------------------------
#  ## Basic example code for modeling the sample dataset
# 
# library(cNORM)
# 
# # Start the graphical user interface (needs shiny installed)
# # The GUI includes the most important functions. For specific cases,
# # please use cNORM on the console.
# 
# cNORM.GUI()
# 
# # Using the syntax on the console: The function 'cnorm' performs
# # all steps automatically. Please specify the raw score and the
# # grouping variable. The resulting object contains the ranked data
# # via object$data and the model via object$model.
# 
# cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
# 
# # Plot different indicators of model fit depending on the number of
# # predictors
# 
# plot(cnorm.elfe, "subset", type=0) # plot R2
# plot(cnorm.elfe, "subset", type=3) # plot MSE
# 
# # NOTE! At this point, you usually select a good fitting model and rerun
# # the process with a fixed number of terms, e. g. 4. Avoid models
# # with a high number of terms:
# 
# cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group, terms = 4)
# 
# # Powers of age can be specified via the parameter 't'.
# # Cubic modeling is usually sufficient, i.e., t = 3.
# # In contrast, 'k' specifies the power of the person location.
# # This parameter should be somewhat higher, e.g., k = 5.
# 
# cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group, k = 5, t = 3)
# 
# # Visual inspection of the percentile curves of the fitted model
# 
# plot(cnorm.elfe, "percentiles")
# 
# # Visual inspection of the observed and fitted raw and norm scores
# 
# plot(cnorm.elfe, "norm")
# plot(cnorm.elfe, "raw")
# 
# # In order to compare different models, generate a series of percentile
# # plots with an ascending number of predictors, in this example between
# # 5 and 14 predictors.
# 
# plot(cnorm.elfe, "series", start=5, end=14)
# 
# # Cross validation in order to choose appropriate number of terms
# # with 80% of the data for training and 20% for validation. Due to
# # the time consumption, the maximum number of terms is limited to 10
# # in this example with 3 repetitions of the cross validation.
# 
# cnorm.cv(cnorm.elfe$data, max=10, repetitions=3)
# 
# # Cross validation with prespecified terms of an already
# # existing model
# 
# cnorm.cv(cnorm.elfe, repetitions=3)
# 
# # Print norm table (in this case: 0, 3 or 6 months at grade level 3)
# # (Note: The data is coded such that 3.0 represents the beginning and
# # 3.5 the middle of the third school year)
# 
# normTable(c(3, 3.25, 3.5), cnorm.elfe)
# 

## ----eval=FALSE-------------------------------------------
# # Application of cNORM for the generation of conventional norms
# # for a specific age group (in this case age group 3):
# 
# data <- elfe[elfe$group == 3,]
# cnorm(raw=data$raw)

## ----eval=FALSE-------------------------------------------
#     # Creates a grouping variable for the variable 'age'
#     # of the ppvt data set. In this example, 12 equidistant
#     # subgroups are generated.
# 
#     group <- getGroups(ppvt$age, 12)

## ----fig0, fig.height = 4, fig.width = 7------------------
library(cNORM)
# Models the 'raw' variable as a function of the discrete 'group' variable
model <- cnorm(raw = elfe$raw, group = elfe$group)

## ---------------------------------------------------------
print(model)

## ----eval=FALSE-------------------------------------------
# # Generates a series of percentile plots with increasing number of predictors
# 
# plotPercentileSeries(model, start = 1, end = 15)

## ----fig1, fig.height = 4, fig.width = 7------------------
    plot(model, "subset", type = 0)

## ---------------------------------------------------------
predictNorm(15, 4.7, model, minNorm = 25, maxNorm = 75)


## ---------------------------------------------------------
predictRaw(55, 4.5, model, minRaw = 0, maxRaw = 28)

# ... or for several norm scores and age levels ...
predictRaw(c(45, 50, 55), c(2.5, 3, 3.5), model) 

## ---------------------------------------------------------
normTable(3, model, minRaw = 0, maxRaw = 28, minNorm=30.5, maxNorm=69.5, step = 1)

## ---------------------------------------------------------
rawTable(3.5, model, minRaw = 0, maxRaw = 28, minNorm = 25, maxNorm = 75, step = 1, CI = .95, reliability = .89)

# generate several raw tables
table <- rawTable(c(2.5, 3.5, 4.5), model, minRaw = 0, maxRaw = 28)

## ----irt-cnorm, eval = FALSE------------------------------
# library(TAM)
# 
# # 1. Simulate a Rasch dataset with 700 persons and 30 items
# data <- simulateRasch(n = 700, items.n = 30)
# data <- data$sim[, (ncol(data$sim) - 30):(ncol(data$sim) - 1)] # only use item data
# 
# # 2. Fit a Rasch model and extract WLE person abilities
# tam.model <- tam(data)
# theta     <- tam.wle(tam.model)$theta
# 
# # 3. Model continuous norms for the latent ability scores
# cnorm.model <- cnorm(raw = theta, group = data$sim$group)

## ----fig2, fig.height = 7, fig.width = 7------------------
plot(model, "raw", group = TRUE)

## ----fig3, fig.height = 7, fig.width = 7------------------
plot(model, "norm", group = TRUE, minNorm = 25, maxNorm = 75)

## ----fig4, fig.height = 4, fig.width = 7------------------
plot(model, "density", group = c (2, 3, 4))

## ----fig5, fig.height = 4, fig.width = 7------------------
plot(model, "derivative")

# ---------------------------------------------------------------------------
# modelling.R -- model estimation, selection, consistency screening and
#                validation for cNORM
# ---------------------------------------------------------------------------

#' Determine Regression Model
#'
#' Computes Taylor polynomial regression models by evaluating a series of models
#' with increasing number of predictors. It aims to find a consistent model that
#' effectively captures the variance in the data. It draws on the regsubsets
#' function from the leaps package, builds up to 10 models per number of
#' predictors, screens them for model consistency (monotonicity of the mapping
#' between norm score and raw score over the complete age range) and selects the
#' consistent model with the highest R^2. This automatic model selection should
#' usually be accompanied by visual inspection of the percentile plots and
#' assessment of fit statistics. Set R^2 or the number of terms manually to
#' retrieve a more parsimonious model, if desired.
#'
#' The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
#' \code{computePowers} and \code{prepareData} are usually not called directly,
#' but accessed through other functions like \code{cnorm}.
#'
#' Additional functions like \code{plotSubset(model)} and \code{cnorm.cv} can
#' aid in model evaluation.
#'
#' If \code{averaging = TRUE}, the final coefficients are not taken from a
#' single selected model, but computed as a BIC-weighted average across all
#' consistency-screened candidate models (weights \eqn{w_j \propto
#' exp(-\Delta BIC_j / 2)}). Since a convex combination of functions that are
#' all monotone in the same direction is itself monotone, the averaged model is
#' guaranteed to remain consistent, while reducing model selection variance.
#' This replaces the deprecated \code{subsampling} approach.
#'
#' @param data Preprocessed dataset with 'raw' scores, powers, interactions, and
#'   usually an explanatory variable (like age).
#' @param raw Name of the raw score variable (default: 'raw').
#' @param terms Desired number of terms in the model.
#' @param R2 Adjusted R^2 stopping criterion for model building.
#' @param k Power constant influencing model complexity (default: taken from the
#'   data preparation, max: 6).
#' @param t Age power parameter. If unset, taken from the data preparation
#'   (default 3).
#' @param predictors List of predictors or regression formula for model
#'   selection. Overrides 'k' and can include additional variables.
#' @param force.in Variables forcibly included in the regression.
#' @param weights Optional case weights. If set to FALSE, default weights
#'   (if any) are ignored.
#' @param plot If TRUE (default), displays a percentile plot of the model and
#'   information about the regression object. FALSE turns off plotting and report.
#' @param extensive If TRUE (default), screen models for consistency and - if
#'   possible - exclude inconsistent ones.
#' @param averaging If TRUE (default FALSE), apply BIC-weighted model averaging
#'   across the consistency-screened candidate models instead of selecting a
#'   single model. Requires \code{extensive = TRUE} and age-based norming.
#' @param minDip Tolerance for monotonicity check. Allow small violations
#'   (default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
#'   for strict checking.
#' @return The model. Further exploration can be done using
#'   \code{plotSubset(model)} and \code{plotPercentiles(data, model)}.
#' @examples
#' \dontrun{
#' # It is not recommended to directly use this function. Rather use 'cnorm' instead.
#' normData <- prepareData(elfe)
#' model <- bestModel(normData)
#' plotSubset(model)
#' plotPercentiles(buildCnormObject(normData, model))
#'
#' # Specifying variables explicitly
#' preselectedModel <- bestModel(normData, predictors = c("L1", "L3", "L1A3", "A2", "A3"))
#' print(regressionFunction(preselectedModel))
#' }
#' @seealso plotSubset, plotPercentiles, plotPercentileSeries, checkConsistency
#' @export
#' @family model
bestModel <- function(data,
                      raw = NULL,
                      R2 = NULL,
                      k = NULL,
                      t = NULL,
                      predictors = NULL,
                      terms = 0,
                      weights = NULL,
                      force.in = NULL,
                      plot = TRUE,
                      extensive = TRUE,
                      averaging = FALSE,
                      minDip = .01) {

  # --- retrieve attributes and consolidate defaults -------------------------
  if (is.null(raw)) raw <- attr(data, "raw")
  if (is.null(raw)) raw <- "raw"

  # consolidate case weights into a single vector 'w'
  w <- NULL
  if (is.null(weights)) {
    if (!is.null(attr(data, "weights")) && !is.null(data$weights))
      w <- data$weights
  } else if (isFALSE(weights)) {
    w <- NULL
  } else if (is.numeric(weights) && length(weights) == nrow(data)) {
    w <- weights
  } else {
    warning("Invalid 'weights' argument (must be numeric and of length nrow(data)); ignored.")
  }

  kData <- attr(data, "k")
  if (is.null(k)) {
    k <- if (!is.null(kData)) kData else 5
  } else if (!is.null(kData) && k > kData) {
    warning("k parameter exceeds the power degrees in the dataset. ",
            "Setting to default of k = ", kData)
    k <- kData
  }

  if (is.null(t)) {
    t <- if (!is.null(attr(data, "t"))) attr(data, "t") else 3
  }

  if (is.null(predictors) && (k < 1 || k > 6)) {
    warning("k parameter out of bounds. Please specify a value between 1 and 6. ",
            "Setting to default = 5.")
    k <- 5
  }

  # number of candidate terms
  if (is.null(predictors)) {
    nvmax <- (t + 1) * (k + 1) - 1
  } else if (inherits(predictors, "formula")) {
    nvmax <- length(attr(stats::terms(predictors), "term.labels"))
  } else {
    nvmax <- length(predictors)
  }

  # --- check parameter ranges ------------------------------------------------
  if (!is.null(R2) && (R2 <= 0 || R2 >= 1)) {
    warning("R2 parameter out of bounds. Setting to default R2 = .99")
    R2 <- .99
  }

  if (terms < 0 || terms > nvmax) {
    warning("terms parameter out of bounds. Setting to 5.")
    terms <- 5
  }

  if (!(raw %in% colnames(data)) && !inherits(predictors, "formula")) {
    stop("Raw value variable '", raw, "' does not exist in data object.")
  }

  if (!is.null(predictors) && !inherits(predictors, "formula") &&
      !all(predictors %in% colnames(data))) {
    stop("Missing variables from predictors variable. Please check variable list.")
  }

  # --- set up regression formula ---------------------------------------------
  useAge <- isTRUE(attr(data, "useAge"))
  if (is.null(predictors)) {
    lmX <- buildFunction(raw = raw, k = k, t = t, age = useAge)
  } else if (inherits(predictors, "formula")) {
    lmX <- predictors
  } else {
    lmX <- stats::reformulate(predictors, response = raw)
  }

  big <- nvmax > 50
  if (big && plot)
    message("The computation might take some time ...")

  index <- NULL
  if (!is.null(force.in)) {
    rhs <- attr(stats::terms(lmX), "term.labels")
    index <- match(force.in, rhs)
    if (anyNA(index))
      stop("'force.in' contains variables that are not part of the model formula.")
  }

  nbest <- if (extensive && useAge) 10 else 1

  # --- best subset search -----------------------------------------------------
  # regsubsets() evaluates its 'weights' argument via model.frame(), i.e.
  # first in 'data', then in the environment of the formula. Since lmX was
  # created in buildFunction() (or supplied by the user), 'w' is not visible
  # there. We therefore wrap the formula environment: 'w' becomes findable,
  # while the original environment remains accessible as parent.
  if (!is.null(w)) {
    wEnv <- new.env(parent = environment(lmX))
    wEnv$w <- w
    environment(lmX) <- wEnv

    subsets <- regsubsets(lmX, data = data, nbest = nbest, nvmax = nvmax,
                          force.in = index, really.big = big,
                          weights = w, method = "exhaustive")
  } else {
    subsets <- regsubsets(lmX, data = data, nbest = nbest, nvmax = nvmax,
                          force.in = index, really.big = big,
                          method = "exhaustive")
  }

  results <- summary(subsets)
  highestConsistent <- NULL
  if (extensive && useAge) {
    results <- screenSubset(data, results, data[[raw]], weights = w,
                            minDip =  (max(data[[raw]])-min(data[[raw]])) * minDip)
    highestConsistent <- results$highestConsistent
  }

  # --- model selection strategy ------------------------------------------------
  # 1. If no criterion is specified, take the largest consistent model, if available
  # 2. else take the first model exceeding R2 > .99
  # 3. else fall back to BIC-optimal model
  # 4. if terms are specified, take terms
  # 5. selection based on R2
  n_terms_per_row <- rowSums(results$outmat == "*")
  selectionStrategy <- 0

  if (is.null(R2) && terms == 0) {
    if (!is.null(results$highestConsistent)) {
      i <- results$highestConsistent
      selectionStrategy <- 1
      report <- paste0("Final solution: ", n_terms_per_row[i],
                       " terms (highest consistent model)")
    } else {
      i <- which(results$adjr2 > 0.99)[1]
      if (is.na(i)) {
        # fall back to the BIC-optimal model instead of the smallest one
        i <- which.min(results$bic)
        selectionStrategy <- 3
        report <- paste0(
          "Final solution: ", n_terms_per_row[i],
          " terms (BIC-optimal fallback; R2 > .99 not reached)")
    } else {
        selectionStrategy <- 2
        report <- paste0("Final solution: ", n_terms_per_row[i],
                         " terms (model exceeding R2 > .99)")
      }
    }
  } else if (terms > 0) {
    candidate <- which(n_terms_per_row == terms)[1]
    if (is.na(candidate)) {
      candidate <- which.min(abs(n_terms_per_row - terms))
      warning("No model with exactly ", terms,
              " terms is available after consistency screening; ",
              "using closest match (", n_terms_per_row[candidate], " terms).")
    }
    i <- candidate
    selectionStrategy <- 4
    report <- paste0("User specified solution: ", n_terms_per_row[i], " terms")
  } else {
    candidates <- which(results$adjr2 > R2)
    if (length(candidates) == 0L) {
      i <- which.max(results$adjr2)
      selectionStrategy <- 3
      report <- paste0("User specified solution: R2 > ", R2,
                       ", but value not reached. Using best available model (",
                       n_terms_per_row[i], " terms).")
    } else {
      i <- candidates[which.min(n_terms_per_row[candidates])]
      selectionStrategy <- 5
      report <- paste0("User specified solution: R2 > ", R2,
                       " resulting in ", n_terms_per_row[i], " terms")
    }
  }

  # after selection, independent of strategy:
  if (extensive && useAge && is.null(highestConsistent)) {
    warning("No model passed the monotonicity screening. The selected model ",
            "violates consistency somewhere in the norm score / age range. ",
            "Inspect 'plotPercentileSeries(model)', restrict the norm score ",
            "range (e.g. +/- 2.5 SD via minNorm/maxNorm), or reduce k/t.")
  }

  report[2] <- paste0("R-Square Adj. = ", round(results$adjr2[i], digits = 6))

  # --- final model fit -----------------------------------------------------------
  selectedTerms <- colnames(results$outmat)[results$outmat[i, ] == "*"]
  text <- paste0(raw, " ~ ", paste(selectedTerms, collapse = " + "))
  report[3] <- paste0("Final regression model: ", text)

  if (averaging && !is.null(results$consistent) &&
      any(results$consistent, na.rm = TRUE)) {
    bestformula <- weightedAverageModel(results, data, raw, weights = w)
    report[3] <- paste0(report[3],
                        " (BIC-weighted averaging over consistent models applied)")
  } else {
    if (averaging)
      warning("Model averaging requires consistency screening (extensive = TRUE) ",
              "and age-based norming with at least one consistent model. ",
              "Falling back to single model fit.")
    if (is.null(w))
      bestformula <- stats::lm(stats::as.formula(text), data = data)
    else
      bestformula <- stats::lm(stats::as.formula(text), data = data, weights = w)
  }

  # --- fit statistics ---------------------------------------------------------
  tab <- data.frame(raw = data[[raw]], fitted = bestformula$fitted.values)
  tab <- tab[stats::complete.cases(tab), ]
  rmse <- sqrt(mean((tab$raw - tab$fitted)^2))

  # --- model information --------------------------------------------------------
  bestformula$ideal.model <- i
  bestformula$cutoff <- R2
  bestformula$subsets <- results
  bestformula$useAge <- useAge

  if (is.null(data$A1)) {          # conventional norming
    bestformula$minA1 <- 0
    bestformula$maxA1 <- 0
  } else {                         # continuous norming
    bestformula$minA1 <- min(data$A1)
    bestformula$maxA1 <- max(data$A1)
  }

  bestformula$minL1 <- min(data$L1)
  bestformula$maxL1 <- max(data$L1)
  bestformula$minRaw <- min(data[[raw]])
  bestformula$maxRaw <- max(data[[raw]])
  bestformula$raw <- raw
  bestformula$rmse <- rmse
  bestformula$scaleSD <- attr(data, "scaleSD")
  bestformula$scaleM <- attr(data, "scaleM")
  bestformula$descend <- attr(data, "descend")
  bestformula$group <- attr(data, "group")
  bestformula$age <- attr(data, "age")
  bestformula$k <- attr(data, "k")
  bestformula$A <- attr(data, "A")
  bestformula$highestConsistent <- highestConsistent
  bestformula$selectionStrategy <- selectionStrategy

  # --- report ----------------------------------------------------------------
  report[4] <- paste0("Regression function: ",
                      regressionFunction(bestformula, digits = 10))
  report[5] <- paste0("Raw Score RMSE = ", round(rmse, digits = 5))
  if (!is.null(w)) {
    report[6] <- paste0(
      "Post stratification was applied. The weights range from ",
      round(min(w), digits = 3), " to ", round(max(w), digits = 3),
      " (m = ", round(mean(w), digits = 3),
      ", sd = ", round(stats::sd(w), digits = 3), ").")
  }

  bestformula$report <- report

  if (plot) cat(report, sep = "\n")

  if (anyNA(bestformula$coefficients)) {
    warning("The regression contains missing coefficients. No fitting model ",
            "could be found. Please try a different number of terms.")
  }

  nSelected <- length(bestformula$coefficients) - 1L
  if (nSelected < 4) {
    message("\nThe model includes a low number of terms. Models with four or ",
            "more predictors are usually more robust. The low number is probably ",
            "the consequence of the rather strict monotonicity checks in cNORM. ",
            "Information functions with 'plotSubset' and 'plotPercentileSeries' ",
            "might help to identify a balanced number of terms. Consider fixing ",
            "the 'terms' parameter to a higher number.")
  }

  if (plot) {
    if (!is.null(data$A1)) {
      cat("\nUse 'printSubset(model)' to get detailed information on the ",
          "different solutions, 'plotPercentiles(model)' to display the ",
          "percentile plot and 'plotSubset(model)' to inspect model fit.", sep = "")
    } else {
      cat("\nConventional norming was applied. Use 'normTable(0, model)' or ",
          "'rawTable(0, model)' to retrieve norm scores. If you would like to ",
          "achieve a closer fit, increase the terms parameter.", sep = "")
    }
  }

  class(bestformula) <- "cnormModel"

  if (plot && bestformula$useAge) {
    tmp <- list(data = data, model = bestformula)
    class(tmp) <- "cnormTemp"
    plotPercentiles(tmp)
  }

  return(bestformula)
}


#' BIC-Weighted Model Averaging Across Consistent Candidate Models
#'
#' Computes a weighted average of the regression coefficients across the
#' consistency-screened candidate models returned by \code{regsubsets} and
#' \code{screenSubset}. Weights are information-theoretic model weights
#' \eqn{w_j \propto exp(-\Delta BIC_j / 2)}, restricted to models within
#' \code{deltaBIC} of the best model. Each candidate is refitted on the
#' complete sample via (weighted) least squares; coefficients of terms not
#' included in a candidate are treated as zero.
#'
#' Because all candidates passed the monotonicity screening (same direction),
#' and the averaging weights are positive and sum to one, the averaged model is
#' itself monotone and thus consistent. In contrast to averaging coefficients
#' over subsamples (deprecated \code{subsample_lm}), this approach targets the
#' actual source of variance - model selection - while each component estimate
#' remains a full-sample least squares fit.
#'
#' @param results A (screened) \code{summary.regsubsets} object, including a
#'   \code{consistent} flag per row.
#' @param data The preprocessed norm data.
#' @param raw Name of the raw score variable.
#' @param weights Optional numeric vector of case weights.
#' @param deltaBIC Only models within this BIC distance of the best candidate
#'   are averaged (default 10; Burnham & Anderson, 2002).
#' @return An \code{lm}-type object with averaged coefficients over the union of
#'   the candidate terms; \code{fitted.values} and \code{residuals} are updated
#'   accordingly.
#' @references Burnham, K. P., & Anderson, D. R. (2002). Model Selection and
#'   Multimodel Inference. Springer.
#' @export
#' @family model
weightedAverageModel <- function(results, data, raw, weights = NULL,
                                 deltaBIC = 10) {
  candidates <- if (!is.null(results$consistent) &&
                    any(results$consistent, na.rm = TRUE)) {
    which(results$consistent)
  } else {
    seq_along(results$bic)
  }

  bic <- results$bic[candidates]
  candidates <- candidates[bic - min(bic) <= deltaBIC]
  bic <- results$bic[candidates]
  wgt <- exp(-0.5 * (bic - min(bic)))
  wgt <- wgt / sum(wgt)

  unionTerms <- colnames(results$outmat)[
    colSums(results$outmat[candidates, , drop = FALSE] == "*") > 0]

  coefMat <- matrix(0, nrow = length(candidates), ncol = length(unionTerms) + 1L,
                    dimnames = list(NULL, c("(Intercept)", unionTerms)))

  for (j in seq_along(candidates)) {
    trms <- colnames(results$outmat)[results$outmat[candidates[j], ] == "*"]
    f <- stats::reformulate(trms, response = raw)
    fitj <- if (is.null(weights)) stats::lm(f, data = data)
    else stats::lm(f, data = data, weights = weights)
    cf <- stats::coef(fitj)
    cf[is.na(cf)] <- 0
    coefMat[j, names(cf)] <- cf
  }

  avg <- as.vector(crossprod(coefMat, wgt))
  names(avg) <- colnames(coefMat)

  fUnion <- stats::reformulate(unionTerms, response = raw)
  final <- if (is.null(weights)) stats::lm(fUnion, data = data)
  else stats::lm(fUnion, data = data, weights = weights)

  X <- stats::model.matrix(final)
  final$coefficients <- avg[colnames(X)]
  final$fitted.values <- as.vector(X %*% final$coefficients)
  final$residuals <- stats::model.response(stats::model.frame(final)) -
    final$fitted.values
  final$averaged <- TRUE
  final$averagingWeights <- wgt
  final
}


#' Print Model Selection Information
#'
#' Displays R^2 and other metrics for models with varying predictors, aiding in
#' choosing the best-fitting model after model fitting. The F-test compares each
#' model to the preceding (smaller) one, with degrees of freedom based on the
#' actual difference in the number of parameters.
#' @param x Model output from 'bestModel' or a cnorm object.
#' @param ... Additional parameters.
#' @return Table with model information criteria.
#' @export
#'
#' @examples
#' \dontrun{
#'   result <- cnorm(raw = elfe$raw, group = elfe$group)
#'   printSubset(result)
#' }
#' @family model
printSubset <- function(x, ...) {
  if (isTaylor(x)) {
    x <- x$model
  }

  rss <- x$subsets$rss
  adjr2 <- x$subsets$adjr2
  nModels <- length(rss)
  nParams <- rowSums(x$subsets$outmat == "*") + 1L   # + intercept
  n <- length(x$fitted.values)

  Fvals <- rep(NA_real_, nModels)
  pvals <- rep(NA_real_, nModels)
  if (nModels > 1L) {
    df1 <- diff(nParams)
    df1[df1 < 1L] <- NA                              # non-nested comparison
    df2 <- n - nParams[-1L]
    Fs <- (-diff(rss) / df1) / (rss[-1L] / df2)
    Fvals[-1L] <- Fs
    pvals[-1L] <- stats::pf(Fs, df1, df2, lower.tail = FALSE)
  }

  table <- data.frame(
    R2adj = adjr2,
    BIC = x$subsets$bic,
    CP = x$subsets$cp,
    RSS = rss,
    RMSE = sqrt(rss / n),
    DeltaR2adj = c(NA, diff(adjr2)),
    F = Fvals,
    p = pvals,
    nr = seq_len(nModels)
  )

  if (!is.null(x$subsets$consistent))
    table$consistent <- x$subsets$consistent

  return(table)
}


#' Check the consistency of the norm data model
#'
#' While abilities increase and decline over age, within one age group, the
#' norm scores always have to show a monotonic increase or decrease with
#' increasing raw scores. Violations of this assumption are an indication for
#' problems in modeling the relationship between raw and norm scores. There are
#' several reasons, why this might occur:
#' \enumerate{
#'   \item Vertical extrapolation: Choosing extreme norm scores, e. g. values
#'   -3 <= x and x >= 3. In order to model these extreme values, a large sample
#'   dataset is necessary.
#'   \item Horizontal extrapolation: Taylor polynomials converge in a certain
#'   radius. Using the model values outside the original dataset may
#'   lead to inconsistent results.
#'   \item The data cannot be modeled with Taylor polynomials, or you need
#'   another power parameter (k) or R2 for the model.
#'  }
#'
#'  In general, extrapolation (point 1 and 2) can carefully be done to a
#'  certain degree outside the original sample, but it should in general
#'  be handled with caution. Please note that at extreme values, the models
#'  most likely become independent and it is thus recommended to restrict the
#'  norm score range to the relevant range of abilities, e.g. +/- 2.5 SD via
#'  the minNorm and maxNorm parameter.
#'
#'  With \code{method = "analytic"} (default), monotonicity within each age is
#'  verified exactly: since the model is polynomial in the norm score L, the
#'  real roots of its derivative are computed via \code{polyroot} and the sign
#'  of the derivative is checked between consecutive roots. This detects even
#'  narrow violations that a coarse norm score grid can miss. Age remains
#'  discretized; violations that are entirely clipped by the raw score bounds
#'  are ignored. \code{method = "grid"} reproduces the classical numerical
#'  check and is used automatically as a fallback for models with non-Taylor
#'  predictors.
#'
#' @param model The model from the bestModel function or a cnorm object
#' @param minAge Age to start with checking
#' @param maxAge Upper end of the age check
#' @param stepAge Stepping parameter for the age check. If NULL, 8 equidistant
#'   age points are used. Lower values indicate higher precision.
#' @param minNorm Lower end of the norm value range
#' @param maxNorm Upper end of the norm value range
#' @param minRaw clipping parameter for the lower bound of raw scores
#' @param maxRaw clipping parameter for the upper bound of raw scores
#' @param stepNorm Stepping parameter for the norm table check within age
#'   (only used with \code{method = "grid"})
#' @param method Monotonicity cgeck. Either "analytic" (default; exact within
#'   age via polynomial derivative roots) or "grid" (numerical check on a norm
#'   score grid)
#' @param warn Retained for backwards compatibility (violations below numerical
#'   tolerance are always suppressed)
#' @param silent turn off messages
#' @param minDip Tolerance for monotonicity check. Allow small violations
#'   (default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
#'   for strict checking.
#' @return Boolean, indicating model violations (TRUE) or no problems (FALSE)
#' @examples
#' \dontrun{
#'   model <- cnorm(raw = elfe$raw, group = elfe$group, plot = FALSE)
#'   modelViolations <- checkConsistency(model, minNorm = 25, maxNorm = 75)
#'   plotDerivative(model, minNorm = 25, maxNorm = 75)
#' }
#' @export
#' @family model
checkConsistency <- function(model,
                             minAge = NULL,
                             maxAge = NULL,
                             minNorm = NULL,
                             maxNorm = NULL,
                             minRaw = NULL,
                             maxRaw = NULL,
                             stepAge = NULL,
                             stepNorm = 1,
                             method = c("analytic", "grid"),
                             warn = FALSE,
                             silent = FALSE,
                             minDip = .01) {
  if (isTaylor(model)) {
    model <- model$model
  }

  if (!inherits(model, "cnormModel")) {
    stop("Please provide a cnorm model.")
  }

  method <- match.arg(method)

  # fallbacks to model limits
  if (is.null(minAge)) minAge <- model$minA1
  if (is.null(maxAge)) maxAge <- model$maxA1
  if (is.null(minNorm)) minNorm <- model$minL1
  if (is.null(maxNorm)) maxNorm <- model$maxL1
  if (is.null(minRaw)) minRaw <- model$minRaw
  if (is.null(maxRaw)) maxRaw <- model$maxRaw

  descend <- isTRUE(model$descend)

  # age evaluation points (handles conventional norming with minAge == maxAge)
  if (maxAge > minAge) {
    ages <- if (is.null(stepAge)) seq(minAge, maxAge, length.out = 4)
    else seq(minAge, maxAge, by = stepAge)
  } else {
    ages <- minAge
  }

  # analytic check requires pure Taylor coefficient names
  B <- taylorCoefficientMatrix(model$coefficients)
  if (is.null(B) || anyNA(model$coefficients)) method <- "grid"

  if (method == "analytic") {
    tPow <- ncol(B) - 1L
    absDip <- (maxRaw - minRaw) * minDip

    violations <- vapply(ages, function(a) {
      pcoef <- as.vector(B %*% a^(0:tPow))
      polyViolatesMonotonicity(pcoef, minNorm, maxNorm,
                               descend = descend,
                               minRaw = minRaw, maxRaw = maxRaw,
                               minDip = absDip)
    }, logical(1))
  } else {
    norms <- seq(minNorm, maxNorm, by = stepNorm)
    if (length(norms) < 2)
      stop("Range of norm scores too small to check consistency.")

    n_ages <- length(ages)
    n_norms <- length(norms)
    grid_norms <- rep(norms, times = n_ages)
    grid_ages <- rep(ages, each = n_norms)

    raw_preds <- predictRaw(norm = grid_norms, age = grid_ages,
                            coefficients = model$coefficients,
                            minRaw = minRaw, maxRaw = maxRaw)

    pred_mat <- matrix(raw_preds, nrow = n_norms, ncol = n_ages)
    diffs <- pred_mat[-1, , drop = FALSE] - pred_mat[-n_norms, , drop = FALSE]

    tol <- 1e-6
    violations <- if (descend) !apply(diffs <= tol, 2, all)
    else !apply(diffs >= -tol, 2, all)
  }

  results <- round(ages[violations], digits = 1)

  if (length(results) == 0) {
    if (!silent) cat("No relevant violations of model consistency found.\n")
    return(FALSE)
  } else {
    if (!silent) {
      message(paste0(
        "Violations of monotonicity found within the specified range of age ",
        "and norm score at age points: ",
        paste(results, collapse = ", "),
        "\n\nUse 'plotPercentiles' to visually inspect the norm curve or ",
        "'plotDerivative' to identify regions violating the consistency. ",
        "Rerun the modeling with adjusted parameters or restrict the valid ",
        "value range accordingly.\n"))
      cat(rangeCheck(model, minAge, maxAge, minNorm, maxNorm))
      cat("\n")
    }
    return(TRUE)
  }
}


#' Regression function
#'
#' The method builds the regression function for the regression model,
#' including the beta weights.
#' It can be used to predict the raw scores based on age and location.
#' @param model The regression model from the bestModel function or a cnorm object
#' @param raw The name of the raw value variable (default 'raw')
#' @param digits Number of digits for formatting the coefficients
#' @return The regression formula as a string
#'
#' @examples
#' \dontrun{
#'   result <- cnorm(raw = elfe$raw, group = elfe$group)
#'   regressionFunction(result)
#' }
#'
#' @export
#' @family model
regressionFunction <- function(model, raw = NULL, digits = NULL) {
  if (isTaylor(model)) {
    raw <- "raw"
    model <- model$model
  } else if (is.null(raw)) {
    raw <- model$raw
  }

  coefs <- model$coefficients

  if (!is.null(digits)) {
    formatted_coefs <- format(coefs, digits = digits, trim = TRUE,
                              scientific = FALSE)
  } else {
    formatted_coefs <- as.character(coefs)
  }

  intercept <- formatted_coefs[1]

  if (length(coefs) == 1) {
    return(paste(raw, "~", intercept))
  }

  term_strings <- paste0("(", formatted_coefs[-1], "*", names(coefs)[-1], ")")

  formulA <- paste(raw, intercept, sep = " ~ ")
  formulA <- paste(formulA, paste(term_strings, collapse = " + "), sep = " + ")

  return(formulA)
}


#' Derivative of regression model
#'
#' Calculates the derivative of the location / norm value from the regression
#' model with the first derivative as the default. This is useful for finding
#' violations of model assumptions and problematic distribution features as
#' f. e. bottom and ceiling effects, non-progressive norm scores within an age
#' group or in general intersecting percentile curves.
#' @param model The regression model or a cnorm object
#' @param order The degree of the derivative, default: 1
#' @return The derived coefficients
#' @examples
#' \dontrun{
#'   m <- cnorm(raw = elfe$raw, group = elfe$group)
#'   derivedCoefficients <- derive(m, order = 1)
#' }
#' @export
#' @family model
derive <- function(model, order = 1) {
  if (isTaylor(model)) model <- model$model

  coeff <- model$coefficients

  for (o in seq_len(order)) {
    # only terms containing an L power survive derivation w.r.t. L
    coeff <- coeff[grepl("L\\d+", names(coeff))]
    if (length(coeff) == 0) return(coeff)

    nms <- names(coeff)
    lpow <- as.integer(sub("^.*?L(\\d+).*$", "\\1", nms))
    apart <- sub("^L\\d+", "", nms)          # "" or "A<j>"

    newVal <- as.numeric(coeff) * lpow
    newPow <- lpow - 1L
    newNms <- ifelse(newPow == 0L,
                     ifelse(apart == "", "(Intercept)", apart),
                     paste0("L", newPow, apart))

    coeff <- newVal
    names(coeff) <- newNms
  }

  return(coeff)
}


#' Prints the results and regression function of a cnorm model
#'
#' @param object A regression model or cnorm object
#' @param ... additional parameters
#' @return A report on the regression function, weights, R2 and RMSE
#' @export
#' @family model
modelSummary <- function(object, ...) {
  if (isTaylor(object)) {
    object <- object$model
  }
  strat <- c(
    "largest consistent model",
    "first model exceeding R2 > .99",
    "fallback: best / smallest available model",
    "terms specified manually",
    "selection based on R2"
  )

  terms <- length(object$coefficients) - 1
  adj_r_squared <- object$subsets$adjr2[object$ideal.model]
  rmse <- object$rmse
  selection_strategy <- object$selectionStrategy
  highest_consistent <- object$highestConsistent

  cat("cNORM Model Summary\n")
  cat("-------------------\n")
  cat("Number of terms:", terms, "\n")
  cat("Adjusted R-squared:", round(adj_r_squared, 4), "\n")
  cat("RMSE:", round(rmse, 4), "\n")
  cat("Selection strategy:", selection_strategy)
  if (!is.null(selection_strategy) && selection_strategy > 0 &&
      selection_strategy <= length(strat)) {
    cat(",", strat[selection_strategy])
  }
  cat("\nHighest consistent model:",
      if (is.null(highest_consistent)) "none" else highest_consistent, "\n")
  if (isTRUE(object$averaged)) {
    cat("BIC-weighted model averaging was applied across",
        length(object$averagingWeights), "consistent models.\n")
  }
  cat("Raw score variable:", object$raw, "\n")
  cat("Raw score range:", object$minRaw, "to", object$maxRaw, "\n")
  if (isTRUE(object$useAge)) {
    cat("Age range:", object$minA1, "to", object$maxA1, "\n")
  }
  cat("\nRegression function:\n")
  cat(regressionFunction(object), "\n")

  cat(object$report, sep = "\n")
}


#' Check for horizontal and vertical extrapolation
#'
#' Regression models only work in a specific range and extrapolation
#' horizontally (outside the original range) or vertically (extreme norm
#' scores) might lead to inconsistent results. The function generates a
#' message, indicating extrapolation and the range of the original data.
#' @param object The regression model or a cnorm object
#' @param minAge The lower age bound
#' @param maxAge The upper age bound
#' @param minNorm The lower norm value bound
#' @param maxNorm The upper norm value bound
#' @param digits The precision for rounding the norm and age data
#' @param ... additional parameters
#' @return the report
#' @export
#' @examples
#' \dontrun{
#'   m <- cnorm(raw = elfe$raw, group = elfe$group)
#'   rangeCheck(m)
#' }
#' @family model
rangeCheck <- function(object,
                       minAge = NULL,
                       maxAge = NULL,
                       minNorm = NULL,
                       maxNorm = NULL,
                       digits = 3,
                       ...) {
  if (isTaylor(object)) {
    object <- object$model
  }

  summary <- paste0(
    "The original data for the regression model spanned from age ",
    round(object$minA1, digits), " to ", round(object$maxA1, digits),
    ", with a norm score range from ",
    round(object$minL1, digits), " to ", round(object$maxL1, digits),
    ". The raw scores range from ", object$minRaw, " to ", object$maxRaw, ".")

  if (isTRUE(object$descend)) {
    summary <- paste0(summary, " The ranking was done in descending order.")
  }

  reportOnly <- (is.null(minAge) || is.null(maxAge) ||
                   is.null(minNorm) || is.null(maxNorm))

  if (!reportOnly &&
      (minAge < object$minA1 || maxAge > object$maxA1) &&
      (minNorm < object$minL1 || maxNorm > object$maxL1)) {
    summary <- paste(
      "Horizontal and vertical extrapolation detected. Be careful using age groups and extreme norm scores outside the original sample.",
      summary, sep = "\n")
  } else if (!reportOnly && (minAge < object$minA1 || maxAge > object$maxA1)) {
    summary <- paste(
      "Horizontal extrapolation detected. Be careful using age groups outside the original sample.",
      summary, sep = "\n")
  } else if (!reportOnly && (minNorm < object$minL1 || maxNorm > object$maxL1)) {
    summary <- paste(
      "Vertical extrapolation detected. Be careful using extreme norm scores exceeding the scores of the original sample.",
      summary, sep = "\n")
  }

  return(summary)
}


#' Cross-validation for Term Selection in cNORM
#'
#' Assists in determining the optimal number of terms for the regression model
#' using repeated Monte Carlo cross-validation. It leverages an 80-20 split
#' between training and validation data, with stratification by norm group or
#' random sample in case of using sliding window ranking.
#'
#' Successive models, with an increasing number of terms, are evaluated, and the
#' RMSE for raw scores plotted. This encompasses the training, validation, and
#' entire dataset. If `norms` is set to TRUE (default), the function will also
#' calculate the mean norm score reliability and crossfit measures. Note that
#' due to the computational requirements of norm score calculations, execution
#' can be slow, especially with numerous repetitions or terms.
#'
#' When `cv` is set to "full" (default), both test and validation datasets are
#' ranked separately, providing comprehensive cross-validation. For a more
#' streamlined validation process focused only on modeling, a pre-ranked dataset
#' can be used. The output comprises RMSE for raw score models, norm score R^2,
#' delta R^2, crossfit, and the norm score SE according to Oosterhuis, van der
#' Ark, & Sijtsma (2016).
#'
#' This function is not yet prepared for the 'extensive' search strategy,
#' introduced in version 3.3, but instead relies on the first model per number
#' of terms, without consistency check.
#'
#' For assessing overfitting:
#' \deqn{CROSSFIT = R(Training; Model)^2 / R(Validation; Model)^2}
#' A CROSSFIT > 1 suggests overfitting, < 1 suggests potential underfitting, and
#' values around 1 are optimal, given a low raw score RMSE and high norm score
#' validation R^2.
#'
#' Suggestions for ideal model selection:
#' \itemize{
#'   \item Visual inspection of percentiles with `plotPercentiles` or `plotPercentileSeries`.
#'   \item Pair visual inspection with repeated cross-validation (e.g., 10 repetitions).
#'   \item Aim for low raw score RMSE and high norm score R^2, avoiding terms with significant overfit (e.g., crossfit > 1.1).
#' }
#'
#' @param data Data frame of norm sample or a cnorm object. Should have ranking,
#'   powers, and interaction of L and A.
#' @param formula Formula from an existing regression model; min/max functions
#'   ignored. If using a cnorm object, this is automatically fetched.
#' @param repetitions Number of repetitions for cross-validation.
#' @param norms If TRUE, computes norm score crossfit and R^2. Note:
#'   Computationally intensive.
#' @param min Start with a minimum number of terms (default = 4).
#' @param max Maximum terms in model, up to (k + 1) * (t + 1) - 1 (default = 20).
#' @param cv "full" (default) splits data into training/validation, then ranks.
#'   Otherwise, expects a pre-ranked dataset.
#' @param pCutoff Checks stratification for unbalanced data. Performs a t-test
#'   per group. Default set to 0.2 to minimize beta error.
#' @param width If provided, ranking done via `rankBySlidingWindow`. Otherwise,
#'   by group.
#' @param raw Name of the raw score variable.
#' @param age Name of the age variable.
#' @param group Name of the grouping variable.
#' @param weights Name of the weighting parameter.
#'
#' @return Table with results per term number: RMSE for raw scores, R^2 for
#'   norm scores, and crossfit measure.
#' @export
#'
#' @examples
#' \dontrun{
#' # Example: Plot cross-validation RMSE by number of terms (up to 9) with three repetitions.
#' result <- cnorm(raw = elfe$raw, group = elfe$group)
#' cnorm.cv(result$data, min = 2, max = 9, repetitions = 3)
#'
#' # Using a cnorm object examines the predefined formula.
#' cnorm.cv(result, repetitions = 1)
#' }
#'
#' @references Oosterhuis, H. E. M., van der Ark, L. A., & Sijtsma, K. (2016).
#'   Sample Size Requirements for Traditional and Regression-Based Norms.
#'   Assessment, 23(2), 191-202. https://doi.org/10.1177/1073191115580638
#' @family model
cnorm.cv <- function(data,
                     formula = NULL,
                     repetitions = 5,
                     norms = TRUE,
                     min = 4,
                     max = 20,
                     cv = "full",
                     pCutoff = NULL,
                     width = NA,
                     raw = NULL,
                     group = NULL,
                     age = NULL,
                     weights = NULL) {

  if (isTaylor(data)) {
    formula <- data$model$terms
    data    <- data$data
  }

  if (is.null(pCutoff)) {
    pCutoff <- if (nrow(data) < 10000) 0.2 else 0.1
  }

  d <- data

  if (is.null(raw)) raw <- attr(d, "raw")
  if (is.null(raw)) stop("Please provide a raw score variable name.")
  if (!(raw %in% colnames(data)))
    stop("The specified raw score variable '", raw,
         "' is not present in the dataset.")

  if (is.null(group)) group <- attr(d, "group")
  if (is.null(age))   age   <- attr(d, "age")

  if (is.na(width) && !is.null(attr(d, "width")))
    width <- attr(d, "width")

  if (is.null(group) || (is.null(age) && is.na(width)))
    stop("Please provide either a grouping variable or age and width.")

  if (is.null(weights)) weights <- attr(d, "weights")

  if (!is.null(weights) && !(weights %in% colnames(data))) {
    warning("Weighting variable not found in dataset. Continuing without weighting.\n")
    weights <- NULL
  } else if (!is.null(weights)) {
    cat("Applying weighting ...\n")
  }

  scaleM  <- attr(d, "scaleMean")
  scaleSD <- attr(d, "scaleSD")
  if (is.null(scaleM)  || is.na(scaleM)  || cv == "full") scaleM  <- 50
  if (is.null(scaleSD) || is.na(scaleSD) || cv == "full") scaleSD <- 10

  k <- attr(d, "k"); if (is.null(k)) k <- 5
  t <- attr(d, "t"); if (is.null(t)) t <- 3

  n.models <- (k + 1) * (t + 1) - 1
  if (is.na(max) || max > n.models || max < 1) max <- n.models

  if (is.null(formula)) {
    lmX <- buildFunction(raw = raw, k = k, t = t, age = TRUE)
  } else {
    lmX <- formula
    n_formula_terms <- length(attr(stats::terms(formula), "term.labels"))
    min <- n_formula_terms
    max <- n_formula_terms
  }

  val.errors      <- rep(0, max)
  train.errors    <- rep(0, max)
  complete.errors <- rep(0, max)
  r2.train        <- rep(0, max)
  r2.test         <- rep(0, max)
  delta           <- rep(NA, max)
  norm.rmse       <- rep(0, max)
  norm.se         <- rep(0, max)
  norm.rmse.min   <- rep(0, max)

  terms_list <- vector("list", repetitions * (max - min + 1L))
  terms_idx  <- 1L

  rankGroup <- TRUE
  if (!is.null(age) && !is.na(width)) {
    cat("Age and width parameters available, switching to rankBySlidingWindow() ...\n")
    rankGroup <- FALSE
  }

  full_formula <- stats::formula(lmX)

  for (a in seq_len(repetitions)) {
    p.value <- 0.01
    n       <- 1L
    train   <- NA
    test    <- NA

    # establish a balanced 80/20 split
    while (p.value < pCutoff) {
      if (n > 100L)
        stop("Could not establish balanced data sets. Try decreasing pCutoff.")
      n <- n + 1L

      if (rankGroup) {
        d  <- d[sample(nrow(d)), ]
        d  <- d[order(d[, group]), ]
        sp <- split(d, list(d[, group]))
        sp <- lapply(sp, function(x) x[sample(nrow(x)), ])

        train <- lapply(sp, function(x) x[c(FALSE, rep(TRUE, 4)), ])
        test  <- lapply(sp, function(x) x[c(TRUE, rep(FALSE, 4)), ])

        p <- vapply(seq_along(train), function(z)
          stats::t.test(train[[z]][, raw], test[[z]][, raw])$p.value, numeric(1))

        p.value <- min(p)
        if (p.value < pCutoff) next

        train <- do.call(rbind, train)
        test  <- do.call(rbind, test)

        if (cv == "full") {
          train <- prepareData(train, raw = raw, group = group, age = age,
                               width = width, weights = weights, silent = TRUE)
          test  <- prepareData(test, raw = raw, group = group, age = age,
                               width = width, weights = weights, silent = TRUE)
        }
      } else {
        d      <- d[sample(nrow(d)), ]
        number <- floor(nrow(d) * 0.8)
        train  <- d[seq_len(number), ]
        test   <- d[(number + 1L):nrow(d), ]

        p.value <- stats::t.test(train[, age], test[, age])$p.value
        if (p.value < pCutoff) next

        train <- rankBySlidingWindow(train, age = age, raw = raw,
                                     weights = weights, width = width,
                                     silent = TRUE)
        test  <- rankBySlidingWindow(test, age = age, raw = raw,
                                     weights = weights, width = width,
                                     silent = TRUE)
        train <- computePowers(train, age = age, k = k, t = t, silent = TRUE)
        test  <- computePowers(test, age = age, k = k, t = t, silent = TRUE)
      }
    }

    subsets <- regsubsets(lmX, data = train, nbest = 1, nvmax = max,
                          really.big = n.models > 25)

    if (norms && is.null(formula)) cat(paste0("Cycle ", a, "\n"))

    # design matrices, built once per repetition
    X_train_full <- stats::model.matrix(full_formula, train)
    X_test_full  <- stats::model.matrix(full_formula, test)
    y_train      <- train[, raw]

    for (i in min:max) {
      variables <- names(coef(subsets, id = i))

      X_train_sub <- X_train_full[, variables, drop = FALSE]
      X_test_sub  <- X_test_full[, variables, drop = FALSE]

      # lm.fit (NOT .lm.fit): returns named, unpivoted coefficients
      # and fitted values
      fit <- stats::lm.fit(X_train_sub, y_train)
      cf <- fit$coefficients
      cf[is.na(cf)] <- 0                      # guard against aliased columns

      test.fitted <- as.vector(X_test_sub %*% cf)

      # lightweight mock object for norm score prediction
      mock_model <- list(
        coefficients = cf,
        k = k,
        minRaw = min(train[, raw], na.rm = TRUE),
        maxRaw = max(train[, raw], na.rm = TRUE),
        scaleM = scaleM,
        scaleSD = scaleSD,
        descend = isTRUE(attr(d, "descend")),
        raw = raw,
        age = age
      )
      class(mock_model) <- "cnormModel"

      terms_list[[terms_idx]] <- variables[variables != "(Intercept)"]
      terms_idx <- terms_idx + 1L

      train.errors[i] <- train.errors[i] +
        sqrt(mean((fit$fitted.values - y_train)^2, na.rm = TRUE))
      val.errors[i]   <- val.errors[i] +
        sqrt(mean((test.fitted - test[, raw])^2, na.rm = TRUE))

      if (norms) {
        train$T <- predictNorm(train[, raw], train[, age], mock_model,
                               min(train$normValue), max(train$normValue),
                               silent = TRUE)
        test$T  <- predictNorm(test[, raw], test[, age], mock_model,
                               min(train$normValue), max(train$normValue),
                               silent = TRUE)

        r2.train[i]  <- r2.train[i] +
          stats::cor(train$normValue, train$T, use = "pairwise.complete.obs")^2
        r2.test[i]   <- r2.test[i] +
          stats::cor(test$normValue, test$T, use = "pairwise.complete.obs")^2
        norm.rmse[i] <- norm.rmse[i] +
          sqrt(mean((test$T - test$normValue)^2, na.rm = TRUE))

        n_valid    <- sum(!is.na(test$T))
        norm.se[i] <- norm.se[i] +
          sqrt(sum((test$T - test$normValue)^2, na.rm = TRUE) /
                 max(n_valid - 2L, 1L))
      }
    }
  }

  norm.rmse.min[1] <- NA
  complete <- regsubsets(lmX, data = d, nbest = 1, nvmax = n.models,
                         really.big = n.models > 25)

  X_full_dataset <- stats::model.matrix(full_formula, d)
  y_full_dataset <- d[, raw]

  for (i in seq_len(max)) {
    variables <- names(coef(complete, id = i))

    fit_complete <- stats::lm.fit(
      X_full_dataset[, variables, drop = FALSE], y_full_dataset)
    complete.errors[i] <- sqrt(mean(
      (fit_complete$fitted.values - y_full_dataset)^2, na.rm = TRUE))

    train.errors[i] <- train.errors[i] / repetitions
    val.errors[i]   <- val.errors[i]   / repetitions

    if (norms) {
      r2.train[i]  <- r2.train[i]  / repetitions
      r2.test[i]   <- r2.test[i]   / repetitions
      norm.rmse[i] <- norm.rmse[i] / repetitions
      norm.se[i]   <- norm.se[i]   / repetitions

      if (i > min) {
        delta[i]         <- r2.test[i] - r2.test[i - 1L]
        norm.rmse.min[i] <- if (norm.rmse[i] > 0)
          norm.rmse[i] - norm.rmse[i - 1L] else NA
      }
    }

    if (i < min) {
      r2.train[i] <- NA
      r2.test[i] <- NA
      val.errors[i] <- NA
      train.errors[i] <- NA
      complete.errors[i] <- NA
      norm.rmse[i] <- NA
    }
    if (i <= min) norm.rmse.min[i] <- NA
  }

  Terms <- unlist(terms_list)

  tab <- data.frame(
    RMSE.raw.train    = train.errors,
    RMSE.raw.test     = val.errors,
    RMSE.raw.complete = complete.errors,
    R2.norm.train     = r2.train,
    R2.norm.test      = r2.test,
    Delta.R2.test     = delta,
    Crossfit          = r2.train / r2.test,
    RMSE.norm.test    = norm.rmse,
    SE.norm.test      = norm.se,
    terms             = seq_len(length(train.errors))
  )

  theme_custom <- theme_minimal() +
    theme(
      plot.title = element_text(face = "bold", size = 16, hjust = 0.5),
      axis.title = element_text(face = "bold", size = 12),
      axis.title.x = element_text(margin = margin(t = 10)),
      axis.title.y = element_text(margin = margin(r = 10)),
      axis.text = element_text(size = 10),
      legend.title = element_blank(),
      legend.text = element_text(size = 10),
      legend.position = "bottom",
      panel.grid.major = element_line(color = "gray90"),
      panel.grid.minor = element_line(color = "gray95")
    )

  breaks_step_1 <- function(x) seq(floor(min(x)), ceiling(max(x)), by = 1)

  if (is.null(formula)) {
    p1 <- ggplot(tab) + theme_custom +
      geom_line(aes(x = .data$terms, y = .data$RMSE.raw.complete, color = "Complete"), linewidth = .75, na.rm = TRUE) +
      geom_point(aes(x = .data$terms, y = .data$RMSE.raw.complete), size = 2.5, color = "#33aa55", na.rm = TRUE) +
      geom_line(aes(x = .data$terms, y = .data$RMSE.raw.test, color = "Validation"), linewidth = .75, na.rm = TRUE) +
      geom_point(aes(x = .data$terms, y = .data$RMSE.raw.test), size = 2.5, color = "#1f77b4", na.rm = TRUE) +
      geom_line(aes(x = .data$terms, y = .data$RMSE.raw.train, color = "Training"), linewidth = .75, na.rm = TRUE) +
      geom_point(aes(x = .data$terms, y = .data$RMSE.raw.train), size = 2.5, color = "#d62728", na.rm = TRUE) +
      labs(title = "Raw Score RMSE (1)", x = "Number of terms", y = "Root Mean Squared Error") +
      scale_color_manual(values = c("Training" = "#d62728", "Validation" = "#1f77b4", "Complete" = "#33aa55")) +
      scale_x_continuous(breaks = breaks_step_1)
    print(p1)

    if (norms) {
      p2 <- ggplot(tab) + theme_custom +
        geom_line(aes(x = .data$terms, y = .data$R2.norm.test, color = "Validation"), linewidth = .75, na.rm = TRUE) +
        geom_point(aes(x = .data$terms, y = .data$R2.norm.test), size = 2.5, color = "#1f77b4", na.rm = TRUE) +
        geom_line(aes(x = .data$terms, y = .data$R2.norm.train, color = "Training"), linewidth = .75, na.rm = TRUE) +
        geom_point(aes(x = .data$terms, y = .data$R2.norm.train), size = 2.5, color = "#d62728", na.rm = TRUE) +
        labs(title = expression(paste("Norm Score ", R^2, " (2)")), x = "Number of terms", y = expression(R^2)) +
        scale_color_manual(values = c("Training" = "#d62728", "Validation" = "#1f77b4", "Complete" = "#33aa55")) +
        scale_x_continuous(breaks = breaks_step_1)
      print(p2)

      p3 <- ggplot(tab) + theme_custom +
        geom_line(aes(x = .data$terms, y = .data$Crossfit, color = "Crossfit"), linewidth = .75, na.rm = TRUE) +
        geom_point(aes(x = .data$terms, y = .data$Crossfit), size = 2.5, color = "#1f77b4", na.rm = TRUE) +
        geom_hline(aes(yintercept = 1.10, color = "Overfit"), linetype = "dashed", linewidth = 1, na.rm = TRUE) +
        geom_hline(aes(yintercept = 0.90, color = "Underfit"), linetype = "dashed", linewidth = 1, na.rm = TRUE) +
        labs(title = "Norm Score CROSSFIT (3)", x = "Number of terms", y = "Crossfit") +
        scale_color_manual(values = c("Underfit" = "#FF2728", "Crossfit" = "#1f77b4", "Overfit" = "#AA00AA")) +
        scale_x_continuous(breaks = breaks_step_1)
      print(p3)

      p4 <- ggplot(tab) + theme_custom +
        geom_line(aes(x = .data$terms, y = .data$Delta.R2.test, color = "Delta R2"), linewidth = .75, na.rm = TRUE) +
        geom_point(aes(x = .data$terms, y = .data$Delta.R2.test), size = 2.5, color = "#1f77b4", na.rm = TRUE) +
        geom_hline(aes(yintercept = 0.00, color = "Equal R2"), linetype = "dashed", linewidth = 1, na.rm = TRUE) +
        labs(title = expression(paste("Norm Score ", Delta, R^2, " in Validation (4)")), x = "Number of terms", y = "Delta R2") +
        scale_color_manual(values = c("Equal R2" = "#33aa55", "Delta R2" = "#1f77b4")) +
        scale_x_continuous(breaks = breaks_step_1)
      print(p4)
    } else {
      tab$R2.norm.train <- NULL
      tab$R2.norm.test <- NULL
      tab$Delta.R2.test <- NULL
      tab$Crossfit <- NULL
      tab$RMSE.norm.test <- NULL
    }

    cat("\nOccurrence of selected terms, sorted by frequency:\n")
    print(sort(table(Terms), decreasing = TRUE))

    cat("\nThe simulation yielded the following optimal settings:\n")
    if (norms) {
      cat(paste0("\nNumber of terms with best crossfit: ",
                 which.min((1 - tab$Crossfit)^2)))
      best.norm <- which.max(r2.test)
      FirstNegative <- which(tab$Delta.R2.test <= 0)[1]

      cat(paste0("\nNumber of terms with best norm validation R2: ",
                 best.norm, "\n"))
      cat(paste0("First negative norm score R2 delta in validation: ",
                 FirstNegative))
      cat(paste0("\nNumber of terms with best norm validation RMSE: ",
                 which.min(tab$RMSE.norm.test)))

      cat(paste0("\nChoosing a model with ", (FirstNegative - 1),
                 " terms might be a good choice. For this, use the parameter ",
                 "'terms = ", (FirstNegative - 1), "' in the cnorm-function.\n"))
      cat("\nPlease investigate the plots and the summary table, as the results might vary within a narrow range.")
      cat("\nEspecially pay attention to where RMSE.norm.test and delta R2 stop to progress.")
    }

    cat("\n\n")
    return(tab[min:max, ])

  } else {
    cat("\n\n")
    cat(paste0("Repeated cross validation with prespecified formula and ",
               repetitions, " repetitions yielded the following results:\n\n"))
    tab$Delta.R2.test <- NULL
    return(tab[stats::complete.cases(tab), ])
  }
}


#' Calculates the standard error (SE) or root mean square error (RMSE) of the
#' norm scores. In case of large datasets, both results should be almost
#' identical.
#'
#' @param model a cnorm object
#' @param type either '1' for the standard error sensu Oosterhuis et al. (2016)
#'   or '2' for the RMSE (default)
#'
#' @return The standard error (SE) of the norm scores sensu Oosterhuis et al.
#'   (2016) or the RMSE
#' @export
#'
#' @references Oosterhuis, H. E. M., van der Ark, L. A., & Sijtsma, K. (2016).
#'   Sample Size Requirements for Traditional and Regression-Based Norms.
#'   Assessment, 23(2), 191-202. https://doi.org/10.1177/1073191115580638
getNormScoreSE <- function(model, type = 2) {
  if (!isTaylor(model)) {
    stop("Please provide cnorm object as the model parameter")
  }

  if (type != 1 && type != 2) type <- 2

  data <- model$data
  model <- model$model
  minNorm <- model$minL1
  maxNorm <- model$maxL1
  raw <- data[[model$raw]]
  age <- data[[model$age]]

  fitted <- predictNorm(raw, age, model, minNorm = minNorm, maxNorm = maxNorm)

  diff <- fitted - data$normValue
  diff <- diff[!is.na(diff)]

  if (type == 1)
    return(sqrt(sum(diff^2) / (length(diff) - 2)))
  else
    return(sqrt(mean(diff^2)))
}


#' Build regression function for bestModel
#'
#' @param raw name of the raw score variable
#' @param k the power degree for location
#' @param t the power degree for age
#' @param age use age
#'
#' @return regression function
#' @keywords internal
buildFunction <- function(raw, k, t, age) {
  terms <- paste0("L", 1:k)
  if (age) {
    terms <- c(terms, paste0("A", 1:t))
    terms <- c(terms, as.vector(outer(1:k, 1:t, function(i, j)
      paste0("L", i, "A", j))))
  }
  stats::formula(paste(raw, paste(terms, collapse = " + "), sep = " ~ "))
}


# ---------------------------------------------------------------------------
# Internal helpers for analytic monotonicity checks
# ---------------------------------------------------------------------------

#' Parse Taylor coefficients into an (L power x A power) matrix
#'
#' Maps named regression coefficients following the Taylor naming scheme
#' (\code{(Intercept)}, \code{L#}, \code{A#}, \code{L#A#}) onto a matrix
#' \code{B} with \code{B[l + 1, a + 1]} holding the coefficient of
#' \eqn{L^l A^a}. The intercept is deliberately placed at \code{B[1, 1]}:
#' while a plain monotonicity check would be invariant to the constant term,
#' the clipping-aware check in \code{polyViolatesMonotonicity()} operates on
#' the absolute raw score scale, so the intercept determines which reversals
#' fall inside \code{[minRaw, maxRaw]}.
#'
#' Returns \code{NULL} if the coefficient names do not exclusively follow the
#' Taylor naming scheme (e.g. with custom predictors), or if two names map to
#' the same (L, A) power cell (duplicated names, or degenerate aliases such
#' as \code{L0}), which would otherwise be overwritten silently.
#' @keywords internal
#' @noRd
taylorCoefficientMatrix <- function(coefficients) {
  nm <- names(coefficients)
  if (is.null(nm) ||
      !all(grepl("^(\\(Intercept\\)|L\\d+|A\\d+|L\\d+A\\d+)$", nm)))
    return(NULL)

  lp <- integer(length(nm))
  ap <- integer(length(nm))

  hasL <- grepl("L\\d+", nm)
  hasA <- grepl("A\\d+", nm)

  # conversion restricted to matching names -> no NA coercion warnings
  lp[hasL] <- as.integer(sub("^.*?L(\\d+).*$", "\\1", nm[hasL]))
  ap[hasA] <- as.integer(sub("^.*?A(\\d+).*$", "\\1", nm[hasA]))

  # guard against colliding cells (duplicated names, "L0"/"A0" aliases)
  if (anyDuplicated(cbind(lp, ap)) > 0L) return(NULL)

  B <- matrix(0, nrow = max(lp) + 1L, ncol = max(ap) + 1L)
  B[cbind(lp + 1L, ap + 1L)] <- as.numeric(coefficients)
  B
}


#' Evaluate polynomial (ascending coefficients) via Horner scheme
#' @keywords internal
#' @noRd
evalPolynomial <- function(cf, x) {
  r <- rep(0, length(x))
  for (coefI in rev(cf)) r <- r * x + coefI
  r
}


#' Exact, clipping-aware monotonicity violation check
#'
#' Checks whether p(L) violates monotonicity on \code{[minL, maxL]} in the
#' direction implied by \code{descend}. The interval is partitioned at the
#' real roots of p'(L); within each subinterval p is strictly monotone, so
#' \code{|p(a) - p(b)|} is the exact local reversal depth. A wrong-direction
#' segment only counts as a violation if the part of its raw score span that
#' is *visible after clipping* to \code{[minRaw, maxRaw]} (the effective dip)
#' exceeds \code{minDip}. Reversals entirely outside the clipping bounds thus
#' have an effective dip <= 0 and are ignored.
#'
#' @param pcoef Polynomial coefficients in L, ascending order (constant
#'   first). The constant term must be included for the clipping logic to be
#'   meaningful.
#' @param minL,maxL Interval of the latent location L to check.
#' @param descend If \code{TRUE}, p must be non-increasing; otherwise
#'   non-decreasing.
#' @param minRaw,maxRaw Raw score clipping bounds.
#' @param minDip Minimum raw-score reversal depth (in raw score units) for a
#'   sign violation to be considered a genuine, reportable inconsistency.
#'   Defaults to 1\% of the raw score range (intentional tolerance factor);
#'   falls back to an absolute epsilon if the range is unknown/degenerate.
#' @param dTol Numeric tolerance factor for the derivative sign test
#'   (intentional; scaled by the local magnitude of p).
#' @keywords internal
#' @noRd
polyViolatesMonotonicity <- function(pcoef, minL, maxL, descend = FALSE,
                                     minRaw = -Inf, maxRaw = Inf,
                                     minDip = NULL, dTol = 1e-8) {

  deg <- length(pcoef) - 1L
  if (deg < 1L) return(FALSE)

  dcoef <- pcoef[-1L] * seq_len(deg)

  # strip numerically-zero leading terms; threshold is scale-relative so the
  # behaviour does not depend on the raw score / predictor scaling
  zeroTol <- 1e-14 * max(1, abs(dcoef))
  while (length(dcoef) > 0L && abs(dcoef[length(dcoef)]) < zeroTol)
    dcoef <- dcoef[-length(dcoef)]
  if (length(dcoef) == 0L) return(FALSE)   # flat in L -> nothing to violate

  # tolerance factor (intentional): reversals smaller than 1% of the raw
  # score range are treated as practically irrelevant
  if (is.null(minDip)) {
    rawRange <- maxRaw - minRaw
    minDip <- if (is.finite(rawRange) && rawRange > 0) 1e-2 * rawRange else 1e-6
  }

  breaks <- c(minL, maxL)
  if (length(dcoef) > 1L) {
    rts <- polyroot(dcoef)
    re <- Re(rts)[abs(Im(rts)) < 1e-8 * (1 + abs(Re(rts)))]
    re <- re[re > minL & re < maxL]

    # merge roots that are numerically indistinguishable (root-finding
    # jitter); dips hidden by this merging are far below minDip by design
    if (length(re) > 1L) {
      re <- sort(re)
      minSep <- 1e-6 * (maxL - minL)
      keep <- c(TRUE, diff(re) > minSep)
      re <- re[keep]
    }
    breaks <- sort(unique(c(minL, re, maxL)))
  }

  for (s in seq_len(length(breaks) - 1L)) {
    a <- breaks[s]
    b <- breaks[s + 1L]
    dv <- evalPolynomial(dcoef, (a + b) / 2)
    pa <- evalPolynomial(pcoef, a)
    pb <- evalPolynomial(pcoef, b)

    # numeric tolerance on the derivative sign test (intentional factor),
    # scaled by the local magnitude of p
    tol <- dTol * max(1, abs(pa), abs(pb))

    wrong <- if (descend) dv > tol else dv < -tol
    if (wrong) {
      lo <- min(pa, pb)
      hi <- max(pa, pb)
      # clipping-aware: only the part of the reversal visible inside
      # [minRaw, maxRaw] can matter; entirely clipped reversals yield
      # effDip <= 0 and are skipped automatically
      effDip <- min(hi, maxRaw) - max(lo, minRaw)
      if (effDip > minDip) return(TRUE)
    }
  }
  FALSE
}


#' Subset or reorder rows of a summary.regsubsets object
#'
#' \code{keep} may be a logical vector (filtering) or an integer permutation
#' (reordering). Fields absent from the object are skipped.
#' @keywords internal
#' @noRd
filterSubsetRows <- function(results, keep) {
  results$which  <- results$which[keep, , drop = FALSE]
  results$outmat <- results$outmat[keep, , drop = FALSE]
  for (f in c("adjr2", "cp", "bic", "rss", "rsq"))
    if (!is.null(results[[f]])) results[[f]] <- results[[f]][keep]
  results
}


#' Screen `regsubsets` output for monotonic consistency
#'
#' For each model size, candidate models are checked in order of decreasing
#' R^2 (rows are explicitly re-sorted by size and ascending RSS, so this
#' holds regardless of input order). A model is consistent if its prediction
#' is monotone in L, in the direction implied by \code{descend}, across all
#' evaluated ages -- verified analytically in L via
#' \code{polyViolatesMonotonicity()} (the same routine used by
#' \code{checkConsistency(method = "analytic")}), which is clipping-aware
#' (only the reversal depth visible inside \code{[minRaw, maxRaw]} counts)
#' and tolerant of practically negligible reversals via \code{minDip}.
#'
#' Note that the check is exact in L but *sampled* in age: the age grid is
#' automatically densified to at least \code{2 * maxApow + 1} points (where
#' \code{maxApow} is the highest age power in the candidate terms) as a
#' heuristic safeguard against interaction-driven violations between grid
#' points. The final \code{checkConsistency()} call remains the
#' authoritative verdict.
#'
#' Models that do not depend on L at all are treated as
#' degenerate/inconsistent (they cannot serve for norming), as are
#' rank-deficient fits. If no consistent model exists for a size, the best
#' model of that size is retained as fallback (flagged inconsistent). At
#' most one model per size is retained. \code{highestConsistent} reports the
#' *number of terms* of the largest consistent model (not a row index), or
#' is absent if no model is consistent.
#'
#' @param data1 Data frame containing the Taylor basis columns (including
#'   \code{L1} and \code{A1}); its \code{"descend"} attribute supplies the
#'   default direction.
#' @param results A \code{summary.regsubsets} object.
#' @param raw Numeric vector of raw scores (response used in the subset
#'   selection).
#' @param nAgePoints Minimum number of age grid points; automatically raised
#'   to \code{2 * maxApow + 1}.
#' @param weights Optional observation weights (must match \code{raw} in
#'   length); used to refit each candidate via \code{lm.wfit}.
#' @param minRaw,maxRaw Raw score clipping bounds used for the
#'   clipping-awareness of the monotonicity check. Default to the observed
#'   raw score range.
#' @param descend Expected direction of the raw~L relationship. Defaults to
#'   \code{attr(data1, "descend")}; if that attribute is absent, ascending
#'   is assumed.
#' @param minDip Minimum practically relevant raw-score reversal depth passed
#'   through to \code{polyViolatesMonotonicity()}. \code{NULL} (default) lets
#'   that function pick a self-scaling default based on the raw score range.
#' @keywords internal
#' @noRd
screenSubset <- function(data1, results, raw, nAgePoints = 4,
                         weights = NULL, minRaw = NULL, maxRaw = NULL,
                         descend = NULL, minDip = NULL) {
  all_vars <- colnames(results$outmat)
  y <- as.numeric(raw)

  if (!is.null(weights) && length(weights) != length(y))
    stop("'weights' must have the same length as 'raw'.")

  nTerms <- as.integer(rowSums(results$outmat == "*"))

  # the screening logic requires rows sorted by size, best (lowest RSS)
  # model first within each size -- enforce rather than assume
  ord <- order(nTerms, results$rss)
  if (!identical(ord, seq_along(ord))) {
    results <- filterSubsetRows(results, ord)
    nTerms <- nTerms[ord]
  }
  nModels <- length(nTerms)

  # screening requires pure, numeric Taylor terms and the L1/A1 columns
  taylorOnly <- all(grepl("^(L\\d+|A\\d+|L\\d+A\\d+)$", all_vars)) &&
    all(all_vars %in% colnames(data1)) &&
    all(vapply(data1[all_vars], is.numeric, logical(1))) &&
    !is.null(data1$L1) && !is.null(data1$A1)

  if (!taylorOnly) {
    # cannot screen custom predictors: keep the best model per size
    keepFirst <- !duplicated(nTerms)
    results1 <- filterSubsetRows(results, keepFirst)
    results1$consistent <- rep(NA, sum(keepFirst))
    results1$highestConsistent <- NULL
    return(results1)
  }

  X_full <- cbind(`(Intercept)` = 1,
                  as.matrix(data1[, all_vars, drop = FALSE]))
  Lvals <- data1$L1
  Avals <- data1$A1

  # lm.fit()/lm.wfit() cannot handle missing values -- restrict to
  # complete cases (consistently for X, y, weights, and the L/A ranges)
  cc <- stats::complete.cases(X_full, y)
  if (!all(cc)) {
    X_full <- X_full[cc, , drop = FALSE]
    y      <- y[cc]
    Lvals  <- Lvals[cc]
    Avals  <- Avals[cc]
    if (!is.null(weights)) weights <- weights[cc]
  }

  minL <- min(Lvals)
  maxL <- max(Lvals)
  ages <- seq(min(Avals), max(Avals), length.out = nAgePoints)

  if (is.null(minRaw)) minRaw <- min(y)
  if (is.null(maxRaw)) maxRaw <- max(y)
  if (is.null(descend)) descend <- isTRUE(attr(data1, "descend"))

  consistent      <- rep(FALSE, nModels)
  currentNumber   <- 0L
  consistentFound <- FALSE

  for (i in seq_len(nModels)) {
    if (nTerms[i] > currentNumber) {
      currentNumber   <- nTerms[i]
      consistentFound <- FALSE
    }
    if (consistentFound) next

    sel <- c(TRUE, results$outmat[i, ] == "*")
    fit <- if (is.null(weights)) stats::lm.fit(X_full[, sel, drop = FALSE], y)
    else stats::lm.wfit(X_full[, sel, drop = FALSE], y, w = weights)
    cf <- fit$coefficients

    if (anyNA(cf)) next                       # rank deficient -> inconsistent

    B <- taylorCoefficientMatrix(cf)
    if (is.null(B) || nrow(B) < 2L) next      # no L-dependency -> degenerate

    # invariant: the intercept must survive into B[1, 1] -- it anchors the
    # absolute raw score scale on which clipping and minDip operate
    stopifnot(isTRUE(all.equal(unname(B[1L, 1L]),
                               unname(cf[["(Intercept)"]]))))

    tPow <- ncol(B) - 1L
    violated <- FALSE
    for (a in ages) {
      pcoef <- as.vector(B %*% a^(0:tPow))
      if (polyViolatesMonotonicity(pcoef, minL, maxL, descend = descend,
                                   minRaw = minRaw, maxRaw = maxRaw,
                                   minDip = minDip)) {
        violated <- TRUE
        break
      }
    }

    ok <- !violated
    consistent[i]   <- ok
    consistentFound <- ok
  }

  # fallback: if no consistent model exists for a size, keep the best one
  keepFlag <- consistent
  for (term in unique(nTerms)) {
    idx <- which(nTerms == term)
    if (!any(consistent[idx])) keepFlag[idx[1]] <- TRUE
  }

  results1 <- filterSubsetRows(results, keepFlag)
  results1$consistent <- consistent[keepFlag]

  # number of terms of the largest consistent model (not a row position)
  keptTerms <- nTerms[keepFlag]
  results1$highestConsistent <- if (any(results1$consistent))
    max(keptTerms[results1$consistent]) else NULL

  results1
}

---
title: "Modelling Norms with the Sinh-Arcsinh (shash) Distribution"
author: "Wolfgang Lenhard & Alexandra Lenhard"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
#output: 
#  pdf_document:
#    latex_engine: xelatex
#    toc: true  # Optional table of contents
#    number_sections: true  # Optional section numbering
vignette: >
  %\VignetteIndexEntry{Modelling Norms with the Sinh-Arcsinh (shash) Distribution}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
options(width = 60)

#devtools::load_all(".")
library(cNORM)
```

The Sinh-Arcsinh (shash) distribution is very flexible and can adjust to a wide range of distributional shapes. This makes it particularly useful for modeling psychometric data, which often exhibit non-normal characteristics such as skewness and heavy tails. Unlike discrete distributions (e.g., beta-binomial), shash naturally handles continuous scores, decimal values, and unbounded measures. It can accommodate scores of zero and negative values without transformation, unlike Box-Cox approaches that require strictly positive data. Finally, skewness (ε) and tail weight (δ) can be controlled independently, allowing precise modeling of distributional characteristics. This makes it possible to model complex developmental patterns where skewness and tail weight change with age.

# Mathematical Foundation
The Sinh-Arcsinh (shash; Jones & Pewsey, 2009) distribution is defined by a transformation of a standard normal variable. If $Y$ follows a standard normal distribution, $Y \sim N(0,1)$, then the shash-distributed variable $X$ is generated by:

$$X = \mu + \sigma \cdot \sinh\left(\frac{\text{arcsinh}(Y) - \epsilon}{\delta}\right)$$

This transformation allows the resulting variable $X$ to have its location, scale, skewness, and tail weight controlled by the four parameters:

-   **μ** (mu): location parameter, shifts the distribution horizontally (similar to mean)
-   **σ** (sigma): scale parameter ($\sigma > 0$), controls the spread of the distribution (similar to standard deviation)
-   **ε** (epsilon): skewness parameter ($\epsilon = 0$ for symmetry, $\epsilon > 0$ for right skew, $\epsilon < 0$ for left skew)
-   **δ** (delta): tail weight parameter ($\delta = 1$ produces normal-like tails, $\delta > 1$ produces heavier tails, $\delta < 1$ produces lighter tails)

The probability density function involves:

$$f(x|\mu,\sigma,\epsilon,\delta) = \frac{\delta}{\sigma\sqrt{2\pi}} \cdot \frac{\cosh(\delta \cdot \text{arcsinh}(z) + \epsilon)}{\sqrt{1 + z^2}} \cdot \exp\left(-\frac{1}{2}[\sinh(\delta \cdot \text{arcsinh}(z) + \epsilon)]^2\right)$$

where $z = \frac{x - \mu}{\sigma}$. The cumulative distribution function (CDF) does not have a closed-form expression but can be computed numerically. For a given value $x$, the CDF is:

$$F(x|\mu,\sigma,\epsilon,\delta) = P(X \leq x) = \Phi[\sinh(\delta \cdot \text{arcsinh}(z) + \epsilon)]$$

where $\Phi$ is the standard normal CDF and $z = (x - \mu)/\sigma$. The quantile function (inverse CDF) can be expressed as:

$$Q(p|\mu,\sigma,\epsilon,\delta) = \mu + \sigma \cdot \sinh\left(\frac{\text{arcsinh}(\Phi^{-1}(p)) - \epsilon}{\delta}\right)$$

## Modeling the Sinh-Arcsinh over Age

In cNORM's shash implementation, we model the four parameters as polynomial functions of standardized age with mu_degree = 3, sigma_degree = 2, epsilon_degree = 2 and delta_degree = 1 as the default settings. It is advisable to keep the delta_degree parameter low, for example 1 or 2, to avoid overfitting. The tail weight parameter δ can as well be held constant by setting the 'delta_degree' parameter to NULL. In this case, a fixed delta (default = 1) will be used to reflect population characteristics, e. g. by increasing it to values delta > 1 for heterogenous samples or delta < 1 for homogenuous samples. Age is standardized as: $\text{age}_{std} = \frac{\text{age} - \overline{\text{age}}}{\text{SD}(\text{age})}$ for numerical stability during optimization. The parameters are estimated using maximum likelihood estimation.


## Application and Prerequisites

The shash distribution is particularly well-suited for continuous performance measures (reaction times, achievement scores with decimal precision), tests with floor or ceiling effects where skewness varies across age groups, heterogeneous populations with varying degrees of individual differences, change or difference scores that may include negative values, adaptive tests or measures with variable stopping rules and large-scale assessments, where distributional assumptions are complex. And of course, it is very suited for developmental studies, since the distribution shape can change systematically with age.

When applied to normative data, there are some prerequisites and considerations to ensure optimal model performance and valid results. The test scores should exhibit systematic (though not necessarily monotonic) development across the predictor variable. The shash model can capture complex non-linear relationships through polynomial fitting of the distribution parameters. The norming sample should include a sufficient sample size. It is usually much lower than in conventional norming. Nonetheless, as a rule of thumb, we recommend a minimum of 100 cases per major age group, with larger samples needed for higher polynomial degrees or complex developmental patterns. Despite empirical validation of this rule is ongoing, these sample sizes have proven effective in previous norming studies (Lenhard et al., 2019).

While shash handles continuous scores optimally, it can also model discrete scores effectively, especially when the number of possible values is large (e. g. value range > 20). As always, representativeness of the sample is important. If the sample deviates from the target population, though post-stratification weighting can help address moderate deviations from representativeness.


## Model Selection Considerations

It is advisable to explore the norm data sample prior to modeling. The polynomial degrees for the four parameters should be selected based on theoretical expectations about developmental trajectories, sample size, and model comparison criteria (AIC, BIC, cross-validation). Visual inspection of fitted curves is essential to ensure realistic patterns. Overly complex models may lead to overfitting, especially with limited data. Thus, choose polynomial degrees based on:

- Theoretical expectations about developmental trajectories
- Sample size (higher degrees require more data to avoid overfitting)
- Model comparison criteria (AIC, BIC, cross-validation)
- Visual inspection of fitted curves for realistic patterns

When adjusting the δ parameter, you can either keep it constant or model it as a polynomial function of age using the delta_degree parameter. If the delta_degree parameter is used, keep it low (1 or 2, default = 1) to avoid overfitting. In that case, the default δ is not used. To explicitely keep delta constant, please set 'delta_degree = NULL'. The delta parameter should reflect population characteristics:

- δ = 0.7-0.9: Homogeneous populations, selected samples
- δ = 1.0: General population samples with normal-like variability
- δ = 1.2-2.0: Heterogeneous populations with high individual differences


# Modeling Example

We demonstrate Sinh-Arcsinh (shash) modeling using the PPVT-4 vocabulary development dataset. Please have a look at the beta-binomial vignette for a comparison with discrete models.

## Basic Model Fitting

```{r fig1, fig.height = 4, fig.width = 7}
# Fit shash model with custom settings
# The function automatically displays percentile plots
model.shash <- cnorm.shash(age = ppvt$age, score = ppvt$raw)


# Use print(model.shash), diagnostics(model.shash) or summary(model.shash)
# to retrieve information on the data fit.
```

The model uses reasonable default polynomial degrees per default and provides immediate visual feedback through percentile plots. The output includes fitted parameters, convergence information, and basic model statistics. Model diagnostics (summary, diagnostics) provides:

- Model fit statistics (log-likelihood, AIC, BIC, R²)
- Parameter estimates with standard errors and significance tests
- Convergence diagnostics
- Separate tables for location, scale, and skewness parameters

Please pay attention to convergence (should be successful, if not, please inspect model visually), R² (correlation between fitted and manifest percentiles; > 0.95 desirable) and general fit statistics.


## Custom Model Specifications

For datasets with complex patterns, adjust polynomial degrees and distributional parameters:

```{r, eval=FALSE}
# Conservative parameterization with fixed delta across age
model.simple <- cnorm.shash(age = ppvt$age, score = ppvt$raw,
                           mu_degree = 2,        # Quadratic location pattern
                           sigma_degree = 1,     # Linear variability change  
                           epsilon_degree = 1,   # Linear skewness change
                           delta_degree = NULL,  # deactivates polynimial fitting for delta
                           delta = 1.1)          # Slightly heavy tails,
                                                 # kept constant across age

# Example with more complex parameterization
model.complex <- cnorm.shash(age = ppvt$age, score = ppvt$raw,
                           mu_degree = 4,        # Quadric pattern
                           sigma_degree = 3,     # Complex variability changes  
                           epsilon_degree = 2,   # Quadratic age-varying skewness
                           delta_degree = 2)     # Changing tail weights across age (quadratic)

# Compare models
compare(model.simple, model.complex, age = ppvt$age, score = ppvt$raw,
        title = "ShaSh Model Comparison")


```

**Note**: Higher polynomial degrees increase model flexibility but may lead to overfitting with insufficient data. Always validate complex models through visual inspection and preferable cross-validation.


## Post-Stratification and Weighting

Like betabinomial and distribution free models, cNORM supports post stratification in shash distributions via iterative post stratification to approximate representativity:

```{r, eval=FALSE}
# Calculate post-stratification weights
margins <- data.frame(variables = c("sex", "sex", "migration", "migration"),
                     levels = c(1, 2, 0, 1),
                     share = c(.52, .48, .7, .3))

weights <- computeWeights(ppvt, margins)

# Fit weighted ShaSh model
model.weighted <- cnorm.shash(ppvt$age, ppvt$raw, weights = weights)

# Compare weighted vs. unweighted
compare(model.shash, model.weighted, age = ppvt$age, score = ppvt$raw,
        title = "Unweighted vs. Weighted ShaSh Models")
```


# Norm Score Generation

Norm scores can be calculated for individual cases (or vectors of cases) or as comprehensive norm tables for specified ages.

```{r eval=FALSE}
# Individual Norm Score Prediction:
# Generate norm scores for specific age-score combinations
ages <- c(10.25, 10.75, 11.25, 11.75)
raw_scores <- c(180, 185, 190, 195)

norm_scores <- predict(model.shash, ages, raw_scores)
prediction_table <- data.frame(
  Age = ages, 
  Raw_Score = raw_scores, 
  Norm_Score = round(norm_scores, 1)
)
print(prediction_table)


# Norm Score tables:
# Generate detailed norm tables for multiple ages
tables <- normTable.shash(model.shash, 
                         ages = c(10.25, 10.75), 
                         start = 150, 
                         end = 220,
                         step = 1,
                         CI = 0.95, 
                         reliability = 0.94)

# Display head from first table
head(tables[[1]], 10)
```

The norm tables provide:

- **x**: Raw scores
- **Px**: Probability density values
- **Pcum**: Cumulative probabilities
- **Percentile**: Percentile ranks (0-100)
- **z**: Standardized z-scores  
- **norm**: Norm scores in specified scale
- **Confidence intervals**: When reliability is specified


# Model Comparison and Selection

## Comparing Distributional Approaches

```{r fig5, fig.height = 4, fig.width = 7}
# Compare shash with betabinomial models (BB). BB models should work worse,
# since the test has stop rules, leading to non-binomial distributions.
model.bb <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228, plot = FALSE)

# Model comparisons
compare(model.shash, model.bb, age = ppvt$group, score = ppvt$raw,
        title = "SinH-ArcSinH vs. Beta-Binomial")


# Compare distribution free Taylor model
model.taylor <- cnorm(group = ppvt$group, raw = ppvt$raw, plot=FALSE)

# Model comparisons shash versus taylor
compare(model.shash, model.taylor, age = ppvt$group, score = ppvt$raw,
        title = "SinH-ArcSinH vs. Taylor")
```

## Decision Framework

The different continuous models have advantages for specific use cases, with the distribution free approach (Taylor polynomials) being the most flexible, the betabinomial approach being optimal for discrete item counts (e.g. 1PL IRT models), and shash being ideal for continuous scores with complex distributional shapes.

Use shash models for

- continuous scores with decimal precision
- complex skewness patterns that change across age
- floor or ceiling effects requiring flexible shape modeling
- zero or negative scores are present


Use Beta-Binomial models for

- discrete item counts from binary scoring
- fixed maximum score (number of test items)
- unspeeded tests
- 1PL IRT-based psychometric instruments
- small to moderate number of possible scores


Use distribution-free (Taylor Polynomial) models when

- maximum flexibility is paramount
- no strong distributional assumptions can be made
- quick implementation is needed and manual adjustment is desired


In selecting models, please compare models using:

1. Information criteria: AIC, BIC (lower is better)
2. Fit statistics: R², RMSE, bias
3. Visual inspection: Smoothness, realism of percentile curves
4. Cross-validation: Out-of-sample prediction accuracy
5. Theoretical appropriateness: Match between model assumptions and data characteristics


# Conclusion

The Sinh-Arcsinh (shash) distribution provides a sophisticated yet practical framework for modeling psychometric norm data. Its ability to independently control location, scale, skewness, and tail weight makes it particularly valuable for applications where traditional parametric approaches are inadequate due to distributional complexity. Best practices for shash modeling:

1. Carefully assess data characteristics before model selection
2. Use appropriate sample sizes for model complexity
3. Validate models through visually and via statistical indicators (R2, AIC, Bias)
4. Compare alternative distributional assumptions. Is distribution-free or beta-binomial more appropriate?
5. Document model selection rationale


# References

Jones, M. C., & Pewsey, A. (2009). Sinh-arcsinh distributions. *Biometrika*, 96(4), 761-780.

Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. *PLoS ONE*, 14(9), e0222279. https://doi.org/10.1371/journal.pone.0222279

---
title: "Weighted Regression-Based Norming"
author: "Sebastian Gary, Wolfgang Lenhard, Alexandra Lenhard"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
#output: 
#  pdf_document:
#    latex_engine: xelatex
#    toc: true  # Optional table of contents
#    number_sections: true  # Optional section numbering
vignette: >
  %\VignetteIndexEntry{Weighted Regression-Based Norming}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
options(width = 60)

library(cNORM)
library(ggplot2)
```

# Enhancing Representativeness with Weighted Regression-Based Norming in cNORM
Acquiring valid norm scores fundamentally relies on the representativeness of the norm sample. Traditionally, random sampling is employed to meet this end, but even this approach may result in a sample that diverges from the population structure. The *cNORM* R package provides a potent solution by incorporating sampling weights into the norming process, thereby diminishing the impact of non-representative norm samples on the norm score quality. A key component of this process is raking, or iterative proportional fitting, which allows for the post-stratification of the norm sample based on stratification variables (SVs), considering their population marginals.

## Impact of Non-Representative Norm Samples
When a norm sample inadequately represents the target population, especially regarding pertinent stratification variables, it can compromise the quality of the resulting norm scores (Kruskal & Mosteller, 1979). An illustration of this issue is the neglect of parental educational background when determining norm scores for a children's intelligence test. Such oversight can lead to skewed norm scores, thus distorting the child's actual intelligence level (Hernandez et al., 2017). As norm scores commonly inform significant decisions, like school placement or the diagnosis of learning disabilities (Gary & Lenhard, 2021; Lenhard, Lenhard & Gary, 2019; Lenhard & Lenhard, 2021), any bias can potentially disadvantage those being evaluated. Therefore, techniques such as sample weighting methods become crucial to mitigate non-representativeness in norm samples.

## Raking: Enhancing Sample Representativeness through Iterative Proportional Fitting
Raking, also called iterative proportional fitting, is a post-stratification approach targeted to enhance sample representativeness with respect to two or more stratification variables . For this purpose, sample weights are computed for every case in the norm sample based on the ratio between the proportion of the corresponding strata in the target population and the proportion in the actual norm sample (Lumley, 2011). The procedure can be described as an iterative post-stratification with respect to one variable in each step. For example, let's assume a target population containing 49% female as well as 51% male persons, while the resulting norm sample contains 45% female and 55% male subjects. To enhance the representativeness of the norm sample with respect to the SV sex (female/male), every single female case would be weighted with $w_{female}=\frac{49\%}{45\%}=1.09$ and every single male case with $w_{male}=\frac{55\%}{51\%}=0.93$. 
For stratifying a norm sample with respect to two or more variables, for example sex(female/male) and education(low/medium/high), the before described adaptation is applied several times regarding the marginals of one variable by time iteratively. For example, if the weights are adapted with respect to the variable sex first, the weights would be adapted regarding education in the second step. Since the weights no longer represent the population with respect to variable sex after the second step, the weights are computed to SV sex in the third step respectively to education in the fourth step and so on until the corresponding raking weights are converged. Finally, the resulting raking weights respectively the weighted norm sample represents the target population with respect to the marginal proportions of the used SVs. Each case is assigned with an according weight in a way that the proportions of the strata in the norm sample aligns with the composition of the representative population.

## Integration of raking weights in regression-based norming in cNORM
The integration of raking weights in *cNORM* is accomplished in three steps.

*  Computation and standardization of raking weights
*  Initial ranking of test raw scores using standardized raking weights with weighted percentile estimation
*  Regression-based norming with standardized regression weights

### Step 1: Computation and standardization of raking weights
Raking weights are computed regarding the proportions of the SVs in the target population and the actual norm sample. Afterwards, the resulting raking weights are standardized by dividing every weight by the smallest resulting raking weight, i.e., the smallest weight is set to 1.0, while the ratio between one weight and each other remains the same. Consequently, underrepresented cases in the sample are weighted with a factor larger 1.0. To compute the weights, please provide a data frame with three columns to specify the population marginals. The first column specifies the stratification variable, the second the factor level of the stratification variable and the third the proportion for the representative population. The function 'computeWeights()' is used to retrieve the weights. The original data and the marginals have to be passed as function parameters.

### Step 2: Weighted percentile estimation
Secondly, the norm sample is ranked with respect to the raking weights using weighted percentile. This step is the actual start of the further regression-based norming approach and it is automatically applied in the 'cnorm()' function, as soon as weights are specified.

### Step 3: Regression-based norming with standardized regression weights
Finally, the standardized raking weights are used in the weighted best-subset regression to obtain an adequate norm model. While the former steps can be seen as kind of data preparation, the computation of the regression-based norm model represents the actual norming process, since the resulting regression model is used for the actual mapping between achieved raw score and assigned norm score. By using the standardized raking weights in weighted regression, an overfit of the regression model with respect to overrepresented data points should be reduced. This third step is as well applied automatically when using the 'cnorm()' function.

## Example
In the following, the usage of raking weights in regression-based norming with *cNORM* is illustrated in detail based the on a not representative norm sample for the German version of the *Peabody Picture Vocabulary Test* (PPVT-IV)
```{r message = FALSE}
library(cNORM)
# Assign data to object norm.data
norm.data <- ppvt
head(norm.data)
```
For the post-stratification, we need population marginals for the relevant stratification variables as a data frame, with each level of each stratification variable in a row. The data frame must contain the names of the SVs (column 1), the single levels (column 2) and the corresponding proportion in the target population (column 3).

```{r}
# Generate population marginals
marginals <- data.frame(var = c("sex", "sex", "migration", "migration"),
                             level = c(1,2,0,1),
                             prop = c(0.51, 0.49, 0.65, 0.35))
head(marginals)
```

To calculate raking weights, the cNORM's 'computeWeights()' function is used, with the norm sample data and the population marginals as function parameters. 
```{r message=FALSE}
weights <- computeWeights(data = norm.data, population.margins = marginals)
```

Using the 'cnorm()' function passing the raking weights by function parameter 'weights', the initial weighted ranking and the actual norming process is started.
```{r message = FALSE, results = 'hide', warning=FALSE, fig0, fig.height = 4, fig.width = 7}
norm.model <- cnorm(raw = norm.data$raw, group = norm.data$group,
					weights = weights, terms=6)
```
The resulting model contains four predictors with a RMSE of 3.54212.

```{r}
summary(norm.model)
```
Moreover, the percentile plot reveals no hints on model violation, like intersecting percentile curves.
```{r message = FALSE, results = 'hide', warning=FALSE, fig1, fig.height = 4, fig.width = 7}
plot(norm.model, "subset")
plot(norm.model, "norm")
```

## Caveats and recommendation for use
We extensively simulated biased distributions and assessed, if our approach can mitigate the effects of unrepresentative samples. cNORM itself already corrects for several types of sampling eror, namely if deviations occur in specific age groups or if joint probabilities of stratification variables are unbalanced (while preserving the marginals). Weighted Continuous Norming as well works very well in most, but not all use cases. Please note the following: 

*  Non-representativeness in most cases leads to (moderately) increased error of the normed scores. It is - of  course - always better to ensure the highest feasible degree of representativeness in the data collection.
*  The data collection should be as random as possible.
*  In most but not in all cases, Weighted Continuous Norming reduces negative effects of non-representative norm samples. If the mean of the standardized weights exceeds a value of $m_{weights}=2$, this is an indication to rather not use weighting.  
*  With cNORM, representativeness need not necessarily be established in every single age group. If the marginals are more or less correct, weighting is unnecessary.
*  Only use stratification for variables with considerable influence on the dependent variable.
*  If available, the probabilities of cross-classifications of the stratification variables can be used. You can recode several variables into one and directly specify the according population marginals (especially in combination with the next point).
*  Avoid too many stratification variables with many fine-grained levels. This leads to high weights in specific subgroups. Rather combine different levels of stratification variables, if the according subgroups do not differ in the outcome variable.


# References 

* Gary, S., & Lenhard, W., (2021). In norming we trust - Verfahren zur statistischen Modellierung kontinuierlicher Testnormen auf dem Prüfstand. *Diagnostica*, 67(2), 75 - 86.
* Hernandez, A., Aguilar, C., Paradell, È., Munoz, M., Vannier, L.C., & Vallar, F. (2017). The effect of demographic variables on the assesment of cognitive ability. *Psicothema*. *29*(4), 469 - 474.
* Kruskal, W., & Mosteller, F. (1979). Representative sampling, III: The current statistical literature. *International Statistical Review/Revue Internationale de Statistique*, 245 - 265
* Lenhard, A., Lenhard, W., Segerer, R., & Suggate, S. (2015). *Peabody Picture Vocabulary Test (PPVT-4)*. Frankfurt: Pearson Clinical Assesment. 
* Lenhard, A., Lenhard, W., & Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. *PloS one*, *14*(9), e0222279.
* Lenhard, W., & Lenhard, A. (2021). Improvement of norm score quality via regression-based continuous norming. *Educational and Psychological Measurement*, *81*(2), 229 - 261.
* Lumley, T. (2011). *Complex surveys: A guide to analysis using R* (Vol. 565); John Wiley & Sons.
* Mercer, A., Lau, A., & Kennedy, C. (2018). For weighting online opt-in samples, what matters most. *Pew Research Center*.

---
title: "Modelling Norms with the Beta-Binomial Distribution"
author: "Wolfgang Lenhard & Alexandra Lenhard"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
#output: 
#  pdf_document:
#    latex_engine: xelatex
#    toc: true  # Optional table of contents
#    number_sections: true  # Optional section numbering
vignette: >
  %\VignetteIndexEntry{Modelling Norms with the Beta-Binomial Distribution}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(tidy.opts=list(width.cutoff=60),
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

An alternative to distribution-free modeling of norm data is parametric modeling. In this approach, the parameters of specific distribution functions (e.g., normal distribution, Box-Cox power function, beta-binomial distribution, etc.) are adjusted across age. One challenge with this approach is finding a function that fits the manifest data well. Note that for the normal distribution, this is usually not the case (cf. A. Lenhard et al., 2019).

In certain cases, however, the beta-binomial distribution is excellently suited for modeling norm scores (Timmermann, Voncken & Albers, 2021). This is especially true for psychometric test scales with a fixed number of questions or items that are completed without time constraints, and where each individual item can only be scored as 0 or 1.

As the name suggests, the beta-binomial distribution combines the binomial distribution with the beta distribution. The binomial distribution can be used to determine the probability of achieving x successes in n trials. The binomial distribution assumes a constant probability of success. However, this is usually not the case in psychometric test scales, especially in performance tests. This is where the beta distribution comes into play. In the beta-binomial distribution, it is assumed that the probability of success is not constant, but that the item difficulties follow a beta distribution. The latter does not represent a specific distribution, but a family of distributions determined by two shape parameters $\alpha$ and $\beta$. Although the item difficulties in most test scales are probably not exactly beta-distributed, test results can usually be modeled very well with the beta-binomial distribution, at least for unspeeded tests.

Since version 3.2, cNORM is not only able to model psychometric data across age using Taylor polynomials, but also using the beta-binomial distribution (additionally, since 3.5, modelling with the four parametric ShaSh distribution is possible). In the following we will show the mathematical derivation, explain the requirements, demonstrate a practical modeling example and present the results of a simulation study. If you prefer a graphical
user interface, you can start the included shiny app via cNORM.GUI2().


# Mathematical Derivation
The beta-binomial distribution is a combination of the binomial and the beta distribution. For a test with n items, which are scored dichotomously with 0 (= false) and 1 (= correct), let X be the number of correct responses. The probability function of the beta-binomial distribution is given by: 

\[P(X = k | n, \alpha, \beta) = \binom{n}{k} \frac{B(k + \alpha, n - k + \beta)}{B(\alpha, \beta)}\]

where $B(\cdot,\cdot)$ is the beta function, and $\alpha$ and $\beta$ are shape parameters of the beta distribution.

In cNORM, we model the shape parameters $\alpha$ and $\beta$ of the beta-binomial distribution as functions of age (or another explanatory variable) using polynomial regression. Specifically: 

\[\log(\alpha(age)) = a_0 + a_1age + a_2age^2 + ... + a_mage^m\]
\[\log(\beta(age)) = b_0 + b_1age + b_2age^2 + ... + b_sage^s\]

where m and s are the degrees of the polynomials for $\alpha$ and $\beta$ respectively. We use the logarithm of $\alpha$ and β to ensure that they are positive, since the beta-binomial distribution is only defined for positive $\alpha$ and $\beta$. This transformation also helps in stabilizing the variance and improving the optimization process. The mean $\mu$ and variance $\sigma^2$ of the beta-binomial distribution can be derived from $\alpha$ and $\beta$ as follows: 

\[\mu = \frac{n\alpha}{\alpha + \beta}\]
\[\sigma^2 = \frac{n\alpha\beta(\alpha + \beta + n)}{(\alpha + \beta)^2(\alpha + \beta + 1)}\]

To estimate the parameters ($a_0, ..., a_m, b_0, ..., b_s$), we use maximum likelihood estimation. The log-likelihood function for N observations is:

\[L(a, b | X, Age) = \sum_{i=1}^N \log[P(X_i | n, \alpha(Age_i), \beta(Age_i))]\]

where $X_i$ is the score and $Age_i$ is the age for the i-th observation.

The data fitting is performed using numerical optimization techniques, specifically the L-BFGS-B (Limited-memory BFGS) algorithm of the 'optim' function. The latter is a quasi-Newton method for solving large nonlinear optimization problems with simple bounds. By approximating the Hessian matrix, it simultaneously determines the coefficients of the regression equations for $\alpha$ and $\beta$ that maximize the log-likelihood, thus providing the best fit to the observed data.


# Prerequisites
To our knowledge, it has not yet been systematically investigated empirically under which circumstances the beta-binomial distribution yields the best results and when distribution-free modeling should be preferred. Nevertheless, there are some requirements that should be fulfilled if you want to use the beta-binomial distribution when modeling your data with the cNORM package.



- **Systematic development across age or the explanatory variable**: The test raw scores must show a systematic, but not necessarily monotonic, development across age (resp. the explanatory variable). By default, the modeling can capture nonlinear relationships across age. This applies to both the development of the mean and the standard deviation of the raw scores. For more complex trajectories, the polynomial degrees for age adjustment of both shape parameters can be increased. However, ensure that sufficiently large sample sizes are available in these cases to avoid overfitting.

- **Discrete outcomes**: Only positive integers including zero are permitted as test raw scores. Decimal numbers, fractions or negative scores are not suitable for fitting with the beta-binomial distribution.

- **Fixed maximum score**: The test should have a predefined maximum score or a fixed number of items. It cannot be ruled out that the beta-binomial distribution also works for test scales with an open-ended score. However, the beta-binomial distribution assumes, at least theoretically, that there is a discrete number of trials and therefore also a maximum number of successes.

- **No time contraints**: The test in question should be unspeeded. At least theoretically, the beta-binomial distribution assumes that the raw score distribution depends solely on the distribution of the item difficulties and the number of items. However, in speeded tests, the probability of solving an item also depends on its position within the test. For example, people with low processing speed are less likely to solve an item far back in the test than would be expected based on the item difficulty alone. In line with this hypothesis, we observed in simulation studies that the fit of the beta-binomial distribution deteriorates as the speed component of a test increases.

- **Size and representativeness of the norm sample**: As always, the result of norm score modeling depends on the quality of the norm data. The norm sample should be representative of the reference population and it should be large enough to sufficiently capture the variability of the personality trait to be measured. Slight to moderate deviations of the standard sample from representativeness can be partially corrected by post-stratification and weighting. However, this is not necessarily guaranteed in the case of strong deviations. For continuous norming, a minimum of 100 cases per age cohort is generally recommended. This recommendation results from the need to be able to estimate the parameters of the model reliably enough. However, if several stratification variables are used or if the relation between raw score distribution and explanatory variable is complex, larger samples may also be necessary.

- **Absence of outliers or influential observations**: It is crucial to examine the data for outliers prior to modeling, as these anomalous points could disproportionately influence the model fit. Such outliers may arise from data entry errors or invalid test conditions. If such reasons are identified, the affected data points should be excluded from the modeling process. (Note that it is important to document any data exclusions and provide justification for their removal to maintain transparency in the analysis.)


# Modeling Example

In the following, we will demonstrate parametric continous norming based on the beta-binomial distribution using the example of the development of receptive vocabulary in childhood and adolescence. For this purpose, we have integrated a data set into cNORM that was collected using the German version of the Peabody Picture Vocabulary Test 4 (PPVT-4). If you would like to reproduce the modeling yourself please make sure that you have installed a current version of cNORM.

## Data Preparation
With the following code you can load the cNORM library and view the structure of the ppvt data set:

```{r fig0, fig.height = 4, fig.width = 7}
 ## Loads the package and displays the data

library(cNORM)
str(ppvt)
plot(ppvt$age, ppvt$raw, main="PPVT Raw Scores by Age",
              xlab="Age", ylab="Raw score")
```

The dataset contains two columns that encode the explanatory variable: a continuous age variable and a discrete group variable (see also Data Preparation). The latter is primarily used for visualization purposes in parametric norming. Furthermore, the dataset includes the raw score and three background variables that can be used for stratification or weighting, namely sex, migration, and region.

The 'Raw Scores by Age' plot exhibits a pronounced curvilinear pattern, which is characteristic of vocabulary development. A rapid increase in scores is observed during the preschool years, followed by a deceleration of growth during adolescence.

We will initially conduct the modeling without stratification or weighting.

## Model Fitting
To fit the beta-binomial distribution to the data, we employ the 'cnorm.betabinomial()' function. You have to specify the variables 'age' for the explanatory variable and 'score' for the raw score to be modeled. Additionally, it is necessary to indicate the number of items, which for the PPVT-4, is 228. Note that it is important to employ the actual number of items here, not the maximum score achieved. 

```{r fig1, fig.height = 4, fig.width = 7}
# Models the data across a continuos explanatroy variable such as age,
# thereby assuming that the raw scores follow a beta-binomial distribution
# at a given age level. Use autoselect.betabinomial (same syntax) to find
# the model with the lowest BIC automatically.

model.betabinomial <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
```

Fit indices can be obtained using the 'diagnostic.betabinomial()' function or the 'summary()' function. If the age variable and raw score variable are additionally specified, the output will also include R2, RMSE, and bias in relation to the observed data. 

```{r}
# Provides fit indices

diagnostics.betabinomial(model.betabinomial, age = ppvt$age, score = ppvt$raw)
```

By default, the function fits a third-degree polynomial for both $\alpha$ and $\beta$. However, you can also specify the degrees of the polynomials yourself. Use the parameters 'alpha' and 'beta' for this purpose. When the 'mode' parameter of the function is set to 1, the adjustment is performed using the parameters for mean and standard deviation ($\mu$ and $\sigma$) instead of $\alpha$ and $\beta$. If the number of items is not specified using the 'n' parameter, the function will use the maximum raw score (which we do not recommend). For an automatic selection of the model, you can use the autoselect.betabinomial function (and for shash accordingly autoselect.shah), which applies a full grid search for the model with the lowest BIC.

Norm scores are returned as T-scores by default, but IQ scores, z-scores, percentiles, or customized norm scores specified by a vector 'c(M, SD)' are also possible.

By default, a plot of the manifest and modeled percentile curves is returned, too. Diagramms can also be generated using the 'plot()' function. As with distribution-free modeling, it is always important to have a close look at the percentile curves. Wavy patterns indicate overfitting.

To determine norm scores, please use the 'predict()' function and provide the model object along with a vector for the desired age levels and raw scores. Please use the midpoint of the desired age interval to specify age. In the example, T-scores for 10-year-old children are provided in 3-month intervals for the raw score 200.

```{r}
# Provides norm scores for specified age levels and raw scores.
# If not specified otherwise in the model, the norm scores will
# be returned as T-scores.

predict(model.betabinomial, c(10.125, 10.375, 10.625, 10.875), c(200, 200, 200, 200))
```


## Post stratification and weighting
Post-stratification is a method in which individual cases are weighted after data collection is complete, aiming to minimize deviations from representativeness in the sample. In parametric modeling, such weights can be applied in the same way as in distribution-free modeling. The 'computeWeights()' function calculates the necessary weights using a technique called 'raking'. In this example, we weight individual cases to account for population proportions of gender and immigration status when computing the model. Consequently, the resulting norm scores should better represent the target population compared to the unweighted model.

```{r fig2, fig.height = 4, fig.width = 7}
# Calculates weights and models the data:

margins <- data.frame(variables = c("sex", "sex",
                    "migration", "migration"),
                    levels = c(1, 2, 0, 1),
                    share = c(.52, .48, .7, .3))
weights <- computeWeights(ppvt, margins)
model_weighted <- cnorm.betabinomial(ppvt$age, ppvt$raw, weights = weights)

```

These norm tables provide the raw scores (x), their probabilities (Px), the cumulative probabilities (Pcum), percentiles, z-scores, and norm scores (if not specified otherwise: T-scores) and in this specific case the 95%-confidence intervals with respect to the norm score and percentile.

Please consult the vignette on `WeightedRegression` for more information on post-stratification and weighting techniques.



## Generation of Norm tables
Norm tables can be generated using the 'normTable.betabinomial()' function. This function requires the model object and an age specification or a vector with multiple age specifications for which you want to generate norm tables. When specifying ages, please note that the midpoint of the desired age interval should be used. In this example, norms for 14-year-old children are provided in 6-month intervals. If a confidence coefficient and a reliability coefficient are additionally specified in the 'normTable.betabinomial()' function, the table will be returned with confidence intervals. By default, the function limits the norm scores to +/- 3 standard deviations.

```{r}
# Generates norm tables for age 14.25 and 14.75 and computes 95%-confidence
# intervals with a reliability of .97.

tables <- normTable.betabinomial(model.betabinomial, c(14.25, 14.75), CI = .95, reliability = .97)
head(tables[[1]]) # head is used to show only the first few rows of the table
```

These norm tables provide the raw scores (x), their probabilities (Px), cumulative probabilities (Pcum), percentiles, z-scores, and norm scores (in this case, T-scores) and in this specific case the confidence intervals for the norm score and the percentiles for each age. This allows for the interpretation of an individual's raw score in relation to their age group.


## Simulation Study

In simulation studies (W. Lenhard & Lenhard, 2021), we demonstrated that distribution-free continuous norming using Taylor polynomials, as applied in cNORM, is significantly superior to conventional norming methods. This superiority was more pronounced with smaller normative samples. Only with unrealistically large samples did the relevant difference between continuous and conventional norming disappear. In another simulation study (A. Lenhard et al., 2019), we compared distribution-free with parametric norming. These investigations showed no general superiority of one method over the other. Instead, the quality of norming depended on the raw score distribution and sample size. A crucial factor for norming quality was whether an appropriate distribution could be found in parametric norming that fit the raw score distributions within each individual age group. However, the beta-binomial distribution was not included in our simulation study at that time.

Recent publications by Urban et al. (2024) have now shown that parametric modeling using the beta-binomial distribution can very effectively approximate the norm scores of some typical psychological test scales. Our own (unpublished) simulations with 1-PL IRT scales also pointed in this direction.

With the simulation study presented below, we aim to extend our previous results comparing parametric and distribution-free modeling to include the beta-binomial distribution. The simulations exclusively modeled scales without time limitations (i.e., without speed effects). In theory, the beta-binomial distribution should be particularly well-suited for modeling these scales.

Our simulation includes scales of varying difficulty and item numbers, and normative samples of different sizes.

We conducted five simulation runs with the following parameters:

- Number of items: 10, 30, and 50

- Scale difficulty: easy, medium, and hard

- Sample sizes: 700, 1,400, 2,800, and 5,600 cases across 7 age groups

- Cross-validation sample: 70,000 cases

In the results, we compare conventional norming with 6-month intervals, distribution-free continuous norming, and parametric continuous norming by means of beta-binomial distribution. The resulting norm scores for each method are compared with manifest T-scores simulated for an ideally representative, large population using an IRT process. This approach allows us to assess model quality while minimizing biases that occur in real norming studies. Since the person parameters are known in the simulation, we can calculate the mean bias, Root Mean Square Error (RMSE), R2, and Mean Absolute Difference (MAD) for each norming method. For simplicity, we will focus on R2 in the cross-validation sample in the following. However, detailed results and the complete simulation code are available in our [OSF repository](https://osf.io/hcg7v/).



## Results and discussion

```{r, echo=FALSE, out.height="450px", out.width="555px", fig.align='center'}
knitr::include_graphics("beta.png")
```

Our results demonstrate that parametric continuous norming with beta-binomial distribution performs exceptionally well for the simulated scales. Even with relatively small samples, it closely approximates the 'ideal' norm scores simulated for the large and representative cross-validation sample. Parametric norming with the beta-binomial distribution is markedly superior to conventional norming with 6-month intervals. In fact, for the simple 1-PL scales simulated here, it performs even slightly better than distribution-free continuous norming with Taylor polynomials. However, the differences between the two continuous methods are minimal that they are unlikely to have practical significance in most cases. At the moment (2026), we are systeamtically evaluation, which approach is best suited in what use case. So far, our results point towards using beta binomial modelling in the case of
1PL IRT based tests, while Taylor polynomials are better suited for speeded tests.


## Conclusion: When to Use What?

A diverse methodological toolkit is invaluable in psychometrics. Taylor polynomials, as implemented in cNORM, have proven effective in a wide range of norming scenarios. Their distribution-free nature makes them applicable to various tests, including those with continuous or negative raw scores. Parametric modeling with beta-binomial distribution, on the other hand, is particularly well-suited for unspeeded tests with dichotomous items and fixed maximum scores. Both methods yield significantly better results than conventional norming for typical sample sizes and should therefore be preferentially applied in test construction.

The choice between these approaches should be guided by the following factors:

1. Characteristics of raw scores (e.g., discrete vs. continuous raw scores, negative values)
2. Test scale features (e.g., speeded vs. unspeeded)

Regardless of the method used, it is always necessary to thoroughly evaluate the model fit and transparently report its quality in the test manuals. Particular attention should be paid to the shape of the percentile curves, which should align with theoretical expectations. For instance, wavy percentile curves are generally not expected. Such a pattern indicates overfitting and should be avoided. In such cases, review the model parameters or employ an alternative modeling approach.

In summary, both distribution-free and parametric continuous norming with cNORM offer significant advantages over conventional norming. The specific choice between the two continuous norming methods should be made through careful consideration of the test type and available data.

# References
* Lenhard, A., Lenhard, W. & Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. PLoS ONE, 14(9), e0222279. https://doi.org/10.1371/journal.pone.0222279
* Lenhard, W., & Lenhard, A. (2021). Improvement of Norm Score Quality via Regression-Based Continuous Norming. Educational and Psychological Measurement, 81(2), 229-261. https://doi.org/10.1177/0013164420928457
* Timmerman, M. E., Voncken, L., & Albers, C. J. (2021). A tutorial on regression-based norming of psychological tests with GAMLSS. Psychological Methods, 26(3), 357–373. https://doi.org/10.1037/met0000348
* Urban, J., Scherrer, V., Strobel, A., & Preckel, F. (2024). Continuous Norming Approaches: A Systematic Review and Real Data Example. Assessment. https://doi.org/10.1177/10731911241260545

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{log_likelihood}
\alias{log_likelihood}
\title{Calculate the negative log-likelihood for a beta binomial regression model}
\usage{
log_likelihood(params, X, Z, y, weights = NULL)
}
\arguments{
\item{params}{A numeric vector containing all model parameters. The first
n_beta elements are coefficients for the mean model, and the remaining
elements are coefficients for the log-standard deviation model.}

\item{X}{A matrix of predictors for the mean model.}

\item{Z}{A matrix of predictors for the log-standard deviation model.}

\item{y}{A numeric vector of response values.}

\item{weights}{A numeric vector of weights for each observation
(NULL = equal weights).}
}
\value{
The negative log-likelihood of the model (large finite penalty
  if non-finite).
}
\description{
This function computes the negative log-likelihood for a beta binomial
regression model where both the mean and standard deviation are modeled
as functions of predictors (mode 1, normal approximation on standardized
scores).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotDerivative}
\alias{plotDerivative}
\title{Plot first order derivative of regression model}
\usage{
plotDerivative(
  model,
  minAge = NULL,
  maxAge = NULL,
  minNorm = NULL,
  maxNorm = NULL,
  stepAge = NULL,
  stepNorm = NULL,
  order = 1
)
}
\arguments{
\item{model}{The model from the bestModel function, a cnorm object.}

\item{minAge}{Minimum age to start checking. If NULL, it's automatically determined from the model.}

\item{maxAge}{Maximum age for checking. If NULL, it's automatically determined from the model.}

\item{minNorm}{Lower end of the norm score range. If NULL, it's automatically determined from the model.}

\item{maxNorm}{Upper end of the norm score range. If NULL, it's automatically determined from the model.}

\item{stepAge}{Stepping parameter for the age check, usually 1 or 0.1; lower values indicate higher precision.}

\item{stepNorm}{Stepping parameter for norm scores.}

\item{order}{Degree of the derivative (default = 1).}
}
\value{
A ggplot object representing the derivative of the regression function.
}
\description{
This function plots the scores obtained via the first order derivative of the regression model
in dependence of the norm score.
}
\details{
The results indicate the progression of the norm scores within each age group. The regression-based
modeling approach relies on the assumption of a linear progression of the norm scores. Negative scores
in the first order derivative indicate a violation of this assumption. Scores near zero are typical
for bottom and ceiling effects in the raw data.

The regression models usually converge within the range of the original values. In case of vertical
and horizontal extrapolation, with increasing distance to the original data, the risk of assumption
violation increases as well.
}
\note{
This function is currently incompatible with reversed raw score scales ('descent' option).
}
\examples{
\dontrun{
# For traditional continuous norming model
result <- cnorm(raw = elfe$raw, group = elfe$group)
plotDerivative(result, minAge=2, maxAge=5, stepAge=.2, minNorm=25, maxNorm=75, stepNorm=1)
}

}
\seealso{
\code{\link{checkConsistency}}, \code{\link{bestModel}}, \code{\link{derive}}

Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{rangeCheck}
\alias{rangeCheck}
\title{Check for horizontal and vertical extrapolation}
\usage{
rangeCheck(
  object,
  minAge = NULL,
  maxAge = NULL,
  minNorm = NULL,
  maxNorm = NULL,
  digits = 3,
  ...
)
}
\arguments{
\item{object}{The regression model or a cnorm object}

\item{minAge}{The lower age bound}

\item{maxAge}{The upper age bound}

\item{minNorm}{The lower norm value bound}

\item{maxNorm}{The upper norm value bound}

\item{digits}{The precision for rounding the norm and age data}

\item{...}{additional parameters}
}
\value{
the report
}
\description{
Regression models only work in a specific range and extrapolation
horizontally (outside the original range) or vertically (extreme norm
scores) might lead to inconsistent results. The function generates a
message, indicating extrapolation and the range of the original data.
}
\examples{
\dontrun{
  m <- cnorm(raw = elfe$raw, group = elfe$group)
  rangeCheck(m)
}
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cNORM.R
\name{cnorm}
\alias{cnorm}
\title{Continuous Norming}
\usage{
cnorm(
  raw = NULL,
  group = NULL,
  age = NULL,
  width = NA,
  weights = NULL,
  scale = "T",
  method = 4,
  descend = FALSE,
  k = NULL,
  t = NULL,
  terms = 0,
  R2 = NULL,
  plot = TRUE,
  extensive = TRUE,
  averaging = FALSE,
  minDip = 0.01
)
}
\arguments{
\item{raw}{Numeric vector of raw scores}

\item{group}{Numeric vector of grouping variable, e.g. grade. If no group
or age variable is provided, conventional norming is applied.}

\item{age}{Numeric vector with chronological age. If used without `group`,
please additionally specify `width`.}

\item{width}{Size of the sliding window in case an age vector is used.}

\item{weights}{Optional numeric vector of case weights for post-stratification.}

\item{scale}{Type of norm scale, either "T" (default), "IQ", "z" or
"percentile" (= no transformation); a numeric vector with mean and SD can
also be provided, e.g. `c(10, 3)` for Wechsler scale index points.}

\item{method}{Ranking method in case of ties; an integer index from 1
(Blom 1958) through 7 (Yu & Huang 2001). Default is 4 (Rankit).}

\item{descend}{If TRUE, inverts the ranking order so that higher raw scores
receive lower norm scores (e.g. for error scores).}

\item{k}{Power degree for the location dimension (max 6).}

\item{t}{Power degree for the age dimension (max 6).}

\item{terms}{If > 0, fix the model to this number of terms.}

\item{R2}{Stopping criterion (adjusted R-squared) for model selection.}

\item{plot}{If TRUE (default), display percentile plot and report.}

\item{extensive}{If TRUE (default), screen models for monotonic consistency.}

\item{averaging}{If TRUE (default FALSE), apply BIC-weighted model averaging
across the consistency-screened candidate models instead of selecting a
single model. Requires \code{extensive = TRUE} and age-based norming.}

\item{minDip}{Tolerance for monotonicity check. Allow small violations
(default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
for strict checking.}
}
\value{
cnorm object including the ranked raw data and the regression model.
}
\description{
Conducts continuous norming in one step and returns an object including ranked
raw data and the continuous norming model. Please consult the function
description of 'rankByGroup', 'rankBySlidingWindow' and 'bestModel' for specifics
of the steps in the data preparation and modeling process. In addition to the
raw scores, either provide
\itemize{
 \item{a numeric vector for the grouping information (group)}
 \item{a numeric age vector and the width of the sliding window (age, width)}
}
for the ranking of the raw scores. You can
adjust the grade of smoothing of the regression model by setting the k and terms
parameter. In general, increasing k to more than 4 and the number of terms lead
to a higher fit, while lower values lead to more smoothing. The power parameter
for the age trajectory can be specified independently by 't'. If both parameters
are missing, cnorm uses k = 5 and t = 3 by default.
}
\examples{
\dontrun{
# Conventional norming
cnorm(raw = elfe$raw)

# Continuous norming via group
m1 <- cnorm(raw = elfe$raw, group = elfe$group)

# Continuous norming via continuous age + sliding window
m2 <- cnorm(raw = ppvt$raw, age = ppvt$age, width = 1)

# Norm tables with confidence intervals
normTable(c(2.125, 2.375, 2.625), m1, CI = .9, reliability = .95)
rawTable (c(2.125, 2.375, 2.625), m1, CI = .9, reliability = .95)
}
}
\references{
\enumerate{
  \item Gary, S. & Lenhard, W. (2021). In norming we trust. Diagnostica.
  \item Gary, S., Lenhard, W. & Lenhard, A. (2021). Modelling Norm Scores with the cNORM Package in R. Psych, 3(3), 501-521. https://doi.org/10.3390/psych3030033
  \item Gary, S., Lenhard, W., Lenhard, A., & Herzberg, D. (2023). A tutorial on automatic post-stratification and weighting in conventional and regression-based norming of psychometric tests. Behavior Research Methods. https://doi.org/10.3758/s13428-023-02207-0
  \item Gary, S., Lenhard, A., Lenhard, W., & Herzberg, D. S. (2023). Reducing the Bias of Norm Scores in Non-Representative Samples: Weighting as an Adjunct to Continuous Norming Methods. Assessment, 30(8), 2491–2509. https://doi.org/10.1177/10731911231153832
  \item Lenhard, A., Lenhard, W., Suggate, S. & Segerer, R. (2016). A continuous solution to the norming problem. Assessment, Online first, 1-14. doi:10.1177/1073191116656437
  \item Lenhard, A., Lenhard, W., Gary, S. (2018). Continuous Norming (cNORM). The Comprehensive R Network, Package cNORM, available: https://CRAN.R-project.org/package=cNORM
  \item Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. PLoS ONE, 14(9),  e0222279. doi:10.1371/journal.pone.0222279
  \item Lenhard, W., & Lenhard, A. (2020). Improvement of Norm Score Quality via Regression-Based Continuous Norming. Educational and Psychological Measurement(Online First), 1-33. https://doi.org/10.1177/0013164420928457
}
}
\seealso{
rankByGroup, rankBySlidingWindow, computePowers, bestModel
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predictNorm}
\alias{predictNorm}
\title{Retrieve norm value for raw score at a specific age}
\usage{
predictNorm(
  raw,
  A,
  model,
  minNorm = NULL,
  maxNorm = NULL,
  force = TRUE,
  silent = FALSE
)
}
\arguments{
\item{raw}{The raw value, either single numeric or numeric vector}

\item{A}{the explanatory variable (e. g. age), either single numeric or numeric vector}

\item{model}{The regression model or a cnorm object}

\item{minNorm}{The lower bound of the norm score range}

\item{maxNorm}{The upper bound of the norm score range}

\item{force}{Try to resolve missing norm scores in case of inconsistent models}

\item{silent}{set to TRUE to suppress messages}
}
\value{
The predicted norm score for a raw score, either single value or vector
}
\description{
This functions numerically determines the norm score for raw scores depending on the
level of the explanatory variable A, e. g. norm scores for raw scores at given ages.
}
\examples{
\dontrun{
# Generate cnorm object from example data
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)

# return norm value for raw value 21 for grade 2, month 9
specificNormValue <- predictNorm(raw = 21, A = 2.75, cnorm.elfe)
}

}
\seealso{
Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{plot.cnorm}
\alias{plot.cnorm}
\title{S3 function for plotting cnorm objects}
\usage{
\method{plot}{cnorm}(x, y, ...)
}
\arguments{
\item{x}{the cnorm object}

\item{y}{the type of plot as a string, can be one of
'raw' (1), 'norm' (2), 'curves' (3), 'percentiles' (4), 'series' (5), 'subset' (6),
or 'derivative' (7), either as a string or the according index}

\item{...}{additional parameters for the specific plotting function}
}
\description{
S3 function for plotting cnorm objects
}
\seealso{
Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{CDC}
\alias{CDC}
\title{BMI growth curves from age 2 to 25}
\format{
A data frame with 45053 rows and 7 variables:
\describe{
  \item{age}{continuous age in years, based on the month variable}
  \item{group}{age group; chronological age in years at the time of examination}
  \item{month}{chronological age in month at the time of examination}
  \item{sex}{sex of the participant, 1 = male, 2 = female}
  \item{height}{height of the participants in cm}
  \item{weight}{weight of the participants in kg}
  \item{bmi}{the body mass index, computed by (weight in kg)/(height in m)^2}
}

A data frame with 45035 rows and 7 columns
}
\usage{
CDC
}
\description{
By the courtesy of the Center of Disease Control (CDC), cNORM includes human growth data for children and adolescents
age 2 to 25 that can be used to model trajectories of the body mass index and to estimate percentiles for clinical
definitions of under- and overweight. The data stems from the NHANES surveys in the US and was published in 2012
as public domain. The data was cleaned by removing missing values and it includes the following variables from or
based on the original dataset.
}
\references{
Center for Disease Control and Prevention (2012).
  National Health and Nutrition Examination Survey: Questionnaires,
  datasets and related documentation. U.S. Department of Health and Human Services
  (original source not available anymore).
}
\concept{Body Mass Index growth curves weight height}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{normTable}
\alias{normTable}
\title{Create a norm table based on model for specific age}
\usage{
normTable(
  A,
  model,
  minNorm = NULL,
  maxNorm = NULL,
  minRaw = NULL,
  maxRaw = NULL,
  step = NULL,
  monotonuous = TRUE,
  CI = 0.9,
  reliability = NULL,
  pretty = TRUE
)
}
\arguments{
\item{A}{the age as single value or a vector of age values}

\item{model}{The regression model from the cnorm function}

\item{minNorm}{The lower bound of the norm score range}

\item{maxNorm}{The upper bound of the norm score range}

\item{minRaw}{clipping parameter for the lower bound of raw scores}

\item{maxRaw}{clipping parameter for the upper bound of raw scores}

\item{step}{Stepping parameter with lower values indicating higher precision}

\item{monotonuous}{corrects for decreasing norm scores in case of model inconsistencies (default)}

\item{CI}{confidence coefficient, ranging from 0 to 1, default .9}

\item{reliability}{coefficient, ranging between  0 to 1}

\item{pretty}{Format table by collapsing intervals and rounding to meaningful precision}
}
\value{
either data.frame with norm scores, predicted raw scores and percentiles in case of simple A
value or a list of norm tables if vector of A values was provided
}
\description{
This function generates a norm table for a specific age based on the regression
model by assigning raw scores to norm scores. Please specify the
range of norm scores, you want to cover. A T value of 25 corresponds to a percentile
of .6. As a consequence, specifying a range of T = 25 to T = 75 would cover 98.4 % of
the population. Please be careful when extrapolating vertically (at the lower and
upper end of the age specific distribution). Depending on the size of your standardization
sample, extreme values with T < 20 or T > 80 might lead to inconsistent results.
In case a confidence coefficient (CI, default .9) and the reliability is specified,
confidence intervals are computed for the true score estimates, including a correction for
regression to the mean (Eid & Schmidt, 2012, p. 272).
}
\examples{
\dontrun{
# Generate cnorm object from example data
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)

# create single norm table
norms <- normTable(3.5, cnorm.elfe, minNorm = 25, maxNorm = 75, step = 0.5)

# create list of norm tables
norms <- normTable(c(2.5, 3.5, 4.5), cnorm.elfe,
  minNorm = 25, maxNorm = 75,
  step = 1, minRaw = 0, maxRaw = 26
)

# conventional norming, set age to arbitrary value
model <- cnorm(raw = elfe$raw)
normTable(0, model)
}

}
\references{
Eid, M. & Schmidt, K. (2012). Testtheorie und Testkonstruktion. Hogrefe.
}
\seealso{
rawTable

Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{gradient_log_likelihood2}
\alias{gradient_log_likelihood2}
\title{Analytic gradient of the negative log-likelihood (mode 2)}
\usage{
gradient_log_likelihood2(params, X, Z, y, n, weights = NULL, lch = NULL)
}
\arguments{
\item{params}{A numeric vector containing all model parameters. The first
n_alpha elements are coefficients for the alpha model, and the remaining
elements are coefficients for the beta model.}

\item{X}{A matrix of predictors for the alpha model.}

\item{Z}{A matrix of predictors for the beta model.}

\item{y}{A numeric vector of response values.}

\item{n}{The maximum score (number of trials).}

\item{weights}{A numeric vector of weights for each observation
(NULL = equal weights).}

\item{lch}{Optional precomputed \code{lchoose(n, y)}. Since this term does
not depend on the parameters, passing it once avoids recomputation in
every optimizer iteration.}
}
\value{
Numeric vector: gradient of the negative log-likelihood with
  respect to \code{params}.
}
\description{
Computes the exact gradient of \code{\link{log_likelihood2}} with respect
to all regression coefficients. With the per-observation log-probability
\deqn{\log p = \mathrm{lchoose}(n, y) + \mathrm{lbeta}(y+\alpha, n-y+\beta)
  - \mathrm{lbeta}(\alpha, \beta),}
the partial derivatives on the natural scale are (with digamma function
\eqn{\psi}):
\deqn{\partial \log p/\partial\alpha = \psi(y+\alpha) - \psi(\alpha)
  + \psi(\alpha+\beta) - \psi(n+\alpha+\beta)}
\deqn{\partial \log p/\partial\beta = \psi(n-y+\beta) - \psi(\beta)
  + \psi(\alpha+\beta) - \psi(n+\alpha+\beta)}
The chain rule through the log link multiplies by \eqn{\alpha} resp.
\eqn{\beta}; coefficient gradients follow as
\eqn{-X^\top(w \cdot \partial\ell/\partial\eta)} for the negative
log-likelihood. Contributions of observations whose linear predictor is
clamped (see \code{log_likelihood2}) are zeroed, which is the exact
subgradient of the clamped objective.
}
\details{
Supplying this gradient to \code{\link[stats]{optim}} (L-BFGS-B) replaces
the finite-difference approximation, which required
\code{length(params) + 1} likelihood evaluations per gradient.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/roots.R
\name{calcPolyInLBase2}
\alias{calcPolyInLBase2}
\title{Internal function for retrieving regression function coefficients at specific
age}
\usage{
calcPolyInLBase2(raw, age, coeff, k)
}
\arguments{
\item{raw}{The raw value (subtracted from the intercept)}

\item{age}{The age}

\item{coeff}{The cNORM regression model coefficients}

\item{k}{The cNORM regression model power parameter}
}
\value{
The coefficients
}
\description{
The function is an inline for searching zeros in the inverse regression
function. It collapses the regression function at a specific age and
simplifies the coefficients. Optimized version of the prior 'calcPolyInLBase'
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moments.R
\name{predictMoments}
\alias{predictMoments}
\alias{predictMoments.default}
\alias{predictMoments.cnorm}
\alias{predictMoments.cnormBetaBinomial}
\alias{predictMoments.cnormBetaBinomial2}
\alias{predictMoments.cnormShash}
\title{Model-Implied Distributional Moments at Specific Ages}
\usage{
predictMoments(model, age, ...)

\method{predictMoments}{default}(model, age, ...)

\method{predictMoments}{cnorm}(model, age, nNodes = 100, ...)

\method{predictMoments}{cnormBetaBinomial}(model, age, ...)

\method{predictMoments}{cnormBetaBinomial2}(model, age, ...)

\method{predictMoments}{cnormShash}(model, age, nNodes = 100, ...)
}
\arguments{
\item{model}{A model object of class \code{cnorm},
\code{cnormBetaBinomial}, \code{cnormBetaBinomial2} or
\code{cnormShash}.}

\item{age}{A numeric vector of ages (values of the explanatory variable)
at which to compute the moments.}

\item{...}{Additional parameters passed to the methods, e.g.
\code{nNodes}.}

\item{nNodes}{Number of Gauss-Hermite quadrature nodes (default 100).
Only relevant for the Taylor and SHASH methods; ignored for
beta-binomial models, which are computed exactly by summation.}
}
\value{
A \code{data.frame} with one row per age and the columns
  \code{age}, \code{mean}, \code{sd}, \code{variance}, \code{skewness}
  and \code{kurtosis} (excess). The computation method is stored in the
  attribute \code{"method"}.
}
\description{
Computes the mean, standard deviation, variance, skewness and (excess)
kurtosis of the raw score distribution implied by a fitted cNORM model
at one or more ages (or, more generally, values of the explanatory
variable). The moments are model-implied population moments of the
conditional raw score distribution, censored at the bounds of the raw
score range \code{[minRaw, maxRaw]} for consistency across model families.
}
\details{
The computation strategy depends on the model family:
\describe{
  \item{Taylor polynomial (\code{cnorm})}{The bivariate regression function
    is collapsed at the specified age into a univariate polynomial in the
    norm score (location) variable. Moments are then obtained by
    Gauss-Hermite quadrature, which is mathematically exact for
    polynomial quantile functions (up to the censoring at
    \code{minRaw}/\code{maxRaw}).}
  \item{Beta-binomial (\code{cnormBetaBinomial},
    \code{cnormBetaBinomial2})}{Moments are computed exactly by summation
    over the discrete probability mass function on the support
    \code{0:n}, using the age-specific predicted \eqn{\alpha} and
    \eqn{\beta} parameters. This respects the discreteness of the
    distribution; no continuity approximation is involved.}
  \item{SHASH (\code{cnormShash})}{Moments are obtained by Gauss-Hermite
    quadrature of the quantile function \code{qshash} evaluated at the
    age-specific distribution parameters, censored at
    \code{minRaw}/\code{maxRaw}.}
}

Kurtosis is reported as \emph{excess} kurtosis (0 for the normal
distribution).

Note that the skewness and kurtosis of the \emph{censored} distribution
are reported. For well-fitting models whose raw score range covers the
probability mass of the conditional distribution, censoring effects are
negligible; for distributions with substantial floor or ceiling effects,
the censored moments are the substantively meaningful ones.
}
\examples{
\dontrun{
# Taylor polynomial model
model <- cnorm(raw = elfe$raw, group = elfe$group)
predictMoments(model, age = c(2.25, 2.75, 3.25, 3.75, 4.25))

# Beta-binomial model
bb <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
predictMoments(bb, age = seq(4, 16, by = 2))
}

}
\references{
Isserlis, L. (1918). On a formula for the product-moment coefficient of
any order of a normal frequency distribution in any number of variables.
Biometrika, 12(1/2), 134-139.

Jones, M. C. & Pewsey, A. (2009). Sinh-arcsinh distributions.
Biometrika, 96(4), 761-780.
}
\seealso{
\code{\link{normTable}}, \code{\link{predictNorm}},
  \code{\link{predictRaw}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotNorm}
\alias{plotNorm}
\title{Plot manifest and fitted norm scores}
\usage{
plotNorm(
  model,
  age = NULL,
  score = NULL,
  width = NULL,
  weights = NULL,
  group = FALSE,
  minNorm = NULL,
  maxNorm = NULL,
  type = 0
)
}
\arguments{
\item{model}{The regression model, usually from the 'cnorm', 'cnorm.betabinomial', or 'cnorm.shash' function}

\item{age}{In case of parametric models, please provide the age vector}

\item{score}{In case of parametric models, please provide the score vector}

\item{width}{In case of parametric models, please provide the width for the sliding window.
If null, the function tries to determine a sensible setting.}

\item{weights}{Vector or variable name in the dataset with weights for each
individual case. If NULL, no weights are used.}

\item{group}{On optional grouping variable, use empty string for no group, the variable name
for Taylor polynomial models or a vector with the groups for parametric models}

\item{minNorm}{lower bound of fitted norm scores}

\item{maxNorm}{upper bound of fitted norm scores}

\item{type}{Type of display: 0 = plot manifest against fitted values, 1 = plot
manifest against difference values}
}
\value{
A ggplot object representing the norm scores plot.
}
\description{
This function plots the manifest norm score against the fitted norm score from
the inverse regression model per group. This helps to inspect the precision
of the modeling process. The scores should not deviate too far from
the regression line. Applicable for Taylor polynomial, beta-binomial, and shash models.
}
\examples{
\dontrun{
# Load example data set, compute model and plot results

# Taylor polynomial model
model <- cnorm(raw = elfe$raw, group = elfe$group)
plot(model, "norm")

# Beta binomial models; maximum number of items in elfe is n = 28
model.bb <- cnorm.betabinomial(elfe$group, elfe$raw, n = 28)
plotNorm(model.bb, age = elfe$group, score = elfe$raw)
}

}
\seealso{
Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{checkConsistency}
\alias{checkConsistency}
\title{Check the consistency of the norm data model}
\usage{
checkConsistency(
  model,
  minAge = NULL,
  maxAge = NULL,
  minNorm = NULL,
  maxNorm = NULL,
  minRaw = NULL,
  maxRaw = NULL,
  stepAge = NULL,
  stepNorm = 1,
  method = c("analytic", "grid"),
  warn = FALSE,
  silent = FALSE,
  minDip = 0.01
)
}
\arguments{
\item{model}{The model from the bestModel function or a cnorm object}

\item{minAge}{Age to start with checking}

\item{maxAge}{Upper end of the age check}

\item{minNorm}{Lower end of the norm value range}

\item{maxNorm}{Upper end of the norm value range}

\item{minRaw}{clipping parameter for the lower bound of raw scores}

\item{maxRaw}{clipping parameter for the upper bound of raw scores}

\item{stepAge}{Stepping parameter for the age check. If NULL, 8 equidistant
age points are used. Lower values indicate higher precision.}

\item{stepNorm}{Stepping parameter for the norm table check within age
(only used with \code{method = "grid"})}

\item{method}{Monotonicity cgeck. Either "analytic" (default; exact within
age via polynomial derivative roots) or "grid" (numerical check on a norm
score grid)}

\item{warn}{Retained for backwards compatibility (violations below numerical
tolerance are always suppressed)}

\item{silent}{turn off messages}

\item{minDip}{Tolerance for monotonicity check. Allow small violations
(default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
for strict checking.}
}
\value{
Boolean, indicating model violations (TRUE) or no problems (FALSE)
}
\description{
While abilities increase and decline over age, within one age group, the
norm scores always have to show a monotonic increase or decrease with
increasing raw scores. Violations of this assumption are an indication for
problems in modeling the relationship between raw and norm scores. There are
several reasons, why this might occur:
\enumerate{
  \item Vertical extrapolation: Choosing extreme norm scores, e. g. values
  -3 <= x and x >= 3. In order to model these extreme values, a large sample
  dataset is necessary.
  \item Horizontal extrapolation: Taylor polynomials converge in a certain
  radius. Using the model values outside the original dataset may
  lead to inconsistent results.
  \item The data cannot be modeled with Taylor polynomials, or you need
  another power parameter (k) or R2 for the model.
 }
}
\details{
In general, extrapolation (point 1 and 2) can carefully be done to a
 certain degree outside the original sample, but it should in general
 be handled with caution. Please note that at extreme values, the models
 most likely become independent and it is thus recommended to restrict the
 norm score range to the relevant range of abilities, e.g. +/- 2.5 SD via
 the minNorm and maxNorm parameter.

 With \code{method = "analytic"} (default), monotonicity within each age is
 verified exactly: since the model is polynomial in the norm score L, the
 real roots of its derivative are computed via \code{polyroot} and the sign
 of the derivative is checked between consecutive roots. This detects even
 narrow violations that a coarse norm score grid can miss. Age remains
 discretized; violations that are entirely clipped by the raw score bounds
 are ignored. \code{method = "grid"} reproduces the classical numerical
 check and is used automatically as a fallback for models with non-Taylor
 predictors.
}
\examples{
\dontrun{
  model <- cnorm(raw = elfe$raw, group = elfe$group, plot = FALSE)
  modelViolations <- checkConsistency(model, minNorm = 25, maxNorm = 75)
  plotDerivative(model, minNorm = 25, maxNorm = 75)
}
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{plot.cnormShash}
\alias{plot.cnormShash}
\title{Plot Sinh-Arcsinh Model with Data and Percentile Lines}
\usage{
\method{plot}{cnormShash}(x, ...)
}
\arguments{
\item{x}{A fitted model object of class "cnormShash"}

\item{...}{Additional arguments including age, score, weights, percentiles, points}
}
\value{
A ggplot object
}
\description{
Plot Sinh-Arcsinh Model with Data and Percentile Lines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cNORM.R
\name{cNORM.GUI}
\alias{cNORM.GUI}
\title{Launcher for the graphical user interface of cNORM for distribution free
continuous norming}
\usage{
cNORM.GUI(launch.browser = TRUE)
}
\arguments{
\item{launch.browser}{Default TRUE; automatically open browser for GUI}
}
\description{
Launcher for the graphical user interface of cNORM for distribution free
continuous norming
}
\examples{
\dontrun{
# Launch graphical user interface
cNORM.GUI()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{cnorm.betabinomial1}
\alias{cnorm.betabinomial1}
\title{Fit a beta binomial regression model (mode 1, mu/sigma parameterization)}
\usage{
cnorm.betabinomial1(
  age,
  score,
  n = NULL,
  weights = NULL,
  mu = 3,
  sigma = 3,
  control = NULL,
  scale = "T",
  plot = TRUE
)
}
\arguments{
\item{age}{A numeric vector of predictor values (e.g., age).}

\item{score}{A numeric vector of response values.}

\item{n}{Number of items in the test, resp. maximum score to be achieved.}

\item{weights}{A numeric vector of weights for each observation.
Default is NULL (equal weights).}

\item{mu}{Integer specifying the degree of the polynomial for the mean
model. Default is 3.}

\item{sigma}{Integer specifying the degree of the polynomial for the
standard deviation model. Default is 3.}

\item{control}{A list of control parameters to be passed to the
\code{optim} function. Any parameters not specified are filled with
the defaults \code{list(reltol = 1e-8, maxit = 1000)}.}

\item{scale}{Type of norm scale, either T (default), IQ, z or percentile
(= no transformation); a double vector with the mean and standard
deviation can be provided as well, e.g. c(10, 3) for Wechsler scale
index points.}

\item{plot}{Logical indicating whether to plot the model. Default is TRUE.}
}
\value{
A list of class "cnormBetaBinomial" containing:
  \item{beta_est}{Estimated coefficients for the mean model}
  \item{gamma_est}{Estimated coefficients for the log-standard deviation model}
  \item{se}{Standard errors of the estimated coefficients}
  \item{mu}{Degree of the polynomial for the mean model}
  \item{sigma}{Degree of the polynomial for the standard deviation model}
  \item{result}{Full result from the optimization procedure}
}
\description{
This function fits a beta binomial regression model where both the mean and
standard deviation of the response variable are modeled as polynomial
functions of the predictor variable. While 'cnorm.betabinomial2' fits a
beta-binomial model directly on the basis of \eqn{\alpha} and \eqn{\beta},
this function fits \eqn{\mu} and \eqn{\sigma}, which are then used to
estimate the beta binomial distribution parameters. Kept for backwards
compatibility; 'cnorm.betabinomial2' is the recommended default.
}
\details{
The function standardizes the input variables, fits polynomial models for
both the mean and standard deviation, and uses maximum likelihood
estimation with analytic gradients to find the optimal parameters. The
optimization is performed using the BFGS method.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{log_likelihood_shash}
\alias{log_likelihood_shash}
\title{Calculate the negative log-likelihood for a shash regression model}
\usage{
log_likelihood_shash(
  params,
  X_mu,
  X_sigma,
  X_epsilon,
  X_delta = NULL,
  y,
  weights = NULL,
  fixed_delta = NULL
)
}
\arguments{
\item{params}{A numeric vector containing all model parameters}

\item{X_mu}{Design matrix for location parameter}

\item{X_sigma}{Design matrix for scale parameter}

\item{X_epsilon}{Design matrix for skewness parameter}

\item{X_delta}{Design matrix for tail weight parameter (or NULL for fixed delta)}

\item{y}{Response vector}

\item{weights}{Observation weights}

\item{fixed_delta}{If not NULL, the delta parameter is fixed to this value}
}
\value{
The negative log-likelihood of the model
}
\description{
This function computes the negative log-likelihood for a Sinh-Arcsinh regression model
where the location, scale, skewness, and (optionally) tail weight parameters
are modeled as functions of predictors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preparation.R
\name{computePowers}
\alias{computePowers}
\title{Compute powers of the explanatory variable a as well as of the person
location l (data preparation)}
\usage{
computePowers(data, k = 5, norm = NULL, age = NULL, t = 3, silent = FALSE)
}
\arguments{
\item{data}{data.frame with the norm data}

\item{k}{degree of the location polynomial (1..6)}

\item{norm}{name of the norm variable in the data.frame (T scores, IQ, percentiles, ...).
If `NULL`, the `"normValue"` attribute of `data` is used.}

\item{age}{explanatory variable (e.g. age or grade). May be a column name, a numeric
vector of `nrow(data)`, `FALSE` to disable age handling, or `NULL` to fall back
to the `"age"` attribute of `data`.}

\item{t}{age power parameter (1..6). If `NULL`, falls back to `k`.}

\item{silent}{set to TRUE to suppress messages}
}
\value{
data.frame with the powers and interactions of location and explanatory
  variable / age
}
\description{
The function computes powers of the norm variable e. g. T scores (location, L),
an explanatory variable, e. g. age or grade of a data frame (age, A) and the
interactions of both (L X A). The k variable indicates the degree up to which
powers and interactions are build. These predictors can be used later on in the
\code{\link{bestModel}} function to model the norm sample. Higher values of k
allow for modeling the norm sample closer, but might lead to over-fit. In general
k = 3 or k = 4 (default) is sufficient to model human performance data. For example,
k = 2 results in the variables L1, L2, A1, A2, and their interactions L1A1, L2A1, L1A2
and L2A2 (but k = 2 is usually not sufficient for the modeling). Please note, that
you do not need to use a normal rank transformed scale like T or IQ; you can
use the percentiles for the 'normValue' as well.
}
\details{
The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
\code{computePowers} and \code{prepareData} are usually not called directly, but accessed
through other functions like \code{cnorm}.
}
\examples{
\dontrun{
# Dataset with grade levels as grouping
data.elfe <- rankByGroup(elfe)
data.elfe <- computePowers(data.elfe)

# Dataset with continuous age variable and k = 5
data.ppvt <- rankByGroup(ppvt)
data.ppvt <- computePowers(data.ppvt, age = "age", k = 5)
}
}
\seealso{
bestModel

Other prepare:
\code{\link[=prepareData]{prepareData()}},
\code{\link[=rankByGroup]{rankByGroup()}},
\code{\link[=rankBySlidingWindow]{rankBySlidingWindow()}}
}
\concept{prepare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{regressionFunction}
\alias{regressionFunction}
\title{Regression function}
\usage{
regressionFunction(model, raw = NULL, digits = NULL)
}
\arguments{
\item{model}{The regression model from the bestModel function or a cnorm object}

\item{raw}{The name of the raw value variable (default 'raw')}

\item{digits}{Number of digits for formatting the coefficients}
}
\value{
The regression formula as a string
}
\description{
The method builds the regression function for the regression model,
including the beta weights.
It can be used to predict the raw scores based on age and location.
}
\examples{
\dontrun{
  result <- cnorm(raw = elfe$raw, group = elfe$group)
  regressionFunction(result)
}

}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{summary.cnormBetaBinomial2}
\alias{summary.cnormBetaBinomial2}
\title{Summarize a Beta-Binomial Continuous Norming Model}
\usage{
\method{summary}{cnormBetaBinomial2}(object, ...)
}
\arguments{
\item{object}{An object of class "cnormBetaBinomial" or
"cnormBetaBinomial2", typically the result of a call to
\code{\link{cnorm.betabinomial}}.}

\item{...}{Additional arguments passed to the summary method:
\itemize{
   \item age An optional numeric vector of age values.
   \item score An optional numeric vector of raw scores.
   \item weights An optional numeric vector of weights.
 }}
}
\value{
Invisibly returns a list containing detailed diagnostic
  information about the model.
}
\description{
Summary method for objects of class "cnormBetaBinomial2". See
\code{\link{summary.cnormBetaBinomial}} for details.
}
\seealso{
\code{\link{cnorm.betabinomial}}, \code{\link{diagnostics.betabinomial}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotNormCurves}
\alias{plotNormCurves}
\title{Plot norm curves}
\usage{
plotNormCurves(
  model,
  normList = NULL,
  minAge = NULL,
  maxAge = NULL,
  step = 0.1,
  minRaw = NULL,
  maxRaw = NULL
)
}
\arguments{
\item{model}{The model from the bestModel function, a cnorm object, or a cnormBetaBinomial / cnormBetaBinomial2 / cnormShash object.}

\item{normList}{Vector with norm scores to display. If NULL, default values are used.}

\item{minAge}{Age to start with checking. If NULL, it's automatically determined from the model.}

\item{maxAge}{Upper end of the age check. If NULL, it's automatically determined from the model.}

\item{step}{Stepping parameter for the age check, usually 1 or 0.1; lower scores indicate higher precision.}

\item{minRaw}{Lower end of the raw score range, used for clipping implausible results. If NULL, it's automatically determined from the model.}

\item{maxRaw}{Upper end of the raw score range, used for clipping implausible results. If NULL, it's automatically determined from the model.}
}
\value{
A ggplot object representing the norm curves.
}
\description{
This function plots the norm curves based on the regression model. It supports both
Taylor polynomial models, beta-binomial models, and shash models.
}
\details{
Please check the function for inconsistent curves: The different curves should not intersect.
Violations of this assumption are a strong indication of violations of model assumptions in
modeling the relationship between raw and norm scores.

Common reasons for inconsistencies include:
1. Vertical extrapolation: Choosing extreme norm scores (e.g., scores <= -3 or >= 3).
2. Horizontal extrapolation: Using the model scores outside the original dataset.
3. The data cannot be modeled with the current approach, or you need another power parameter (k) or R2 for the model.
}
\examples{
\dontrun{
# For Taylor continuous norming model
m <- cnorm(raw = ppvt$raw, group = ppvt$group)
plotNormCurves(m, minAge=2, maxAge=5)

# For beta-binomial model
bb_model <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
plotNormCurves(bb_model)
}
}
\seealso{
\code{\link{checkConsistency}}, \code{\link{plotDerivative}}, \code{\link{plotPercentiles}}

Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{print.cnormShash}
\alias{print.cnormShash}
\title{Print method for Sinh-Arcsinh objects}
\usage{
\method{print}{cnormShash}(x, ...)
}
\arguments{
\item{x}{A cnormShash object}

\item{...}{Additional arguments}
}
\description{
Print method for Sinh-Arcsinh objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{rawTable}
\alias{rawTable}
\title{Create a table with norm scores assigned to raw scores for a specific age based on the regression model}
\usage{
rawTable(
  A,
  model,
  minRaw = NULL,
  maxRaw = NULL,
  minNorm = NULL,
  maxNorm = NULL,
  step = 1,
  monotonuous = TRUE,
  CI = 0.9,
  reliability = NULL,
  pretty = TRUE
)
}
\arguments{
\item{A}{the age, either single value or vector with age values}

\item{model}{The regression model or a cnorm object}

\item{minRaw}{The lower bound of the raw score range}

\item{maxRaw}{The upper bound of the raw score range}

\item{minNorm}{Clipping parameter for the lower bound of norm scores (default 25)}

\item{maxNorm}{Clipping parameter for the upper bound of norm scores (default 25)}

\item{step}{Stepping parameter for the raw scores (default 1)}

\item{monotonuous}{corrects for decreasing norm scores in case of model inconsistencies (default)}

\item{CI}{confidence coefficient, ranging from 0 to 1, default .9}

\item{reliability}{coefficient, ranging between  0 to 1}

\item{pretty}{Format table by collapsing intervals and rounding to meaningful precision}
}
\value{
either data.frame with raw scores and the predicted norm scores in case of simple A value or a list
of norm tables if vector of A values was provided
}
\description{
This function is comparable to 'normTable', despite it reverses the assignment:
A table with raw scores and the according norm scores for a specific age based on the regression
model is generated. This way, the inverse function of the regression model is solved numerically with
brute force. Please specify the range of raw values, you want to cover. With higher precision
and smaller stepping, this function becomes computational intensive.
In case a confidence coefficient (CI, default .9) and the reliability is specified,
confidence intervals are computed for the true score estimates, including a correction for
regression to the mean (Eid & Schmidt, 2012, p. 272).
}
\examples{
\dontrun{
# Generate cnorm object from example data
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
# generate a norm table for the raw value range from 0 to 28 for the time point month 7 of grade 3
table <- rawTable(3 + 7 / 12, cnorm.elfe, minRaw = 0, maxRaw = 28)

# generate several raw tables
table <- rawTable(c(2.5, 3.5, 4.5), cnorm.elfe, minRaw = 0, maxRaw = 28)

# additionally compute confidence intervals
table <- rawTable(c(2.5, 3.5, 4.5), cnorm.elfe, minRaw = 0, maxRaw = 28, CI = .9, reliability = .94)

# conventional norming, set age to arbitrary value
model <- cnorm(raw = elfe$raw)
rawTable(0, model)
}

}
\references{
Eid, M. & Schmidt, K. (2012). Testtheorie und Testkonstruktion. Hogrefe.
}
\seealso{
normTable

Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{bb_prepare_data}
\alias{bb_prepare_data}
\title{Validate and clean age / score / weights input}
\usage{
bb_prepare_data(age, score, weights = NULL)
}
\arguments{
\item{age}{Numeric vector of predictor values.}

\item{score}{Numeric vector of response values.}

\item{weights}{Optional numeric vector of weights.}
}
\value{
A list with the (possibly filtered) elements age, score, weights.
}
\description{
Performs the shared input validation for the beta-binomial fitting
functions: length checks, removal of non-finite cases, weight validation,
and checks for negative or non-integer scores as well as degenerate
predictor variance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{predictCoefficients_shash}
\alias{predictCoefficients_shash}
\title{Predict parameters for a shash regression model}
\usage{
predictCoefficients_shash(model, ages)
}
\arguments{
\item{model}{An object of class "cnormShash"}

\item{ages}{A numeric vector of age points for prediction}
}
\value{
A data frame with predicted parameters and statistics
}
\description{
Predict parameters for a shash regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{cnorm.cv}
\alias{cnorm.cv}
\title{Cross-validation for Term Selection in cNORM}
\usage{
cnorm.cv(
  data,
  formula = NULL,
  repetitions = 5,
  norms = TRUE,
  min = 4,
  max = 20,
  cv = "full",
  pCutoff = NULL,
  width = NA,
  raw = NULL,
  group = NULL,
  age = NULL,
  weights = NULL
)
}
\arguments{
\item{data}{Data frame of norm sample or a cnorm object. Should have ranking,
powers, and interaction of L and A.}

\item{formula}{Formula from an existing regression model; min/max functions
ignored. If using a cnorm object, this is automatically fetched.}

\item{repetitions}{Number of repetitions for cross-validation.}

\item{norms}{If TRUE, computes norm score crossfit and R^2. Note:
Computationally intensive.}

\item{min}{Start with a minimum number of terms (default = 4).}

\item{max}{Maximum terms in model, up to (k + 1) * (t + 1) - 1 (default = 20).}

\item{cv}{"full" (default) splits data into training/validation, then ranks.
Otherwise, expects a pre-ranked dataset.}

\item{pCutoff}{Checks stratification for unbalanced data. Performs a t-test
per group. Default set to 0.2 to minimize beta error.}

\item{width}{If provided, ranking done via `rankBySlidingWindow`. Otherwise,
by group.}

\item{raw}{Name of the raw score variable.}

\item{group}{Name of the grouping variable.}

\item{age}{Name of the age variable.}

\item{weights}{Name of the weighting parameter.}
}
\value{
Table with results per term number: RMSE for raw scores, R^2 for
  norm scores, and crossfit measure.
}
\description{
Assists in determining the optimal number of terms for the regression model
using repeated Monte Carlo cross-validation. It leverages an 80-20 split
between training and validation data, with stratification by norm group or
random sample in case of using sliding window ranking.
}
\details{
Successive models, with an increasing number of terms, are evaluated, and the
RMSE for raw scores plotted. This encompasses the training, validation, and
entire dataset. If `norms` is set to TRUE (default), the function will also
calculate the mean norm score reliability and crossfit measures. Note that
due to the computational requirements of norm score calculations, execution
can be slow, especially with numerous repetitions or terms.

When `cv` is set to "full" (default), both test and validation datasets are
ranked separately, providing comprehensive cross-validation. For a more
streamlined validation process focused only on modeling, a pre-ranked dataset
can be used. The output comprises RMSE for raw score models, norm score R^2,
delta R^2, crossfit, and the norm score SE according to Oosterhuis, van der
Ark, & Sijtsma (2016).

This function is not yet prepared for the 'extensive' search strategy,
introduced in version 3.3, but instead relies on the first model per number
of terms, without consistency check.

For assessing overfitting:
\deqn{CROSSFIT = R(Training; Model)^2 / R(Validation; Model)^2}
A CROSSFIT > 1 suggests overfitting, < 1 suggests potential underfitting, and
values around 1 are optimal, given a low raw score RMSE and high norm score
validation R^2.

Suggestions for ideal model selection:
\itemize{
  \item Visual inspection of percentiles with `plotPercentiles` or `plotPercentileSeries`.
  \item Pair visual inspection with repeated cross-validation (e.g., 10 repetitions).
  \item Aim for low raw score RMSE and high norm score R^2, avoiding terms with significant overfit (e.g., crossfit > 1.1).
}
}
\examples{
\dontrun{
# Example: Plot cross-validation RMSE by number of terms (up to 9) with three repetitions.
result <- cnorm(raw = elfe$raw, group = elfe$group)
cnorm.cv(result$data, min = 2, max = 9, repetitions = 3)

# Using a cnorm object examines the predefined formula.
cnorm.cv(result, repetitions = 1)
}

}
\references{
Oosterhuis, H. E. M., van der Ark, L. A., & Sijtsma, K. (2016).
  Sample Size Requirements for Traditional and Regression-Based Norms.
  Assessment, 23(2), 191-202. https://doi.org/10.1177/1073191115580638
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotCnorm}
\alias{plotCnorm}
\title{General convenience plotting function}
\usage{
plotCnorm(x, y, ...)
}
\arguments{
\item{x}{a cnorm object}

\item{y}{the type of plot as a string, can be one of
'raw' (1), 'norm' (2), 'curves' (3), 'percentiles' (4), 'series' (5), 'subset' (6),
or 'derivative' (7), either as a string or the according index}

\item{...}{additional parameters for the specific plotting function}
}
\description{
General convenience plotting function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{weightedAverageModel}
\alias{weightedAverageModel}
\title{BIC-Weighted Model Averaging Across Consistent Candidate Models}
\usage{
weightedAverageModel(results, data, raw, weights = NULL, deltaBIC = 10)
}
\arguments{
\item{results}{A (screened) \code{summary.regsubsets} object, including a
\code{consistent} flag per row.}

\item{data}{The preprocessed norm data.}

\item{raw}{Name of the raw score variable.}

\item{weights}{Optional numeric vector of case weights.}

\item{deltaBIC}{Only models within this BIC distance of the best candidate
are averaged (default 10; Burnham & Anderson, 2002).}
}
\value{
An \code{lm}-type object with averaged coefficients over the union of
  the candidate terms; \code{fitted.values} and \code{residuals} are updated
  accordingly.
}
\description{
Computes a weighted average of the regression coefficients across the
consistency-screened candidate models returned by \code{regsubsets} and
\code{screenSubset}. Weights are information-theoretic model weights
\eqn{w_j \propto exp(-\Delta BIC_j / 2)}, restricted to models within
\code{deltaBIC} of the best model. Each candidate is refitted on the
complete sample via (weighted) least squares; coefficients of terms not
included in a candidate are treated as zero.
}
\details{
Because all candidates passed the monotonicity screening (same direction),
and the averaging weights are positive and sum to one, the averaged model is
itself monotone and thus consistent. In contrast to averaging coefficients
over subsamples (deprecated \code{subsample_lm}), this approach targets the
actual source of variance - model selection - while each component estimate
remains a full-sample least squares fit.
}
\references{
Burnham, K. P., & Anderson, D. R. (2002). Model Selection and
  Multimodel Inference. Springer.
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ppvt}
\alias{ppvt}
\title{Vocabulary development from 2.5 to 17}
\format{
A data frame with 4542 rows and 6 variables:
\describe{
  \item{age}{the chronological age of the child}
  \item{sex}{the sex of the test taker, 1=male, 2=female}
  \item{migration}{migration status of the family, 0=no, 1=yes}
  \item{region}{factor specifying the region, the data were collected; grouped into south, north, east and west}
  \item{raw}{the raw score of the student, spanning values from 0 to 228}
  \item{group}{age group of the child, determined by the getGroups()-function with 12 equidistant age groups}
}

A data frame with 5600 rows and 9 columns
}
\source{
\url{https://www.psychometrica.de/ppvt4.html}
}
\usage{
ppvt
}
\description{
A dataset based on an unstratified sample of PPVT4 data (German adaption). The PPVT4 consists of blocks of items with
12 items each. Each item consists of 4 pictures. The test taker is given a word orally and he or she has to point out
the picture matching the oral word. Bottom and ceiling blocks of items are determined according to age and performance. For
instance, when a student knows less than 4 word from a block of 12 items, the testing stops. The sample is not identical
with the norm sample and includes doublets of cases in order to align the sample size per age group. It is
primarily intended for running the cNORM analyses with regard to modeling and stratification.
}
\examples{
\dontrun{
# Example with continuous age variable, ranked with sliding window
model.ppvt.sliding <- cnorm(age=ppvt$age, raw=ppvt$raw, width=1)

# Example with age groups; you might first want to experiment with
# the granularity of the groups via the 'getGroups()' function
model.ppvt.group <- cnorm(group=ppvt$group, raw=ppvt$raw) # with predefined groups
model.ppvt.group <- cnorm(group=getGroups(ppvt$age, n=15, equidistant = T),
                          raw=ppvt$raw) # groups built 'on the fly'


# plot information function
plot(model.ppvt.group, "subset")

# check model consistency
checkConsistency(model.ppvt.group)

# plot percentiles
plot(model.ppvt.group, "percentiles")
}
}
\references{
Lenhard, A., Lenhard, W., Segerer, R. & Suggate, S. (2015). Peabody Picture Vocabulary Test - Revision IV (Deutsche Adaption). Frankfurt a. M./Germany: Pearson Assessment.
}
\concept{vocabulary acquisition development receptive}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{weighted.rank}
\alias{weighted.rank}
\title{Weighted rank estimation}
\usage{
weighted.rank(x, weights = NULL)
}
\arguments{
\item{x}{A numerical vector}

\item{weights}{A numerical vector with weights; should have the same length as x}
}
\value{
the weighted absolute ranks
}
\description{
Conducts weighted ranking on the basis of sums of weights per unique raw score.
Please provide a vector with raw values and an additional vector with the weight
of each observation. In case the weight vector is NULL, a normal ranking is done.
The vectors may not include NAs and the weights should be positive non-zero
values. With equal weights, the result reproduces the mid-ranks of base R's
\code{rank()} (\code{ties.method = "average"}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotPercentileSeries}
\alias{plotPercentileSeries}
\title{Generates a series of plots with percentile curves for different models}
\usage{
plotPercentileSeries(
  model,
  start = 1,
  end = NULL,
  group = NULL,
  percentiles = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975),
  filename = NULL
)
}
\arguments{
\item{model}{The Taylor polynomial regression model object or a cnorm object}

\item{start}{Number of terms to start with (default 1)}

\item{end}{Number of terms to end with; defaults to the largest available
model}

\item{group}{The name of the grouping variable; the distinct groups are
automatically determined}

\item{percentiles}{Vector with percentile scores, ranging from 0 to 1
(exclusive)}

\item{filename}{Prefix of the filename. If specified, the plots are saved as
png files in the directory of the workspace, instead of only displaying
them. The number of terms is appended to the prefix.}
}
\value{
A named list of plots (names indicate the number of terms),
  returned invisibly
}
\description{
This function makes use of 'plotPercentiles' to generate a series of plots
for models with an increasing number of terms. It draws on the information
provided by the model object to determine the bounds of the modeling (age
and standard score range). It can be used as an additional model check to
determine the best fitting model. Please have a look at the
'plotPercentiles' function for further information.
}
\details{
Each model of the series is refitted on the complete norm sample (applying
case weights, if the norm data were post stratified). Models are identified
by their actual number of terms, which - after consistency screening in
\code{bestModel} - is not necessarily identical to the row number of the
model selection table. The subtitle of each plot reports the number of
terms, the adjusted R2 and, if available, the result of the consistency
check.
}
\examples{
\dontrun{
  # Load example data set, compute model and plot results
  result <- cnorm(raw = elfe$raw, group = elfe$group)
  plotPercentileSeries(result, start = 4, end = 6)
}

}
\seealso{
plotPercentiles

Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{predictCoefficients}
\alias{predictCoefficients}
\title{Predict mean and standard deviation for a beta binomial regression model}
\usage{
predictCoefficients(model, ages, n = NULL)
}
\arguments{
\item{model}{An object of class "cnormBetaBinomial".}

\item{ages}{A numeric vector of age points at which to make predictions.}

\item{n}{The maximum score to be achieved.}
}
\value{
A data frame with columns:
  \item{age}{The input age points}
  \item{mu}{Predicted mean values}
  \item{sigma}{Predicted standard deviation values}
  \item{a}{Alpha parameters}
  \item{b}{Beta parameters}
}
\description{
This function generates predictions from a fitted beta binomial regression
model (mode 1) for new age points and converts them into alpha and beta
parameters of the beta-binomial distribution via method of moments.
}
\details{
The function applies the same standardization used in model fitting,
generates predictions on the standardized scale, and transforms these back
to the original scale. If the method of moments yields invalid (i.e.
non-positive or non-finite) parameters, a mean-preserving low-precision
fallback is used instead of arbitrarily clamping both parameters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotPercentiles}
\alias{plotPercentiles}
\title{Plot norm curves against actual percentiles}
\usage{
plotPercentiles(
  model,
  minRaw = NULL,
  maxRaw = NULL,
  minAge = NULL,
  maxAge = NULL,
  raw = NULL,
  group = NULL,
  percentiles = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975),
  scale = NULL,
  title = NULL,
  subtitle = NULL,
  points = FALSE
)
}
\arguments{
\item{model}{The Taylor polynomial regression model object from the cNORM}

\item{minRaw}{Lower bound of the raw score (default = 0)}

\item{maxRaw}{Upper bound of the raw score}

\item{minAge}{Variable to restrict the lower bound of the plot to a specific age}

\item{maxAge}{Variable to restrict the upper bound of the plot to a specific age}

\item{raw}{The name of the raw variable}

\item{group}{The name of the grouping variable; the distinct groups are automatically
determined}

\item{percentiles}{Vector with percentile scores, ranging from 0 to 1 (exclusive)}

\item{scale}{The norm scale, either 'T', 'IQ', 'z', 'percentile' or
self defined with a double vector with the mean and standard deviation,
f. e. c(10, 3) for Wechsler scale index points; if NULL, scale information from the
data preparation is used (default)}

\item{title}{custom title for plot}

\item{subtitle}{custom title for plot}

\item{points}{Logical indicating whether to plot the data points. Default is TRUE.}
}
\description{
The function plots the norm curves based on the regression model against
the actual percentiles from the raw data. As in 'plotNormCurves',
please check for inconsistent curves, especially intersections.
Violations of this assumption are a strong indication for problems
in modeling the relationship between raw and norm scores.
In general, extrapolation (point 1 and 2) can carefully be done to a
certain degree outside the original sample, but it should in general
be handled with caution.
The original percentiles are displayed as distinct points in the according
color, the model based projection of percentiles are drawn as lines.
Please note, that the estimation of the percentiles of the raw data is done with
the quantile function with the default settings.
In case, you get 'jagged' or disorganized percentile curve, try to reduce the 'k'
and/or 't' parameter in modeling.
}
\examples{
\dontrun{
  # Load example data set, compute model and plot results
  result <- cnorm(raw = elfe$raw, group = elfe$group)
  plotPercentiles(result)
}
}
\seealso{
plotNormCurves, plotPercentileSeries

Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{summary.cnorm}
\alias{summary.cnorm}
\title{S3 method for printing the results and regression function of a cnorm model}
\usage{
\method{summary}{cnorm}(object, ...)
}
\arguments{
\item{object}{A regression model or cnorm object}

\item{...}{additional parameters}
}
\value{
A report on the regression function, weights, R2 and RMSE
}
\description{
S3 method for printing the results and regression function of a cnorm model
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{simMean}
\alias{simMean}
\title{Simulate mean per age}
\usage{
simMean(age)
}
\arguments{
\item{age}{the age variable}
}
\value{
return predicted means
}
\description{
Simulate mean per age
}
\examples{
\dontrun{
x <- simMean(a)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{normTable.shash}
\alias{normTable.shash}
\title{Calculate Norm Tables for Sinh-Arcsinh Distribution}
\usage{
normTable.shash(
  model,
  ages,
  start = NULL,
  end = NULL,
  step = 1,
  CI = 0.9,
  reliability = NULL
)
}
\arguments{
\item{model}{Fitted shash model object of class "cnormShash"}

\item{ages}{Numeric vector of age points for norm table generation}

\item{start}{Minimum raw score value for the norm table}

\item{end}{Maximum raw score value for the norm table}

\item{step}{Step size between consecutive raw scores (default: 1)}

\item{CI}{Confidence coefficient (0-1, default: 0.9) for confidence intervals}

\item{reliability}{Reliability coefficient (0-1) for true score confidence intervals}
}
\value{
List of data frames (one per age) containing:
  \item{x}{Raw scores}
  \item{Px}{Approximate interval probabilities (density times step width).
    These are consistent with \code{Pcum} but do not necessarily sum to 1
    over the tabulated score range.}
  \item{Pcum}{Cumulative probabilities}
  \item{Percentile}{Percentile ranks (0-100)}
  \item{z}{Standardized z-scores}
  \item{norm}{Norm scores in specified scale}
  \item{lowerCI, upperCI}{Confidence intervals (if reliability provided)}
  \item{lowerCI_PR, upperCI_PR}{CI as percentile ranks (if reliability provided)}
}
\description{
Generates norm tables for specific ages based on a fitted Sinh-Arcsinh (shash) regression model.
Computes interval probabilities, percentiles, z-scores, and norm scores for a specified range of raw scores.
Optionally includes confidence intervals when reliability is provided.
}
\details{
For continuous shash distributions, densities are converted to interval probabilities
(density times step width) and cumulative probabilities. When reliability is specified,
confidence intervals include correction for regression to the mean.
}
\examples{
\dontrun{
# Basic norm table
model <- cnorm.shash(age, score)
tables <- normTable.shash(model, ages = c(7, 8, 9), start = 0, end = 50)

# With confidence intervals and finer granularity
tables_ci <- normTable.shash(model, ages = c(8, 9), start = 10, end = 40,
                             step = 0.5, CI = 0.95, reliability = 0.85)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{derivationTable}
\alias{derivationTable}
\title{Create a table based on first order derivative of the regression model for specific age}
\usage{
derivationTable(A, model, minNorm = NULL, maxNorm = NULL, step = 0.1)
}
\arguments{
\item{A}{the age}

\item{model}{The regression model or a cnorm object}

\item{minNorm}{The lower bound of the norm value range}

\item{maxNorm}{The upper bound of the norm value range}

\item{step}{Stepping parameter with lower values indicating higher precision}
}
\value{
data.frame with norm scores and the predicted scores based on the
derived regression function
}
\description{
In order to check model assumptions, a table of the first order derivative of the model
coefficients is created.
}
\examples{
\dontrun{
# Generate cnorm object from example data
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)

# retrieve function for time point 6
d <- derivationTable(6, cnorm.elfe, step = 0.5)
}

}
\seealso{
plotDerivative, derive

Other predict:
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{bb_resolve_scale}
\alias{bb_resolve_scale}
\title{Resolve a norm scale specification}
\usage{
bb_resolve_scale(scale)
}
\arguments{
\item{scale}{Either a character ("T", "IQ", "z", anything else = percentile)
or a numeric vector of length 2 with mean and sd.}
}
\value{
A numeric vector c(mean, sd); c(NA, NA) indicates percentile output.
}
\description{
Resolve a norm scale specification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{elfe}
\alias{elfe}
\title{Sentence completion test from ELFE 1-6}
\format{
A data frame with 1400 rows and 3 variables:
\describe{
  \item{personID}{ID of the student}
  \item{group}{grade level, with x.5 indicating the end of the school year and x.0 indicating the middle of the school year}
  \item{raw}{the raw score of the student, spanning values from 0 to 28}
}

A data frame with 1400 rows and 3 columns
}
\source{
\url{https://www.psychometrica.de/elfe2.html}
}
\usage{
elfe
}
\description{
A dataset containing the raw data of 1400 students from grade 2 to 5 in the sentence
comprehension test from ELFE 1-6 (Lenhard & Schneider, 2006). In this test, students
are presented lists of sentences with one gap. The student has to fill in the correct
solution by selecting from a list of 5 alternatives per sentence. The alternatives
include verbs, adjectives, nouns, pronouns and conjunctives. Each item stems from
the same word type. The text is speeded, with a time cutoff of 180 seconds. The
variables are as follows:
}
\examples{
\dontrun{
  # prepare data, retrieve model and plot percentiles
  model <- cnorm(group = elfe$group, raw = elfe$raw)
}
}
\references{
Lenhard, W. & Schneider, W.(2006). Ein Leseverstaendnistest fuer Erst- bis Sechstklaesser. Goettingen/Germany: Hogrefe.
}
\concept{reading comprehension}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{plot.cnormBetaBinomial}
\alias{plot.cnormBetaBinomial}
\title{Plot cnormBetaBinomial Model with Data and Percentile Lines}
\usage{
\method{plot}{cnormBetaBinomial}(x, ...)
}
\arguments{
\item{x}{A fitted model object of class "cnormBetaBinomial" or
"cnormBetaBinomial2".}

\item{...}{Additional arguments passed to the plot method.
\itemize{
   \item age A vector of the age data.
   \item score A vector of the score data.
   \item weights An optional numeric vector of weights for each observation.
   \item percentiles An optional vector with the percentiles to plot.
   \item points Logical indicating whether to plot the data points.
     Default is TRUE.
   \item discrete Logical indicating whether to plot the exact discrete
     beta-binomial quantiles (TRUE, default) or a smooth continuous
     approximation via the underlying beta distribution (FALSE).
 }}
}
\value{
A ggplot object.
}
\description{
This function creates a visualization of a fitted cnormBetaBinomial model,
including the original data points, the manifest percentiles (diamond-shaped
dots per age group) and the model-implied percentile lines.
}
\details{
By default (\code{discrete = TRUE}), the percentile lines represent the
exact quantiles of the fitted discrete beta-binomial distribution and are
therefore step functions. Setting \code{discrete = FALSE} draws smooth
lines based on the quantiles of the underlying beta (mixing) distribution.
Note that this continuous approximation omits the binomial stage of the
variance and therefore displays less spread than the fitted model actually
implies, particularly in the outer percentiles. Use it only for visual
comparability with continuous modelling approaches.

Cases with non-finite values (NA, NaN, Inf) in age, score or weights are
removed prior to plotting, mirroring the behavior of the fitting functions.
}
\seealso{
Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotSubset}
\alias{plotSubset}
\title{#' Evaluate information criteria for regression model}
\usage{
plotSubset(model, type = 0)
}
\arguments{
\item{model}{The regression model from the bestModel function or a cnorm object.}

\item{type}{Integer specifying the type of plot to generate:
\itemize{
  \item 0: Adjusted R2 by number of terms (default)
  \item 1: Log-transformed Mallows's Cp by adjusted R2
  \item 2: Bayesian Information Criterion (BIC) by adjusted R2
  \item 3: Root Mean Square Error (RMSE) by number of terms
  \item 4: Residual Sum of Squares (RSS) by number of terms
  \item 5: F-test statistic for consecutive models by number of terms
  \item 6: p-value for model tests by number of terms
}}
}
\value{
A ggplot object representing the selected information criterion plot.
}
\description{
This function plots various information criteria and model fit statistics
against the number of terms or the adjusted R-squared, depending on the type
of plot selected. It helps in model selection by visualizing different
aspects of model performance. Models which did not pass the consistency
check are depicted with an empty circle; the automatically selected model is
highlighted in red. If BIC-weighted model averaging was applied
(\code{averaging = TRUE}), a caption indicates that the final coefficients
are a weighted combination of the consistent candidate models shown.
}
\details{
The function generates different plots to help in model selection:

- For types 1 and 2 (Mallows's Cp and BIC), look for the "elbow" in the curve
  where the information criterion begins to drop. This often indicates a good
  balance between model fit and complexity.
- For type 0 (Adjusted R2), higher values indicate better fit, but be cautious
  of overfitting with values approaching 1.
- For types 3 and 4 (RMSE and RSS), lower values indicate better fit.
- For type 5 (F-test), higher values suggest significant improvement with
  added terms.
- For type 6 (p-values), values below the significance level (typically 0.05)
  suggest significant improvement with added terms.

The F-tests and p-values compare each model with the preceding (smaller) one,
with degrees of freedom based on the actual difference in the number of
parameters. After consistency screening, consecutive models may differ by
more than one term.
}
\note{
It's important to balance statistical measures with practical considerations
and to visually inspect the model fit using functions like
\code{plotPercentiles}.
}
\examples{
\dontrun{
# Compute model with example data and plot information function
cnorm.model <- cnorm(raw = elfe$raw, group = elfe$group)
plotSubset(cnorm.model)

# Plot BIC against adjusted R-squared
plotSubset(cnorm.model, type = 2)

# Plot RMSE against number of terms
plotSubset(cnorm.model, type = 3)
}

}
\seealso{
\code{\link{bestModel}}, \code{\link{plotPercentiles}}, \code{\link{printSubset}}

Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{shash}
\alias{shash}
\alias{dshash}
\alias{pshash}
\alias{qshash}
\alias{rshash}
\title{Sinh-Arcsinh (shash) Distribution}
\usage{
dshash(x, mu = 0, sigma = 1, epsilon = 0, delta = 1, log = FALSE)

pshash(
  q,
  mu = 0,
  sigma = 1,
  epsilon = 0,
  delta = 1,
  lower.tail = TRUE,
  log.p = FALSE
)

qshash(
  p,
  mu = 0,
  sigma = 1,
  epsilon = 0,
  delta = 1,
  lower.tail = TRUE,
  log.p = FALSE
)

rshash(n, mu = 0, sigma = 1, epsilon = 0, delta = 1)
}
\arguments{
\item{x, q}{vector of quantiles}

\item{mu}{location parameter (default: 0)}

\item{sigma}{scale parameter (must be > 0, default: 1)}

\item{epsilon}{skewness parameter (default: 0, symmetric distribution)}

\item{delta}{tail weight parameter (must be > 0, default: 1 for normal-like
tails; \code{delta < 1} gives heavier tails, \code{delta > 1} gives
lighter tails)}

\item{log, log.p}{logical; if TRUE, probabilities p are given as log(p)}

\item{lower.tail}{logical; if TRUE (default), probabilities are P[X <= x]
otherwise, P[X > x]}

\item{p}{vector of probabilities}

\item{n}{number of observations. If \code{length(n) > 1}, the length is
taken to be the number required.}
}
\value{
\code{dshash} gives the density, \code{pshash} gives the distribution
function, \code{qshash} gives the quantile function, and \code{rshash}
generates random deviates.

The length of the result is determined by \code{n} for \code{rshash}, and
is the maximum of the lengths of the numerical arguments for the other functions.
}
\description{
Density, distribution function, quantile function and random generation
for the Sinh-Arcsinh distribution with location parameter \code{mu},
scale parameter \code{sigma}, skewness parameter \code{epsilon}, and
tail weight parameter \code{delta}.
}
\details{
The Sinh-Arcsinh distribution (Jones & Pewsey, 2009) is defined by the transformation:
\deqn{X = \mu + \sigma \cdot \sinh\left(\frac{\text{asinh}(Z) - \epsilon}{\delta}\right)}
where \eqn{Z \sim N(0,1)} is a standard normal variable.

The four parameters control:
\itemize{
  \item \code{mu}: Location (similar to mean)
  \item \code{sigma}: Scale (similar to standard deviation)
  \item \code{epsilon}: Skewness (\code{epsilon = 0} gives symmetry)
  \item \code{delta}: Tail weight (\code{delta = 1} gives normal-like tails,
        \code{delta < 1} gives heavier tails, \code{delta > 1} gives
        lighter tails)
}
}
\examples{
# Generate random samples
x <- rshash(1000, mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2)

# Density
plot(density(x))
curve(dshash(x, mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2),
      add = TRUE, col = "red")

# Cumulative probability
pshash(0, mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2)

# Quantiles
qshash(c(0.025, 0.5, 0.975), mu = 0, sigma = 1, epsilon = 0.5, delta = 1.2)

# Compare with normal distribution (epsilon = 0, delta = 1)
oldpar <- par(mfrow = c(2, 2))
x_vals <- seq(-4, 4, length.out = 200)
plot(x_vals, dshash(x_vals), type = "l", main = "Symmetric (like normal)")
plot(x_vals, dshash(x_vals, epsilon = 1), type = "l", main = "Right skewed")
plot(x_vals, dshash(x_vals, delta = 0.5), type = "l", main = "Heavy tails")
plot(x_vals, dshash(x_vals, delta = 2), type = "l", main = "Light tails")
par(oldpar)

}
\references{
Jones, M. C., & Pewsey, A. (2009). Sinh-arcsinh distributions.
\emph{Biometrika}, 96(4), 761-780. \doi{10.1093/biomet/asp053}
}
\seealso{
\link[stats]{Normal} for the normal distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{predict.cnormShash}
\alias{predict.cnormShash}
\title{Predict Norm Scores from Raw Scores}
\usage{
\method{predict}{cnormShash}(object, ...)
}
\arguments{
\item{object}{A fitted model object of class 'cnormShash'.}

\item{...}{Additional arguments passed to the prediction method:
\itemize{
   \item age A numeric vector of ages, same length as score.
   \item score A numeric vector of raw scores.
   \item range The range of the norm scores in standard deviations. Default is 3.
     Thus, scores in the range of +/- 3 standard deviations are considered.
 }}
}
\value{
A numeric vector of norm scores.
}
\description{
This function calculates norm scores based on raw scores, age, and a fitted cnormShash model.
}
\details{
The function predicts the Sinh-Arcsinh (shash) distribution parameters (mu, sigma, epsilon, delta)
for each age using the provided model. It then calculates the cumulative probability for each
raw score given these parameters using the continuous shash distribution. Finally,
it converts these probabilities to the norm scale specified in the model. Raw scores
at the extreme floor or ceiling of the distribution are mapped to the boundary of the
norm score range (mean +/- \code{range} standard deviations), not to the mean.
}
\examples{
\dontrun{
# Assuming you have a fitted model named 'shash_model':
model <- cnorm.shash(children$age, children$score)
raw_scores <- c(25.5, 30.2, 18.7, 45.3)
ages <- c(7, 8, 9, 10)
norm_scores <- predict(model, ages, raw_scores)
}

}
\seealso{
Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{plot.cnormBetaBinomial2}
\alias{plot.cnormBetaBinomial2}
\title{Plot cnormBetaBinomial2 Model with Data and Percentile Lines}
\usage{
\method{plot}{cnormBetaBinomial2}(x, ...)
}
\arguments{
\item{x}{A fitted model object of class "cnormBetaBinomial" or
"cnormBetaBinomial2".}

\item{...}{Additional arguments passed to the plot method.
\itemize{
   \item age A vector of the age data.
   \item score A vector of the score data.
   \item weights An optional numeric vector of weights for each observation.
   \item percentiles An optional vector with the percentiles to plot.
   \item points Logical indicating whether to plot the data points.
     Default is TRUE.
   \item discrete Logical indicating whether to plot the exact discrete
     beta-binomial quantiles (TRUE, default) or a smooth continuous
     approximation via the underlying beta distribution (FALSE).
 }}
}
\value{
A ggplot object.
}
\description{
This function creates a visualization of a fitted cnormBetaBinomial2
model, including the original data points, manifest percentiles and
specified percentile lines. See \code{\link{plot.cnormBetaBinomial}}.
}
\seealso{
Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{bb_distribution}
\alias{bb_distribution}
\title{Beta-binomial distribution on the full support 0:n}
\usage{
bb_distribution(a, b, n)
}
\arguments{
\item{a}{Alpha parameter (> 0).}

\item{b}{Beta parameter (> 0).}

\item{n}{Number of trials / maximum score.}
}
\value{
A list with elements x (support 0:n), Px (pmf), cum (cdf) and
  perc (mid-p percentiles, i.e. P(X < x) + 0.5 * P(X = x)).
}
\description{
Computes the (normalized) probability mass function, the cumulative
distribution function and mid-p percentiles of a beta-binomial
distribution in a numerically stable way (log space).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{predict.cnormBetaBinomial2}
\alias{predict.cnormBetaBinomial2}
\title{Predict Norm Scores from Raw Scores}
\usage{
\method{predict}{cnormBetaBinomial2}(object, ...)
}
\arguments{
\item{object}{A fitted model object of class 'cnormBetaBinomial' or
'cnormBetaBinomial2'.}

\item{...}{Additional arguments passed to the prediction method:
\itemize{
   \item age A numeric vector of ages, same length as score.
   \item score A numeric vector of raw scores.
   \item range The range of the norm scores in standard deviations.
     Default is 3.
 }}
}
\value{
A numeric vector of norm scores.
}
\description{
This function calculates norm scores based on raw scores, age, and a
fitted cnormBetaBinomial2 model. See
\code{\link{predict.cnormBetaBinomial}} for details.
}
\seealso{
Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raking.R
\name{standardizeRakingWeights}
\alias{standardizeRakingWeights}
\title{Function for standardizing raking weights}
\usage{
standardizeRakingWeights(weights)
}
\arguments{
\item{weights}{Raking weights computed by computeWeights()}
}
\value{
the standardized weights
}
\description{
Raking weights get divided by the smallest weight. Thereby, all weights
become larger or equal to 1 without changing the ratio of the weights
to each other.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{simSD}
\alias{simSD}
\title{Simulate sd per age}
\usage{
simSD(age)
}
\arguments{
\item{age}{the age variable}
}
\value{
return predicted sd
}
\description{
Simulate sd per age
}
\examples{
\dontrun{
x <- simSD(a)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{autoselect.shash}
\alias{autoselect.shash}
\title{Automatic model selection for SinH-ArcSinH continuous norming via BIC}
\usage{
autoselect.shash(
  age,
  score,
  weights = NULL,
  max_mu = 4,
  max_sigma = 3,
  max_epsilon = 3,
  max_delta = 1,
  min_mu = 1,
  min_sigma = 1,
  min_epsilon = 1,
  min_delta = 0,
  delta = 1,
  control = NULL,
  scale = "T",
  parallel = TRUE,
  n_cores = NULL,
  plot = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{age, score}{Numeric vectors of predictor and response values.}

\item{weights}{Optional numeric vector of observation weights.}

\item{max_mu, max_sigma, max_epsilon, max_delta}{Maximum polynomial degrees for
the four shash parameters. Defaults: \code{4, 3, 3, 1}.}

\item{min_mu, min_sigma, min_epsilon}{Minimum polynomial degrees (default 1).}

\item{min_delta}{Minimum degree for \eqn{\delta}. \code{0} (default) means
the fixed-delta variant is included in the search.}

\item{delta}{Value of \eqn{\delta} used whenever the candidate uses
\code{delta_degree = 0} (i.e. fixed delta). Default 1.}

\item{control}{Optional control list passed to \code{\link[stats]{optim}}.}

\item{scale}{Norm scale (default \code{"T"}).}

\item{parallel}{Logical; attempt parallel execution. Default \code{TRUE}.}

\item{n_cores}{Number of cores. Defaults to all logical cores.}

\item{plot}{Logical; plot the selected model. Default \code{TRUE}.}

\item{verbose}{Logical; print progress. Default \code{TRUE}.}
}
\value{
The selected fitted \code{cnormShash} model with an additional
  element \code{$selection} containing:
  \itemize{
    \item \code{evaluated}: data frame of every combination tried, sorted by BIC.
    \item \code{selected}: list with the chosen degrees and BIC.
  }
}
\description{
Selects the polynomial degrees for the four components of a
\code{\link{cnorm.shash}} model (\eqn{\mu}, \eqn{\sigma}, \eqn{\epsilon},
\eqn{\delta}) by minimizing BIC over a full grid.
}
\details{
A degree of \code{0} for \eqn{\delta} is interpreted as a **fixed**
\eqn{\delta} (passed to \code{\link{cnorm.shash}} as \code{delta_degree = NULL})
with value taken from the \code{delta} argument. This allows the grid
search to choose between a fixed and a varying tail-weight model.

Parallel execution is attempted by default. If the workers cannot access
the \pkg{cNORM} namespace, the function transparently falls
back to sequential execution.
}
\examples{
\dontrun{
m <- autoselect.shash(elfe$group, elfe$raw)
m$selection$evaluated
summary(m, age = elfe$group, score = elfe$raw)
}

}
\seealso{
\code{\link{cnorm.shash}}, \code{\link{autoselect.betabinomial}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{log_likelihood2}
\alias{log_likelihood2}
\title{Calculate the negative log-likelihood for a beta-binomial regression model}
\usage{
log_likelihood2(params, X, Z, y, n, weights = NULL, lch = NULL)
}
\arguments{
\item{params}{A numeric vector containing all model parameters. The first
n_alpha elements are coefficients for the alpha model, and the remaining
elements are coefficients for the beta model.}

\item{X}{A matrix of predictors for the alpha model.}

\item{Z}{A matrix of predictors for the beta model.}

\item{y}{A numeric vector of response values.}

\item{n}{The maximum score (number of trials).}

\item{weights}{A numeric vector of weights for each observation
(NULL = equal weights).}

\item{lch}{Optional precomputed \code{lchoose(n, y)}. Since this term does
not depend on the parameters, passing it once avoids recomputation in
every optimizer iteration.}
}
\value{
The negative log-likelihood of the model (large finite penalty
  if non-finite).
}
\description{
This function computes the negative log-likelihood for a beta-binomial
regression model where both the alpha and beta parameters are modeled as
functions of predictors (mode 2).
}
\details{
Uses a numerically stable implementation of the beta-binomial
log-probability via \code{lbeta}. The linear predictors of
log(alpha) and log(beta) are clamped to [-20, 20].
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{predictCoefficients2}
\alias{predictCoefficients2}
\title{Predict alpha and beta parameters for a beta-binomial regression model}
\usage{
predictCoefficients2(model, ages, n = NULL)
}
\arguments{
\item{model}{An object of class "cnormBetaBinomial2".}

\item{ages}{A numeric vector of age points at which to make predictions.}

\item{n}{The maximum score to be achieved.}
}
\value{
A data frame with columns:
  \item{age}{The input age points}
  \item{a}{Predicted alpha values}
  \item{b}{Predicted beta values}
  \item{mu}{Predicted mean values}
  \item{sigma}{Predicted standard deviation values}
}
\description{
This function generates predictions from a fitted beta-binomial regression
model (mode 2) for new age points.
}
\details{
The function applies the same standardization used in model fitting and
the same clamping of the linear predictors (\eqn{\pm 20} on the log scale)
as the likelihood function, ensuring consistent behavior also under
(mild) extrapolation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{buildFunction}
\alias{buildFunction}
\title{Build regression function for bestModel}
\usage{
buildFunction(raw, k, t, age)
}
\arguments{
\item{raw}{name of the raw score variable}

\item{k}{the power degree for location}

\item{t}{the power degree for age}

\item{age}{use age}
}
\value{
regression function
}
\description{
Build regression function for bestModel
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{cnorm.betabinomial2}
\alias{cnorm.betabinomial2}
\title{Fit a beta-binomial regression model for continuous norming (mode 2)}
\usage{
cnorm.betabinomial2(
  age,
  score,
  n = NULL,
  weights = NULL,
  alpha_degree = 3,
  beta_degree = 3,
  control = NULL,
  scale = "T",
  plot = TRUE
)
}
\arguments{
\item{age}{A numeric vector of predictor values (e.g., age).}

\item{score}{A numeric vector of response values.}

\item{n}{The maximum score (number of trials in the beta-binomial
distribution). If NULL, max(score) is used.}

\item{weights}{A numeric vector of weights for each observation.
Default is NULL (equal weights).}

\item{alpha_degree}{Integer specifying the degree of the polynomial for
the alpha model. Default is 3.}

\item{beta_degree}{Integer specifying the degree of the polynomial for
the beta model. Default is 3.}

\item{control}{A list of control parameters to be passed to the
\code{optim} function (method "L-BFGS-B"). Any parameters not specified
are filled with adaptive defaults based on problem size.}

\item{scale}{Type of norm scale, either "T" (default), "IQ", "z" or a
double vector with the mean and standard deviation.}

\item{plot}{Logical indicating whether to plot the model. Default is TRUE.}
}
\value{
A list of class "cnormBetaBinomial2" containing:
  \item{alpha_est}{Estimated coefficients for the alpha model}
  \item{beta_est}{Estimated coefficients for the beta model}
  \item{se}{Standard errors of the estimated coefficients}
  \item{alpha_degree}{Degree of the polynomial for the alpha model}
  \item{beta_degree}{Degree of the polynomial for the beta model}
  \item{result}{Full result from the optimization procedure}
}
\description{
This function fits a beta-binomial regression model where both the alpha
and beta parameters of the beta-binomial distribution are modeled as
polynomial functions of the predictor variable (typically age). While
'cnorm.betabinomial1' fits a beta-binomial model on the basis of \eqn{\mu}
and \eqn{\sigma}, this function fits the model directly on the basis of
\eqn{\alpha} and \eqn{\beta}. This is the recommended default approach
(see \code{\link{cnorm.betabinomial}}).
}
\details{
The function standardizes the input variables, fits polynomial models for
both the alpha and beta parameters, and uses maximum likelihood estimation
with analytic (digamma-based) gradients to find the optimal parameters.
The optimization is performed using the L-BFGS-B method with parameter
bounds to prevent numerical issues.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3methods.R
\name{print.cnorm}
\alias{print.cnorm}
\title{S3 method for printing model selection information}
\usage{
\method{print}{cnorm}(x, ...)
}
\arguments{
\item{x}{The model from the 'bestModel' function or a cnorm object}

\item{...}{additional parameters}
}
\value{
A table with information criteria
}
\description{
After conducting the model fitting procedure on the data set, the best fitting
model has to be chosen. The print function shows the R2 and other information
on the different best fitting models with increasing number of predictors.
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{standardize}
\alias{standardize}
\title{Standardize a numeric vector}
\usage{
standardize(x)
}
\arguments{
\item{x}{A numeric vector to be standardized.}
}
\value{
A numeric vector of the same length as x, containing the standardized values.
}
\description{
This function standardizes a numeric vector by subtracting the mean
and dividing by the standard deviation. The resulting vector will have
a mean of 0 and a standard deviation of 1.
}
\examples{
\dontrun{
data <- c(1, 2, 3, 4, 5)
standardized_data <- standardize(data)
print(standardized_data)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{compare}
\alias{compare}
\title{Compare Two Norm Models Visually}
\usage{
compare(
  model1,
  model2,
  percentiles = c(0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975),
  age = NULL,
  score = NULL,
  weights = NULL,
  title = NULL,
  subtitle = NULL,
  discrete = TRUE
)
}
\arguments{
\item{model1}{First model object (distribution free, beta-binomial, or shash)}

\item{model2}{Second model object (distribution free, beta-binomial, or shash)}

\item{percentiles}{Vector with percentile scores, ranging from 0 to 1 (exclusive)}

\item{age}{Optional vector with manifest age or group values}

\item{score}{Optional vector with manifest raw score values}

\item{weights}{Optional vector with manifest weights}

\item{title}{Custom title for plot (optional)}

\item{subtitle}{Custom subtitle for plot (optional)}

\item{discrete}{Logical indicating whether beta-binomial models are displayed
with their exact discrete quantiles as step functions (TRUE, default) or
with a smooth continuous approximation via the underlying beta
distribution (FALSE). Ignored for other model types.}
}
\value{
A ggplot object showing the comparison of both models
}
\description{
This function creates a visualization comparing two norm models by displaying
their percentile curves. The first model is shown with solid lines, the second
with dashed lines. If age and score vectors are provided, manifest percentiles
are displayed as dots. The function works with regular cnorm models, beta-binomial
models, and shash models, allowing comparison between different model types.
}
\details{
For beta-binomial models, the exact quantiles of the discrete beta-binomial
distribution are displayed by default as step functions (\code{discrete = TRUE}).
Setting \code{discrete = FALSE} draws smooth lines based on the quantiles of
the underlying beta (mixing) distribution instead. Note that this continuous
approximation omits the binomial stage of the variance and therefore displays
less spread than the fitted model actually implies, particularly in the outer
percentiles. The parameter has no effect on Taylor polynomial or shash models.
}
\examples{
\dontrun{
# Compare different types of models
model1 <- cnorm(group = elfe$group, raw = elfe$raw)
model2 <- cnorm.betabinomial(elfe$group, elfe$raw)
model3 <- cnorm.shash(elfe$group, elfe$raw)

# Compare traditional cnorm with shash
compare(model1, model3, age = elfe$group, score = elfe$raw)

# Compare beta-binomial with shash
compare(model2, model3, age = elfe$group, score = elfe$raw)
}

}
\seealso{
Other plot:
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predictRaw}
\alias{predictRaw}
\title{Predict raw values}
\usage{
predictRaw(norm, age, coefficients, minRaw = -Inf, maxRaw = Inf)
}
\arguments{
\item{norm}{The norm score, e. g. a specific T score or a vector of scores}

\item{age}{The age value or a vector of scores}

\item{coefficients}{The a cnorm object or the coefficients from the regression model}

\item{minRaw}{Minimum score for the results; can be used for clipping unrealistic outcomes,
usually set to the lower bound of the range of values of the test (default: 0)}

\item{maxRaw}{Maximum score for the results; can be used for clipping unrealistic outcomes
usually set to the upper bound of the range of values of the test}
}
\value{
the predicted raw score or a data.frame of scores in case, lists of norm scores or age is used
}
\description{
Most elementary function to predict raw score based on Location (L, T score),
Age (grouping variable) and the coefficients from a regression model.
}
\examples{
\dontrun{
# Prediction of single scores
model <- cnorm(raw = elfe$raw, group = elfe$group)
predictRaw(35, 3.5, model)
}

}
\seealso{
Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{cnorm.betabinomial}
\alias{cnorm.betabinomial}
\title{Fit a beta-binomial regression model for continuous norming}
\usage{
cnorm.betabinomial(
  age,
  score,
  n = NULL,
  weights = NULL,
  mode = 2,
  alpha = 3,
  beta = 3,
  control = NULL,
  scale = "T",
  plot = TRUE
)
}
\arguments{
\item{age}{A numeric vector of predictor values (e.g., age).}

\item{score}{A numeric vector of response values.}

\item{n}{The maximum score (number of trials in the beta-binomial
distribution). If NULL, max(score) is used.}

\item{weights}{A numeric vector of weights for each observation.
Default is NULL (equal weights).}

\item{mode}{Integer specifying the mode of the model, either 1 or 2.
Default is 2 (direct modelling of \eqn{\alpha} and \eqn{\beta};
recommended). If set to 1, the model is fitted on the basis of
\eqn{\mu} and \eqn{\sigma}, the predicted mean and standard deviation
over age (kept for backwards compatibility).}

\item{alpha}{Integer specifying the degree of the polynomial for the alpha
model. Default is 3. If mode is set to 1, this parameter is used to
specify the degree of the polynomial for the \eqn{\mu} model.}

\item{beta}{Integer specifying the degree of the polynomial for the beta
model. Default is 3. If mode is set to 1, this parameter is used to
specify the degree of the polynomial for the \eqn{\sigma} model.}

\item{control}{A list of control parameters to be passed to the
\code{optim} function. Parameters not specified are filled with
defaults, namely \code{list(reltol = 1e-8, maxit = 1000)} for mode 1
and adaptive L-BFGS-B settings (factr between 1e4 and 1e6, maxit
proportional to the number of parameters) for mode 2.}

\item{scale}{Type of norm scale, either "T" (default), "IQ", "z" or a
double vector with the mean and standard deviation.}

\item{plot}{Logical indicating whether to plot the model. Default is TRUE.}
}
\value{
A list of class "cnormBetaBinomial" or "cnormBetaBinomial2".
  In case of mode 2 containing:
  \item{alpha_est}{Estimated coefficients for the alpha model}
  \item{beta_est}{Estimated coefficients for the beta model}
  \item{se}{Standard errors of the estimated coefficients}
  \item{alpha_degree}{Degree of the polynomial for the alpha model}
  \item{beta_degree}{Degree of the polynomial for the beta model}
  \item{result}{Full result from the optimization procedure}
}
\description{
This function fits a beta-binomial regression model where both the
\eqn{\alpha} and \eqn{\beta} parameters of the beta-binomial distribution
are modeled as polynomial functions of the predictor variable (typically
age). Setting mode to 1 fits a beta-binomial model on the basis of
\eqn{\mu} and \eqn{\sigma}, setting it to 2 (default and recommended)
fits a beta-binomial model directly on the basis of \eqn{\alpha} and
\eqn{\beta}.
}
\details{
The function standardizes the input variables, fits polynomial models for
both distribution parameters, and uses maximum likelihood estimation with
analytic gradients to find the optimal parameters. In mode 2, the
optimization is performed using the L-BFGS-B method.

Please note that the model fit indices (log-likelihood, AIC, BIC) of
mode 1 and mode 2 models are not directly comparable, since mode 1
maximizes a normal likelihood on standardized scores, whereas mode 2
maximizes the discrete beta-binomial likelihood on the raw scores.
}
\examples{
\dontrun{
# Fit a beta-binomial regression model to the PPVT data
model <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228)
summary(model)

# Use weights for post-stratification
marginals <- data.frame(var = c("sex", "sex", "migration", "migration"),
                        level = c(1,2,0,1),
                        prop = c(0.51, 0.49, 0.65, 0.35))
weights <- computeWeights(ppvt, marginals)
model <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228, weights = weights)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{bestModel}
\alias{bestModel}
\title{Determine Regression Model}
\usage{
bestModel(
  data,
  raw = NULL,
  R2 = NULL,
  k = NULL,
  t = NULL,
  predictors = NULL,
  terms = 0,
  weights = NULL,
  force.in = NULL,
  plot = TRUE,
  extensive = TRUE,
  averaging = FALSE,
  minDip = 0.01
)
}
\arguments{
\item{data}{Preprocessed dataset with 'raw' scores, powers, interactions, and
usually an explanatory variable (like age).}

\item{raw}{Name of the raw score variable (default: 'raw').}

\item{R2}{Adjusted R^2 stopping criterion for model building.}

\item{k}{Power constant influencing model complexity (default: taken from the
data preparation, max: 6).}

\item{t}{Age power parameter. If unset, taken from the data preparation
(default 3).}

\item{predictors}{List of predictors or regression formula for model
selection. Overrides 'k' and can include additional variables.}

\item{terms}{Desired number of terms in the model.}

\item{weights}{Optional case weights. If set to FALSE, default weights
(if any) are ignored.}

\item{force.in}{Variables forcibly included in the regression.}

\item{plot}{If TRUE (default), displays a percentile plot of the model and
information about the regression object. FALSE turns off plotting and report.}

\item{extensive}{If TRUE (default), screen models for consistency and - if
possible - exclude inconsistent ones.}

\item{averaging}{If TRUE (default FALSE), apply BIC-weighted model averaging
across the consistency-screened candidate models instead of selecting a
single model. Requires \code{extensive = TRUE} and age-based norming.}

\item{minDip}{Tolerance for monotonicity check. Allow small violations
(default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
for strict checking.}
}
\value{
The model. Further exploration can be done using
  \code{plotSubset(model)} and \code{plotPercentiles(data, model)}.
}
\description{
Computes Taylor polynomial regression models by evaluating a series of models
with increasing number of predictors. It aims to find a consistent model that
effectively captures the variance in the data. It draws on the regsubsets
function from the leaps package, builds up to 10 models per number of
predictors, screens them for model consistency (monotonicity of the mapping
between norm score and raw score over the complete age range) and selects the
consistent model with the highest R^2. This automatic model selection should
usually be accompanied by visual inspection of the percentile plots and
assessment of fit statistics. Set R^2 or the number of terms manually to
retrieve a more parsimonious model, if desired.
}
\details{
The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
\code{computePowers} and \code{prepareData} are usually not called directly,
but accessed through other functions like \code{cnorm}.

Additional functions like \code{plotSubset(model)} and \code{cnorm.cv} can
aid in model evaluation.

If \code{averaging = TRUE}, the final coefficients are not taken from a
single selected model, but computed as a BIC-weighted average across all
consistency-screened candidate models (weights \eqn{w_j \propto
exp(-\Delta BIC_j / 2)}). Since a convex combination of functions that are
all monotone in the same direction is itself monotone, the averaged model is
guaranteed to remain consistent, while reducing model selection variance.
This replaces the deprecated \code{subsampling} approach.
}
\examples{
\dontrun{
# It is not recommended to directly use this function. Rather use 'cnorm' instead.
normData <- prepareData(elfe)
model <- bestModel(normData)
plotSubset(model)
plotPercentiles(buildCnormObject(normData, model))

# Specifying variables explicitly
preselectedModel <- bestModel(normData, predictors = c("L1", "L3", "L1A3", "A2", "A3"))
print(regressionFunction(preselectedModel))
}
}
\seealso{
plotSubset, plotPercentiles, plotPercentileSeries, checkConsistency

Other model:
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{getNormScoreSE}
\alias{getNormScoreSE}
\title{Calculates the standard error (SE) or root mean square error (RMSE) of the
norm scores. In case of large datasets, both results should be almost
identical.}
\usage{
getNormScoreSE(model, type = 2)
}
\arguments{
\item{model}{a cnorm object}

\item{type}{either '1' for the standard error sensu Oosterhuis et al. (2016)
or '2' for the RMSE (default)}
}
\value{
The standard error (SE) of the norm scores sensu Oosterhuis et al.
  (2016) or the RMSE
}
\description{
Calculates the standard error (SE) or root mean square error (RMSE) of the
norm scores. In case of large datasets, both results should be almost
identical.
}
\references{
Oosterhuis, H. E. M., van der Ark, L. A., & Sijtsma, K. (2016).
  Sample Size Requirements for Traditional and Regression-Based Norms.
  Assessment, 23(2), 191-202. https://doi.org/10.1177/1073191115580638
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{weighted.quantile.inflation}
\alias{weighted.quantile.inflation}
\title{Weighted quantile estimator through case inflation}
\usage{
weighted.quantile.inflation(
  x,
  probs,
  weights = NULL,
  degree = 3,
  cutoff = 1e+07
)
}
\arguments{
\item{x}{A numerical vector}

\item{probs}{Numerical vector of quantiles}

\item{weights}{A numerical vector with weights; should have the same length as x.}

\item{degree}{power parameter for case inflation (default = 3, equaling factor 1000)
If no weights are provided (NULL), it falls back to the base quantile function, type 7}

\item{cutoff}{stop criterion for the sum of standardized weights to switch to Harrell-Davis,
default = 1000000}
}
\value{
the quantiles
}
\description{
Applies weighted ranking numerically by inflating cases according to weight. This function
will be resource intensive, if inflated cases get too high and in this cases, it switches
to the parametric Harrell-Davis estimator.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{normTable.betabinomial}
\alias{normTable.betabinomial}
\title{Calculate Cumulative Probabilities, Density, Percentiles, and Z-Scores for
Beta-Binomial Distribution}
\usage{
normTable.betabinomial(
  model,
  ages,
  n = NULL,
  m = NULL,
  range = 3,
  CI = 0.9,
  reliability = NULL
)
}
\arguments{
\item{model}{The model, fitted with \code{\link{cnorm.betabinomial}}.}

\item{ages}{A numeric vector of age points at which to compute the tables.}

\item{n}{The number of items resp. the maximum score.}

\item{m}{An optional stop criterion in table generation. Positive integer
lower than n. Please note: The probabilities are always computed on the
full support 0 to n; m only truncates the output table.}

\item{range}{The range of the norm scores in standard deviations.
Default is 3. Thus, scores in the range of +/- 3 standard deviations
are considered.}

\item{CI}{Confidence coefficient, ranging from 0 to 1, default .9.}

\item{reliability}{Reliability coefficient, ranging between 0 to 1.}
}
\value{
A list of data frames with columns: x, Px, Pcum, Percentile, z,
  norm score and possibly confidence intervals.
}
\description{
This function generates a norm table for specific ages based on the beta
binomial regression model. In case a confidence coefficient (CI, default
.9) and the reliability is specified, confidence intervals are computed
for the true score estimates, including a correction for regression to
the mean (Eid & Schmidt, 2012, p. 272).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotRaw}
\alias{plotRaw}
\title{Plot manifest and fitted raw scores}
\usage{
plotRaw(model, group = FALSE, type = 0)
}
\arguments{
\item{model}{The regression model from the 'cnorm' function}

\item{group}{Should the fit be displayed by group?}

\item{type}{Type of display: 0 = plot manifest against fitted values, 1 = plot
manifest against difference values}
}
\description{
The function plots the raw data against the fitted scores from
the regression model per group. This helps to inspect the precision
of the modeling process. The scores should not deviate too far from
regression line.
}
\examples{
\dontrun{
  # Compute model with example dataset and plot results
  result <- cnorm(raw = elfe$raw, group = elfe$group)
  plotRaw(result)
}
}
\seealso{
Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDensity]{plotDensity()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{subsample_lm}
\alias{subsample_lm}
\title{K-fold Resampled Coefficient Estimation for Linear Regression}
\usage{
subsample_lm(text, data, weights, k = 10)
}
\arguments{
\item{text}{A character string or formula specifying the model to be fitted}

\item{data}{A data frame containing the variables in the model}

\item{weights}{Optional numeric vector of weights. If NULL, unweighted regression is performed}

\item{k}{Integer specifying the number of resampling folds (default = 10)}
}
\value{
An object of class 'lm' with averaged coefficients from k-fold resampling.
For small samples, returns a standard lm object.
}
\description{
Performs k-fold resampling to estimate averaged coefficients for linear regression.
The coefficients are averaged across k different subsets of the data to provide
more stable estimates. For small samples (n < 100), returns a standard linear model instead.
}
\details{
The function splits the data into k subsets, fits a linear model on k-1 subsets,
and stores the coefficients. This process is repeated k times, and the final
coefficients are averaged across all iterations to provide more stable estimates.
}
\keyword{deprecated}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{diagnostics.betabinomial}
\alias{diagnostics.betabinomial}
\title{Diagnostic Information for Beta-Binomial Model}
\usage{
diagnostics.betabinomial(model, age = NULL, score = NULL, weights = NULL)
}
\arguments{
\item{model}{An object of class "cnormBetaBinomial" or
"cnormBetaBinomial2", typically the result of a call to
cnorm.betabinomial().}

\item{age}{An optional vector with age values.}

\item{score}{An optional vector with raw values.}

\item{weights}{An optional vector with weights.}
}
\value{
A list containing diagnostic information, including convergence,
  number of evaluations, log-likelihood, AIC, BIC, the maximum absolute
  analytic gradient at the solution (first-order optimality), parameter
  estimates with standard errors, z and p values, and - if age and score
  are provided - R2, RMSE and bias of the norm scores.
}
\description{
This function provides diagnostic information for a fitted beta-binomial
model from the cnorm.betabinomial function. It returns various metrics
related to model convergence, fit, and complexity. In case age and raw
scores are provided, the function as well computes R2, RMSE and bias for
the norm scores based on the manifest and predicted norm scores.
}
\details{
The AIC and BIC are calculated as:
AIC = 2k - 2ln(L)
BIC = ln(n)k - 2ln(L)
where k is the number of parameters, L is the maximum likelihood, and n
is the number of observations.

Please note that fit indices of mode 1 and mode 2 models are not directly
comparable, since the two modes maximize different likelihoods (normal
on standardized scores vs. discrete beta-binomial).
}
\examples{
\dontrun{
model <- cnorm.betabinomial(ppvt$age, ppvt$raw)
diag_info <- diagnostics.betabinomial(model)
print(diag_info)
summary(diag_info)

if(diag_info$converged) {
  cat("Model converged successfully.\n")
} else {
  cat("Warning: Model did not converge.\n")
}

cat("AIC:", diag_info$AIC, "\n")
cat("BIC:", diag_info$BIC, "\n")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{simulateRasch}
\alias{simulateRasch}
\title{Simulate raw test scores based on Rasch model}
\usage{
simulateRasch(
  data = NULL,
  n = 100,
  minAge = 1,
  maxAge = 7,
  items.n = 21,
  items.m = 0,
  items.sd = 1,
  Theta = "random",
  width = 1
)
}
\arguments{
\item{data}{data.frame from previous simulations for recomputation (overrides n, minAge, maxAge)}

\item{n}{The sample size per age group}

\item{minAge}{The minimum age (default 1)}

\item{maxAge}{The maximum age (default 7)}

\item{items.n}{The number of items of the test}

\item{items.m}{The mean difficulty of the items}

\item{items.sd}{The standard deviation of the item difficulty}

\item{Theta}{irt scales difficulty parameters, either "random" for drawing a random sample,
"even" for evenly distributed or a set of predefined values, which then overrides the item.n
parameters}

\item{width}{The width of the window size for the continuous age per group; +- 1/2 width around group
center
on items.m and item.sd; if set to FALSE, the distribution is not drawn randomly but normally nonetheless}
}
\value{
a list containing the simulated data and thetas
\describe{
  \item{data}{the data.frame with only age, group and raw}
  \item{sim}{the complete simulated data with item level results}
  \item{theta}{the difficulty of the items}
}
}
\description{
For testing purposes only:
The function simulates raw test scores based on a virtual Rasch based test with n results per
age group, an evenly distributed age variable, items.n test items with a simulated difficulty and
standard deviation. The development trajectories over age group are modeled by a curve linear
function of age, with at first fast progression, which slows down over age, and a slightly increasing
standard deviation in order to model a scissor effects. The item difficulties can be accessed via $theta
and the raw data via $data of the returned object.
}
\examples{
\dontrun{
# simulate data for a rather easy test (m = -1.0)
sim <- simulateRasch(n=150, minAge=1,
                     maxAge=7, items.n = 30, items.m = -1.0,
                     items.sd = 1, Theta = "random", width = 1.0)

# Show item difficulties
mean(sim$theta)
sd(sim$theta)
hist(sim$theta)

# Plot raw scores
boxplot(raw~group, data=sim$data)

# Model data
data <- prepareData(sim$data, age="age")
model <- bestModel(data, k = 4)
printSubset(model)
plotSubset(model, type=0)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{printSubset}
\alias{printSubset}
\title{Print Model Selection Information}
\usage{
printSubset(x, ...)
}
\arguments{
\item{x}{Model output from 'bestModel' or a cnorm object.}

\item{...}{Additional parameters.}
}
\value{
Table with model information criteria.
}
\description{
Displays R^2 and other metrics for models with varying predictors, aiding in
choosing the best-fitting model after model fitting. The F-test compares each
model to the preceding (smaller) one, with degrees of freedom based on the
actual difference in the number of parameters.
}
\examples{
\dontrun{
  result <- cnorm(raw = elfe$raw, group = elfe$group)
  printSubset(result)
}
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=derive]{derive()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cNORM.R
\name{cNORM.GUI2}
\alias{cNORM.GUI2}
\title{Launch the cNORM Parametric Modeling Shiny Application}
\usage{
cNORM.GUI2(launch.browser = TRUE)
}
\arguments{
\item{launch.browser}{Logical, whether to launch browser automatically (default: TRUE)}
}
\description{
Launch the cNORM Parametric Modeling Shiny Application
}
\examples{
\dontrun{
# Launch graphical user interface
cNORM.GUI2()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{gradient_log_likelihood}
\alias{gradient_log_likelihood}
\title{Analytic gradient of the negative log-likelihood (mode 1)}
\usage{
gradient_log_likelihood(params, X, Z, y, weights = NULL)
}
\arguments{
\item{params}{A numeric vector containing all model parameters. The first
n_beta elements are coefficients for the mean model, and the remaining
elements are coefficients for the log-standard deviation model.}

\item{X}{A matrix of predictors for the mean model.}

\item{Z}{A matrix of predictors for the log-standard deviation model.}

\item{y}{A numeric vector of response values.}

\item{weights}{A numeric vector of weights for each observation
(NULL = equal weights).}
}
\value{
Numeric vector: gradient of the negative log-likelihood with
  respect to \code{params}.
}
\description{
Computes the exact gradient of \code{\link{log_likelihood}} with respect
to all regression coefficients. With \eqn{r = (y - \mu)/\sigma}:
\deqn{\partial\ell/\partial\mu = r/\sigma, \qquad
      \partial\ell/\partial\log\sigma = r^2 - 1.}
Coefficient gradients follow via the chain rule through the design
matrices. Contributions of observations whose log(sigma) linear predictor
is clamped (see \code{log_likelihood}) are zeroed, which is the exact
subgradient of the clamped objective. Non-finite per-observation terms
(matching the penalty branch of the objective) are zeroed as well.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelling.R
\name{derive}
\alias{derive}
\title{Derivative of regression model}
\usage{
derive(model, order = 1)
}
\arguments{
\item{model}{The regression model or a cnorm object}

\item{order}{The degree of the derivative, default: 1}
}
\value{
The derived coefficients
}
\description{
Calculates the derivative of the location / norm value from the regression
model with the first derivative as the default. This is useful for finding
violations of model assumptions and problematic distribution features as
f. e. bottom and ceiling effects, non-progressive norm scores within an age
group or in general intersecting percentile curves.
}
\examples{
\dontrun{
  m <- cnorm(raw = elfe$raw, group = elfe$group)
  derivedCoefficients <- derive(m, order = 1)
}
}
\seealso{
Other model:
\code{\link[=bestModel]{bestModel()}},
\code{\link[=checkConsistency]{checkConsistency()}},
\code{\link[=cnorm.cv]{cnorm.cv()}},
\code{\link[=modelSummary]{modelSummary()}},
\code{\link[=print.cnorm]{print.cnorm()}},
\code{\link[=printSubset]{printSubset()}},
\code{\link[=rangeCheck]{rangeCheck()}},
\code{\link[=regressionFunction]{regressionFunction()}},
\code{\link[=summary.cnorm]{summary.cnorm()}},
\code{\link[=weightedAverageModel]{weightedAverageModel()}}
}
\concept{model}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{weighted.quantile.type7}
\alias{weighted.quantile.type7}
\title{Weighted type7 quantile estimator}
\usage{
weighted.quantile.type7(x, probs, weights = NULL)
}
\arguments{
\item{x}{A numerical vector}

\item{probs}{Numerical vector of quantiles}

\item{weights}{A numerical vector with weights; should have the same length as x.
If no weights are provided (NULL), it falls back to the base quantile function, type 7}
}
\value{
the quantiles
}
\description{
Computes weighted quantiles; code from Andrey Akinshin (2023) "Weighted quantile estimators" arXiv:2304.07265 [stat.ME]
Code made available via the CC BY-NC-SA 4.0 license
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{bb_design_matrix}
\alias{bb_design_matrix}
\title{Build a raw polynomial design matrix with intercept}
\usage{
bb_design_matrix(x, degree)
}
\arguments{
\item{x}{A numeric vector.}

\item{degree}{Degree of the raw polynomial.}
}
\value{
A numeric matrix with columns 1, x, x^2, ..., x^degree.
}
\description{
Build a raw polynomial design matrix with intercept
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{cnorm.shash}
\alias{cnorm.shash}
\title{Fit a Sinh-Arcsinh (shash) Regression Model for Continuous Norming}
\usage{
cnorm.shash(
  age,
  score,
  weights = NULL,
  mu_degree = 3,
  sigma_degree = 2,
  epsilon_degree = 2,
  delta_degree = 1,
  delta = 1,
  control = NULL,
  scale = "T",
  plot = TRUE
)
}
\arguments{
\item{age}{A numeric vector of predictor values (typically age, but can be any continuous predictor).}

\item{score}{A numeric vector of response values (raw test scores). Must be the same length as age.
The value range is unrestricted and it can include zeros and negative values.}

\item{weights}{An optional numeric vector of weights for each observation.
Useful for incorporating sampling weights. If NULL (default), all observations are weighted equally.}

\item{mu_degree}{Integer specifying the degree of the polynomial for modeling the location parameter mu(age).
Default is 3. Higher degrees allow more flexible modeling of how the central tendency changes with age,
but may lead to overfitting with small samples. Common choices:
\itemize{
  \item 1: Linear change with age
  \item 2: Quadratic change (allows one inflection point)
  \item 3: Cubic change (allows two inflection points, suitable for most developmental curves)
  \item 4+: Higher-order changes (use cautiously, mainly for large samples)
}}

\item{sigma_degree}{Integer specifying the degree of the polynomial for modeling the scale parameter sigma(age).
Default is 2. This controls how the variability (spread) of scores changes with age.
Lower degrees are often sufficient as variability typically changes more smoothly than location.}

\item{epsilon_degree}{Integer specifying the degree of the polynomial for modeling the skewness parameter epsilon(age).
Default is 2. This controls how the asymmetry of the distribution changes with age.}

\item{delta_degree}{Integer specifying the degree of the polynomial for modeling the tail weight parameter delta(age).
Default is 1. The tail weight can be fixed as well in case of numerical instability. In that case, set
\code{delta_degree} to NULL and specify a value for \code{delta} instead. Recommendation: Keep
\code{delta_degree} low to avoid overfitting.}

\item{delta}{Fixed tail weight parameter (must be > 0). Default is 1. This parameter controls the
heaviness of the distribution tails and is kept constant across all ages when \code{delta_degree}
is set to NULL. Common values:
\itemize{
  \item delta = 1: Normal-like tail behavior (baseline)
  \item delta < 1: Heavier tails, higher kurtosis (more extreme scores than normal distribution)
  \item delta > 1: Lighter tails, lower kurtosis (fewer extreme scores than normal distribution)
}}

\item{control}{An optional list of control parameters passed to the \code{optim} function for
maximum likelihood estimation. Any parameters not specified are filled with sensible defaults
based on the model complexity. Common parameters to adjust:
\itemize{
  \item \code{factr}: Controls precision of optimization (default: 1e-8)
  \item \code{maxit}: Maximum number of iterations (default: n_parameters * 200)
  \item \code{lmm}: Memory limit for L-BFGS-B (default: min(n_parameters, 20))
}
Increase \code{maxit} or decrease \code{factr} if optimization fails to converge.}

\item{scale}{Character string or numeric vector specifying the type of norm scale for output.
This affects the scaling of derived norm scores but does not influence model fitting:
\itemize{
  \item "T": T-scores (mean = 50, SD = 10) - default
  \item "IQ": IQ-like scores (mean = 100, SD = 15)
  \item "z": z-scores (mean = 0, SD = 1)
  \item c(M, SD): Custom scale with specified mean M and standard deviation SD
}}

\item{plot}{Logical indicating whether to automatically display a diagnostic plot of the fitted model.
Default is TRUE.}
}
\value{
An object of class "cnormShash" containing the fitted model results. This is a list with components:
  \item{mu_est}{Numeric vector of estimated coefficients for the location parameter mu(age).
    The first coefficient is the intercept, subsequent coefficients correspond to polynomial terms.}
  \item{sigma_est}{Numeric vector of estimated coefficients for the scale parameter log(sigma(age)).
    Note: These are coefficients for log(sigma) to ensure sigma > 0.}
  \item{epsilon_est}{Numeric vector of estimated coefficients for the skewness parameter epsilon(age).}
  \item{delta}{The fixed tail weight parameter value used in fitting (relevant if \code{delta_degree} is NULL).}
  \item{delta_est}{Numeric vector of estimated coefficients for log(delta(age)) -
     in case a degree has been set, otherwise NULL.}
  \item{se}{Numeric vector of standard errors for all estimated coefficients (if Hessian computation succeeds).}
  \item{mu_degree, sigma_degree, epsilon_degree}{The polynomial degrees used for each parameter.}
  \item{delta_degree}{The polynomial degree used for delta, or NULL if delta was fixed.}
  \item{result}{Complete output from the \code{optim} function, including convergence information,
    log-likelihood value, and other optimization details.}
}
\description{
This function fits a Sinh-Arcsinh (shash; Jones & Pewsey, 2009) regression model for continuous norm
score modeling, where the distribution parameters vary smoothly as polynomial functions of age or other
predictors. The shash distribution is well-suited for psychometric data as it can flexibly model
skewness and tail weight independently, making it ideal for handling floor effects, ceiling effects,
and varying degrees of individual differences across age groups. In a simulation study (Lenhard et
al, 2019), the shash model demonstrated superior performance compared to other parametric approaches
from the Box Cox family of functions. In contrast to Box Cox, Sinh-Arcsinh can model distributions
including zero and negative values.
}
\details{
This implementation uses the Jones & Pewsey (2009) parameterization of the Sinh-Arcsinh distribution.
Parameters are estimated using maximum likelihood via the L-BFGS-B algorithm with analytic
gradients. In case optimization fails, try reducing model complexity by reducing polynomial
degrees or fixing the delta parameter.

\subsection{The Sinh-Arcsinh Distribution}{
The shash distribution is defined by the transformation:
\deqn{X = \mu + \sigma \cdot \sinh\left(\frac{\text{arcsinh}(Y) - \epsilon}{\delta}\right)}
where Y is a standard normal variable, Y ~ N(0,1).

This transformation provides:
\itemize{
  \item mu: Location parameter (similar to mean)
  \item sigma: Scale parameter (similar to standard deviation)
  \item epsilon: Skewness parameter (epsilon = 0 for symmetry)
  \item delta: Tail weight parameter (delta = 1 for normal-like tails)
}
}

\subsection{Model Selection}{
Choose polynomial degrees based on:
\itemize{
  \item Sample size (higher degrees need more data)
  \item Theoretical expectations about developmental trajectories
  \item Model comparison criteria (AIC, BIC)
  \item Visual inspection of fitted curves
}

For most applications, mu_degree = 3, sigma_degree = 2, epsilon_degree = 2, delta_degree = 1 provides
a good balance of flexibility and parsimony.
}
}
\note{
\itemize{
  \item The function requires the input data to have sufficient variability. Very small datasets
    or datasets with little age spread may cause convergence problems.
  \item Polynomial models can exhibit edge effects at the boundaries of the age range.
    Predictions outside the observed age range should be made cautiously.
  \item If convergence fails, try: (1) reducing polynomial degrees, (2) adjusting the delta parameter,
    (3) providing custom control parameters, or (4) checking for data quality issues.
  \item If parameters end up at their box constraints, the reported standard errors and
    Wald tests for these parameters are not valid; a warning is issued in this case.
}
}
\examples{
\dontrun{
# Basic usage with default settings
model <- cnorm.shash(age = children$age, score = children$raw_score)

# Custom polynomial degrees for complex developmental pattern
model_complex <- cnorm.shash(
  age = adolescents$age,
  score = adolescents$vocabulary_score,
  mu_degree = 4,         # Complex mean trajectory
  sigma_degree = 3,      # Changing variability pattern
  epsilon_degree = 2,    # Skewness shifts
  delta_degree = NULL,   # set to NULL to activate fixed delta
  delta = 0.8            # Slightly heavy tails
)

# With sampling weights
model_weighted <- cnorm.shash(
  age = survey$age,
  score = survey$score,
  weights = survey$sample_weight
)

# Custom optimization control for difficult convergence
model_robust <- cnorm.shash(
  age = mixed$age,
  score = mixed$score,
  control = list(factr = 1e-6, maxit = 2000),
  delta = 1.5
)

# Compare model fit
compare(model, model_complex)
}

}
\references{
Jones, M. C., & Pewsey, A. (2009). Sinh-arcsinh distributions. *Biometrika*, 96(4), 761-780.

Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A
simulation study of parametric and semi-parametric approaches. *PLoS ONE*, 14(9), e0222279.
https://doi.org/10.1371/journal.pone.0222279
}
\seealso{
\code{\link{plot}} for plotting fitted models,
\code{\link{predict}} for generating predictions,
\code{\link{cnorm.betabinomial2}} for discrete beta-binomial alternative
}
\author{
Wolfgang Lenhard
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/roots.R
\name{calcPolyInL}
\alias{calcPolyInL}
\title{Internal function for retrieving regression function coefficients at specific age}
\usage{
calcPolyInL(raw, age, model)
}
\arguments{
\item{raw}{The raw value (subtracted from the intercept)}

\item{age}{The age}

\item{model}{The cNORM regression model}
}
\value{
The coefficients
}
\description{
The function is an inline for searching zeros in the inverse regression
function. It collapses the regression function at a specific age and simplifies
the coefficients.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{getNormCurve}
\alias{getNormCurve}
\title{Computes the curve for a specific T value}
\usage{
getNormCurve(
  norm,
  model,
  minAge = NULL,
  maxAge = NULL,
  step = 0.1,
  minRaw = NULL,
  maxRaw = NULL
)
}
\arguments{
\item{norm}{The specific norm score, e. g. T value}

\item{model}{The model from the regression modeling obtained with the cnorm function}

\item{minAge}{Age to start from}

\item{maxAge}{Age to stop at}

\item{step}{Stepping parameter for the precision when retrieving of the values, lower
values indicate higher precision (default 0.1).}

\item{minRaw}{lower bound of the range of raw scores (default = 0)}

\item{maxRaw}{upper bound of raw scores}
}
\value{
data.frame of the variables raw, age and norm
}
\description{
As with this continuous norming regression approach, raw scores are modeled as a function of age and norm score
(location), getNormCurve is a straightforward approach to show the raw score development over
age, while keeping the norm value constant. This way, e. g. academic performance or intelligence development
of a specific ability is shown.
}
\examples{
\dontrun{
# Generate cnorm object from example data
cnorm.elfe <- cnorm(raw = elfe$raw, group = elfe$group)
getNormCurve(35, cnorm.elfe)
}

}
\seealso{
Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial]{predict.cnormBetaBinomial()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{weighted.quantile}
\alias{weighted.quantile}
\title{Weighted quantile estimator}
\usage{
weighted.quantile(x, probs, weights = NULL, type = "Harrell-Davis")
}
\arguments{
\item{x}{A numerical vector}

\item{probs}{Numerical vector of quantiles}

\item{weights}{A numerical vector with weights; should have the same length as x}

\item{type}{Type of estimator, can either be "inflation", "Harrell-Davis" using a beta function to
approximate the weighted percentiles (Harrell & Davis, 1982) or "Type7" (default; Hyndman & Fan, 1996), an adaption
of the generic quantile function in R, including weighting. The inflation procedure is essentially
a numerical, non-parametric solution that gives the same results as Harrel-Davis. It requires less
ressources with small datasets and always finds a solution (e. g. 1000 cases with
weights between 1 and 10). If it becomes too resource intense, it switches to Harrell-Davis automatically.
Harrel-Davis and Type7 code is based on the work of Akinshin (2023).}
}
\value{
the weighted quantiles
}
\description{
Computes weighted quantiles (code from Andrey Akinshin (2023) "Weighted quantile estimators" arXiv:2304.07265 [stat.ME]
Code made available via the CC BY-NC-SA 4.0 license) on the basis of either the weighted Harrell-Davis
quantile estimator or an adaption of the type 7 quantile estimator of the generic quantile function in
the base package. Please provide a vector with raw values, the probabilities for the quantiles and an
additional vector with the weight of each observation. In case the weight vector is NULL, a normal
quantile estimation is done. The vectors may not include NAs and the weights should be positive non-zero
values. Please draw on the computeWeights() function for retrieving weights in post stratification.
}
\references{
\enumerate{
  \item Harrell, F.E. & Davis, C.E. (1982). A new distribution-free quantile estimator. Biometrika, 69(3), 635-640.
  \item Hyndman, R. J. & Fan, Y. (1996). Sample quantiles in statistical packages, American Statistician 50, 361–365.
  \item Akinshin, A. (2023). Weighted quantile estimators arXiv:2304.07265 [stat.ME]
}
}
\seealso{
weighted.quantile.inflation, weighted.quantile.harrell.davis, weighted.quantile.type7
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cNORM.R
\name{taylorSwift}
\alias{taylorSwift}
\title{Swiftly compute Taylor regression models for distribution free continuous norming}
\usage{
taylorSwift(
  raw = NULL,
  group = NULL,
  age = NULL,
  width = NA,
  weights = NULL,
  scale = "T",
  method = 4,
  descend = FALSE,
  k = NULL,
  t = NULL,
  terms = 0,
  R2 = NULL,
  plot = TRUE,
  extensive = TRUE,
  averaging = FALSE,
  minDip = 0.01
)
}
\arguments{
\item{raw}{Numeric vector of raw scores}

\item{group}{Numeric vector of grouping variable, e. g. grade. If no group
or age variable is provided, conventional norming is applied}

\item{age}{Numeric vector with chronological age, please additionally specify
width of window}

\item{width}{Size of the sliding window in case an age vector is used}

\item{weights}{Vector or variable name in the dataset with weights for each
individual case. It can be used to compensate for moderate imbalances due to
insufficient norm data stratification. Weights should be numerical and positive.}

\item{scale}{type of norm scale, either T (default), IQ, z or percentile (= no
transformation); a double vector with the mean and standard deviation can as
well, be provided f. e. c(10, 3) for Wechsler scale index points}

\item{method}{Ranking method in case of bindings, please provide an index,
choosing from the following methods: 1 = Blom (1958), 2 = Tukey (1949),
3 = Van der Warden (1952), 4 = Rankit (default), 5 = Levenbach (1953),
6 = Filliben (1975), 7 = Yu & Huang (2001)}

\item{descend}{ranking order (default descent = FALSE): inverses the
ranking order with higher raw scores getting lower norm scores; relevant
for example when norming error scores, where lower scores mean higher
performance}

\item{k}{The power constant. Higher values result in more detailed approximations
but have the danger of over-fit (max = 6). If not set, it uses t and if both
parameters are NULL, k is set to 5.}

\item{t}{The age power parameter (max = 6). If not set, it uses k and if both
parameters are NULL, k is set to 3, since age trajectories are most often well
captured by cubic polynomials.}

\item{terms}{Selection criterion for model building. The best fitting model with
this number of terms is used}

\item{R2}{Adjusted R square as a stopping criterion for the model building
(default R2 = 0.99)}

\item{plot}{Default TRUE; plots the regression model and prints report}

\item{extensive}{If TRUE, screen models for consistency and - if possible, exclude inconsistent ones}

\item{averaging}{If TRUE (default FALSE), apply BIC-weighted model averaging
across the consistency-screened candidate models instead of selecting a
single model. Requires \code{extensive = TRUE} and age-based norming.}

\item{minDip}{Tolerance for monotonicity check. Allow small violations
(default: .01 or 1% of the raw score range). Decrease e. g. to 1e-6
for strict checking.}
}
\value{
cnorm object including the ranked raw data and the regression model
}
\description{
Conducts distribution free continuous norming and aims to find a fitting model. Raw data are modelled as a Taylor polynomial
of powers of age and location and their interactions. In addition to the
raw scores, either provide a numeric vector for the grouping information (group)
for the ranking of the raw scores. You can adjust the grade of smoothing of the regression model by setting the k, t and terms
parameter. In general, increasing k and t leads to a higher fit, while lower values lead to more smoothing. If both parameters
are missing, taylorSwift uses k = 5 and t = 3 by default.
}
\examples{
\dontrun{
# Using this function with the example dataset 'ppvt'
# You can use the 'getGroups()' function to set up grouping variable in case,
# you have a continuous age variable.
model <- taylorSwift(raw = ppvt$raw, group = ppvt$group)

# return norm tables including 90\% confidence intervals for a
# test with a reliability of r = .85; table are set to mean of quartal
# in grade 3 (children completed 2 years of schooling)
normTable(c(5, 15), model, CI = .90, reliability = .95)

# ... or instead of raw scores for norm scores, the other way round
rawTable(c(8, 12), model, CI = .90, reliability = .95)
}
}
\references{
\enumerate{
  \item Gary, S. & Lenhard, W. (2021). In norming we trust. Diagnostica.
  \item Gary, S., Lenhard, W. & Lenhard, A. (2021). Modelling Norm Scores with the cNORM Package in R. Psych, 3(3), 501-521. https://doi.org/10.3390/psych3030033
  \item Lenhard, A., Lenhard, W., Suggate, S. & Segerer, R. (2016). A continuous solution to the norming problem. Assessment, Online first, 1-14. doi:10.1177/1073191116656437
  \item Lenhard, A., Lenhard, W., Gary, S. (2018). Continuous Norming (cNORM). The Comprehensive R Network, Package cNORM, available: https://CRAN.R-project.org/package=cNORM
  \item Lenhard, A., Lenhard, W., Gary, S. (2019). Continuous norming of psychometric tests: A simulation study of parametric and semi-parametric approaches. PLoS ONE, 14(9),  e0222279. doi:10.1371/journal.pone.0222279
  \item Lenhard, W., & Lenhard, A. (2020). Improvement of Norm Score Quality via Regression-Based Continuous Norming. Educational and Psychological Measurement(Online First), 1-33. https://doi.org/10.1177/0013164420928457

}
}
\seealso{
rankByGroup, rankBySlidingWindow, computePowers, bestModel
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preparation.R
\name{rankByGroup}
\alias{rankByGroup}
\title{Determine the norm scores of the participants in each subsample}
\usage{
rankByGroup(
  data = NULL,
  group = "group",
  raw = "raw",
  weights = NULL,
  method = 4,
  scale = "T",
  descend = FALSE,
  descriptives = TRUE,
  na.rm = TRUE,
  silent = FALSE
)
}
\arguments{
\item{data}{data.frame with norm sample data. If no data.frame is provided, the raw score
and group vectors are directly used}

\item{group}{name of the grouping variable (default 'group') or numeric vector, e. g. grade, setting
group to FALSE cancels grouping (data is treated as one group)}

\item{raw}{name of the raw value variable (default 'raw') or numeric vector}

\item{weights}{Vector or variable name in the dataset with weights for each individual case. It can be used
to compensate for moderate imbalances due to insufficient norm data stratification. Weights should be numerical
and positive.  Please use the 'computeWeights' function for this purpose.}

\item{method}{Ranking method in case of bindings, please provide an index,
choosing from the following methods: 1 = Blom (1958), 2 = Tukey (1949),
3 = Van der Warden (1952), 4 = Rankit (default), 5 = Levenbach (1953),
6 = Filliben (1975), 7 = Yu & Huang (2001)}

\item{scale}{type of norm scale, either T (default), IQ, z or percentile (= no
transformation); a double vector with the mean and standard deviation can as well,
be provided f. e. c(10, 3) for Wechsler scale index points}

\item{descend}{ranking order (default descent = FALSE): inverses the
ranking order with higher raw scores getting lower norm scores; relevant
for example when norming error scores, where lower scores mean higher
performance}

\item{descriptives}{If set to TRUE (default), information in n, mean, median and
standard deviation per group is added to each observation}

\item{na.rm}{remove values, where the percentiles could not be estimated,
most likely happens in the context of weighting}

\item{silent}{set to TRUE to suppress messages}
}
\value{
the dataset with the percentiles and norm scales per group
}
\description{
This is the initial step, usually done in all kinds of test norming projects,
after the scale is constructed and the norm sample is established. First,
the data is grouped according to a grouping variable and afterwards, the percentile
for each raw value is retrieved. The percentile can be used for the modeling
procedure, but in case, the samples to not deviate too much from normality,
T, IQ or z scores can be computed via a normal rank procedure based on the
inverse cumulative normal distribution. In case of bindings, we use the medium rank
and there are different methods for estimating the percentiles (default RankIt).
}
\section{Remarks on using covariates}{

So far the inclusion of a binary covariate is experimental and far from optimized.
The according variable name has to be specified in the ranking procedure
and the modeling includes this in the further process. At the moment, during ranking
the data are split into the according cells group x covariate, which leads to small
sample sizes. Please take care to have enough cases in each combination. Additionally,
covariates can lead to unstable modeling solutions. The question, if it is really
reasonable to include covariates when norming a test is a decision beyond the pure data
modeling. Please use with care or alternatively split the dataset into the two groups
beforehand and model them separately.

The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
\code{computePowers} and \code{prepareData} are usually not called directly, but accessed
through other functions like \code{cnorm}.
}

\examples{
\dontrun{
# Transformation with default parameters: RankIt and converting to T scores
data.elfe <- rankByGroup(elfe, group = "group") # using a data frame with vector names
data.elfe2 <- rankByGroup(raw=elfe$raw, group=elfe$group) # use vectors for raw score and group

# Transformation into Wechsler scores with Yu & Huang (2001) ranking procedure
data.elfe <- rankByGroup(raw = elfe$raw, group = elfe$group, method = 7, scale = c(10, 3))

# cNORM can as well be used for conventional norming, in case no group is given
d <- rankByGroup(raw = elfe$raw)
d <- computePowers(d)
m <- bestModel(d)
rawTable(0, m) # please use an arbitrary value for age when generating the tables
}

}
\seealso{
rankBySlidingWindow, computePowers, computeWeights, weighted.rank

Other prepare:
\code{\link[=computePowers]{computePowers()}},
\code{\link[=prepareData]{prepareData()}},
\code{\link[=rankBySlidingWindow]{rankBySlidingWindow()}}
}
\concept{prepare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{diagnostics.shash}
\alias{diagnostics.shash}
\title{Diagnostic Statistics for Sinh-Arcsinh Continuous Norming Model}
\usage{
diagnostics.shash(object, age = NULL, score = NULL, weights = NULL)
}
\arguments{
\item{object}{An object of class "cnormShash"}

\item{age}{An optional numeric vector of age values for computing fit statistics}

\item{score}{An optional numeric vector of raw scores for computing fit statistics}

\item{weights}{An optional numeric vector of observation weights}
}
\value{
A list containing comprehensive diagnostic information
}
\description{
This function computes detailed diagnostic statistics for a fitted shash model,
including fit statistics, parameter estimates, and convergence information.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preparation.R
\name{rankBySlidingWindow}
\alias{rankBySlidingWindow}
\title{Determine the norm scores of the participants by sliding window}
\usage{
rankBySlidingWindow(
  data = NULL,
  age = "age",
  raw = "raw",
  weights = NULL,
  width,
  method = 4,
  scale = "T",
  descend = FALSE,
  descriptives = TRUE,
  nGroup = 0,
  group = NA,
  na.rm = TRUE,
  silent = FALSE
)
}
\arguments{
\item{data}{data.frame with norm sample data}

\item{age}{the continuous age variable. Setting 'age' to FALSE inhibits computation of
powers of age and the interactions}

\item{raw}{name of the raw value variable (default 'raw')}

\item{weights}{Vector or variable name in the dataset with weights for each individual case. It can be used
to compensate for moderate imbalances due to insufficient norm data stratification. Weights should be numerical
and positive. It can be resource intense when applied to the sliding window. Please use the 'computeWeights' function for this purpose.}

\item{width}{the width of the sliding window}

\item{method}{Ranking method in case of bindings, please provide an index,
choosing from the following methods: 1 = Blom (1958), 2 = Tukey (1949),
3 = Van der Warden (1952), 4 = Rankit (default), 5 = Levenbach (1953),
6 = Filliben (1975), 7 = Yu & Huang (2001)}

\item{scale}{type of norm scale, either T (default), IQ, z or percentile (= no
transformation); a double vector with the mean and standard deviation can as well,
be provided f. e. c(10, 3) for Wechsler scale index points}

\item{descend}{ranking order (default descent = FALSE): inverses the
ranking order with higher raw scores getting lower norm scores; relevant
for example when norming error scores, where lower scores mean higher
performance}

\item{descriptives}{If set to TRUE (default), information in n, mean, median and
standard deviation per group is added to each observation}

\item{nGroup}{If set to a positive value, a grouping variable is created with the desired number of
equi distant groups, named by the group mean age of each group. It creates the
column 'group' in the data.frame and in case, there is already one with that name,
overwrites it.}

\item{group}{Optional parameter for providing the name of the grouping variable (if present; overwritten
if ngroups is used)}

\item{na.rm}{remove values, where the percentiles could not be estimated,
most likely happens in the context of weighting}

\item{silent}{set to TRUE to suppress messages}
}
\value{
the dataset with the individual percentiles and norm scores
}
\description{
The function retrieves all individuals in the predefined age range (x +/- width/2)
around each case and ranks that individual based on this individually drawn sample.
This function can be directly used with a continuous age variable in order to avoid
grouping. When collecting data on the basis of a continuous age variable, cases
located far from the mean age of the group receive distorted percentiles when building
discrete groups and generating percentiles with the traditional approach. The distortion
increases with distance from the group mean and this effect can be avoided by the
sliding window. Nonetheless, please ensure, that the optional grouping variable in fact
represents the correct mean age of the respective age groups, as this variable is
later on used for displaying the manifest data in the percentile plots.
}
\details{
In case of bindings, the function uses the medium rank and applies the algorithms
already described in the \code{\link{rankByGroup}} function. At the upper and lower end of the
data sample, the sliding stops and the sample is drawn from the interval min + width and
max - width, respectively.
}
\section{Remarks on using covariates}{

So far the inclusion of a binary covariate is experimental and far from optimized.
The according variable name has to be specified in the ranking procedure
and the modeling includes this in the further process. At the moment, during ranking
the data are split into the according degrees of the covariate and the ranking is done
separately. This may lead to small sample sizes. Please take care to have enough cases in each combination. Additionally,
covariates can lead to unstable modeling solutions. The question, if it is really
reasonable to include covariates when norming a test is a decision beyond the pure data
modeling. Please use with care or alternatively split the dataset into the two groups
beforehand and model them separately.

The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
\code{computePowers} and \code{prepareData} are usually not called directly, but accessed
through other functions like \code{cnorm}.
}

\examples{
\dontrun{
# Transformation using a sliding window
data.elfe2 <- rankBySlidingWindow(elfe, raw = "raw", age = "group", width = 0.5)

# Comparing this to the traditional approach should give us exactly the same
# values, since the sample dataset only has a grouping variable for age
data.elfe <- rankByGroup(elfe, group = "group")
mean(data.elfe$normValue - data.elfe2$normValue)
}
}
\seealso{
rankByGroup, computePowers, computeWeights, weighted.rank, weighted.quantile

Other prepare:
\code{\link[=computePowers]{computePowers()}},
\code{\link[=prepareData]{prepareData()}},
\code{\link[=rankByGroup]{rankByGroup()}}
}
\concept{prepare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{summary.cnormShash}
\alias{summary.cnormShash}
\title{Summarize a Sinh-Arcsinh Continuous Norming Model}
\usage{
\method{summary}{cnormShash}(object, ...)
}
\arguments{
\item{object}{An object of class "cnormShash", typically the result of a call to
\code{\link{cnorm.shash}}.}

\item{...}{Additional arguments passed to the summary method:
\itemize{
   \item age An optional numeric vector of age values corresponding to the raw scores.
     If provided along with \code{score}, additional fit statistics (R-squared, RMSE, bias)
     will be calculated.
   \item score An optional numeric vector of raw scores. Must be provided if \code{age} is given.
   \item weights An optional numeric vector of weights for each observation.
 }}
}
\value{
Invisibly returns a list containing detailed diagnostic information about the model.
  The function primarily produces printed output summarizing the model.
}
\description{
This function provides a summary of a fitted Sinh-Arcsinh (shash) continuous norming model,
including model fit statistics, convergence information, and parameter estimates.
}
\details{
The summary includes:
\itemize{
  \item Basic model information (polynomial degrees, delta parameter, number of observations)
  \item Model fit statistics (log-likelihood, AIC, BIC)
  \item R-squared, RMSE, and bias (if age and raw scores are provided)
  \item Convergence information
  \item Parameter estimates with standard errors, z-values, and p-values
}
}
\examples{
\dontrun{
model <- cnorm.shash(children$age, children$score)
summary(model)

# Including R-squared, RMSE, and bias in the summary:
summary(model, age = children$age, score = children$score)
}

}
\seealso{
\code{\link{cnorm.shash}}, \code{\link{diagnostics.shash}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{betaCoefficients}
\alias{betaCoefficients}
\title{Compute Parameters of a Beta Binomial Distribution}
\usage{
betaCoefficients(x, n = NULL)
}
\arguments{
\item{x}{A numeric vector of non-negative integers representing observed
counts.}

\item{n}{The maximum number or the maximum possible value of \code{x}.
If not specified, uses max(x) instead.}
}
\value{
A numeric vector containing the calculated parameters in the
  following order: alpha (a), beta (b), mean (m), standard deviation (sd),
  and the maximum number (n). If the method of moments is degenerate
  (mean at the boundary of the support or a non-positive moment
  denominator), alpha and beta are returned as \code{NA} with a warning.
  Non-positive estimates (indicating dispersion incompatible with a
  beta-binomial distribution) are returned as-is with a warning.
}
\description{
This function calculates the \eqn{\alpha} (a) and \eqn{\beta} (b)
parameters of a beta binomial distribution, along with the mean (m) and
standard deviation (sd) based on the input vector \code{x} and the maximum
number \code{n}, using the method of moments.
}
\details{
The beta-binomial distribution is a discrete probability distribution that
models the number of successes in a fixed number of trials, where the
probability of success varies from trial to trial. This variability in
success probability is modeled by a beta distribution. Such a calculation
is particularly relevant in scenarios where there is heterogeneity in
success probabilities across trials, which is common in real-world
situations, as for example the number of correct solutions in a
psychometric test, where the test has a fixed number of items.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{buildCnormObject}
\alias{buildCnormObject}
\title{Build cnorm object from data and bestModel model object}
\usage{
buildCnormObject(data, model)
}
\arguments{
\item{data}{A data object from 'prepareData', or from 'rankByGroup' and
'computePower'}

\item{model}{Object obtained from the bestModel function}
}
\value{
A cnorm object
}
\description{
Helper function to build a cnorm object from a data object and
a model object from the bestModel function for compatibility reasons.
}
\examples{
\dontrun{
  data <- prepareData(elfe)
  model <- bestModel(data, k = 4)
  model.cnorm <- buildCnormObject(data, model)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preparation.R
\name{prepareData}
\alias{prepareData}
\title{Prepare data for modeling in one step (convenience method)}
\usage{
prepareData(
  data = NULL,
  group = "group",
  raw = "raw",
  age = "group",
  k = 4,
  t = NULL,
  width = NA,
  weights = NULL,
  scale = "T",
  descend = FALSE,
  silent = FALSE
)
}
\arguments{
\item{data}{data.frame with a grouping variable named 'group' and a raw score variable
named 'raw'.}

\item{group}{grouping variable in the data, e. g. age groups, grades ...
Setting group = FALSE deactivates modeling in dependence of age. Use this in case you do want
conventional norm tables.}

\item{raw}{the raw scores}

\item{age}{the continuous explanatory variable; by default set to "group"}

\item{k}{The power parameter, default = 4}

\item{t}{the age power parameter (default NULL). If not set, cNORM automatically uses k. The age power parameter
can be used to specify the k to produce rectangular matrices and specify the course of scores per  independently from k}

\item{width}{if a width is provided, the function switches to rankBySlidingWindow to determine the
observed raw scores, otherwise, ranking is done by group (default)}

\item{weights}{Vector or variable name in the dataset with weights for each individual case. It can be used
to compensate for moderate imbalances due to insufficient norm data stratification. Weights should be numerical
and positive. Please use the 'computeWeights' function for this purpose.}

\item{scale}{type of norm scale, either T (default), IQ, z or percentile (= no
transformation); a double vector with the mean and standard deviation can as well,
be provided f. e. c(10, 3) for Wechsler scale index point}

\item{descend}{ranking order (default descent = FALSE): inverses the
ranking order with higher raw scores getting lower norm scores; relevant
for example when norming error scores, where lower scores mean higher
performance}

\item{silent}{set to TRUE to suppress messages}
}
\value{
data frame including the norm scores, powers and interactions of the norm score and
grouping variable
}
\description{
This is a convenience method to either load the inbuilt sample dataset, or
to provide a data frame with the variables "raw" (for the raw scores) and "group"
The function ranks the data within groups, computes norm values, powers of the norm
scores and interactions. Afterwards, you can use these preprocessed data to
determine the best fitting model.
}
\details{
The functions \code{rankBySlidingWindow}, \code{rankByGroup}, \code{bestModel},
\code{computePowers} and \code{prepareData} are usually not called directly, but accessed
through other functions like \code{cnorm}.
}
\examples{
\dontrun{
# conducts ranking and computation of powers and interactions with the 'elfe' dataset
data.elfe <- prepareData(elfe)

# use vectors instead of data frame
data.elfe <- prepareData(raw=elfe$raw, group=elfe$group)

# variable names can be specified as well, here with the BMI data included in the package
data.bmi <- prepareData(CDC, group = "group", raw = "bmi", age = "age")
}

# modeling with only one group with the 'elfe' dataset as an example
# this results in conventional norming
data.elfe2 <- prepareData(data = elfe, group = FALSE)
m <- bestModel(data.elfe2)
}
\seealso{
Other prepare:
\code{\link[=computePowers]{computePowers()}},
\code{\link[=rankByGroup]{rankByGroup()}},
\code{\link[=rankBySlidingWindow]{rankBySlidingWindow()}}
}
\concept{prepare}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotDensity}
\alias{plotDensity}
\title{Plot the density function per group by raw score}
\usage{
plotDensity(
  model,
  minRaw = NULL,
  maxRaw = NULL,
  minNorm = NULL,
  maxNorm = NULL,
  group = NULL
)
}
\arguments{
\item{model}{The model from the bestModel function, a cnorm object, a cnormBetaBinomial, a cnormBetaBinomial2 or
cnormShash object.}

\item{minRaw}{Lower bound of the raw score. If NULL, it's automatically determined based on the model type.}

\item{maxRaw}{Upper bound of the raw score. If NULL, it's automatically determined based on the model type.}

\item{minNorm}{Lower bound of the norm score. If NULL, it's automatically determined based on the model type.}

\item{maxNorm}{Upper bound of the norm score. If NULL, it's automatically determined based on the model type.}

\item{group}{Numeric vector specifying the age groups to plot. If NULL, groups are automatically selected.}
}
\value{
A ggplot object representing the density functions.
}
\description{
This function plots density curves based on the regression model against the raw scores.
It supports both traditional continuous norming models and beta-binomial models.
The function allows for customization of the plot range and groups to be displayed.
}
\details{
The function generates density curves for specified age groups, allowing for easy comparison of score distributions
across different ages.

For beta-binomial models, the density is based on the probability mass function, while for
traditional models, it uses a normal distribution based on the norm scores.
}
\note{
Please check for inconsistent curves, especially those showing implausible shapes
such as violations of biuniqueness in the cnorm models.
}
\examples{
\dontrun{
# For traditional continuous norming model
result <- cnorm(raw = elfe$raw, group = elfe$group)
plotDensity(result, group = c(2, 4, 6))

# For beta-binomial model
bb_model <- cnorm.betabinomial(age = ppvt$age, score = ppvt$raw, n = 228)
plotDensity(bb_model)
}

}
\seealso{
\code{\link{plotNormCurves}}, \code{\link{plotPercentiles}}

Other plot:
\code{\link[=compare]{compare()}},
\code{\link[=plot.cnorm]{plot.cnorm()}},
\code{\link[=plot.cnormBetaBinomial]{plot.cnormBetaBinomial()}},
\code{\link[=plot.cnormBetaBinomial2]{plot.cnormBetaBinomial2()}},
\code{\link[=plotDerivative]{plotDerivative()}},
\code{\link[=plotNorm]{plotNorm()}},
\code{\link[=plotNormCurves]{plotNormCurves()}},
\code{\link[=plotPercentileSeries]{plotPercentileSeries()}},
\code{\link[=plotPercentiles]{plotPercentiles()}},
\code{\link[=plotRaw]{plotRaw()}},
\code{\link[=plotSubset]{plotSubset()}}
}
\concept{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raking.R
\name{checkWeights}
\alias{checkWeights}
\title{Check, if NA or values <= 0 occur and issue warning}
\usage{
checkWeights(weights)
}
\arguments{
\item{weights}{Raking weights}
}
\description{
Check, if NA or values <= 0 occur and issue warning
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{autoselect.betabinomial}
\alias{autoselect.betabinomial}
\title{Automatic model selection for beta-binomial continuous norming via BIC}
\usage{
autoselect.betabinomial(
  age,
  score,
  n = NULL,
  weights = NULL,
  mode = 2,
  max_alpha = 4,
  max_beta = 4,
  min_alpha = 1,
  min_beta = 1,
  control = NULL,
  scale = "T",
  parallel = TRUE,
  n_cores = NULL,
  plot = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{age, score}{Numeric vectors of predictor and response values.}

\item{n}{Maximum score. Defaults to \code{max(score)}.}

\item{weights}{Optional numeric vector of weights.}

\item{mode}{1 for \eqn{\mu}/\eqn{\sigma}, 2 for direct
\eqn{\alpha}/\eqn{\beta} (default).}

\item{max_alpha, max_beta}{Maximum polynomial degrees. Default 4.}

\item{min_alpha, min_beta}{Minimum polynomial degrees. Default 1.}

\item{control}{Optional control list passed to \code{\link[stats]{optim}}.}

\item{scale}{Norm scale (default \code{"T"}).}

\item{parallel}{Logical; attempt parallel execution. Default \code{TRUE}.}

\item{n_cores}{Number of cores. Defaults to all logical cores.}

\item{plot}{Logical; plot the selected model. Default \code{TRUE}.}

\item{verbose}{Logical; print progress. Default \code{TRUE}.}
}
\value{
The selected fitted model, with an additional element
  \code{selection} containing the evaluation table and selection details.
}
\description{
Selects polynomial degrees for the two components of a beta-binomial model
(\eqn{\alpha}/\eqn{\beta} when \code{mode = 2} or \eqn{\mu}/\eqn{\sigma}
when \code{mode = 1}) by minimizing BIC over the full grid of degree
combinations.
}
\details{
Parallel execution is attempted by default. If the workers cannot access
the \pkg{cNORM} namespace (e.g. during \code{devtools::load_all()}), the
function transparently falls back to sequential execution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{summary.cnormBetaBinomial}
\alias{summary.cnormBetaBinomial}
\title{Summarize a Beta-Binomial Continuous Norming Model}
\usage{
\method{summary}{cnormBetaBinomial}(object, ...)
}
\arguments{
\item{object}{An object of class "cnormBetaBinomial" or
"cnormBetaBinomial2", typically the result of a call to
\code{\link{cnorm.betabinomial}}.}

\item{...}{Additional arguments passed to the summary method:
\itemize{
   \item age An optional numeric vector of age values corresponding to
     the raw scores. If provided along with \code{score}, additional fit
     statistics (R-squared, RMSE, bias) will be calculated.
   \item score An optional numeric vector of raw scores. Must be
     provided if \code{age} is given.
   \item weights An optional numeric vector of weights for each
     observation.
 }}
}
\value{
Invisibly returns a list containing detailed diagnostic
  information about the model. The function primarily produces printed
  output summarizing the model.
}
\description{
This function provides a summary of a fitted beta-binomial continuous
norming model, including model fit statistics, convergence information,
and parameter estimates.
}
\details{
The summary includes:
\itemize{
  \item Basic model information (type, number of observations, number of parameters)
  \item Model fit statistics (log-likelihood, AIC, BIC)
  \item R-squared, RMSE, and bias (if age and raw scores are provided)
        in comparison to manifest norm scores
  \item Convergence information, including the maximum absolute gradient
        at the solution (first-order optimality)
  \item Parameter estimates with standard errors, z-values, and p-values
}
}
\examples{
\dontrun{
model <- cnorm.betabinomial(ppvt$age, ppvt$raw, n = 228)
summary(model)

# Including R-squared, RMSE, and bias in the summary:
summary(model, age = ppvt$age, score = ppvt$raw)
}
}
\seealso{
\code{\link{cnorm.betabinomial}}, \code{\link{diagnostics.betabinomial}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utilities.R
\name{getGroups}
\alias{getGroups}
\title{Determine groups and group means}
\usage{
getGroups(x, n = NULL, equidistant = FALSE)
}
\arguments{
\item{x}{Numeric vector to be split.}

\item{n}{Desired number of groups. If \code{NULL}, the function chooses
roughly one group per 100 observations, constrained to 3 \eqn{\le} n \eqn{\le} 20.}

\item{equidistant}{If \code{TRUE}, intervals have equal width; otherwise
(default) each interval contains roughly the same number of observations.}
}
\value{
Numeric vector of the same length as \code{x} with the group mean
  assigned to each observation.
}
\description{
Splits a continuous variable into \code{n} groups (either equidistant or
with equal number of observations) and returns the group mean for every
observation.
}
\examples{
\dontrun{
x <- rnorm(1000, mean = 50, sd = 10)
m <- getGroups(x, n = 10)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/betaBinomial.R
\name{predict.cnormBetaBinomial}
\alias{predict.cnormBetaBinomial}
\title{Predict Norm Scores from Raw Scores}
\usage{
\method{predict}{cnormBetaBinomial}(object, ...)
}
\arguments{
\item{object}{A fitted model object of class 'cnormBetaBinomial' or
'cnormBetaBinomial2'.}

\item{...}{Additional arguments passed to the prediction method:
\itemize{
   \item age A numeric vector of ages, same length as score.
   \item score A numeric vector of raw scores.
   \item range The range of the norm scores in standard deviations.
     Default is 3. Thus, scores in the range of +/- 3 standard
     deviations are considered.
 }}
}
\value{
A numeric vector of norm scores (or percentiles, if no norm scale
  was specified in the model). Missing, non-integer or out-of-range raw
  scores yield \code{NA}.
}
\description{
This function calculates norm scores based on raw scores, age, and a
fitted cnormBetaBinomial model.
}
\details{
The function first predicts the alpha and beta parameters of the
beta-binomial distribution for each unique age using the provided model.
It then calculates the mid-p cumulative probability for each raw score
given these parameters and converts these probabilities to the norm scale
specified in the model. The distribution is computed only once per unique
age value, which considerably speeds up predictions for grouped data.
}
\examples{
\dontrun{
# Assuming you have a fitted model named 'bb_model':
model <- cnorm.betabinomial(ppvt$age, ppvt$raw)
raw <- c(100, 121, 97, 180)
ages <- c(7, 8, 9, 10)
norm_scores <- predict(model, ages, raw)
}

}
\seealso{
Other predict:
\code{\link[=derivationTable]{derivationTable()}},
\code{\link[=getNormCurve]{getNormCurve()}},
\code{\link[=normTable]{normTable()}},
\code{\link[=predict.cnormBetaBinomial2]{predict.cnormBetaBinomial2()}},
\code{\link[=predict.cnormShash]{predict.cnormShash()}},
\code{\link[=predictNorm]{predictNorm()}},
\code{\link[=predictRaw]{predictRaw()}},
\code{\link[=rawTable]{rawTable()}}
}
\concept{predict}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raking.R
\name{computeWeights}
\alias{computeWeights}
\title{Weighting of cases through iterative proportional fitting (Raking)}
\usage{
computeWeights(data, population.margins, standardized = TRUE)
}
\arguments{
\item{data}{data.frame with norm sample data.}

\item{population.margins}{A data.frame including three columns, specifying the
variable name in the original dataset used for data stratification, the factor
level of the variable and the according population share. Please ensure the
original data does not include factor levels not present in the
population.margins. Additionally, summing up the shares of the different
levels of a variable should result in a value near 1.0. The first column must
specify the name of the stratification variable, the second the level and
the third the proportion.}

\item{standardized}{If TRUE (default), the raking weights are scaled to
weights/min(weights)}
}
\value{
a vector with the standardized weights
}
\description{
Computes and standardizes weights via raking to compensate for non-stratified
samples. It is based on the implementation in the survey R package. It reduces
data collection biases in the norm data by the means of post stratification,
thus reducing the effect of unbalanced data in percentile estimation and norm
data modeling.
}
\details{
This function computes standardized raking weights to overcome biases in norm
samples. It generates weights by drawing on the information of population
shares (e. g. for sex, ethnic group, region ...) and subsequently reduces the
influence of over-represented groups or increases underrepresented cases. The
returned weights are either raw or standardized and scaled to be larger than 0.

Raking in general has a number of advantages over post stratification and it
additionally allows cNORM to draw on larger datasets, since less cases have
to be removed during stratification. To use this function, additionally to the
data, a data frame with stratification variables has to be specified. The data
frame should include a row with (a) the variable name, (b) the level of the
variable and (c) the according population proportion.
}
\examples{
\dontrun{
# cNORM features a dataset on vocabulary development (ppvt)
# that includes variables like sex or migration. In order
# to weight the data, we have to specify the population shares.
# According to census, the population includes 52\% boys
# (factor level 1 in the ppvt dataset) and 70\% / 30\% of persons
# without / with a history of migration (= 0 / 1 in the dataset).
# First we set up the population margins with all shares of the
# different levels:

margins <- data.frame(variables = c("sex", "sex",
                                    "migration", "migration"),
                      levels = c(1, 2, 0, 1),
                      share = c(.52, .48, .7, .3))
head(margins)

# Now we use the population margins to generate weights
# through raking

weights <- computeWeights(ppvt, margins)


# There are as many different weights as combinations of
# factor levels, thus only four in this specific case

unique(weights)


# To include the weights in the cNORM modelling, we have
# to pass them as weights. They are then used to set up
# weighted quantiles and as weights in the regression.

model <- cnorm(raw = ppvt$raw,
               group = ppvt$group,
               weights = weights)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shash.R
\name{gradient_shash}
\alias{gradient_shash}
\title{Analytic gradient of the negative log-likelihood for a shash regression model}
\usage{
gradient_shash(
  params,
  X_mu,
  X_sigma,
  X_epsilon,
  X_delta = NULL,
  y,
  weights = NULL,
  fixed_delta = NULL
)
}
\arguments{
\item{params}{A numeric vector containing all model parameters}

\item{X_mu}{Design matrix for location parameter}

\item{X_sigma}{Design matrix for scale parameter}

\item{X_epsilon}{Design matrix for skewness parameter}

\item{X_delta}{Design matrix for tail weight parameter (or NULL for fixed delta)}

\item{y}{Response vector}

\item{weights}{Observation weights}

\item{fixed_delta}{If not NULL, the delta parameter is fixed to this value}
}
\value{
Numeric vector: gradient of the negative log-likelihood with
  respect to \code{params}.
}
\description{
Computes the exact gradient of \code{\link{log_likelihood_shash}} with
respect to all regression coefficients. Supplying this gradient to
\code{\link[stats]{optim}} (method \code{"L-BFGS-B"}) avoids costly and
noisy finite-difference approximations, typically yielding a speedup
and more reliable convergence.
}
\details{
With \eqn{z = (y - \mu)/\sigma} and \eqn{A = \delta \, \mathrm{asinh}(z) + \epsilon},
the per-observation log-density is
\deqn{\ell = \log\delta - \log\sigma - \tfrac{1}{2}\log(2\pi)
  - \tfrac{1}{2}\log(1+z^2) - \tfrac{1}{2}\sinh^2(A) + \log\cosh(A).}
The gradient uses the chain rule through the linear predictors:
\deqn{\partial\ell/\partial A = \tanh(A) - \sinh(A)\cosh(A)}
\deqn{\partial\ell/\partial z = -z/(1+z^2) + (\partial\ell/\partial A)\,\delta/\sqrt{1+z^2}}
\deqn{\partial\ell/\partial\mu = -(\partial\ell/\partial z)/\sigma}
\deqn{\partial\ell/\partial\log\sigma = -1 - z\,(\partial\ell/\partial z)}
\deqn{\partial\ell/\partial\epsilon = \partial\ell/\partial A}
\deqn{\partial\ell/\partial\log\delta = 1 + (\partial\ell/\partial A)\,\delta\,\mathrm{asinh}(z)}
Coefficient gradients follow as \eqn{-X^\top (w \cdot \partial\ell/\partial\eta)}
for the negative log-likelihood. Contributions of observations whose linear
predictor is clamped (see \code{log_likelihood_shash}) are zeroed, which is
the exact subgradient of the clamped objective. In extreme regions where
\eqn{\sinh(A)\cosh(A)} would overflow, the gradient magnitude is capped at
a large finite value; these regions coincide with the penalty branch of the
objective, so the capped direction remains correct.
}
\keyword{internal}
